//cpuchk.c
//CPU check functions for Stump DOS loader
//Bryan E. Topp <betopp@betopp.com> 2022

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "cpuchk.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

//Returns the EDX value from the CPUID result, given the specified EAX and ECX input.
//This looks super ugly because it has to take parameters and give results as pairs of 16-bit values.
uint32_t cpuid_edx(uint32_t eax, uint32_t ecx);
#pragma aux cpuid_edx = \
	"xchg EDX, EAX" \
	"and EDX, 0xFFFF" \
	"shl EAX, 16" \
	"or EAX, EDX" \
	"and EBX, 0xFFFF" \
	"shl ECX, 16" \
	"or ECX, EBX" \
	"xor EBX, EBX" \
	"xor EDX, EDX" \
	"db 0x0f, 0xa2" \
	"mov EAX, EDX" \
	"shr EDX, 16" \
	parm [AX DX] [CX BX] \
	value [AX DX] \
	modify [AX DX CX BX]
	
//Returns the EBX value from the CPUID result, given the specified EAX and ECX input.
uint32_t cpuid_ebx(uint32_t eax, uint32_t ecx);
#pragma aux cpuid_ebx = \
	"xchg EDX, EAX" \
	"and EDX, 0xFFFF" \
	"shl EAX, 16" \
	"or EAX, EDX" \
	"and EBX, 0xFFFF" \
	"shl ECX, 16" \
	"or ECX, EBX" \
	"xor EBX, EBX" \
	"xor EDX, EDX" \
	"db 0x0f, 0xa2" \
	"xchg EDX, EBX" \
	"mov EAX, EDX" \
	"shr EDX, 16" \
	parm [AX DX] [CX BX] \
	value [AX DX] \
	modify [AX DX CX BX]
	
//Returns the ECX value from the CPUID result, given the specified EAX and ECX input.
uint32_t cpuid_ecx(uint32_t eax, uint32_t ecx);
#pragma aux cpuid_ecx = \
	"xchg EDX, EAX" \
	"and EDX, 0xFFFF" \
	"shl EAX, 16" \
	"or EAX, EDX" \
	"and EBX, 0xFFFF" \
	"shl ECX, 16" \
	"or ECX, EBX" \
	"xor EBX, EBX" \
	"xor EDX, EDX" \
	"db 0x0f, 0xa2" \
	"xchg EDX, ECX" \
	"mov EAX, EDX" \
	"shr EDX, 16" \
	parm [AX DX] [CX BX] \
	value [AX DX] \
	modify [AX DX CX BX]
	
	

void cpuchk(void)
{
	//Make sure our CPU configuration seems like it can run Stump.
	//We need a multicore 64-bit CPU with a few extensions.	
	char cpuid_vendor[13] = {0};
	*((uint32_t*)(cpuid_vendor + 0)) = cpuid_ebx(0, 0);
	*((uint32_t*)(cpuid_vendor + 4)) = cpuid_edx(0, 0);
	*((uint32_t*)(cpuid_vendor + 8)) = cpuid_ecx(0, 0);
	printf("CPUID vendor string: %s\n", cpuid_vendor);

	if(!(cpuid_edx(1, 0) & (1ull<<6)))
	{
		printf("No PAE support (can't be a 64-bit CPU).\n");
		exit(-1);
	}
	
	if(!(cpuid_edx(1, 0) & (1ull<<9)))
	{
		printf("No Local APIC (can't be a multicore CPU).\n");
		exit(-1);
	}	

	if(!(cpuid_edx(0x80000001, 0) & (1ull<<29)))
	{
		printf("No Long Mode support (not a 64-bit CPU).\n");
		exit(-1);
	}
	
	if(!(cpuid_edx(0x80000001, 0) & (1ull<<20)))
	{
		printf("No Non-Executable page support.\n");
		exit(-1);
	}
	
	if(!(cpuid_edx(0x80000001, 0) & (1ull<<11)))
	{
		printf("No SYSCALL instruction support.\n");
		exit(-1);
	}
	
	if(!(cpuid_edx(0x80000001, 0) & (1ull<<26)))
	{
		printf("No 1GByte page support.\n");
		exit(-1);
	}

	if(!(cpuid_ebx(7, 0) & (1ull<<0)))
	{
		printf("No direct FS-/GS-BASE manipulation support.\n");
		exit(-1);
	}
	
	if(!(cpuid_edx(1, 0) & (1ull<<5)))
	{
		printf("No support for Model-Specific Registers.\n");
		exit(-1);
	}
}
