//dmem.c
//DOS page-aligned memory allocation
//Bryan E. Topp <betopp@betopp.com> 2022

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "dmem.h"
#include <assert.h>
#include <dos.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>

//Ranges of memory allocated from DOS
typedef struct dmem_pool_s
{
	//Block as originally allocated
	void huge *block;
	
	//Beginning of remaining memory, in 16-byte units
	uint16_t addr_left;
	
	//Size of remaining memory, in 16-byte units
	uint16_t size_left;
	
} dmem_pool_t;
#define DMEM_POOL_MAX 16
static dmem_pool_t dmem_pool_table[DMEM_POOL_MAX];

static void dmem_atexit(void)
{
	for(int pp = 0; pp < DMEM_POOL_MAX; pp++)
	{
		if(dmem_pool_table[pp].block != NULL)
		{
			hfree(dmem_pool_table[pp].block);
			dmem_pool_table[pp].block = NULL;
		}
	}
}

void dmem_init(void)
{
	atexit(&dmem_atexit);
}

void far *dmem_alloc(uint32_t nbytes)
{
	assert((nbytes % 4096) == 0);
	assert(nbytes < 65536ull);
	
	//May have to search multiple times, if we allocate more in our pools
	bool keep_looking = true;
	while(keep_looking)
	{
		keep_looking = false;
		
		//Check if we have enough space in one of our pools.
		for(int pp = 0; pp < DMEM_POOL_MAX; pp++)
		{
			dmem_pool_t *pptr = &(dmem_pool_table[pp]);
			if(pptr->block == NULL)
				continue;
			
			//Make sure the pool is page-aligned
			uint16_t misalign = pptr->addr_left % (4096 / 16);
			if(misalign != 0)
			{
				uint16_t correction = (4096/16) - misalign;
				if(correction > pptr->size_left)
				{
					pptr->addr_left = 0;
					pptr->size_left = 0;
				}
				else
				{
					pptr->addr_left += correction;
					pptr->size_left -= correction;
				}
			}
			
			//See if there's enough room
			if(pptr->size_left >= nbytes / 16)
			{
				//printf("Allocating from DOS pool %d\n", pp);
				//This pool has enough room - allocate from its beginning.
				void far *retval = MK_FP(pptr->addr_left, 0);
				_fmemset(retval, 0, nbytes);
				pptr->addr_left += nbytes / 16;
				pptr->size_left -= nbytes / 16;
				return retval;
			}
		}
		
		//We don't have memory in our pools. Try to allocate more.
		for(int pp = 0; pp < DMEM_POOL_MAX; pp++)
		{
			dmem_pool_t *pptr = &(dmem_pool_table[pp]);
			if(pptr->block != NULL)
				continue;
			
			pptr->block = halloc(65536, 1);
			if(pptr->block == NULL)
				break;
			
			printf("Acquired 64KBytes of DOS memory at %Fp\n", pptr->block);

			assert(FP_OFF(pptr->block) == 0);
			pptr->addr_left = FP_SEG(pptr->block);
			pptr->size_left = 65536/16;
			keep_looking = true;
			break;
		}
	}
	
	//No memory and no room to allocate any
	printf("Ran out of DOS memory while setting up kernel.\n");
	exit(-1);
	return NULL;
}

