//keyb.c
//Keyboard hook for real-mode loop
//Bryan E. Topp <betopp@betopp.com> 2022

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <dos.h>
#include "keyb.h"
#include "ring.h"

//Definition of scancodes in kernel; lines up with system-call library
typedef enum _sc_con_scancode_e
{
	_SC_CON_SCANCODE_UNKNOWN = 0,
	_SC_CON_SCANCODE_A = 4,
	_SC_CON_SCANCODE_B = 5,
	_SC_CON_SCANCODE_C = 6,
	_SC_CON_SCANCODE_D = 7,
	_SC_CON_SCANCODE_E = 8,
	_SC_CON_SCANCODE_F = 9,
	_SC_CON_SCANCODE_G = 10,
	_SC_CON_SCANCODE_H = 11,
	_SC_CON_SCANCODE_I = 12,
	_SC_CON_SCANCODE_J = 13,
	_SC_CON_SCANCODE_K = 14,
	_SC_CON_SCANCODE_L = 15,
	_SC_CON_SCANCODE_M = 16,
	_SC_CON_SCANCODE_N = 17,
	_SC_CON_SCANCODE_O = 18,
	_SC_CON_SCANCODE_P = 19,
	_SC_CON_SCANCODE_Q = 20,
	_SC_CON_SCANCODE_R = 21,
	_SC_CON_SCANCODE_S = 22,
	_SC_CON_SCANCODE_T = 23,
	_SC_CON_SCANCODE_U = 24,
	_SC_CON_SCANCODE_V = 25,
	_SC_CON_SCANCODE_W = 26,
	_SC_CON_SCANCODE_X = 27,
	_SC_CON_SCANCODE_Y = 28,
	_SC_CON_SCANCODE_Z = 29,
	_SC_CON_SCANCODE_1 = 30,
	_SC_CON_SCANCODE_2 = 31,
	_SC_CON_SCANCODE_3 = 32,
	_SC_CON_SCANCODE_4 = 33,
	_SC_CON_SCANCODE_5 = 34,
	_SC_CON_SCANCODE_6 = 35,
	_SC_CON_SCANCODE_7 = 36,
	_SC_CON_SCANCODE_8 = 37,
	_SC_CON_SCANCODE_9 = 38,
	_SC_CON_SCANCODE_0 = 39,
	_SC_CON_SCANCODE_RETURN = 40,
	_SC_CON_SCANCODE_ESCAPE = 41,
	_SC_CON_SCANCODE_BACKSPACE = 42,
	_SC_CON_SCANCODE_TAB = 43,
	_SC_CON_SCANCODE_SPACE = 44,
	_SC_CON_SCANCODE_MINUS = 45,
	_SC_CON_SCANCODE_EQUALS = 46,
	_SC_CON_SCANCODE_LEFTBRACKET = 47,
	_SC_CON_SCANCODE_RIGHTBRACKET = 48,
	_SC_CON_SCANCODE_BACKSLASH = 49,
	_SC_CON_SCANCODE_NONUSHASH = 50,
	_SC_CON_SCANCODE_SEMICOLON = 51,
	_SC_CON_SCANCODE_APOSTROPHE = 52,
	_SC_CON_SCANCODE_GRAVE = 53,
	_SC_CON_SCANCODE_COMMA = 54,
	_SC_CON_SCANCODE_PERIOD = 55,
	_SC_CON_SCANCODE_SLASH = 56,
	_SC_CON_SCANCODE_CAPSLOCK = 57,
	_SC_CON_SCANCODE_F1 = 58,
	_SC_CON_SCANCODE_F2 = 59,
	_SC_CON_SCANCODE_F3 = 60,
	_SC_CON_SCANCODE_F4 = 61,
	_SC_CON_SCANCODE_F5 = 62,
	_SC_CON_SCANCODE_F6 = 63,
	_SC_CON_SCANCODE_F7 = 64,
	_SC_CON_SCANCODE_F8 = 65,
	_SC_CON_SCANCODE_F9 = 66,
	_SC_CON_SCANCODE_F10 = 67,
	_SC_CON_SCANCODE_F11 = 68,
	_SC_CON_SCANCODE_F12 = 69,
	_SC_CON_SCANCODE_PRINTSCREEN = 70,
	_SC_CON_SCANCODE_SCROLLLOCK = 71,
	_SC_CON_SCANCODE_PAUSE = 72,
	_SC_CON_SCANCODE_INSERT = 73,
	_SC_CON_SCANCODE_HOME = 74,
	_SC_CON_SCANCODE_PAGEUP = 75,
	_SC_CON_SCANCODE_DELETE = 76,
	_SC_CON_SCANCODE_END = 77,
	_SC_CON_SCANCODE_PAGEDOWN = 78,
	_SC_CON_SCANCODE_RIGHT = 79,
	_SC_CON_SCANCODE_LEFT = 80,
	_SC_CON_SCANCODE_DOWN = 81,
	_SC_CON_SCANCODE_UP = 82,
	_SC_CON_SCANCODE_NUMLOCKCLEAR = 83,
	_SC_CON_SCANCODE_KP_DIVIDE = 84,
	_SC_CON_SCANCODE_KP_MULTIPLY = 85,
	_SC_CON_SCANCODE_KP_MINUS = 86,
	_SC_CON_SCANCODE_KP_PLUS = 87,
	_SC_CON_SCANCODE_KP_ENTER = 88,
	_SC_CON_SCANCODE_KP_1 = 89,
	_SC_CON_SCANCODE_KP_2 = 90,
	_SC_CON_SCANCODE_KP_3 = 91,
	_SC_CON_SCANCODE_KP_4 = 92,
	_SC_CON_SCANCODE_KP_5 = 93,
	_SC_CON_SCANCODE_KP_6 = 94,
	_SC_CON_SCANCODE_KP_7 = 95,
	_SC_CON_SCANCODE_KP_8 = 96,
	_SC_CON_SCANCODE_KP_9 = 97,
	_SC_CON_SCANCODE_KP_0 = 98,
	_SC_CON_SCANCODE_KP_PERIOD = 99,
	_SC_CON_SCANCODE_NONUSBACKSLASH = 100,
	_SC_CON_SCANCODE_APPLICATION = 101,
	_SC_CON_SCANCODE_POWER = 102,
	_SC_CON_SCANCODE_KP_EQUALS = 103,
	_SC_CON_SCANCODE_F13 = 104,
	_SC_CON_SCANCODE_F14 = 105,
	_SC_CON_SCANCODE_F15 = 106,
	_SC_CON_SCANCODE_F16 = 107,
	_SC_CON_SCANCODE_F17 = 108,
	_SC_CON_SCANCODE_F18 = 109,
	_SC_CON_SCANCODE_F19 = 110,
	_SC_CON_SCANCODE_F20 = 111,
	_SC_CON_SCANCODE_F21 = 112,
	_SC_CON_SCANCODE_F22 = 113,
	_SC_CON_SCANCODE_F23 = 114,
	_SC_CON_SCANCODE_F24 = 115,
	_SC_CON_SCANCODE_EXECUTE = 116,
	_SC_CON_SCANCODE_HELP = 117,
	_SC_CON_SCANCODE_MENU = 118,
	_SC_CON_SCANCODE_SELECT = 119,
	_SC_CON_SCANCODE_STOP = 120,
	_SC_CON_SCANCODE_AGAIN = 121,
	_SC_CON_SCANCODE_UNDO = 122,
	_SC_CON_SCANCODE_CUT = 123,
	_SC_CON_SCANCODE_COPY = 124,
	_SC_CON_SCANCODE_PASTE = 125,
	_SC_CON_SCANCODE_FIND = 126,
	_SC_CON_SCANCODE_MUTE = 127,
	_SC_CON_SCANCODE_VOLUMEUP = 128,
	_SC_CON_SCANCODE_VOLUMEDOWN = 129,
	_SC_CON_SCANCODE_KP_COMMA = 133,
	_SC_CON_SCANCODE_KP_EQUALSAS400 = 134,
	_SC_CON_SCANCODE_INTERNATIONAL1 = 135,
	_SC_CON_SCANCODE_INTERNATIONAL2 = 136,
	_SC_CON_SCANCODE_INTERNATIONAL3 = 137,
	_SC_CON_SCANCODE_INTERNATIONAL4 = 138,
	_SC_CON_SCANCODE_INTERNATIONAL5 = 139,
	_SC_CON_SCANCODE_INTERNATIONAL6 = 140,
	_SC_CON_SCANCODE_INTERNATIONAL7 = 141,
	_SC_CON_SCANCODE_INTERNATIONAL8 = 142,
	_SC_CON_SCANCODE_INTERNATIONAL9 = 143,
	_SC_CON_SCANCODE_LANG1 = 144,
	_SC_CON_SCANCODE_LANG2 = 145,
	_SC_CON_SCANCODE_LANG3 = 146,
	_SC_CON_SCANCODE_LANG4 = 147,
	_SC_CON_SCANCODE_LANG5 = 148,
	_SC_CON_SCANCODE_LANG6 = 149,
	_SC_CON_SCANCODE_LANG7 = 150,
	_SC_CON_SCANCODE_LANG8 = 151,
	_SC_CON_SCANCODE_LANG9 = 152,
	_SC_CON_SCANCODE_ALTERASE = 153,
	_SC_CON_SCANCODE_SYSREQ = 154,
	_SC_CON_SCANCODE_CANCEL = 155,
	_SC_CON_SCANCODE_CLEAR = 156,
	_SC_CON_SCANCODE_PRIOR = 157,
	_SC_CON_SCANCODE_RETURN2 = 158,
	_SC_CON_SCANCODE_SEPARATOR = 159,
	_SC_CON_SCANCODE_OUT = 160,
	_SC_CON_SCANCODE_OPER = 161,
	_SC_CON_SCANCODE_CLEARAGAIN = 162,
	_SC_CON_SCANCODE_CRSEL = 163,
	_SC_CON_SCANCODE_EXSEL = 164,
	_SC_CON_SCANCODE_KP_00 = 176,
	_SC_CON_SCANCODE_KP_000 = 177,
	_SC_CON_SCANCODE_THOUSANDSSEPARATOR = 178,
	_SC_CON_SCANCODE_DECIMALSEPARATOR = 179,
	_SC_CON_SCANCODE_CURRENCYUNIT = 180,
	_SC_CON_SCANCODE_CURRENCYSUBUNIT = 181,
	_SC_CON_SCANCODE_KP_LEFTPAREN = 182,
	_SC_CON_SCANCODE_KP_RIGHTPAREN = 183,
	_SC_CON_SCANCODE_KP_LEFTBRACE = 184,
	_SC_CON_SCANCODE_KP_RIGHTBRACE = 185,
	_SC_CON_SCANCODE_KP_TAB = 186,
	_SC_CON_SCANCODE_KP_BACKSPACE = 187,
	_SC_CON_SCANCODE_KP_A = 188,
	_SC_CON_SCANCODE_KP_B = 189,
	_SC_CON_SCANCODE_KP_C = 190,
	_SC_CON_SCANCODE_KP_D = 191,
	_SC_CON_SCANCODE_KP_E = 192,
	_SC_CON_SCANCODE_KP_F = 193,
	_SC_CON_SCANCODE_KP_XOR = 194,
	_SC_CON_SCANCODE_KP_POWER = 195,
	_SC_CON_SCANCODE_KP_PERCENT = 196,
	_SC_CON_SCANCODE_KP_LESS = 197,
	_SC_CON_SCANCODE_KP_GREATER = 198,
	_SC_CON_SCANCODE_KP_AMPERSAND = 199,
	_SC_CON_SCANCODE_KP_DBLAMPERSAND = 200,
	_SC_CON_SCANCODE_KP_VERTICALBAR = 201,
	_SC_CON_SCANCODE_KP_DBLVERTICALBAR = 202,
	_SC_CON_SCANCODE_KP_COLON = 203,
	_SC_CON_SCANCODE_KP_HASH = 204,
	_SC_CON_SCANCODE_KP_SPACE = 205,
	_SC_CON_SCANCODE_KP_AT = 206,
	_SC_CON_SCANCODE_KP_EXCLAM = 207,
	_SC_CON_SCANCODE_KP_MEMSTORE = 208,
	_SC_CON_SCANCODE_KP_MEMRECALL = 209,
	_SC_CON_SCANCODE_KP_MEMCLEAR = 210,
	_SC_CON_SCANCODE_KP_MEMADD = 211,
	_SC_CON_SCANCODE_KP_MEMSUBTRACT = 212,
	_SC_CON_SCANCODE_KP_MEMMULTIPLY = 213,
	_SC_CON_SCANCODE_KP_MEMDIVIDE = 214,
	_SC_CON_SCANCODE_KP_PLUSMINUS = 215,
	_SC_CON_SCANCODE_KP_CLEAR = 216,
	_SC_CON_SCANCODE_KP_CLEARENTRY = 217,
	_SC_CON_SCANCODE_KP_BINARY = 218,
	_SC_CON_SCANCODE_KP_OCTAL = 219,
	_SC_CON_SCANCODE_KP_DECIMAL = 220,
	_SC_CON_SCANCODE_KP_HEXADECIMAL = 221,
	_SC_CON_SCANCODE_LCTRL = 224,
	_SC_CON_SCANCODE_LSHIFT = 225,
	_SC_CON_SCANCODE_LALT = 226,
	_SC_CON_SCANCODE_LGUI = 227,
	_SC_CON_SCANCODE_RCTRL = 228,
	_SC_CON_SCANCODE_RSHIFT = 229,
	_SC_CON_SCANCODE_RALT = 230,
	_SC_CON_SCANCODE_RGUI = 231,
	_SC_CON_SCANCODE_MODE = 257,
	_SC_CON_SCANCODE_AUDIONEXT = 258,
	_SC_CON_SCANCODE_AUDIOPREV = 259,
	_SC_CON_SCANCODE_AUDIOSTOP = 260,
	_SC_CON_SCANCODE_AUDIOPLAY = 261,
	_SC_CON_SCANCODE_AUDIOMUTE = 262,
	_SC_CON_SCANCODE_MEDIASELECT = 263,
	_SC_CON_SCANCODE_WWW = 264,
	_SC_CON_SCANCODE_MAIL = 265,
	_SC_CON_SCANCODE_CALCULATOR = 266,
	_SC_CON_SCANCODE_COMPUTER = 267,
	_SC_CON_SCANCODE_AC_SEARCH = 268,
	_SC_CON_SCANCODE_AC_HOME = 269,
	_SC_CON_SCANCODE_AC_BACK = 270,
	_SC_CON_SCANCODE_AC_FORWARD = 271,
	_SC_CON_SCANCODE_AC_STOP = 272,
	_SC_CON_SCANCODE_AC_REFRESH = 273,
	_SC_CON_SCANCODE_AC_BOOKMARKS = 274,
	_SC_CON_SCANCODE_BRIGHTNESSDOWN = 275,
	_SC_CON_SCANCODE_BRIGHTNESSUP = 276,
	_SC_CON_SCANCODE_DISPLAYSWITCH = 277,
	_SC_CON_SCANCODE_KBDILLUMTOGGLE = 278,
	_SC_CON_SCANCODE_KBDILLUMDOWN = 279,
	_SC_CON_SCANCODE_KBDILLUMUP = 280,
	_SC_CON_SCANCODE_EJECT = 281,
	_SC_CON_SCANCODE_SLEEP = 282,
	_SC_CON_SCANCODE_APP1 = 283,
	_SC_CON_SCANCODE_APP2 = 284,
	
	_SC_CON_SCANCODE_MAX = 512,
} _sc_con_scancode_t;

//Table of DOS make-codes and the SDL scancodes they correspond to
uint16_t keyb_dos_to_kernel[256] = 
{
	[0x01] = _SC_CON_SCANCODE_ESCAPE,
	[0x02] = _SC_CON_SCANCODE_1,
	[0x03] = _SC_CON_SCANCODE_2,
	[0x04] = _SC_CON_SCANCODE_3,
	[0x05] = _SC_CON_SCANCODE_4,
	[0x06] = _SC_CON_SCANCODE_5,
	[0x07] = _SC_CON_SCANCODE_6,
	[0x08] = _SC_CON_SCANCODE_7,
	[0x09] = _SC_CON_SCANCODE_8,
	[0x0A] = _SC_CON_SCANCODE_9,
	[0x0B] = _SC_CON_SCANCODE_0,
	[0x0C] = _SC_CON_SCANCODE_MINUS,
	[0x0D] = _SC_CON_SCANCODE_EQUALS,
	[0x0E] = _SC_CON_SCANCODE_BACKSPACE,
	[0x0F] = _SC_CON_SCANCODE_TAB,
	[0x10] = _SC_CON_SCANCODE_Q,
	[0x11] = _SC_CON_SCANCODE_W,
	[0x12] = _SC_CON_SCANCODE_E,
	[0x13] = _SC_CON_SCANCODE_R,
	[0x14] = _SC_CON_SCANCODE_T,
	[0x15] = _SC_CON_SCANCODE_Y,
	[0x16] = _SC_CON_SCANCODE_U,
	[0x17] = _SC_CON_SCANCODE_I,
	[0x18] = _SC_CON_SCANCODE_O,
	[0x19] = _SC_CON_SCANCODE_P,
	[0x1A] = _SC_CON_SCANCODE_LEFTBRACKET,
	[0x1B] = _SC_CON_SCANCODE_RIGHTBRACKET,
	[0x1C] = _SC_CON_SCANCODE_RETURN,
	[0x1D] = _SC_CON_SCANCODE_LCTRL,
	[0x1E] = _SC_CON_SCANCODE_A,
	[0x1F] = _SC_CON_SCANCODE_S,
	[0x20] = _SC_CON_SCANCODE_D,
	[0x21] = _SC_CON_SCANCODE_F,
	[0x22] = _SC_CON_SCANCODE_G,
	[0x23] = _SC_CON_SCANCODE_H,
	[0x24] = _SC_CON_SCANCODE_J,
	[0x25] = _SC_CON_SCANCODE_K,
	[0x26] = _SC_CON_SCANCODE_L,
	[0x27] = _SC_CON_SCANCODE_SEMICOLON,
	[0x28] = _SC_CON_SCANCODE_APOSTROPHE,
	[0x29] = _SC_CON_SCANCODE_GRAVE,
	[0x2A] = _SC_CON_SCANCODE_LSHIFT,
	[0x2B] = _SC_CON_SCANCODE_BACKSLASH,
	[0x2C] = _SC_CON_SCANCODE_Z,
	[0x2D] = _SC_CON_SCANCODE_X,
	[0x2E] = _SC_CON_SCANCODE_C,
	[0x2F] = _SC_CON_SCANCODE_V,
	[0x30] = _SC_CON_SCANCODE_B,
	[0x31] = _SC_CON_SCANCODE_N,
	[0x32] = _SC_CON_SCANCODE_M,
	[0x33] = _SC_CON_SCANCODE_COMMA,
	[0x34] = _SC_CON_SCANCODE_PERIOD,
	[0x35] = _SC_CON_SCANCODE_SLASH,
	[0x36] = _SC_CON_SCANCODE_RSHIFT,
	[0x37] = _SC_CON_SCANCODE_PRINTSCREEN,
	[0x38] = _SC_CON_SCANCODE_LALT,
	[0x39] = _SC_CON_SCANCODE_SPACE,
	[0x3A] = _SC_CON_SCANCODE_CAPSLOCK,
	[0x3B] = _SC_CON_SCANCODE_F1,
	[0x3C] = _SC_CON_SCANCODE_F2,
	[0x3D] = _SC_CON_SCANCODE_F3,
	[0x3E] = _SC_CON_SCANCODE_F4,
	[0x3F] = _SC_CON_SCANCODE_F5,
	[0x40] = _SC_CON_SCANCODE_F6,
	[0x41] = _SC_CON_SCANCODE_F7,
	[0x42] = _SC_CON_SCANCODE_F8,
	[0x43] = _SC_CON_SCANCODE_F9,
	[0x44] = _SC_CON_SCANCODE_F10,
	[0x45] = _SC_CON_SCANCODE_NUMLOCKCLEAR,
	[0x46] = _SC_CON_SCANCODE_SCROLLLOCK,
	[0x47] = _SC_CON_SCANCODE_HOME,
	[0x48] = _SC_CON_SCANCODE_UP,
	[0x49] = _SC_CON_SCANCODE_PAGEUP,
	[0x4A] = _SC_CON_SCANCODE_KP_MINUS,
	[0x4B] = _SC_CON_SCANCODE_LEFT,
	[0x4C] = _SC_CON_SCANCODE_KP_5,
	[0x4D] = _SC_CON_SCANCODE_RIGHT,
	[0x4E] = _SC_CON_SCANCODE_KP_PLUS,
	[0x4F] = _SC_CON_SCANCODE_END,
	[0x50] = _SC_CON_SCANCODE_DOWN,
	[0x51] = _SC_CON_SCANCODE_PAGEDOWN,
	[0x52] = _SC_CON_SCANCODE_INSERT,
	[0x53] = _SC_CON_SCANCODE_DELETE,
};

//Previous interrupt handler for int 0x15
static void (__interrupt __far *keyb_chained)(void);

//Our interrupt handler for int 0x15 keyboard interception
static void __interrupt __far __loadds keyb_hook(uint16_t ax)
{	
	//Read the AX register at entry to determine what is happening
	#pragma aux keyb_hook __parm [AX]
	if( (ax >> 8) == 0x4F )
	{
		//Int 0x15 Function 0x4F - Keyboard intercept. Send the scancode to the kernel.
		
		//DOS delivers one-byte scancodes, with the 0x80-bit indicating a break rather than a make
		uint8_t scancode = ax & 0xFF;
		uint8_t pressed = 1;
		if(scancode & 0x80)
		{
			pressed = 0;
			scancode &= 0x7F;
		}
		
		//Translate to a kernel scan-code and state
		uint16_t kernel_code = keyb_dos_to_kernel[scancode];
		if(kernel_code != 0)
		{
			//Events come to kernel big-endian so we know they never start with 0xFF.
			static uint8_t event[4];
			event[0] = 0x10;
			event[1] = 0x00;
			event[2] = (pressed ? 0x80 : 0x00) | (kernel_code >> 8);
			event[3] = (kernel_code & 0xFF);
			ring_write(DRING_CHANNEL_INPUT, event, sizeof(event));
		}
		
		//Consume keyboard input
		return;
	}
	
	//Pass along to existing interrupts (Watcom handles restoring registers)
	_chain_intr(keyb_chained);
}

//atexit handler for unhooking keyboard interception
static void keyb_atexit(void)
{
	//Restore old handler
	_dos_setvect(0x15, keyb_chained);
}

void keyb_init(void)
{
	//Hook the keyboard intercept interrupt - interrupt 0x15 function 0x4F, so we get the stream of scancodes
	keyb_chained = _dos_getvect(0x15);
	_dos_setvect(0x15, &keyb_hook);
	atexit(&keyb_atexit);
}

