//kload.c
//Kernel loading at real-mode
//Bryan E. Topp <betopp@betopp.com> 2022

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>
#include <i86.h>

#include "kload.h"
#include "xmem.h"
#include "pae.h"

//Buffer in conventional memory for loading one page of the kernel file.
static uint8_t kload_page[4096];

void kload(FILE *kfile, uint64_t *pml4)
{
	printf("Loading kernel...\n");
	
	//Load all pages in the kernel, mapping into memory starting at the last gigabyte.
	//Todo - actually find how much space the kernel needs. For now just assume BSS ends at +32MBytes.
	uint64_t map_addr = 0xFFFFFFFFC0000000;
	bool done_loading = false;
	for(uint32_t loaded = 0; loaded < 32ull*1024ull*1024ull; loaded += sizeof(kload_page))
	{
		//Allocate a page of extended memory to hold this page of the kernel
		uint32_t xpage = xmem_palloc();
		
		//Read the next page from the file
		if(!done_loading)
		{
			ssize_t nread = fread(kload_page, 1, sizeof(kload_page), kfile);
			if(nread <= 0)
			{
				done_loading = true;
			}
			else
			{
			
				//Clear the rest of the buffer, if it was only partially filled by the read
				memset(kload_page + nread, 0, sizeof(kload_page) - nread);
			
				//Copy the contents into the page
				xmem_pfill(xpage, kload_page);
			}
		}
		
		//Map it
		pae_map(pml4, map_addr, xpage);
		
		//Keep going
		map_addr += sizeof(kload_page);
	}
	
	//Identity-map the page before the PML4
	uint32_t pml4_phys = FP_OFF(pml4) + ((uint32_t)FP_SEG(pml4) << 4);
	pae_map(pml4, pml4_phys - 4096, pml4_phys - 4096);
	
	//Map physical space - use the second-highest entry in the PML4, putting a PDPT there (512GBytes of space).
	//Fill it with 512 entries corresponding to 1GByte physical pages.
	uint64_t identity_pdpt = xmem_palloc();
	pml4[510] = identity_pdpt | 0x3;
	for(int gigabyte = 0; gigabyte < 512; gigabyte++)
	{
		//Use 0x80 flag - stop page translation at 1GByte granularity
		xmem_write64(identity_pdpt + (8 * gigabyte), 0x83ull | (gigabyte * (1024ull * 1024ull * 1024ull)));
	}
}

