;launch.asm
;Assembly routines for abusing unreal mode to launch application cores
;Bryan E. Topp <betopp@betopp.com> 2022

;This file is part of Stump; Copyright Bryan E. Topp, 2023.
;Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
;Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
;You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

.386P

_TEXT16 SEGMENT BYTE PUBLIC USE16 'CODE'

EXTERN puts_ : proc
EXTERN exit_ : proc

public launch_init_
launch_init_ proc

	;Check if we're actually in real-mode
	mov EAX, CR0
	test EAX, 1
	jnz not_real

	;Correct our GDTR to point to the linear address of the GDT
	mov EAX, 0
	mov AX, CS
	shl EAX, 4
	add EAX, [dword ptr offset launch_gdtr + 2]
	mov [dword ptr offset launch_gdtr + 2], EAX
	
	;OK
	mov AX, 0
	ret
	
	not_real:
	
	;Failure
	mov AX, 1
	ret
	
launch_init_ endp


launch_unreal_ proc

	;Enter protected mode with our little GDT
	lgdt fword ptr launch_gdtr
	mov EAX, CR0
	or EAX, 1
	mov CR0, EAX
	
	;Load the 32-bit-long flat data segment in GS
	mov AX, 8
	mov GS, AX
	
	;Drop back to real mode, retaining our "unreal" GS
	mov EAX, CR0
	and EAX, 0xFFFFFFFE
	mov CR0, EAX
	
	ret
launch_unreal_ endp
	

public launch_go_
launch_go_ proc

	;Our one parameter is in AL. This is the address to start-up the other cores.
	;Set it aside in DL.
	mov DL, AL
	
	;Set up an unreal-mode G segment (Can't handle interrupts while doing this)
	pushf
	cli
	call launch_unreal_
	
	;Using the unreal GS, poke at the Local APIC to release the other cores...
	mov EAX, 0xFEE00000
	
	;Clear APIC errors
	mov [GS : dword ptr EAX + 0x280], 0
	
	;Send INIT IPI to all remote cores
	mov [GS : dword ptr EAX + 0x300], 0xC4500
	
	init_wait:
	mov ECX, [GS : dword ptr EAX + 0x300]
	test ECX, 0x1000
	jnz init_wait
	
	;Send a STARTUP IPI to all remote cores
	and EDX, 0xFF
	add EDX, 0xC4600
	mov [GS : dword ptr EAX + 0x300], EDX
	
	startup_wait:
	mov ECX, [GS : dword ptr EAX + 0x300]
	test ECX, 0x1000
	jnz startup_wait
	
	;Wait a little moment and send another STARTUP IPI
	mov ECX, 65536
	idle_wait:
		loop idle_wait

	mov [GS : dword ptr EAX + 0x300], EDX
	
	startup2_wait:
	mov ECX, [GS : dword ptr EAX + 0x300]
	test ECX, 0x1000
	jnz startup2_wait
	
	popf
	ret
launch_go_ endp
	
public launch_stop_
launch_stop_ proc

	pushf
	cli
	call launch_unreal_
	
	;Send INIT with LAPIC
	mov EAX, 0xFEE00000 
	mov [GS : dword ptr EAX + 0x280], 0 ;Clear APIC errors
	mov [GS : dword ptr EAX + 0x300], 0xC4500 ;Send INIT IPI to all remote cores
		
	popf
	ret
	
launch_stop_ endp

public launch_ipi_
launch_ipi_ proc

	;Set aside parameter
	mov DX, AX

	pushf
	cli
	call launch_unreal_
	
	;Retrieve parameter in low byte of EDX, clearing the rest of it
	and EDX, 0xFF
	
	;Send IPI with LAPIC
	mov EAX, 0xFEE00000 ;Local APIC base address
	mov [GS : dword ptr EAX + 0x280], 0 ;Clear APIC errors
	or EDX, 0xC4000 ;Fixed interrupt, positive edge-trigger, to all-except-self - interrupt number in low byte already
	mov [GS : dword ptr EAX + 0x300], EDX ;Send interrupt
	
	popf
	ret

launch_ipi_ endp

;Global Descriptor Table used in entering unreal mode
launch_gdt:
	dq 0 ;Null descriptor
	db 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x92, 0xCF, 0x00 ;Ring-0 32-bit 4GByte data
	
;Contents of Global Descriptor Table Register to point at GDT used for unreal mode
launch_gdtr:
	dw 15 ;Size-1
	dd offset launch_gdt ;Base linear address (needs to be corrected at runtime)

_TEXT16 ENDS


END

