//pae.c
//Functions for using PAE-format 4-level pagetables
//Bryan E. Topp <betopp@betopp.com> 2022

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "pae.h"
#include "xmem.h"
#include <assert.h>
#include <stdio.h>

void pae_map(uint64_t *pml4, uint64_t vaddr, uint64_t paddr)
{	
	//printf("Map %llX = %llX\n", vaddr, paddr);
	
	//Build indexes into each paging structure
	const uint64_t pml4_idx = (vaddr >> 39) % 512;
	const uint64_t pdpt_idx = (vaddr >> 30) % 512;
	const uint64_t pd_idx   = (vaddr >> 21) % 512;
	const uint64_t pt_idx   = (vaddr >> 12) % 512;
	
	//Read PML4 - make sure the PDPT is present. (Can do this directly, as the PML4 is in conventional memory.)
	uint64_t pml4_entry = pml4[pml4_idx];
	if(!(pml4_entry & 1))
	{
		//No entry in PML4. Need to allocate a PDPT and add reference to PML4.
		assert(pml4_entry == 0);
		uint32_t new_pdpt = xmem_palloc();
		pml4_entry = new_pdpt | 0x3;
		pml4[pml4_idx] = pml4_entry;
	}
	
	//Read PDPT - make sure the PD is present. Need to go through XMS to read the memory.
	uint64_t pdpt_addr = pml4_entry & 0x0FFFFFFFFFFFF000;
	uint64_t pdpt_entry = xmem_read64(pdpt_addr + (8 * pdpt_idx));
	if(!(pdpt_entry & 1))
	{
		//No entry in PDPT. Need to allocate a PD and add reference to PDPT.
		assert(pdpt_entry == 0);
		uint32_t new_pd = xmem_palloc();
		pdpt_entry = new_pd | 0x3;
		xmem_write64(pdpt_addr + (8 * pdpt_idx), pdpt_entry);
	}
	
	//Read PD - make sure the PT is present. Need to go through XMS to read the memory.
	uint64_t pd_addr = pdpt_entry & 0x0FFFFFFFFFFFF000;
	uint64_t pd_entry = xmem_read64(pd_addr + (8 * pd_idx));
	if(!(pd_entry & 1))
	{
		//No entry in PD. Need to allocate a PT and add reference to PD.
		assert(pd_entry == 0);
		uint32_t new_pt = xmem_palloc();
		pd_entry = new_pt | 0x3;
		xmem_write64(pd_addr + (8 * pd_idx), pd_entry);
	}
	
	//Read PT, make sure there's no page there, and map the page
	uint64_t pt_addr = pd_entry & 0x0FFFFFFFFFFFF000;
	uint64_t pt_entry = xmem_read64(pt_addr + (8 * pt_idx));
	assert(pt_entry == 0);
	pt_entry = paddr | 0x3;
	xmem_write64(pt_addr + (8 * pt_idx), pt_entry);
}
