//real.c
//Real-mode loop for using DOS drivers in Stump
//Bryan E. Topp <betopp@betopp.com> 2022

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <assert.h>
#include <i86.h>

#include "cpuchk.h"
#include "brk.h"
#include "keyb.h"
#include "mouse.h"
#include "vbe.h"
#include "xmem.h"
#include "dmem.h"
#include "kload.h"
#include "ring.h"
#include "launch.h"


//Ought to be set up in the Makefile, but don't explode otherwise
#ifndef BUILDVERSION
	#define BUILDVERSION "unknown"
#endif
#ifndef BUILDDATE
	#define BUILDDATE "unknown"
#endif
#ifndef BUILDUSER
	#define BUILDUSER "unknown"
#endif

//Memory for application-core startup and PML4.
//Allow one page extra to deal with misalignment, depending on how DOS loads us.
//(Application cores can be started at any page-aligned address in the first megabyte.)
uint8_t startup_buffer[4096 * 3];

//Memory for storing panic strings from the kernel.
uint8_t panic_buffer[4096];

//Message printed when exiting, after video is cleaned up.
void atexit_message(void)
{
	if(panic_buffer[0] != '\0')
	{
		printf("!!! Kernel panic: %s\n", panic_buffer);
	}
	else
	{
		printf("Exiting.\n");
	}
	printf("Real-mode driver loop for Stump / Bryan E. Topp <betopp@betopp.com> 2022\n");
	printf("Version " BUILDVERSION " at " BUILDDATE " by " BUILDUSER "\n");
}

int main(int argc, const char **argv)
{	
	printf("Real-mode driver loop for Stump / Bryan E. Topp <betopp@betopp.com> 2022\n");
	printf("Version " BUILDVERSION " at " BUILDDATE " by " BUILDUSER "\n");
	atexit(atexit_message);
	
	//Check that we have an appropriate CPU for this lunacy
	cpuchk();
	if(launch_init() != 0)
	{
		printf("CPU not running in real mode.\n");
		printf("Physical memory and GDT access are required to launch the kernel.\n");
		exit(-1);
	}
	
	//Set up DOS-side systems
	brk_init();
	dmem_init();
	xmem_init();
	keyb_init();
	mouse_init();
	vbe_init();
	
	//Open this EXE file to read the kernel
	if(argc < 1 || strlen(argv[0]) < 1)
	{
		printf("Failed to determine the name of this executable.\n");
		exit(-1);
	}
	FILE *kfile = fopen(argv[0], "rb");
	if(kfile == NULL)
	{
		printf("Failed to open %s to read kernel.\n", argv[0]);
		exit(-1);
	}
	
	//Look through this EXE to find the start of the kernel.
	//Assume the kernel is page-aligned in the file.
	//Assume it starts in the first megabyte (i.e., assume this loader is <1MByte).
	//We prepend a particular page of noise to just before the kernel. Look for its sum.
	uint32_t kernel_offset = 0;
	for(uint32_t search_offset = 0; search_offset < 1048576; search_offset += 4096)
	{
		//Read a page
		fseek(kfile, search_offset, SEEK_SET);
		if(fread(startup_buffer, 1, sizeof(startup_buffer), kfile) != sizeof(startup_buffer))
			break;
		
		//Sum it as a series of 32-bit values
		uint32_t check = 0;
		uint32_t *values = (uint32_t*)startup_buffer;
		for(int vv = 0; vv < (4096 / sizeof(uint32_t)); vv++)
		{
			check += values[vv];
		}
		
		//See if it matches the magic number
		if(check == 0x0AB84AD8)
		{
			//Found it
			kernel_offset = search_offset + 4096;
			break;
		}
	}
	
	if(kernel_offset == 0)
	{
		printf("Failed to find kernel ELF following this EXE (%s).\n", argv[0]);
		exit(-1);
	}
	
	printf("Found kernel at offset %lXh in this EXE.\n", kernel_offset);
	memset(startup_buffer, 0, sizeof(startup_buffer));

	//Load the first page of the kernel at page-alignment in conventional memory.
	//We already have a buffer - just pick a point in that buffer which is page-aligned.
	uint32_t startup_buffer_phys = FP_LINEAR(startup_buffer);
	uint32_t startup_buffer_misalign = startup_buffer_phys % 4096;
	uint32_t startup_code_phys = startup_buffer_phys + 4096 - startup_buffer_misalign;
	uint8_t *startup_code_ptr = &(startup_buffer[4096 - startup_buffer_misalign]);
	printf("Dropping startup code at %lXh.\n", startup_code_phys);
	
	fseek(kfile, kernel_offset, SEEK_SET);
	if(fread(startup_code_ptr, 1, 4096, kfile) != 4096)
	{
		printf("Failed to read kernel startup page.\n");
		exit(-1);
	}
	
	//Load the entirety of the kernel into extended memory.
	//Make paging structures for it, with its PML4 starting just after the startup code in conventional memory.
	uint64_t *pml4 = (uint64_t*)(startup_code_ptr + 4096);
	fseek(kfile, kernel_offset, SEEK_SET);
	kload(kfile, pml4);
	
	//Done loading kernel data
	fclose(kfile); kfile = NULL;
	
	//Allocate shared memory buffers for communication with kernel.
	//Put them physically in conventional memory; map them virtually into the kernel's memory space.
	ring_init(pml4);
	
	//Hand-off the rest of our memory to the kernel using the shared memory for communication.
	xmem_handoff();
	vbe_handoff();
	
	//The kernel is now set-up and ready to go in memory.
	
	//Diddle the LAPIC in unreal mode to get the other cores running the kernel.
	launch_go(startup_code_phys / 4096);
	
	//Run idle loop on this core to update polled features
	printf("Entering main loop...\n");
	while(1)
	{
		//Check for kernel panic
		if(ring_read(DRING_CHANNEL_PANIC, panic_buffer, sizeof(panic_buffer)) > 0)
		{
			//Panic buffer is filled with message now - just exit and we'll display it from atexit.
			break;
		}
		
		//Check for change in video buffer to display
		uint32_t vpage = 0;
		if(ring_read(DRING_CHANNEL_VMODE, &vpage, sizeof(vpage)) > 0)
		{
			//Call VBE to update the video card
			vbe_flip(vpage);
			
			//Update the kernel on the video buffer now displayed
			ring_write(DRING_CHANNEL_VCURR, &vpage, sizeof(vpage));
		}
		
		//Drain DOS-side keyboard buffer
		while(kbhit())
		{
			getch();
		}
		
	}
	
	//When main loop exits, kill other cores with an INIT, but no STARTUP, IPI.
	launch_stop();
	
	//Todo - there's not really a way to be sure they're stopped, is there?
	
	//Clean up every other DOS thing we hooked/allocated/otherwise messed with.
	exit(0);
}
