//ring.c
//Ringbuffers for DOS loader communication with kernel
//Bryan E. Topp <betopp@betopp.com> 2022

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "ring.h"
#include "pae.h"
#include "dmem.h"
#include "launch.h"
#include <stdint.h>
#include <string.h>
#include <i86.h>
#include <stdio.h>

//Second-highest gigabyte is where ringbuffers are mapped
#define RING_BASE (0xFFFFFFFF80000000ull)

//Each ringbuffer has its own PAE pagetable - 2MBytes of address space
#define RING_OFFS (0x0000000000200000ull)

//Pointer to info in first page
typedef struct dring_info_s
{
	uint32_t size;
	uint32_t rptr;
	uint32_t wptr;
	uint32_t rsvd;
} ring_info_t;
static ring_info_t far *ring_info_ptr;

//Location of each ring's data in conventional memory
static void far *ring_data_ptr[DRING_CHANNEL_MAX];

void ring_init(uint64_t *pml4)
{
	printf("Allocate communication buffers...\n");
	
	//Allocate the ring info page, and map it into the beginning of the kernel's ring-buffer space.
	ring_info_ptr = dmem_alloc(4096);
	pae_map(pml4, RING_BASE, FP_LINEAR(ring_info_ptr));
	
	//Allocate buffers for each ring
	for(uint32_t rr = 1; rr < DRING_CHANNEL_MAX; rr++)
	{
		uint32_t ring_size = 4096;
		ring_data_ptr[rr] = dmem_alloc(ring_size);
		for(uint32_t pageoff = 0; pageoff < ring_size; pageoff += 4096)
		{
			pae_map(pml4, RING_BASE + (rr * RING_OFFS) + pageoff, FP_LINEAR(ring_data_ptr[rr]) + pageoff);
			ring_info_ptr[rr].size += 4096;
		}
	}
}

int ring_write(int channel, const void *data_ptr, int data_len)
{
	if(ring_info_ptr == NULL)
		return 0;
	
	//Info and data located by channel index
	ring_info_t far *infoptr = &(ring_info_ptr[channel]);
	uint8_t far *ringptr = ring_data_ptr[channel];
	
	//Copy into the ringbuffer - possibly twice, if we wrap around the buffer
	const uint8_t *databytes = (const uint8_t *)data_ptr;
	int retval = 0;
	while(data_len > 0)
	{
		//Figure out how much we'll copy in one go. At most, all the data.
		int to_copy = data_len;
		
		//Limit so we don't run off the end of the ringbuffer.
		int to_buf_end = infoptr->size - infoptr->wptr;
		if(to_copy > to_buf_end)
			to_copy = to_buf_end;
		
		//Limit so we don't smash data that's already in the buffer.
		int to_free_end = (infoptr->rptr - 1) - infoptr->wptr;
		if(to_free_end >= 0 && to_copy > to_free_end)
			to_copy = to_free_end;
	
		//If there's just no space, stop trying
		if(to_copy == 0)
			break;
		
		//Do the copy
		_fmemcpy(ringptr + infoptr->wptr, databytes, to_copy);
		
		//Advance src/dst pointers
		databytes += to_copy;
		infoptr->wptr += to_copy;
		if(infoptr->wptr >= infoptr->size)
			infoptr->wptr -= infoptr->size;
		
		//Advance counts
		retval += to_copy;
		data_len -= to_copy;
		
		//Wake kernel if it's halted
		launch_ipi(0x30 + channel);
	}
	return retval;
}

int ring_read(int channel, void *buf_ptr, int buf_len)
{
	if(ring_info_ptr == NULL)
		return 0;
	
	//Info and data located by channel index
	ring_info_t far *infoptr = &(ring_info_ptr[channel]);
	const uint8_t far *ringptr = ring_data_ptr[channel];
	
	//Copy out of the ringbuffer - possibly twice, if we wrap around the buffer
	uint8_t *bufbytes = (uint8_t *)buf_ptr;
	int retval = 0;
	while(buf_len > 0)
	{
		//Figure out how much we'll copy in one go. At most, filling the given buffer.
		int to_copy = buf_len;
		
		//Limit so we don't run off the end of the ringbuffer.
		int to_buf_end = infoptr->size - infoptr->rptr;
		if(to_copy > to_buf_end)
			to_copy = to_buf_end;
		
		//Limit so we don't read empty space in the ringbuffer.
		int to_used_end = infoptr->wptr - infoptr->rptr;
		if(to_used_end >= 0 && to_copy > to_used_end)
			to_copy = to_used_end;
	
		//If there's just no data, stop trying
		if(to_copy == 0)
			break;
		
		//Do the copy
		_fmemcpy(bufbytes, ringptr + infoptr->rptr, to_copy);
		
		//Advance dst/src pointers
		bufbytes += to_copy;
		infoptr->rptr += to_copy;
		if(infoptr->rptr >= infoptr->size)
			infoptr->rptr -= infoptr->size;
		
		//Advance counts
		retval += to_copy;
		buf_len -= to_copy;
		
		//Wake kernel if it's halted
		launch_ipi(0x30 + channel);
	}
	return retval;	
}


