//vbe.c
//Video calls for real-mode loop
//Bryan E. Topp <betopp@betopp.com> 2022

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "vbe.h"
#include "ring.h"
#include <i86.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

//Information about a VBE video mode that is supported
typedef struct vbe_mode_s
{
	uint16_t num; //Number of the mode, to reference when setting modes
	uint32_t paddr; //Physical base address of framebuffer
	uint16_t width; //Active width in pixels
	uint16_t height; //Active height in lines
	uint16_t stride; //Bytes of memory from one line to the next
} vbe_mode_t;
#define VBE_MODE_MAX 128
static vbe_mode_t vbe_mode_table[VBE_MODE_MAX];

//Reads a "vbeFarPtr", assuming it's in segment/offset form.
static void far *vbe_read_farptr(void *farptr)
{
	uint16_t *offptr = (uint16_t*)(farptr);
	uint16_t *segptr = offptr+1;
	return MK_FP(*segptr, *offptr);
}

//Original video mode that was active when we started
static uint16_t vbe_origmode;

//Best mode that was chosen on startup
static vbe_mode_t *vbe_bestmode;

//Total video memory detected on startup;
static uint32_t vbe_vmem;

//Desired video modes to look for
static const uint16_t vbe_desired[][2] = 
{
	//For testing just grab the little resolution first... //todo
	{800, 600},
	
	{1920, 1080},
	{1280, 720},
	{1024, 768},
	{960, 540},
	{800, 600},
	{640, 480},
	{0}
};

//Restores the video mode that we started with
static void vbe_restore(void)
{
	//Switch back to the original video mode we entered at
	union REGS regs = {0};
	struct SREGS sregs = {0};
	segread(&sregs);
	regs.w.ax = 0x4f02;
	regs.w.bx = vbe_origmode;
	int86x(0x10, &regs, &regs, &sregs);
	//Todo - how do we preserve error output while doing this?
	printf("Returned to original video mode %Xh.\n", vbe_origmode);
}

void vbe_init(void)
{	
	//Temp buffer for VBE info
	char vbebuf[512];
	
	//Get controller information from VBE
	union REGS regs = {0};
	struct SREGS sregs = {0};
	segread(&sregs);
	regs.w.ax = 0x4F00;
	regs.w.di = FP_OFF(vbebuf);
	sregs.es = FP_SEG(vbebuf);
	int86x(0x10, &regs, &regs, &sregs);
	if(regs.w.ax != 0x004F)
	{
		printf("Failed to get VBE video adapter info.\n");
		exit(-1);
	}
	
	if(memcmp(vbebuf, "VESA", 4))
	{
		printf("VBE returned invalid signature.\n");
		exit(-1);
	}
	
	printf("VBE provider: %Fs\n", vbe_read_farptr(vbebuf +  6));
	printf("VBE version: %X.%2.2X, video memory: %lu KBytes\n", *(vbebuf + 5), *(vbebuf + 4), 64ull * *((uint16_t*)(vbebuf + 18)));
	
	//Set aside total available memory, as bytes.
	vbe_vmem = 1024ull * 64ull * *((uint16_t*)(vbebuf + 18));
	
	//Query the current video mode and set it aside.
	segread(&sregs);
	memset(&regs, 0, sizeof(regs));
	regs.w.ax = 0x4f03;
	int86x(0x10, &regs, &regs, &sregs);
	vbe_origmode = regs.w.bx;
	//printf("Coming from video mode %d\n", vbe_origmode);
	
	//Restore that original video mode on exit
	atexit(&vbe_restore);
	
	//Copy the list of video mode numbers out of the temp buffer
	uint16_t far *vmodelist = vbe_read_farptr(vbebuf + 14);
	for(int vv = 0; vmodelist[vv] != 0xFFFF && vv < VBE_MODE_MAX; vv++)
	{
		vbe_mode_table[vv].num = vmodelist[vv];
	}
	
	//Query for further information about each of the video modes returned.
	for(int vv = 0; vv < VBE_MODE_MAX; vv++)
	{
		if(vbe_mode_table[vv].num == 0)
			continue;
		
		//Query VBE about this mode
		memset(&regs, 0, sizeof(regs));
		regs.w.ax = 0x4F01;
		regs.w.cx = vbe_mode_table[vv].num;
		regs.w.di = FP_OFF(vbebuf);
		sregs.es = FP_SEG(vbebuf);
		int86x(0x10, &regs, &regs, &sregs);
		if( regs.w.ax != 0x004F)
		{
			printf("Failed to query info about video mode %4.4Xh.\n", vbe_mode_table[vv]);
			exit(-1);
		}
		
		//If the mode isn't usable, don't store its info
		uint16_t attributes = *((uint16_t*)(vbebuf + 0));
		if(!(attributes & 0x0001))
			continue; //Not supported in hardware
		if(!(attributes & 0x0010))
			continue; //Not a graphics mode
		if(!(attributes & 0x0080))
			continue; //No Linear Framebuffer available
		
		uint8_t r_bits = *((uint16_t*)(vbebuf + 31));
		uint8_t g_bits = *((uint16_t*)(vbebuf + 33));
		uint8_t b_bits = *((uint16_t*)(vbebuf + 35));
		if(r_bits != 5 || g_bits != 6 || b_bits != 5)
			continue; //Not RGB565
		
		//Mode looks OK for us, set it aside.
		vbe_mode_table[vv].width = *((uint16_t*)(vbebuf + 18));
		vbe_mode_table[vv].height = *((uint16_t*)(vbebuf + 20));
		vbe_mode_table[vv].paddr = *((uint32_t*)(vbebuf + 40));
		vbe_mode_table[vv].stride = *((uint16_t*)(vbebuf + 16));
		//printf("Mode %4.4Xh: %dx%d (@%lXh,+%Xh)\n", 
		//	vbe_mode_table[vv].num,
		//	vbe_mode_table[vv].width, vbe_mode_table[vv].height, 
		//	vbe_mode_table[vv].paddr, vbe_mode_table[vv].stride);
	}
	
	//Pick the best mode to run in.
	//Look down our list of desired modes, and pick the first we find.
	uint16_t bestmode = 0;
	for(int dd = 0; vbe_desired[dd][0] != 0; dd++)
	{
		for(int vv = 0; vv < VBE_MODE_MAX; vv++)
		{
			if(vbe_mode_table[vv].width != vbe_desired[dd][0])
				continue;
			if(vbe_mode_table[vv].height != vbe_desired[dd][1])
				continue;
			
			//Found it
			printf("Found %dx%d video mode %Xh.\n", vbe_mode_table[vv].width, vbe_mode_table[vv].height, vbe_mode_table[vv].num);
			bestmode = vbe_mode_table[vv].num;
			vbe_bestmode = &(vbe_mode_table[vv]);
			break;
		}
		
		if(bestmode != 0)
			break;
	}
	
	if(bestmode == 0)
	{
		printf("Didn't find an appropriate video mode to use.\n");
		exit(-1);
	}
}

void vbe_handoff()
{
	//Write mode info
	const uint32_t width = vbe_bestmode->width;
	const uint32_t height = vbe_bestmode->height;
	const uint32_t stride = vbe_bestmode->stride;
	const uint32_t pxfmt = 2;
	ring_write(DRING_CHANNEL_VMEMS, &width, sizeof(width));
	ring_write(DRING_CHANNEL_VMEMS, &height, sizeof(height));
	ring_write(DRING_CHANNEL_VMEMS, &stride, sizeof(stride));
	ring_write(DRING_CHANNEL_VMEMS, &pxfmt, sizeof(pxfmt));
	
	//Figure out how many independent framebuffers we could allocate in that mode.
	uint32_t videopage = (uint32_t)(vbe_bestmode->height) * (uint32_t)(vbe_bestmode->stride);
	if(videopage % 4096)
	{
		//Make each framebuffer page, an integer number of MMU pages
		videopage += 4096 - (videopage % 4096);
	}
	uint32_t fbs = vbe_vmem / videopage;
	//printf("Have video memory for %lu framebuffer pages.\n", fbs);
	if(fbs < 4)
	{
		printf("Insufficient video memory (for at least 4 pages).\n");
		exit(-1);
	}
	
	for(uint32_t ff = 0; ff < fbs; ff++)
	{
		uint32_t vmem_base = vbe_bestmode->paddr + (videopage * ff);
		uint32_t vmem_size = videopage;
		//printf("Framebuffer %lu at %lXh.\n", ff, vmem_base);
		ring_write(DRING_CHANNEL_VMEMS, &vmem_base, sizeof(vmem_base));
		ring_write(DRING_CHANNEL_VMEMS, &vmem_size, sizeof(vmem_size));
	}

	//Switch into that mode
	union REGS regs = {0};
	struct SREGS sregs = {0};
	segread(&sregs);
	regs.w.ax = 0x4f02;
	regs.w.bx = vbe_bestmode->num | 0x4000; //Bit 14 set = use linear framebuffer
	int86x(0x10, &regs, &regs, &sregs);
	if(regs.w.ax != 0x004F)
	{
		printf("Failed to switch into desired video mode.\n");
		exit(-1);
	}
	
	printf("Switched to %dx%d VBE video mode %Xh.\n", vbe_bestmode->width, vbe_bestmode->height, vbe_bestmode->num);
	
}

void vbe_flip(uint32_t paddr)
{
	//Compute buffer position, as lines into video memory
	uint32_t offset = paddr - vbe_bestmode->paddr;
	uint32_t linescroll = offset / (vbe_bestmode->stride);
	uint32_t pxscroll = (offset % (vbe_bestmode->stride)) / 2;
	
	//Scroll to that position
	union REGS regs = {0};
	struct SREGS sregs = {0};
	segread(&sregs);
	regs.w.ax = 0x4f07;
	regs.w.bx = 0x0000; //Set display start
	regs.w.cx = pxscroll;
	regs.w.dx = linescroll;
	int86x(0x10, &regs, &regs, &sregs);
	if(regs.w.ax != 0x004F)
	{
		printf("Failed to flip video pages.\n");
		exit(-1);
	}
}
