//xmem.c
//Extended-memory allocation for Stump real-mode loader
//Bryan E. Topp <betopp@betopp.com> 2022

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "xmem.h"
#include "ring.h"
#include <stdint.h>
#include <i86.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

//Ranges of memory allocated from XMS
typedef struct xmem_pool_s
{
	//Allocation from XMS
	uint16_t handle;
	uint32_t paddr_orig;
	uint32_t bytes_orig;
	
	//Current remaining memory
	uint32_t paddr_left;
	uint32_t bytes_left;
} xmem_pool_t;
#define XMEM_POOL_MAX 16
xmem_pool_t xmem_pool_table[XMEM_POOL_MAX];

//Far-call entry point for XMS services
void far * xmem_xmsentry;

//Returns XMS version
uint16_t xmem_xms_version(void);
#pragma aux xmem_xms_version = \
	"mov AX, 0" \
	"callf [xmem_xmsentry]" \
	value [AX] \
	modify [AX BX CX DX]

//Returns largest free XMS block
uint32_t xmem_xms_largest_kb(void);
#pragma aux xmem_xms_largest_kb = \
	"mov AX, 0x8800" \
	"callf [xmem_xmsentry]" \
	"mov EDX, EAX" \
	"shr EDX, 16" \
	value [DX AX] \
	modify [AX BX CX DX]

//Allocates XMS memory
uint16_t xmem_xms_alloc(uint32_t kb);
#pragma aux xmem_xms_alloc = \
	"shl ECX, 16" \
	"and EDX, 0xFFFF" \
	"or ECX, EDX" \
	"xchg ECX, EDX" \
	"mov AX, 0x8900" \
	"callf [xmem_xmsentry]" \
	"cmp AX, 1" \
	"je okay" \
	"mov DX, 0" \
	"okay:" \
	parm [DX CX] \
	value [DX] \
	modify [AX BX CX DX]

//Frees XMS memory
void xmem_xms_free(uint16_t handle);
#pragma aux xmem_xms_free = \
	"mov AX, 0x0A00" \
	"callf [xmem_xmsentry]" \
	parm [DX] \
	modify [AX BX CX DX]
	
//Locks XMS memory and returns its physical address
uint32_t xmem_xms_lock(uint16_t handle);
#pragma aux xmem_xms_lock = \
	"mov AX, 0x0C00" \
	"callf [xmem_xmsentry]" \
	"cmp AX, 1" \
	"je okay" \
	"mov DX, 0" \
	"mov BX, 0" \
	"okay:" \
	parm [DX] \
	value [DX BX] \
	modify [AX BX CX DX]
	
//Unlocks XMS memory
void xmem_xms_unlock(uint16_t handle);
#pragma aux xmem_xms_unlock = \
	"mov AX, 0x0D00" \
	"callf [xmem_xmsentry]" \
	parm [DX] \
	modify [AX BX CX DX]
	
//Structure given to XMS to perform memory moves (XMS 3.0 spec)
#pragma pack ( push, 1 )
typedef struct xmem_xms_move_s
{
	uint32_t nbytes;
	
	uint16_t srchandle;
	uint32_t srcoffset;
	
	uint16_t dsthandle;
	uint32_t dstoffset;
	
} xmem_xms_move_t;
#pragma pack ( pop )
	
//Copies data between XMS-accessible memory regions. Returns 1 on success, 0 on failure.
uint16_t xmem_xms_move(const xmem_xms_move_t *info);
#pragma aux xmem_xms_move = \
	"mov AX, 0x0B00" \
	"callf [xmem_xmsentry]" \
	parm [SI] \
	value [AX] \
	modify [AX BX CX DX SI]

//Frees all allocated XMS memory on exit
static void xmem_freeall(void)
{
	for(int pp = 0; pp < XMEM_POOL_MAX; pp++)
	{	
		if(xmem_pool_table[pp].handle != 0)
		{
			xmem_xms_unlock(xmem_pool_table[pp].handle);
			xmem_xms_free(xmem_pool_table[pp].handle);
		}
	}
}

//Page of 0 for clearing XMS pages
static const uint8_t xmem_zeroes[4096] = {0};

//Cache of 64-bit values read/written using xmem_read64/xmem_write64.
//We use these calls to set up pagetables, and so we end up walking the same tables over and over.
//We therefore cache a single 64-bit word from each page so written.
//Caching like this allows for vastly faster paging setup.
typedef struct xmem_writecache_s
{
	uint32_t addr;
	uint64_t data;
} xmem_writecache_t;
#define XMEM_WRITECACHE_MAX 256
static xmem_writecache_t xmem_writecache_table[XMEM_WRITECACHE_MAX];

void xmem_init(void)
{	
	//Set up atexit() function to free anything we allocate
	atexit(&xmem_freeall);
	
	//Check if XMS is installed
	union REGS regs = {0};
	struct SREGS sregs = {0};
	segread(&sregs);
	regs.w.ax = 0x4300;
	int86x(0x2F, &regs, &regs, &sregs);
	if(regs.h.al != 0x80)
	{
		printf("XMS is not available.\n");
		exit(-1);
	}
	//printf("XMS available.\n");
	
	//Get XMS entry point
	segread(&sregs);
	regs.w.ax = 0x4310;
	int86x(0x2F, &regs, &regs, &sregs);
	xmem_xmsentry = MK_FP(sregs.es, regs.w.bx);
	//printf("XMS entry: %4.4X : %4.4X\n", sregs.es, regs.w.bx);
	
	//Check XMS version number and free memory
	uint16_t ver = xmem_xms_version();
	uint32_t largest_free_kb = xmem_xms_largest_kb();
	printf("XMS version: %X.%2.2X, largest free: %ld KBytes\n", (ver >> 8) & 0xFF, (ver >> 0) & 0xFF, largest_free_kb);

	//Allocate 15/16 of it
	uint32_t to_alloc_kb = largest_free_kb - (largest_free_kb / 16);
	uint16_t handle = xmem_xms_alloc(to_alloc_kb);
	to_alloc_kb -= to_alloc_kb % 4;
	if(handle == 0)
	{
		printf("Failed to allocate %ld KBytes via XMS.\n", to_alloc_kb);
		exit(-1);
	}
	xmem_pool_table[0].handle = handle;
	xmem_pool_table[0].bytes_orig = to_alloc_kb * 1024;
	xmem_pool_table[0].bytes_left = xmem_pool_table[0].bytes_orig;
	
	//Lock the memory so we know its physical address
	uint32_t paddr = xmem_xms_lock(handle);
	if(paddr == 0)
	{
		printf("Failed to lock XMS memory for physical access.\n");
		exit(-1);
	}
	xmem_pool_table[0].paddr_orig = paddr;
	xmem_pool_table[0].paddr_left = xmem_pool_table[0].paddr_orig;
	
	printf("XMS allocation: %ld KBytes locked at %lXh (handle %Xh)\n", to_alloc_kb, paddr, handle);	
}

uint32_t xmem_palloc(void)
{
	//Look for a pool with free memory
	for(int pp = 0; pp < XMEM_POOL_MAX; pp++)
	{
		xmem_pool_t *pptr = &(xmem_pool_table[pp]);
		if(pptr->paddr_left % 4096)
		{
			//Pool is misaligned. Discard bytes to bring us to page alignment.
			uint32_t misalign = pptr->paddr_left % 4096;
			uint32_t discard = 4096 - misalign;
			if(discard < pptr->bytes_left)
			{
				pptr->paddr_left += discard;
				pptr->bytes_left -= discard;
			}
			else
			{
				pptr->paddr_left = 0;
				pptr->bytes_left = 0;
			}
		}
		
		if(pptr->bytes_left % 4096)
		{
			//Pool is of non-page-length. Cut short to bring us to page alignment.
			pptr->bytes_left -= (pptr->bytes_left) % 4096;
		}
		
		if(pptr->bytes_left < 4096)
		{
			//Less than one page remains
			continue;
		}
		
		//Okay, the pool is page-aligned and has at least one page in it.
		//Take the first page away and return it.
		uint32_t retval = pptr->paddr_left;
		pptr->paddr_left += 4096;
		pptr->bytes_left -= 4096;
		
		//Fill it with zeroes first
		xmem_pfill(retval, xmem_zeroes);
		
		return retval;
	}
	
	//No pools have pages left.
	printf("Ran out of memory while setting up kernel.\n");
	exit(-1);
}

void xmem_pfill(uint32_t page, const void * far data)
{
	//printf("copy to %lXh from %Fp\n", page, data);
	
	//Should never have a misaligned address
	assert( (page % 4096) == 0 );
	
	//Invalidate write-caches dealing with this page
	uint16_t cacheentry = (page / 4096) % XMEM_WRITECACHE_MAX;
	xmem_writecache_table[cacheentry].addr = 0;
	xmem_writecache_table[cacheentry].data = 0;
	
	//Find which pool this page came from
	for(int pp = 0; pp < XMEM_POOL_MAX; pp++)
	{
		xmem_pool_t *pptr = &(xmem_pool_table[pp]);
		if(page < pptr->paddr_orig)
			continue;
		if(page + 4096 > (pptr->paddr_orig + pptr->bytes_orig))
			continue;
		
		//Found it. Ask XMS to copy the data in.
		xmem_xms_move_t move_info = {0};
		move_info.nbytes = 4096;
		move_info.srchandle = 0;
		move_info.srcoffset = (((unsigned long)FP_SEG(data)) << 16) | FP_OFF(data);
		move_info.dsthandle = pptr->handle;
		move_info.dstoffset = page - pptr->paddr_orig;		
		if(!xmem_xms_move(&move_info))
		{
			printf("Failed to move data into extended memory.\n");
			printf("nbytes=%lXh srchandle=%Xh srcoffset=%lXh dsthandle=%Xh dstoffset=%lXh\n",
				move_info.nbytes, 
				move_info.srchandle, move_info.srcoffset,
				move_info.dsthandle, move_info.dstoffset);
			exit(-1);
		}
		
		return;
	}
	
	//Should never have a page that didn't come from our allocations
	assert(0);
}

uint64_t xmem_read64(uint32_t addr)
{
	//Check if this address is in our write-cache - allocate one cache entry per page
	uint16_t cacheentry = (addr / 4096) % XMEM_WRITECACHE_MAX;
	if(xmem_writecache_table[cacheentry].addr == addr)
		return xmem_writecache_table[cacheentry].data;
	
	//Find which pool this page came from
	for(int pp = 0; pp < XMEM_POOL_MAX; pp++)
	{
		xmem_pool_t *pptr = &(xmem_pool_table[pp]);
		if(addr < pptr->paddr_orig)
			continue;
		if(addr + 4096 > (pptr->paddr_orig + pptr->bytes_orig))
			continue;
		
		//Found it. Ask XMS to read the data.
		uint64_t buf = 0;
		uint64_t *far bufptr = &buf;
		xmem_xms_move_t move_info = {0};
		move_info.nbytes = sizeof(buf);
		move_info.srchandle = pptr->handle;
		move_info.srcoffset = addr - pptr->paddr_orig;
		move_info.dsthandle = 0;
		move_info.dstoffset = (((uint32_t)FP_SEG(bufptr)) << 16) | (uint32_t)FP_OFF(bufptr);
		if(!xmem_xms_move(&move_info))
		{
			printf("Failed to move data into extended memory.\n");
			printf("nbytes=%lXh srchandle=%Xh srcoffset=%lXh dsthandle=%Xh dstoffset=%lXh\n",
				move_info.nbytes, 
				move_info.srchandle, move_info.srcoffset,
				move_info.dsthandle, move_info.dstoffset);
			exit(-1);
		}
		
		return buf;
	}
	
	//Should never have a page that didn't come from our allocations
	assert(0);	
	return 0;	
}	

void xmem_write64(uint32_t addr, uint64_t value)
{
	//Update in write-cache - allocate one cache entry per page
	uint16_t cacheentry = (addr / 4096) % XMEM_WRITECACHE_MAX;
	xmem_writecache_table[cacheentry].addr = addr;
	xmem_writecache_table[cacheentry].data = value;
	
	//Find which pool this page came from
	for(int pp = 0; pp < XMEM_POOL_MAX; pp++)
	{
		xmem_pool_t *pptr = &(xmem_pool_table[pp]);
		if(addr < pptr->paddr_orig)
			continue;
		if(addr + 4096 > (pptr->paddr_orig + pptr->bytes_orig))
			continue;
		
		//Found it. Ask XMS to write the data.
		uint64_t buf = value;
		uint64_t *far bufptr = &buf;
		xmem_xms_move_t move_info = {0};
		move_info.nbytes = sizeof(buf);
		move_info.dsthandle = pptr->handle;
		move_info.dstoffset = addr - pptr->paddr_orig;
		move_info.srchandle = 0;
		move_info.srcoffset = (((uint32_t)FP_SEG(bufptr)) << 16) | (uint32_t)FP_OFF(bufptr);
		if(!xmem_xms_move(&move_info))
		{
			printf("Failed to move data into extended memory.\n");
			printf("nbytes=%lXh srchandle=%Xh srcoffset=%lXh dsthandle=%Xh dstoffset=%lXh\n",
				move_info.nbytes, 
				move_info.srchandle, move_info.srcoffset,
				move_info.dsthandle, move_info.dstoffset);
			exit(-1);
		}
		return;
	}
	
	//Should never have a page that didn't come from our allocations
	assert(0);	
}

void xmem_handoff(void)
{
	for(int pp = 0; pp < XMEM_POOL_MAX; pp++)
	{
		xmem_pool_t *pptr = &(xmem_pool_table[pp]);
		if(pptr->bytes_left >= 4096)
		{
			assert((pptr->paddr_left % 4096) == 0);
			ring_write(DRING_CHANNEL_MEMRANGES, &(pptr->paddr_left), 4);
			ring_write(DRING_CHANNEL_MEMRANGES, &(pptr->bytes_left), 4);
			pptr->paddr_left = 0;
			pptr->bytes_left = 0;
		}
	}
}
