;cpuinit.asm
;CPU setup and tables for AMD64
;Bryan E. Topp <betopp@betopp.com> 2021

;This file is part of Stump; Copyright Bryan E. Topp, 2023.
;Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
;Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
;You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

section .text
bits 32

;How many CPUs we support (xAPIC for now, 8-bit IDs)
%define CPUINIT_CPU_MAX 256

;Size of task-state segment per CPU
%define CPUINIT_TSS_SIZE 128
%define CPUINIT_TSS_SIZE_LOG2 7

;Size of stack per CPU
%define CPUINIT_STACK_SIZE 4096
%define CPUINIT_STACK_SIZE_LOG2 12

;Entered once we're in 64-bit virtual space.
global cpuinit_entry
cpuinit_entry:

	bits 64
	
	;Alright, we're in virtual kernel space now.
	
	;Use the proper descriptor tables.
	lgdt [cpuinit_gdtr]
	lidt [cpuinit_idtr]
	
	;Use proper data segment descriptors.
	mov AX, cpuinit_gdt.r0data64 - cpuinit_gdt
	mov SS, AX
	mov DS, AX
	mov ES, AX
	
	mov AX, cpuinit_gdt.r3data64 - cpuinit_gdt
	mov FS, AX
	mov GS, AX
	
	;Turn on SYSCALL/SYSRET and set up entry points
	mov RCX, 0xC0000080 ;EFER
	rdmsr
	or RAX, (1<<0) ;SCE
	wrmsr
	
	mov RCX, 0xC0000084 ;SFMASK
	rdmsr
	mov RAX, 0xFFFFFFFF ;Clear all flags on SYSCALL
	wrmsr
	
	mov RCX, 0xC0000083 ;CSTAR - target RIP for compatibility-mode syscalls
	mov RAX, cpuinit_syscall_32
	mov RDX, RAX
	shr RDX, 32
	wrmsr
	
	mov RCX, 0xC0000082 ;LSTAR - target RIP for 64-bit-mode syscalls
	mov RAX, cpuinit_syscall_64
	mov RDX, RAX
	shr RDX, 32
	wrmsr
	
	mov RCX, 0xC0000081 ;STAR
	mov RDX, (cpuinit_gdt.r3dummy - cpuinit_gdt) | 0x3 ;SYSRET CS and SS (+16 and +8 respectively)
	shl RDX, 16
	or RDX, (cpuinit_gdt.r0code64 - cpuinit_gdt) | 0x0 ;SYSCALL CS and SS (+0 and +8 respectively)
	mov RAX, 0 ;32-bit SYSCALL target (unused)
	wrmsr
	
	;Figure out what core ID we are.
	;We could pull this from the APIC or whatever but this is easier.
	mov RAX, 0 ;Compare with the existing variable
	mov RCX, 1 ;Try to replace with this
	.coreloop:
		lock cmpxchg [cpuinit_ncores], RCX
		jz .coredone
		mov RAX, RCX
		inc RCX
		jmp .coreloop
	.coredone:
	
	;Now RAX contains our core number. Set aside in RBX.
	mov RBX, RAX
	
	;Build a task-state segment descriptor, in the appropriate slot in the TSS descriptor array.
	
	;We store the descriptor in the Global Descriptor Table's TSS Descriptor array
	mov RDI, RBX ;Our core number determines which TSS descriptor we use
	shl RDI, 4 ;Each TSS descriptor takes 16 bytes
	add RDI, cpuinit_gdt.ktss_array ;Find location in TSS descriptor array
	
	;We'll point the TSS descriptor at storage in our TSS array.
	mov RSI, RBX ;Our core number determines where our TSS is stored
	shl RSI, CPUINIT_TSS_SIZE_LOG2 ;Offset within array
	add RSI, cpuinit_ktss_storage ;Find location we'll use
	
	;Build the descriptor...
	mov RAX, CPUINIT_TSS_SIZE - 1
	mov [RDI + 0], AX ;Segment limit 15..0
	
	mov RAX, RSI
	mov [RDI + 2], AX ;Base address 15..0
	shr RAX, 16
	mov [RDI + 4], AL ;Base address 23..16
	mov [RDI + 7], AH ;Base address 31..24
	shr RAX, 16
	mov [RDI + 8], EAX ;Base address 63..32
	
	mov AL, 0b10001001 ;Present, DPL=0, available 64-bit TSS
	mov [RDI + 5], AL
	
	mov AL, 0b00000000 ;1-byte limit granularity, 0000 upper bits of limit
	mov [RDI + 6], AL
	
	mov EAX, 0 ;Upper 4 bytes of descriptor unused
	mov [RDI + 12], EAX
	
	;Select the task-state segment descriptor we just built.
	;In the future we can use the task register to determine which CPU core we're on.
	mov RAX, RBX ;Get core number
	shl RAX, 4 ;Each descriptor is 16 bytes
	add RAX, cpuinit_gdt.ktss_array - cpuinit_gdt ;Make into index in GDT
	ltr AX
	
	;Pick our proper per-CPU stack.
	mov RAX, RBX ;Get core number
	shl RAX, CPUINIT_STACK_SIZE_LOG2 ;Make offset in array of stacks
	add RAX, cpuinit_stack_storage ;Position in array
	add RAX, CPUINIT_STACK_SIZE ;Start at top of stack
	mov RSP, RAX
	
	;Store top-of-stack in our task state segment
	mov RAX, RBX ;Get core number
	shl RAX, CPUINIT_TSS_SIZE_LOG2
	add RAX, cpuinit_ktss_storage
	mov [RAX + 4], RSP
	
	;If we're not core 0, run through kernel setup.
	cmp RBX, 0
	jne .notcore0
	
		;Build interrupt descriptor table
		mov RDI, cpuinit_idt
		mov RSI, cpuinit_isrptrs
		mov RCX, 256 ;Fill all vectors
		.idt_loop:
		
			mov AL, 0x8E
			mov [RDI + 5], AL ;P=1, DPL=00, 0, Type=interrupt gate
		
			mov AL, 0
			mov [RDI + 4], AL ;Interrupt Stack Table number (don't use) 000, ignored 00000		
		
			mov AX, cpuinit_gdt.r0code64 - cpuinit_gdt
			mov [RDI + 2], AX ;Target selector
			
			mov RAX, [RSI]
			mov [RDI + 0], AX ;Target Offset[15:0]
			shr RAX, 16
			mov [RDI + 6], AX ;Target Offset[31:16]
			shr RAX, 16
			mov [RDI + 8], EAX ;Target Offset[63:32]
			
			mov EAX, 0
			mov [RDI + 12], EAX ;Reserved
			
			add RDI, 16 ;16 bytes per interrupt descriptor
			add RSI, 8 ;8 byte addresses of ISRs
			loop .idt_loop ;Count down RCX
		.idt_done:
		
		;Store the initial PML4 location that we got from the loader
		mov RAX, CR3
		mov [cpuinit_pml4_phys], RAX
		
		;Set up framebuffer output
		extern _m_video_init
		call _m_video_init
	
		;Set up memory allocator with info from multiboot
		extern _m_frame_init
		call _m_frame_init
		
		;Do uni-processor kernel setup
		extern entry_boot
		call entry_boot
		
		;Indicate that other cores can proceed
		mov RAX, 1
		mov [cpuinit_kready], RAX
	
	.notcore0:
	
	;Wait for core 0 to complete its kernel setup, then go in to run the kernel
	.waitcore0:
	mov RAX, [cpuinit_kready]
	cmp RAX, 0
	pause
	nop
	je .waitcore0
	
	extern entry_smp
	call entry_smp
	
	;Should never return here. Triple-fault.
	lidt [.crash_idtr]
	int 0xFF
	.spin:
	jmp .spin
	.crash_idtr:
		dw 1 ;Limit
		dq 0 ;Base

;Returns the address of our Task-State Segment in RAX.
global cpuinit_gettss
cpuinit_gettss:
	mov RAX, 0 ;LTR doesn't clear high bits, I think
	str AX ;Get our task register - index of task-state-segment descriptor
	sub RAX, cpuinit_gdt.ktss_array - cpuinit_gdt ;Turn into offset from first TSS descriptor selector
	shr RAX, 4 ;Each TSS descriptor is 16 bytes
	shl RAX, CPUINIT_TSS_SIZE_LOG2 ;Turn into offset in TSS storage
	add RAX, cpuinit_ktss_storage ;Turn into pointer in TSS storage
	ret

;Entered on system-call from 32-bit compatibility-mode (not used)
cpuinit_syscall_32:
	jmp cpuinit_syscall_32
	
;Entered on system-call from 64-bit code
cpuinit_syscall_64:
	;All flags are cleared on entry (given our SFMASK MSR setting).
	;So interrupts are already disabled at this point.
	
	;The SYSCALL instruction has saved the old RIP and RFLAGS in RCX and R11, respectively.
	;The user has passed parameters to us in RDI, RSI, RDX, R8, and R9 as usual.
	;The parameter that they'd normally put in RCX is passed in R10 instead.
	;They expect RBX, RBP, RSP, and R12-R15 preserved, as usual.
	;So we can only clobber RAX at this point.
	
	;We need to get back to the kernel stack.
	;Retrieve the one that we saved in this core's task-state segment.
	mov RAX, 0
	str AX ;Get this core's TSS descriptor selector
	sub RAX, (cpuinit_gdt.ktss_array - cpuinit_gdt) ;Make relative to first TSS descriptor
	shr RAX, 4 ;16 bytes per TSS descriptor - turn into index in the descriptor array
	shl RAX, CPUINIT_TSS_SIZE_LOG2 ;Make offset in TSS storage array
	add RAX, cpuinit_ktss_storage ;Make pointer to this CPU's TSS
	mov RAX, [RAX + 4] ;Load the saved RSP0 from when we last dropped to user-mode.	(Yes, it's a 64-bit value at offset +4.)
	
	;Swap over to the kernel stack, preserving the user's then in RAX.
	xchg RAX, RSP
	
	;Put the return-address and stuff on the stack like taking an interrupt
	push qword 0 ;SS (we don't bother saving this)
	push RAX ;RSP
	push R11 ;RFLAGS
	push qword 0 ;CS (we don't bother saving this)
	push RCX ;RIP
	
	;Put the rest of the user's context, at entry, on the stack
	extern m_drop_putstack
	call m_drop_putstack
	sub RSP, 16
	sub RSP, [RSP]
	
	;Swap back to kernel GS
	swapgs
	
	;Restore the fourth parameter in its usual place, per calling convention
	mov RCX, R10
	
	;Call the kernel to handle the system-call.
	push RSP ;With last parameter as saved context
	extern entry_syscall
	call entry_syscall

bits 64

;Macro for saving temporary registers on entry to external interrupt handler.
;We assume the kernel calling convention will preserve registers that are supposed to be preserved.
;The usually-clobbered registers, though, need us to preserve them on entry/exit from ISR.
%macro irq_save 0
	push RAX
	push RCX
	push RDX
	push RSI
	push RDI
	push R8
	push R9
	push R10
	push R11
%endmacro

%macro irq_restore 0
	pop R11
	pop R10
	pop R9
	pop R8
	pop RDI
	pop RSI
	pop RDX
	pop RCX
	pop RAX
%endmacro

extern m_panic

;Entered on any unused interrupt vector
cpuinit_isr_bad:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Unused interrupt vector.", 0

;CPU exceptions follow

;ISR - divide-by-zero error
cpuinit_isr_de:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Divide-by-zero.", 0
	
;ISR - debug exception
cpuinit_isr_db:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Debug exception.", 0
	
;ISR - non-maskable interrupt
cpuinit_isr_nmi:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Nonmaskable interrupt.", 0

;ISR - breakpoint exception
cpuinit_isr_bp:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Breakpoint.", 0

;ISR - overflow exception
cpuinit_isr_of:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Overflow.", 0

;ISR - bound-range exception
cpuinit_isr_br:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Bounds range.", 0

;ISR - undefined opcode exception
cpuinit_isr_ud:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Undefined opcode.", 0

;ISR - device-not-available exception
cpuinit_isr_nm:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Device not available.", 0

;ISR - double-fault
cpuinit_isr_df:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Double fault.", 0

;ISR - coprocessor segment overrun
cpuinit_isr_co:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Coprocessor segment overrun.", 0

;ISR - invalid TSS exception
cpuinit_isr_ts:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Invalid task state segment.", 0

;ISR - segment-not-present exception
cpuinit_isr_np:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Segment not present.", 0

;ISR - stack exception
cpuinit_isr_ss:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Stack exception.", 0

;ISR - general protection exception
cpuinit_isr_gp:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: General protection exception.", 0

;ISR - page fault
cpuinit_isr_pf:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Page fault.", 0

;ISR - x87 floating-point exception pending
cpuinit_isr_mf:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: x87 floating-point exception.", 0

;ISR - alignment check exception
cpuinit_isr_ac:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Alignment check.", 0

;ISR - machine-check exception
cpuinit_isr_mc:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Machine check.", 0
	
;ISR - SIMD floating-point exception
cpuinit_isr_xf:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: SIMD floating-point exception.", 0

;ISR - hypervisor injection exception
cpuinit_isr_hv:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Hypervisor injected exception.", 0

;ISR - VMM communication exception
cpuinit_isr_vc:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: VMM communication exception.", 0

;ISR - security exception
cpuinit_isr_sx:
	mov RDI, .str
	jmp m_panic
	.str: db "CPU fault: Security exception.", 0
	
;Other ISRs follow
	
;ISR - does nothing but returns, to take the CPU out of halt.
cpuinit_isr_woke:
	;APIC EOI
	push RAX
	mov RAX, 0xFFFFFF0000000000 + 0xFEE00000 + 0xB0
	mov [RAX], dword 0
	pop RAX
	iretq
	
;ISR - spurious interrupts
cpuinit_isr_spurious:
	iretq

;ISR - communication from DOS side (x16)
%macro define_dring_isr 1
	irq_save
	
	mov RDI, %1
	
	extern dring_isr
	call dring_isr
	
	;APIC EOI
	mov RDI, 0xFFFFFF0000000000 + 0xFEE00000 + 0xB0
	mov [RDI], dword 0
	
	irq_restore
	iretq
%endmacro
cpuinit_isr_dring_0: define_dring_isr 0x0
cpuinit_isr_dring_1: define_dring_isr 0x1
cpuinit_isr_dring_2: define_dring_isr 0x2
cpuinit_isr_dring_3: define_dring_isr 0x3
cpuinit_isr_dring_4: define_dring_isr 0x4
cpuinit_isr_dring_5: define_dring_isr 0x5
cpuinit_isr_dring_6: define_dring_isr 0x6
cpuinit_isr_dring_7: define_dring_isr 0x7
cpuinit_isr_dring_8: define_dring_isr 0x8
cpuinit_isr_dring_9: define_dring_isr 0x9
cpuinit_isr_dring_a: define_dring_isr 0xa
cpuinit_isr_dring_b: define_dring_isr 0xb
cpuinit_isr_dring_c: define_dring_isr 0xc
cpuinit_isr_dring_d: define_dring_isr 0xd
cpuinit_isr_dring_e: define_dring_isr 0xe
cpuinit_isr_dring_f: define_dring_isr 0xf

section .data
bits 64

;Table of interrupt service routines - just the addresses, so we can swizzle them at runtime.
align 8
cpuinit_isrptrs:
	;First 32 - AMD64 CPU exceptions, 0x00...0x1F
	dq cpuinit_isr_de  ; 0
	dq cpuinit_isr_db  ; 1
	dq cpuinit_isr_nmi ; 2
	dq cpuinit_isr_bp  ; 3
	dq cpuinit_isr_of  ; 4
	dq cpuinit_isr_br  ; 5
	dq cpuinit_isr_ud  ; 6
	dq cpuinit_isr_nm  ; 7
	dq cpuinit_isr_df  ; 8
	dq cpuinit_isr_co  ; 9
	dq cpuinit_isr_ts  ;10
	dq cpuinit_isr_np  ;11
	dq cpuinit_isr_ss  ;12
	dq cpuinit_isr_gp  ;13
	dq cpuinit_isr_pf  ;14
	dq cpuinit_isr_bad ;15
	dq cpuinit_isr_mf  ;16
	dq cpuinit_isr_ac  ;17
	dq cpuinit_isr_mc  ;18
	dq cpuinit_isr_xf  ;19
	dq cpuinit_isr_bad ;20
	dq cpuinit_isr_bad ;21
	dq cpuinit_isr_bad ;22
	dq cpuinit_isr_bad ;23
	dq cpuinit_isr_bad ;24
	dq cpuinit_isr_bad ;25
	dq cpuinit_isr_bad ;26
	dq cpuinit_isr_bad ;27
	dq cpuinit_isr_hv  ;28
	dq cpuinit_isr_vc  ;29
	dq cpuinit_isr_sx  ;30
	dq cpuinit_isr_bad ;31
	
	;Next 16 - unused - 0x20 - 0x2F
	times 16 dq cpuinit_isr_bad
	
	;Next 16 - DOS-side communication rings - 0x30...0x3F
	dq cpuinit_isr_dring_0
	dq cpuinit_isr_dring_1
	dq cpuinit_isr_dring_2
	dq cpuinit_isr_dring_3
	dq cpuinit_isr_dring_4
	dq cpuinit_isr_dring_5
	dq cpuinit_isr_dring_6
	dq cpuinit_isr_dring_7
	dq cpuinit_isr_dring_8
	dq cpuinit_isr_dring_9
	dq cpuinit_isr_dring_a
	dq cpuinit_isr_dring_b
	dq cpuinit_isr_dring_c
	dq cpuinit_isr_dring_d
	dq cpuinit_isr_dring_e
	dq cpuinit_isr_dring_f
	
	;Next 64 - unused - 0x40...0x7F
	times 64 dq cpuinit_isr_bad
	
	;Next 126 - unused - 0x80...0xFD
	times 126 dq cpuinit_isr_bad
	
	;Wakeup - does nothing but brings the CPU out of halt (0xFE)
	dq cpuinit_isr_woke
	
	;Spurious interrupt (0xFF)
	dq cpuinit_isr_spurious

;Global descriptor table that we use once CPUs are set up.
align 8
global cpuinit_gdt
global cpuinit_gdt.r3code64
global cpuinit_gdt.r3data64
cpuinit_gdt:
	;Null descriptor (CPU requires this to be in index 0)
	dq 0

	;Kernel code segment (64-bit long mode, ring-0)
	.r0code64: db 0, 0, 0, 0, 0, 0x98, 0x20, 0
	
	;Kernel data segment (ring-0)
	;Note that a SYSCALL instruction will load CS and SS with a certain value +0 and +8.
	;So the order of these two descriptors matters - r0data64 must follow r0code64.
	.r0data64: db 0, 0, 0, 0, 0, 0x92, 0, 0

	;Similarly, SYSRET loads CS and SS with a certain value +16 and +8.
	;The descriptor that SYSRET points to, +0, is used for returning to 32-bit code. We don't support that.
	;So we must have r3dummy, r3data64, and then r3code64.
	.r3dummy: dq 0 ;Don't support returning to 32-bit mode
	
	;User data segment (ring-3)
	.r3data64: db 0, 0, 0, 0, 0, 0xF2, 0, 0
	
	;User code segment (64-bit long mode, ring-3)
	.r3code64: db 0, 0, 0, 0, 0, 0xF8, 0x20, 0
	
	;Descriptors for task-state segments for each CPU
	;Built at runtime, so we can swizzle the address bytes
	align 16
	.ktss_array:
	times (16 * CPUINIT_CPU_MAX) db 0 ;Each descriptor is 16 bytes, and we need one per CPU
	.end:

;Pointer to global descriptor table, for LGDT.
cpuinit_gdtr:
	dw (cpuinit_gdt.end - cpuinit_gdt) - 1 ;Limit
	dq cpuinit_gdt ;Base
	
;Pointer to interrupt descriptor table, for LIDT
cpuinit_idtr:
	dw (cpuinit_idt.end - cpuinit_idt) - 1 ;Limit
	dq cpuinit_idt ;Base
	
section .bss
bits 32

;Task state segments for every CPU
alignb 4096
cpuinit_ktss_storage:
	resb CPUINIT_TSS_SIZE * CPUINIT_CPU_MAX
	
;Early stacks for every CPU
alignb 4096
cpuinit_stack_storage:
	resb CPUINIT_STACK_SIZE * CPUINIT_CPU_MAX
	
;Interrupt descriptor table - has to be built at runtime because of the insane address swizzling.
;Ideally the linker would support building this at build-time, but that doesn't work.
alignb 4096
cpuinit_idt:
	resb 16*256 ;256 interrupt vectors, needing 16 bytes each. 
	.end:
	
;Location of the PML4 that we got from the loader initially - runs kernel-space only
alignb 8
global cpuinit_pml4_phys
cpuinit_pml4_phys:
	resb 8
	
;Number of cores that have started up. Used with an atomic to figure out "which core am I".
alignb 8
cpuinit_ncores:
	resb 8

;Whether core-0 has finished early init
alignb 8
cpuinit_kready:
	resb 8

