//dring.c
//Kernel side of DOS-to-kernel ringbuffers
//Bryan E. Topp <betopp@betopp.com> 2022

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "dring.h"
#include "shared/kstd.h"
#include "m_cache.h"

//The DOS loader handles mapping the ringbuffers in our memory space before launching the kernel.
//They are placed at a fixed base address so we know where they are at compile time.

//Second-highest gigabyte is where ringbuffers are mapped
#define DRING_BASE (0xFFFFFFFF80000000ull)

//Each ringbuffer has its own PAE pagetable - 2MBytes of address space
#define DRING_OFFS (0x0000000000200000ull)

//Control page in beginning of ringbuffer space
typedef struct dring_info_s
{
	uint32_t size;
	uint32_t rptr;
	uint32_t wptr;
	uint32_t pad0;
} dring_info_t;
static dring_info_t * const dring_info_ptr = (dring_info_t*)(DRING_BASE);

int dring_read(int channel, void *dst, int nbytes)
{
	//Use channel number as an index into info structures and data blocks
	dring_info_t *iptr = &(dring_info_ptr[channel]);
	const uint8_t *ringdata = ((uint8_t*)DRING_BASE) + (channel * DRING_OFFS);
	
	//Read, possibly twice as we wrap around at the end
	uint8_t *dstbytes = (uint8_t *)dst;
	int result = 0;
	while(nbytes > 0)
	{
		//Don't copy more than requested
		int copylen = nbytes;
		
		//Don't run off the end of the buffer
		int until_wrap = iptr->size - iptr->rptr;
		if(copylen > until_wrap)
			copylen = until_wrap;
		
		//Don't run past the valid data inside the buffer
		int until_dataend = iptr->wptr - iptr->rptr;
		if(until_dataend >= 0 && copylen > until_dataend)
			copylen = until_dataend;
		
		if(copylen == 0)
			break;
		
		memcpy(dstbytes, ringdata + iptr->rptr, copylen);
		
		//Make sure we got the data out before the other side sees our pointer move.
		//Otherwise, the data could be clobbered as the other side thinks it's consumed.
		m_cache_finish(ringdata, iptr->size);
		
		result += copylen;
		dstbytes += copylen;
		iptr->rptr += copylen;
		if(iptr->rptr >= iptr->size)
			iptr->rptr -= iptr->size;
		
		nbytes -= copylen;
	}
	
	return result;
}

int dring_write(int channel, const void *src, int nbytes)
{
	//Use channel number as an index into info structures and data blocks
	dring_info_t *iptr = &(dring_info_ptr[channel]);
	uint8_t *ringdata = ((uint8_t*)DRING_BASE) + (channel * DRING_OFFS);
	
	//Write, possibly twice as we wrap around at the end
	const uint8_t *srcbytes = (const uint8_t *)src;
	int result = 0;
	while(nbytes > 0)
	{
		//Don't copy more than requested
		int copylen = nbytes;
		
		//Don't run off the end of the buffer
		int until_wrap = iptr->size - iptr->wptr;
		if(copylen > until_wrap)
			copylen = until_wrap;
		
		//Don't run past the free space inside the buffer.
		//Don't allow rptr==wptr, as this is the empty condition.
		int until_full = (iptr->rptr - iptr->wptr) - 1;
		if(until_full >= 0 && copylen > until_full)
			copylen = until_full;
		
		if(copylen == 0)
			break;
		
		memcpy(ringdata + iptr->wptr, srcbytes, copylen);
		
		//Make sure we got the data in before the other side sees our pointer move.
		//Otherwise, the old contents might be read at the other side, instead of the data just written.
		m_cache_finish(ringdata, iptr->size);
		
		result += copylen;
		srcbytes += copylen;
		iptr->wptr += copylen;
		if(iptr->wptr >= iptr->size)
			iptr->wptr -= iptr->size;
		
		nbytes -= copylen;
	}		

	return result;
}

void dring_isr(int ring)
{
	//Kernel entry points at ISR
	extern void entry_isr_enet();
	extern void entry_isr_uart();
	extern void entry_isr_vbl();
	extern void entry_isr_input();
	switch(ring)
	{
		case DRING_CHANNEL_ENET0_SEND:
		case DRING_CHANNEL_ENET0_RECV:
			entry_isr_enet(0);
			break;
		case DRING_CHANNEL_UART0_SEND:
		case DRING_CHANNEL_UART0_RECV:
			entry_isr_uart(0);
			break;
		case DRING_CHANNEL_VCURR:
			entry_isr_vbl();
			break;
		case DRING_CHANNEL_INPUT:
			entry_isr_input();
			break;
		default:
			break;
	}
}