;m_drop.asm
;Dropping to usermode from kernel
;Bryan E. Topp <betopp@betopp.com> 2021

;This file is part of Stump; Copyright Bryan E. Topp, 2023.
;Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
;Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
;You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

section .text
bits 64

;Structure of user context.
;Put parameters first so they can be specified by index.
%define M_DROP_OFF_RDI (8 *  0)
%define M_DROP_OFF_RSI (8 *  1)
%define M_DROP_OFF_RDX (8 *  2)
%define M_DROP_OFF_RCX (8 *  3)
%define M_DROP_OFF_R8  (8 *  4)
%define M_DROP_OFF_R9  (8 *  5)
;Then return-value
%define M_DROP_OFF_RAX (8 *  6)
;Then scratch registers
%define M_DROP_OFF_R10 (8 *  7)
%define M_DROP_OFF_R11 (8 *  8)
;Then preserved registers
%define M_DROP_OFF_RBX (8 *  9)
%define M_DROP_OFF_RSP (8 * 10)
%define M_DROP_OFF_RBP (8 * 11)
%define M_DROP_OFF_R12 (8 * 12)
%define M_DROP_OFF_R13 (8 * 13)
%define M_DROP_OFF_R14 (8 * 14)
%define M_DROP_OFF_R15 (8 * 15)
;Then special registers
%define M_DROP_OFF_RIP (8 * 16) ;Instruction pointer
%define M_DROP_OFF_FLG (8 * 17) ;Flags
%define M_DROP_OFF_GSB (8 * 18) ;GS-base
%define M_DROP_SIZE    (8 * 24) ;For whole structure, as defined in m_drop.h

global m_drop_sanitize ;void m_drop_sanitize(m_drop_t *drop);
m_drop_sanitize:
	;Operate on flags register...
	mov RAX, [RDI + M_DROP_OFF_FLG]
	and RAX, 0xFF ;Clear all system flags
	or RAX, (1<<9) ;Enable interrupts
	mov [RDI + M_DROP_OFF_FLG], RAX
	ret

global m_drop_entry ;void m_drop_entry(m_drop_t *drop, uint64_t entry);
m_drop_entry:
	mov [RDI + M_DROP_OFF_RIP], RSI
	ret

global m_drop_param ;void m_drop_param(m_drop_t *drop, int which, uint64_t value);
m_drop_param:
	mov [RDI + (8 * RSI)], RDX
	ret

global m_drop_retval ;void m_drop_retval(m_drop_t *drop, uint64_t retval);
m_drop_retval:
	mov [RDI + M_DROP_OFF_RAX], RSI
	ret

global m_drop ;void m_drop(const void *drop) __attribute__((noreturn));
m_drop:
	;Disable interrupts so we don't get caught with the wrong GS-base
	cli
	
	;Set aside kernel GS-base in spare GS-base register
	swapgs
	
	;Set aside the desired RIP, CS, RFLAGS, RSP, SS on our stack.
	;They'll be popped off once we've restored all other registers.
	extern cpuinit_gdt
	extern cpuinit_gdt.r3code64
	extern cpuinit_gdt.r3data64

	;SS
	mov RAX, cpuinit_gdt.r3data64
	sub RAX, cpuinit_gdt
	or RAX, 3
	push RAX
	
	;RSP
	mov RAX, [RDI + M_DROP_OFF_RSP]
	push RAX
	
	;RFLAGS
	mov RAX, [RDI + M_DROP_OFF_FLG]
	or RAX, 1<<9 ;IF
	push RAX
	
	;CS
	mov RAX, cpuinit_gdt.r3code64
	sub RAX, cpuinit_gdt
	or RAX, 3
	push RAX
	
	;RIP
	mov RAX, [RDI + M_DROP_OFF_RIP]
	push RAX
	
	;Restore all other registers - do RDI last, because it's got our pointer in it	
	mov RAX, [RDI + M_DROP_OFF_GSB]
	wrgsbase RAX

	mov RAX, [RDI + M_DROP_OFF_RAX]
	mov RCX, [RDI + M_DROP_OFF_RCX]
	mov RDX, [RDI + M_DROP_OFF_RDX]
	mov RBX, [RDI + M_DROP_OFF_RBX]
	mov RBP, [RDI + M_DROP_OFF_RBP]
	mov RSI, [RDI + M_DROP_OFF_RSI]
	mov R8,  [RDI + M_DROP_OFF_R8 ]
	mov R9,  [RDI + M_DROP_OFF_R9 ]
	mov R10, [RDI + M_DROP_OFF_R10]
	mov R11, [RDI + M_DROP_OFF_R11]
	mov R12, [RDI + M_DROP_OFF_R12]
	mov R13, [RDI + M_DROP_OFF_R13]
	mov R14, [RDI + M_DROP_OFF_R14]
	mov R15, [RDI + M_DROP_OFF_R15]
	
	mov RDI, [RDI + M_DROP_OFF_RDI]
	
	;Pop RIP, CS, RFLAGS, RSP, SS off our stack, finishing the switch to userspace
	iretq

global m_drop_archname ;const char *m_drop_archname(void);
m_drop_archname:
	mov RAX, .str
	ret
	.str:
		db "amd64", 0

global m_drop_elfmachine ;uint16_t m_drop_elfmachine(void);
m_drop_elfmachine:
	mov RAX, 0x3E ;AMD64
	ret
	
global m_drop_elfclass ;uint8_t m_drop_elfclass(void);
m_drop_elfclass:
	mov RAX, 0x2 ;64-bit
	ret

;Called on entry to kernel-mode, to preserve user context on kernel stack
global m_drop_putstack
m_drop_putstack:
	;Put size of structure on the stack, for caller to use
	push qword M_DROP_SIZE
	
	;Leave room on stack for saved context
	sub RSP, M_DROP_SIZE
	
	;Store general purpose registers
	mov [RSP + M_DROP_OFF_RAX], RAX
	mov [RSP + M_DROP_OFF_RCX], RCX
	mov [RSP + M_DROP_OFF_RDX], RDX
	mov [RSP + M_DROP_OFF_RBX], RBX
	mov [RSP + M_DROP_OFF_RSP], RSP
	mov [RSP + M_DROP_OFF_RBP], RBP
	mov [RSP + M_DROP_OFF_RSI], RSI
	mov [RSP + M_DROP_OFF_RDI], RDI
	mov [RSP + M_DROP_OFF_R8 ], R8
	mov [RSP + M_DROP_OFF_R9 ], R9
	mov [RSP + M_DROP_OFF_R10], R10
	mov [RSP + M_DROP_OFF_R11], R11
	mov [RSP + M_DROP_OFF_R12], R12
	mov [RSP + M_DROP_OFF_R13], R13
	mov [RSP + M_DROP_OFF_R14], R14
	mov [RSP + M_DROP_OFF_R15], R15
	
	;Get previously-pushed RIP, just before we were called
	mov RAX, [RSP + (M_DROP_SIZE + 16) + (8 * 0)]
	mov [RSP + M_DROP_OFF_RIP], RAX
	
	;Get flags, pushed before RIP and CS (ignored)
	mov RAX, [RSP + (M_DROP_SIZE + 16) + (8 * 2)]
	mov [RSP + M_DROP_OFF_FLG], RAX
	
	;Get RSP, pushed before RIP, CS, and flags
	mov RAX, [RSP + (M_DROP_SIZE + 16) + (8 * 3)]
	mov [RSP + M_DROP_OFF_RSP], RAX
	
	rdgsbase RAX
	mov [RSP + M_DROP_OFF_GSB], RAX
	
	mov RAX, [RSP + M_DROP_OFF_RAX]
	
	add RSP, M_DROP_SIZE
	add RSP, 8
	ret

