//m_frame.c
//Physical memory allocator on AMD64
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "m_frame.h"
#include "m_spl.h"
#include "shared/panic.h"
#include "pspace.h"
#include "dring.h"

//Spinlock protecting frame allocator
static m_spl_t m_frame_spl;

//Head of free-list
static uintptr_t m_frame_head;

//Number of entries on free-list
static size_t m_frame_count;

//Ranges of memory given by bootloader, not yet put on free-list.
#define M_FRAME_RANGE_MAX 16
static size_t m_frame_range_sizes[M_FRAME_RANGE_MAX];
static uintptr_t m_frame_range_addrs[M_FRAME_RANGE_MAX];

void _m_frame_init(void)
{
	//DOS loader supplies memory ranges to us via a ringbuffer.
	//Read usable memory ranges out of the ringbuffer.
	for(int rr = 0; rr < M_FRAME_RANGE_MAX; rr++)
	{
		dring_read(DRING_CHANNEL_MEMRANGES, &(m_frame_range_addrs[rr]), 4);
		dring_read(DRING_CHANNEL_MEMRANGES, &(m_frame_range_sizes[rr]), 4);
	}
}

size_t m_frame_size(void)
{
	//Normal i386 4K frames
	return 4096;
}

uintptr_t m_frame_alloc(void)
{
	m_spl_acq(&m_frame_spl);
	
	//Head of list, and count of items, should agree on whether we're empty.
	if( (m_frame_head == 0) ^ (m_frame_count == 0) )
		panic("m_frame_alloc corrupt freelist");
	
	if(m_frame_head == 0)
	{
		//No frames on free-list. See if we have any frames from the bootloader we have never used.
		for(int rr = 0; rr < M_FRAME_RANGE_MAX; rr++)
		{
			if(m_frame_range_sizes[rr] > 0)
			{
				//Hack off the last frame of the range and return it
				m_frame_range_sizes[rr] -= m_frame_size();
				uintptr_t retval = m_frame_range_addrs[rr] + m_frame_range_sizes[rr];
				m_spl_rel(&m_frame_spl);
				return retval;
			}
		}
		
		//No frames on the free-list and no ranges from the bootloader yet unused.
		//We're out of memory.
		m_spl_rel(&m_frame_spl);
		return 0;
	}
	
	//We'll return the old head of the free-list.
	uintptr_t retval = m_frame_head;
	
	//The page to which the old head refers, is now the head.
	m_frame_head = pspace_read(m_frame_head);
	
	m_frame_count--;
	
	m_spl_rel(&m_frame_spl);
	return retval;
}

void m_frame_free(uintptr_t frame)
{
	m_spl_acq(&m_frame_spl);
	
	//Write the existing head into the frame being freed
	pspace_write(frame, m_frame_head);
	
	//The frame being freed is now the head
	m_frame_head = frame;
	
	//Keep track
	m_frame_count++;
	
	m_spl_rel(&m_frame_spl);
}

void m_frame_copy(uintptr_t newframe, uintptr_t oldframe)
{
	for(uintptr_t ff = 0; ff < 4096; ff += 8)
	{
		pspace_write(newframe + ff, pspace_read(oldframe + ff));
	}
}
