//m_video.c
//Framebuffer implementation on AMD64
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "shared/panic.h"
#include "shared/kassert.h"
#include "dring.h"
#include "amd64.h"
#include "sc.h"

//Video info handed-off from DOS side
uint32_t _m_video_width;
uint32_t _m_video_height;
uint32_t _m_video_stride;
uint32_t _m_video_pxfmt;

//Video buffers handed-off from DOS side
typedef struct _m_video_bufinfo_s
{
	uint32_t paddr;
	uint32_t bytes;
} _m_video_bufinfo_t;
#define _M_VIDEO_BUFINFO_MAX 16
_m_video_bufinfo_t _m_video_bufinfo_table[_M_VIDEO_BUFINFO_MAX];
int _m_video_bufinfo_count = 0;

//Last video buffer that is active, as told by the DOS side
uint32_t _m_video_lastbuf;

void _m_video_init(void)
{
	//Read format info passed from DOS side
	dring_read(DRING_CHANNEL_VMEMS, &_m_video_width, sizeof(_m_video_width));
	dring_read(DRING_CHANNEL_VMEMS, &_m_video_height, sizeof(_m_video_height));
	dring_read(DRING_CHANNEL_VMEMS, &_m_video_stride, sizeof(_m_video_stride));
	dring_read(DRING_CHANNEL_VMEMS, &_m_video_pxfmt, sizeof(_m_video_pxfmt));	
	
	//Read video buffer info passed from DOS side
	for(int bb = 0; bb < _M_VIDEO_BUFINFO_MAX; bb++)
	{
		uint32_t vmem_base = 0;
		uint32_t vmem_size = 0;
		dring_read(DRING_CHANNEL_VMEMS, &vmem_base, sizeof(vmem_base));
		dring_read(DRING_CHANNEL_VMEMS, &vmem_size, sizeof(vmem_size));
		_m_video_bufinfo_table[bb].paddr = vmem_base;
		_m_video_bufinfo_table[bb].bytes = vmem_size;
		
		if(vmem_size <= 0)
			break;
		
		KASSERT(_m_video_bufinfo_table[0].bytes == _m_video_bufinfo_table[bb].bytes);
		_m_video_bufinfo_count++;
	}
	
	KASSERT(_m_video_bufinfo_count > 0);
	KASSERT(_m_video_bufinfo_table[0].bytes > 0);
}

size_t m_video_fbcount(void)
{
	return _m_video_bufinfo_count;
}

size_t m_video_fbsize(void)
{
	return _m_video_bufinfo_table[0].bytes;
}

uintptr_t m_video_fbaddr(int idx)
{
	return _m_video_bufinfo_table[idx].paddr;
}

void m_video_fbinfo(_sc_con_fbinfo_t *info_out)
{
	//Temp - how should we pass this across from VBE?
	memset(info_out, 0, sizeof(*info_out));
	info_out->pxfmt = _m_video_pxfmt;
	info_out->width = _m_video_width;
	info_out->height = _m_video_height;
	info_out->stride = _m_video_stride;
	info_out->arn = 4;
	info_out->ard = 3;
}

void m_video_setbuf(uintptr_t bufpaddr)
{		
	//Ignore if we're already on this buffer
	if(_m_video_lastbuf == bufpaddr)
		return;
	
	dring_write(DRING_CHANNEL_VMODE, &bufpaddr, 4);
}

uintptr_t m_video_getbuf(void)
{
	//If the DOS side has updated us on the displayed buffer, read it.
	dring_read(DRING_CHANNEL_VCURR, &_m_video_lastbuf, sizeof(_m_video_lastbuf));
	
	//Return the newest known state, regardless
	return _m_video_lastbuf;
}
