//pspace.c
//Physical space access on AMD64
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "pspace.h"
#include "shared/panic.h"

//In the loader, we make a PDPT (512GBytes of virtual space) filled entirely with 1GByte huge-pages.
//We use this to access the first 512GBytes of physical space in a contiguous region just below kernel-space.
#define PSPACE_SIZE (4096ul*512*512*512) //One PDPT
#define PSPACE_BASE (-2l*PSPACE_SIZE) //Second-to-top PDPT in each PML4 - top PDPT is the kernel-space

uint64_t pspace_read(uint64_t addr)
{
	if(addr > PSPACE_SIZE)
		panic("pspace_read bad addr");
	
	return *((volatile uint64_t*)(PSPACE_BASE + addr));
}

uint32_t pspace_read32(uint64_t addr)
{
	if(addr > PSPACE_SIZE)
		panic("pspace_read bad addr");
	
	return *((volatile uint32_t*)(PSPACE_BASE + addr));
}

void pspace_write(uint64_t addr, uint64_t data)
{
	if(addr > PSPACE_SIZE)
		panic("pspace_write bad addr");
	
	*((volatile uint64_t*)(PSPACE_BASE + addr)) = data;
}

void pspace_write32(uint64_t addr, uint32_t data)
{
	if(addr > PSPACE_SIZE)
		panic("pspace_write bad addr");
	
	*((volatile uint32_t*)(PSPACE_BASE + addr)) = data;
}


void pspace_clrframe(uint64_t addr)
{
	if(addr > PSPACE_SIZE)
		panic("pspace_clrframe bad addr");
	if(addr & 0xFFFul)
		panic("pspace_clrframe misalign");
	
	for(int pp = 0; pp < 512; pp++)
	{
		((uint64_t*)(PSPACE_BASE + addr))[pp] = 0;
	}
}
