;stubboot.asm
;Application Core startup + entry to 32-bit mode
;Bryan E. Topp <betopp@betopp.com> 2022

;This file is part of Stump; Copyright Bryan E. Topp, 2023.
;Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
;Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
;You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

bits 16
section .stubboot

;Trampoline copied to conventional memory for starting-up non-bootstrap cores.
global stubboot
stubboot:
	;Entered in conventional memory in real mode.
	;This code ends up somewhere in the first megabyte, with its beginning page-aligned.
	;IP will begin at 0 and CS will be used to point to the code.
	
	;Copy code segment to data segments, so we can refer to this function at 0
	mov AX, CS
	mov SS, AX
	mov DS, AX
	mov ES, AX
	mov FS, AX
	mov GS, AX
	
	;Set aside the linear location where this stub code begins, in EBX
	mov EAX, 0
	mov AX, CS
	shl EAX, 4
	mov EBX, EAX
	
	;Find linear address of GDT.
	mov EAX, EBX
	add EAX, stubboot_gdt - stubboot
	
	;Store linear address of GDT in GDTR, using seg+off address of GDTR
	mov [2 + stubboot_gdtr - stubboot], EAX
	
	;Make a far-pointer (16:16 segment/offset) where we'll jump later, to get into 32-bit mode. Store it.
	mov EAX, EBX
	add EAX, stubboot_jmptarget - stubboot
	mov [stubboot_jmpptr - stubboot], EAX
	
	;Make a far-pointer (16:32 segment/offset) where we'll jump later, to get into 64-bit mode. Store it.
	mov EAX, EBX
	add EAX, stubboot_longtarget - stubboot
	mov [stubboot_longptr - stubboot], EAX
	
	;Load our PML4 (top-level paging structure) pointer - immediately following this code segment.
	mov EAX, EBX
	add EAX, 0x1000
	mov CR3, EAX
	
	;Enter protected mode.
	mov EAX, CR0
	or EAX, 1 ;PE
	mov CR0, EAX
	
	;Enter 32-bit flat-memory mode, but don't turn on paging...
	
	;Load the GDT that we referenced earlier, using the GDTR we filled in.
	;The address of the GDTR is given as a segment-relative, but the GDTR contains linear addresses.
	lgdt [stubboot_gdtr - stubboot]
	
	;Jump into that flat 32-bit code segment
	jmp far dword [stubboot_jmpptr - stubboot]

align 8
stubboot_jmptarget:
	bits 32
	
	;Load flat 32-bit data segments - but leave GS as-is so we can still use it for relative access
	mov AX, 16
	mov SS, AX
	mov DS, AX
	mov ES, AX
	mov FS, AX
	
	;Enable APIC Software Enable bit
	mov ECX, 0x1B ;APIC BAR MSR
	rdmsr
	and EAX, 0xFFFFF000
	mov ECX, [EAX + 0xF0] ;Spurious Interrupt Register
	or ECX, 1<<8 ;APIC Software Enable
	or ECX, 0xFF ;Vector 0xFF for spurious interrupts
	mov [EAX + 0xF0], ECX
	
	;Send EOI to APIC initially
	mov ECX, 0
	mov [EAX + 0xB0], ECX

	;Turn on PAE (8-byte pagetable entries) and FSGSBASE (rdgsbase/wrgsbase instructions)
	mov EAX, CR4
	or EAX, 1<<5 ;PAE
	or EAX, 1<<16 ;FSGSBASE
	mov CR4, EAX
	
	;Turn on Long Mode Enable (so Paging will trigger Long Mode Active) and NX-bit support
	mov ECX, 0xC0000080 ;EFER
	rdmsr
	or EAX, 1<<8 ;LME
	or EAX, 1<<11 ;NX
	wrmsr
	
	;Enable paging, kicking us into long-mode.
	mov EAX, CR0
	or EAX, 1<<31 ;PG
	mov CR0, EAX
	
	;We're using the paging structures that the loader set up for us. We're in an identity-mapped page.
	;Our kernel is now accessible at its proper virtual addresses.
	;But we can't get there yet, because we're only using 32-bit addressing.
	;Jump into a 64-bit code segment with a 32-bit offset first.
	jmp far dword [GS : stubboot_longptr - stubboot]
	
stubboot_longtarget:
	bits 64
	
	;Now we're still in identity-mapped space, but using 64-bit addressing.
	;Jump into virtual space.
	mov RAX, .targetvirt
	jmp RAX
	.targetvirt:
	
	;Jump into the kernel.
	extern cpuinit_entry
	mov RAX, cpuinit_entry
	jmp RAX
	
	bits 16

align 8
stubboot_gdtr:
	dw 31 ;Limit
	dd 0 ;Base - filled at runtime

align 8
stubboot_gdt:
	dq 0 ;Null descriptor
	db 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x9A, 0xCF, 0x00 ;Ring-0 32-bit code
	db 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x92, 0xCF, 0x00 ;Ring-0 32-bit data
	db 0x00, 0x00, 0x00, 0x00, 0x00, 0x98, 0x20, 0x00 ;Ring-0 64-bit code

align 8
stubboot_jmpptr:
	dd 0 ;Offset
	dw 8 ;Selector in GDT
	
align 8
stubboot_longptr:
	dd 0 ;Offset
	dw 24 ;Selector in GDT

align 8
stubboot_end:
