//argenv.c
//Argument/environment loading for new processes
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "argenv.h"
#include "kassert.h"
#include "process.h"
#include "m_frame.h"
#include "m_spl.h"
#include "kstd.h"

//Pointer type used in argv/envp buffers
typedef uint64_t argenv_ptr_t;

//Temporary buffer for making arg/env structures
#define ARGENV_MAX 65536
static char argenv_kbuf[ARGENV_MAX];

//Spinlock protecting the use of the above buffer
static m_spl_t argenv_spl;

//Returns the location of a pointer in the given argv or envp array.
uintptr_t argenv_ptrloc(mem_t *oldmem, char * const * strarr, int idx)
{
	(void)oldmem;
	return ((uintptr_t)strarr) + ((sizeof(argenv_ptr_t)) * idx);
}

//Returns the value of a pointer in the given argv or envp array.
uintptr_t argenv_ptrval(mem_t *oldmem, char * const * strarr, int idx)
{
	uintptr_t loc = argenv_ptrloc(oldmem, strarr, idx);
	return *((argenv_ptr_t*)(loc));
}

//Returns count and total length of strings in an argv or envp array.
int argenv_count(mem_t *oldmem, char * const * strarr, size_t *count_out, size_t *strlen_out)
{
	size_t count = 0;
	size_t totallen = 0;
	while(1)
	{
		uintptr_t ptr_loc = argenv_ptrloc(oldmem, strarr, count);
		if(mem_validate(oldmem, ptr_loc, sizeof(argenv_ptr_t), M_USPC_PROT_R) < 0)
			return -EFAULT;
		
		uintptr_t ptr_val = argenv_ptrval(oldmem, strarr, count);
		if(ptr_val == 0)
			break;
		
		int len = mem_valstr(oldmem, (const char*)ptr_val);
		if(len < 0)
			return len;
		
		totallen += len + 1;
		count++;
	}
	
	*count_out = count;
	*strlen_out = totallen;
	return 0;
}

int argenv_load(mem_t *oldmem, mem_t *mem, char * const * argv, char * const * envp, uintptr_t *loaded_out)
{	
	//Count length of strings.
	size_t argv_count = 0;
	size_t argv_strlen = 0;
	int count_argv_err = argenv_count(oldmem, argv, &argv_count, &argv_strlen);
	if(count_argv_err < 0)
		return count_argv_err;
	
	size_t envp_count = 0;
	size_t envp_strlen = 0;
	int count_envp_err = argenv_count(oldmem, envp, &envp_count, &envp_strlen);
	if(count_envp_err < 0)
		return count_envp_err;
	
	//We need the first two pointers to the pointer-arrays.
	//Then we need the pointer arrays, each NULL-terminated.
	//Then we need the strings, each NUL-terminated.
	size_t argenv_size = (2 * sizeof(argenv_ptr_t)) + ((envp_count + 1 + argv_count + 1) * sizeof(argenv_ptr_t)) + argv_strlen + envp_strlen;
	if(argenv_size > sizeof(argenv_kbuf))
	{
		//Size of arguments and environment too big.
		return -E2BIG;
	}
	
	//Get space in kernel-space, so we can copy across the two user-spaces
	m_spl_acq(&argenv_spl);
	memset(argenv_kbuf, 0, argenv_size);
	
	//Find space and add a segment to the target memory-space
	size_t argenv_segsize = argenv_size;
	size_t pagesize = m_frame_size();
	if(argenv_segsize % pagesize != 0)
		argenv_segsize += pagesize - (argenv_segsize % pagesize);
	
	intptr_t argenv_addr = mem_avail(mem, -1, argenv_segsize);
	if(argenv_addr < 0)
	{
		//No free address space
		m_spl_rel(&argenv_spl);
		return argenv_addr;
	}
	
	int add_err = mem_add(mem, argenv_addr, argenv_segsize, M_USPC_PROT_R | M_USPC_PROT_W);
	if(add_err < 0)
	{
		//Probably ran out of RAM - couldn't make a memory segment to hold the args/env.
		m_spl_rel(&argenv_spl);
		return add_err;
	}
	
	//Copy across - into kernel-space, then out into the new userspace
	
	//Beginning of arg/env blob is pointer to arg-strings array, then pointer to env-strings array.
	//arg-strings array starts right after. env-strings array starts after all arg-string pointers (and NULL).
	//We need to keep track of offset-within-segment, because this is at different locations now (kernel) and finally (user).
	ptrdiff_t ptrptrs_off = 0;
	ptrdiff_t arg_ptrs_off = 2 * sizeof(argenv_ptr_t);
	ptrdiff_t env_ptrs_off = arg_ptrs_off + ((argv_count+1) * sizeof(argenv_ptr_t));
	ptrdiff_t str_data_off = env_ptrs_off + ((envp_count+1) * sizeof(argenv_ptr_t));
	
	//Fill initial two pointers, to the argv pointers and envp pointers
	argenv_ptr_t *array_ptrs = (argenv_ptr_t*)(argenv_kbuf + ptrptrs_off);
	array_ptrs[0] = argenv_addr + arg_ptrs_off;
	array_ptrs[1] = argenv_addr + env_ptrs_off;
	
	//Fill argv pointers
	argenv_ptr_t *arg_ptrs = (argenv_ptr_t *)(argenv_kbuf + arg_ptrs_off);
	for(size_t aa = 0; aa < argv_count; aa++)
	{
		char *ptrval = (char*)argenv_ptrval(oldmem, argv, aa);
		int aalen = mem_valstr(oldmem, ptrval);
		if(aalen > 0)
		{
			arg_ptrs[aa] = argenv_addr + str_data_off;
			memcpy(argenv_kbuf + str_data_off, ptrval, aalen);
			str_data_off += aalen + 1;
		}
	}
	
	argenv_ptr_t *env_ptrs = (argenv_ptr_t*)(argenv_kbuf + env_ptrs_off);
	for(size_t ee = 0; ee < envp_count; ee++)
	{
		char *ptrval = (char*)argenv_ptrval(oldmem, envp, ee);
		int eelen = mem_valstr(oldmem, ptrval);
		if(eelen > 0)
		{
			env_ptrs[ee] = (argenv_ptr_t)(argenv_addr + str_data_off);
			memcpy(argenv_kbuf + str_data_off, ptrval, eelen);
			str_data_off += eelen + 1;
		}
	}
	
	KASSERT(str_data_off == (ssize_t)argenv_size);
	
	//Now that we've got the data into kernel-space, flip over and copy it into the new user-space.
	uintptr_t old_uspc = m_uspc_current();
	m_uspc_activate(mem->uspc);
	memcpy((void*)argenv_addr, argenv_kbuf, argenv_size);
	m_uspc_activate(old_uspc);
	
	//Successo
	*loaded_out = argenv_addr;
	m_spl_rel(&argenv_spl);
	return 0;
}
