//con.c
//Console handling in kernel
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "con.h"
#include "process.h"
#include "thread.h"
#include "kstd.h"
#include "m_input.h"
#include "waiters.h"
#include "kassert.h"
#include <sc_con.h>

//Whether the console should return to the init process.
static bool con_syskey_trigger;

//Partial event read by the console
static uint32_t con_evtpart_buf;
static int con_evtpart_len;

//Threads waiting on console input
static int con_input_waiters[8];

//Trigger keys for stealing the console back to PID1
static const int con_syskey_codes[] = { _SC_CON_SCANCODE_LSHIFT, _SC_CON_SCANCODE_RSHIFT, _SC_CON_SCANCODE_BACKSPACE, 0 };
static bool con_syskey_pressed[sizeof(con_syskey_codes)/sizeof(con_syskey_codes[0])];

//Checks incoming console events against key combo for stealing console
static void con_syskey_check(const _sc_con_input_t *evt)
{
	//Only respond to keyboard events
	if(evt->kbd.type != _SC_CON_INPUT_TYPE_KBD)
		return;
	
	//See if the changed key is one we care about
	int syskey_index = -1;
	for(int ss = 0; con_syskey_codes[ss] != 0; ss++)
	{
		if(evt->kbd.scancode == con_syskey_codes[ss])
		{
			syskey_index = ss;
			break;
		}
	}
	
	if(syskey_index < 0)
		return;
	
	//Update the key's state
	con_syskey_pressed[syskey_index] = evt->kbd.state;
	
	//Check if all necessary keys are pressed
	for(int ss = 0; con_syskey_codes[ss] != 0; ss++)
	{
		if(!con_syskey_pressed[ss])
			return;
	}
	
	//All keys pressed - trigger stealing of console
	con_steal_prepare();
	
	//Reset combo state
	memset(con_syskey_pressed, 0, sizeof(con_syskey_pressed));
}

void con_isr_input(void)
{
	waiters_release(con_input_waiters, sizeof(con_input_waiters));
}

ssize_t con_input(void *buf, ssize_t buflen)
{
	if(buflen % sizeof(_sc_con_input_t))
		return -EINVAL; //Can only read integer numbers of input events
	
	//Console is implemented as UART 0.
	//Read events into a temp buffer and make sure we're aligned correctly.
	_sc_con_input_t *evt_out = buf;
	ssize_t evt_remain = buflen / sizeof(_sc_con_input_t);
	ssize_t evt_count = 0;
	while(evt_remain > 0)
	{
		//Read whole events - skip over misalignments.
		while(con_evtpart_len < 4)
		{
			//Get the single next byte from the UART.
			uint8_t nextbyte = 0;
			ssize_t nread = m_input_recv((char*)&nextbyte, 1);
			if(nread <= 0)
			{
				//No more incoming data from the event UART.
				if(evt_count == 0)
				{
					//No data at all returned 
					waiters_addcur(con_input_waiters, sizeof(con_input_waiters));
					return -EAGAIN;
				}
				else
				{
					//Got at least one whole event.
					return evt_count;
				}
			}

			//Okay, got the next byte.
			//Stick it in the buffer for partially-received events, unless it's 0xFF and the first byte.
			//Events are sent big-end-first and the first byte is therefore the event type and player.
			//0xFF can therefore never be the first byte received in an event - so skip those.
			if(nextbyte == 0xFF && con_evtpart_len == 0)
				continue;

			//Otherwise, put the received byte into the event that we're receiving
			con_evtpart_buf <<= 8; 
			con_evtpart_buf |= nextbyte;
			con_evtpart_len++;
		}

		//Okay, got a complete event. Output it to the user's buffer.
		union {
			_sc_con_input_t inp;
			uint32_t word;
		} evt_in;
		evt_in.word = con_evtpart_buf;
		*evt_out = evt_in.inp;
		evt_out++;
		evt_count++;
		evt_remain--;		
		con_evtpart_buf = 0;
		con_evtpart_len = 0;
		
		//Check if we trigger the "steal the console" key combo
		con_syskey_check(&(evt_in.inp));
	}
	
	//Filled the buffer before we hit the end of the incoming event data from the UART.
	return evt_count;
}

bool con_steal_check(void)
{
	if(con_syskey_trigger)
	{
		con_syskey_trigger = false;
		return true;
	}
	return false;
}

void con_steal_prepare(void)
{
	con_syskey_trigger = true;
	thread_unpause(1);
}

bool con_on(pid_t pid)
{
	//When the user has requested that the console be "stolen" back to PID1, act like that happens
	if(con_steal_check())
	{
		//Reset all processes to either not-participate or pass-to-self
		for(int pp = 0; pp < PROCESS_MAX; pp++)
		{
			if(process_table[pp].conchild > 0)
				process_table[pp].conchild = process_table[pp].pid;
		}
	}
	
	//Figure out who has the console. PID1 gets it by default.
	pid_t holder = 1;
	for(int follows = 0; follows <= PROCESS_MAX; follows++)
	{
		//We think this process (holder) has the console.
		//Figure out if it's been successfully delegated further.
		process_t *holder_pptr = process_lockpid(holder);
		KASSERT(holder_pptr != NULL);
		
		pid_t check_child = holder_pptr->conchild;
		
		process_unlock(holder_pptr);
		holder_pptr = NULL;
		
		if(check_child == holder)
		{
			//This process wants to hold the console itself. It has the console.
			break;
		}
		
		process_t *child_pptr = process_lockpid(check_child);
		if(child_pptr == NULL)
		{
			//This process is delegating to an invalid process. It has the console.
			break;
		}
		
		if(child_pptr->ppid != holder)
		{
			//This process is delegating to a process which isn't its child. It has the console.
			//Todo - check if it's a descendant at all - grandchild etc
			//process_unlock(child_pptr);
			//break;
		}
		
		if(child_pptr->conchild <= 0)
		{
			//This process is delegating to a process which doesn't want to participate. It has the console.
			process_unlock(child_pptr);
			break;
		}
		
		//Okay, this process is delegating to one of its children, and the child wants the console (or wants to delegate it).
		process_unlock(child_pptr);
		holder = check_child;
	}
	
	return (holder == pid);
}


