//devs.c
//Character device functions
//Bryan E. Topp <betopp@betopp.com> 2022

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "devs.h"
#include "kassert.h"
#include "kstd.h"
#include "m_enet.h"
#include "m_uart.h"
#include "m_spl.h"
#include "waiters.h"

#define D_UART_MAX 256
static bool d_uart_busy[D_UART_MAX];
static m_spl_t d_uart_busy_spl;
static pid_t d_uart_waiters[D_UART_MAX][8];

void devs_isr_uart(int minor)
{
	if(minor >= 0 && minor < D_UART_MAX)
		waiters_release(d_uart_waiters[minor], sizeof(d_uart_waiters[minor]));
}

static int d_uart_open(int minor)
{
	if(minor < 0 || minor >= D_UART_MAX)
		return -ENXIO;

	//Disallow opening UART 0 as a device - it's used for the console
	if(minor == 0)
		return -EBUSY;

	m_spl_acq(&d_uart_busy_spl);
	bool was_busy = d_uart_busy[minor];
	d_uart_busy[minor] = true;
	m_spl_rel(&d_uart_busy_spl);

	if(was_busy)
		return -EBUSY;
	else
		return 0;
}

static void d_uart_close(int minor)
{
	KASSERT(minor >= 0 && minor < D_UART_MAX);
	KASSERT(d_uart_busy[minor]);
	d_uart_busy[minor] = false;
}

static ssize_t d_uart_read(int minor, void *buf, ssize_t len)
{
	KASSERT(minor >= 0 && minor < D_UART_MAX);
	KASSERT(buf != NULL);
	KASSERT(len >= 0);

	ssize_t retval = m_uart_recv(minor, buf, len);
	if(retval == 0 || retval == -EAGAIN)
	{
		waiters_addcur(d_uart_waiters[minor], sizeof(d_uart_waiters[minor]));
		retval = -EAGAIN;
	}

	return retval;
}

static ssize_t d_uart_write(int minor, const void *buf, ssize_t len)
{
	KASSERT(minor >= 0 && minor < D_UART_MAX);
	KASSERT(buf != NULL);
	KASSERT(len >= 0);

	ssize_t retval = m_uart_send(minor, buf, len);
	if(retval == 0 || retval == -EAGAIN)
	{
		waiters_addcur(d_uart_waiters[minor], sizeof(d_uart_waiters[minor]));
		retval = -EAGAIN;
	}

	return retval;
}

#define D_ENET_MAX 256
bool d_enet_busy[D_ENET_MAX];
m_spl_t d_enet_busy_spl;
static pid_t d_enet_waiters[D_ENET_MAX][8];

void devs_isr_enet(int minor)
{
	if(minor >= 0 && minor < D_ENET_MAX)
		waiters_release(d_enet_waiters[minor], sizeof(d_enet_waiters[minor]));
}

static int d_enet_open(int minor)
{
	if(minor < 0 || minor >= D_ENET_MAX)
		return -ENXIO;

	m_spl_acq(&d_enet_busy_spl);
	bool was_busy = d_enet_busy[minor];
	d_enet_busy[minor] = true;
	m_spl_rel(&d_enet_busy_spl);

	if(was_busy)
		return -EBUSY;
	else
		return 0;
}

static void d_enet_close(int minor)
{
	KASSERT(minor >= 0 && minor < D_ENET_MAX);
	KASSERT(d_enet_busy[minor]);
	d_enet_busy[minor] = false;
}

static ssize_t d_enet_read(int minor, void *buf, ssize_t len)
{
	KASSERT(minor >= 0 && minor < D_ENET_MAX);
	KASSERT(buf != NULL);
	KASSERT(len >= 0);

	//Receiver must have room for an entire standard Ethernet packet at once
	if(len < 1522)
		return -EINVAL;

	//Note that the receiver may choose to provide more buffer space,
	//and maybe the machine layer can buffer-up a jumbo frame and provide it.

	int retval = m_enet_recv(minor, buf, len);
	if(retval == 0 || retval == -EAGAIN)
	{
		waiters_addcur(d_enet_waiters[minor], sizeof(d_enet_waiters[minor]));
		retval = -EAGAIN;
	}

	return retval;
}

static ssize_t d_enet_write(int minor, const void *buf, ssize_t len)
{
	KASSERT(minor >= 0 && minor < D_ENET_MAX);
	KASSERT(buf != NULL);
	KASSERT(len >= 0);

	//Disallow runt packets
	if(len < 64)
		return -EINVAL;

	//Machine layer may impose restrictions on largest frame it will send.
	//We don't check that here. Obviously it shouldn't fail for 1522-byte or smaller frames.

	int retval = m_enet_send(minor, buf, len);
	if(retval == 0 || retval == -EAGAIN)
	{
		waiters_addcur(d_enet_waiters[minor], sizeof(d_enet_waiters[minor]));
		retval = -EAGAIN;
	}

	return retval;
}

static int d_nxio_open(int minor)
{
	(void)minor;
	return -ENXIO;
}

static void d_nxio_close(int minor)
{
	(void)minor;
}

static ssize_t d_nxio_read(int minor, void *buf, ssize_t nbytes)
{
	(void)minor;
	(void)buf;
	(void)nbytes;
	return -ENXIO;
}

static ssize_t d_nxio_write(int minor, const void *buf, ssize_t nbytes)
{
	(void)minor;
	(void)buf;
	(void)nbytes;
	return -ENXIO;
}

static ssize_t d_null_write(int minor, const void *buf, ssize_t len)
{
	(void)minor;	
	(void)buf;
	return len;
}

static ssize_t d_null_read(int minor, void *buf, ssize_t len)
{
	(void)minor;
	(void)buf;
	(void)len;
	return 0;
}

static ssize_t d_zero_read(int minor, void *buf, ssize_t len)
{
	//Prohibit giant requests
	if(len >= 65536)
		len = 65536;
	
	//Fun hack - minor number can be used to make a /dev/zero for any one byte
	memset(buf, minor, len);
	return len;
}

static ssize_t d_zero_write(int minor, const void *buf, ssize_t len)
{	
	(void)minor;
	(void)buf;
	return len;
}


static ssize_t d_full_write(int minor, const void *buf, ssize_t len)
{
	(void)minor;	
	(void)buf;
	if(len == 0)
		return 0;
	else
		return -ENOSPC;
}

static ssize_t d_full_read(int minor, void *buf, ssize_t len)
{
	if(len > 65536)
		len = 65536;
	
	memset(buf, minor, len);
	return len;
}

//Character device switch - what major numbers mean
typedef enum devs_major_e
{
	DEVS_MAJOR_NXIO   = 0x00, //Always returns -ENXIO for everything - used for bad major numbers
	DEVS_MAJOR_NULL   = 0x01,
	DEVS_MAJOR_ZERO   = 0x02,
	DEVS_MAJOR_FULL   = 0x03,
	DEVS_MAJOR_UART   = 0x40,
	DEVS_MAJOR_ENET   = 0x41,
	DEVS_MAJOR_MAX
} devs_major_t;

//Character device switch
typedef struct devs_s
{
	const char *name;
	int     (*open) (int minor);
	void    (*close)(int minor);
	ssize_t (*read) (int minor, void *buf, ssize_t nbytes);
	ssize_t (*write)(int minor, const void *buf, ssize_t nbytes);
} devs_t;
static const devs_t devs_table[DEVS_MAJOR_MAX] = 
{
	[DEVS_MAJOR_NXIO] = 
	{
		.name = "nxio",
		.open = d_nxio_open,
		.close = d_nxio_close,
		.read = d_nxio_read,
		.write = d_nxio_write,
	},
	[DEVS_MAJOR_NULL] =
	{
		.name = "null",
		.read = d_null_read,
		.write = d_null_write,
	},
	[DEVS_MAJOR_ZERO] = 
	{
		.name = "zero",
		.read = d_zero_read,
		.write = d_zero_write,
	},
	[DEVS_MAJOR_FULL] = 
	{
		.name = "full",
		.read = d_full_read,
		.write = d_full_write,
	},
	[DEVS_MAJOR_UART] = 
	{
		.name = "uart",
		.open = d_uart_open,
		.close = d_uart_close,
		.read = d_uart_read,
		.write = d_uart_write,
	},
	[DEVS_MAJOR_ENET] =
	{
		.name = "enet",
		.open = d_enet_open,
		.close = d_enet_close,
		.read = d_enet_read,
		.write = d_enet_write,
	},
};

//Returns character-device functions for the given character-device number.
//Returns functions that return -NXIO if invalid value is given.
static const devs_t *devs_major(int special)
{
	int major = (special >> 16) & 0xFFFF;
	if(major < 0 || major >= DEVS_MAJOR_MAX || devs_table[major].name == NULL)
		return &(devs_table[DEVS_MAJOR_NXIO]);
	
	return &(devs_table[major]);
}

//Returns the minor number associated with the given character-device number.
static int devs_minor(int special)
{
	return special & 0xFFFF;
}

int devs_open(int rdev)
{
	const devs_t *major = devs_major(rdev);
	if(major->open == NULL)
		return 0; //If we have no particular "open" function, assume it's okay
	
	return (*(major->open))(devs_minor(rdev));
}

void devs_close(int rdev)
{
	const devs_t *major = devs_major(rdev);
	if(major->close == NULL)
		return;
	
	(*(major->close))(devs_minor(rdev));
}

ssize_t devs_read(int rdev, void *buf, ssize_t nbytes)
{
	const devs_t *major = devs_major(rdev);
	if(major->read == NULL)
		return -ENOTTY;
	
	return (*(major->read))(devs_minor(rdev), buf, nbytes);
}

ssize_t devs_write(int rdev, const void *buf, ssize_t nbytes)
{
	const devs_t *major = devs_major(rdev);
	if(major->write == NULL)
		return -ENOTTY;
	
	return (*(major->write))(devs_minor(rdev), buf, nbytes);	
}
