//entry.c
//Entry points to kernel
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "ramfs.h"
#include "fb.h"
#include "systar.h"
#include "process.h"
#include "thread.h"
#include "con.h"
#include "syscalls.h"
#include "panic.h"
#include "m_frame.h"
#include "m_intr.h"
#include "devs.h"
#include "kassert.h"
#include "kstd.h"

//Entered once on bootstrap core. Should set up kernel and return.
void entry_boot(void)
{
	//Set up in-memory systems
	fb_init();
	ramfs_init();
	
	//Unpack and free the TAR file containing initial FS contents
	systar_unpack();
	
	//Make initial process to execute init
	process_init();
	
	//Return to the machine abstraction layer, which will call entry_smp on all cores...
}

//Entered on all cores once entry_boot returns. Does not return (exits kernel to usermode).
void entry_smp(void)
{
	//Schedule the first thread.
	//We'll re-enter the kernel somewhere else after it runs.
	thread_drop();
	panic("thread_drop returned");
}

//Entered on system-call. Does not return (exits kernel to usermode).
uintptr_t entry_syscall(uintptr_t num, uintptr_t p1, uintptr_t p2, uintptr_t p3, uintptr_t p4, uintptr_t p5, m_drop_t *drop)
{	
	//Can once again take interrupts
	m_intr_ei(true);
	
	//Save context in thread that was running and mark it as being serviced
	thread_t *tptr = thread_lockcur();
	
	KASSERT(tptr->state == THREAD_STATE_RUN);
	thread_chstate(tptr, THREAD_STATE_SYSCALL);
	memcpy(&(tptr->drop), drop, sizeof(tptr->drop));
	m_drop_retval(&(tptr->drop), 0);
	thread_unlock(tptr);
	tptr = NULL;
	
	//Run the requested system call
	uintptr_t result = syscalls_handle(num, p1, p2, p3, p4, p5);
	if(result != 0)
	{
		//We initialize the result previously to be 0. Then we run the syscall.
		//Then we change it here, only if the system-call returns nonzero.
		//This is necessary when returning from signals.
		// - Their old context will be restored during syscalls_handle, which then returns 0.
		// - We don't mangle the return value in the interrupted context (now resumed).
		tptr = thread_lockcur();
		m_drop_retval(&(tptr->drop), result);
		thread_unlock(tptr);
	}
	
	//Pick another thread to run
	thread_drop();
	panic("thread_drop returned");
}

//Entered on exception. Does not return (exits kernel to usermode).
void entry_exception(int signum, uintptr_t badaddr, m_drop_t *drop)
{
	//Can once again take interrupts
	m_intr_ei(true);
	
	//Save interrupted context in the thread that was running
	thread_t *tptr = thread_lockcur();
	KASSERT(tptr->state == THREAD_STATE_RUN);
	thread_chstate(tptr, THREAD_STATE_SYSCALL);
	memcpy(&(tptr->drop), drop, sizeof(tptr->drop));
	
	//The CPU's exception is translated into a signal on the thread that caused it.
	tptr->sigpend |= 1ull << signum;
	tptr->sigcause_addr = badaddr;
	tptr->sigcause_tid = tptr->tid;
	tptr->sigcause_pid = tptr->process->pid;
	
	//If the thread is masking the signal it just caused, it cannot continue.
	//(It would just keep running into the same exception, unable to handle it.)
	//So its process dies.
	if(tptr->sigmask & (1ull << signum))
	{
		//Thread has masked the signal it just caused. Its process dies.
		process_t *pptr = process_lockcur();
		if(pptr->state == PROCESS_STATE_ALIVE)
		{
			pptr->state = PROCESS_STATE_DETHREAD;
			pptr->wstatus = _SC_STATUS_SIGNALED_BIT | ((signum << _SC_STATUS_TERMSIG_SHIFT) & _SC_STATUS_TERMSIG_MASK);
		}
		process_unlock(pptr);
	}
	
	//Done with the thread
	thread_unlock(tptr);
	tptr = NULL;
	
	//Pick another thread to run
	thread_drop();
	panic("thread_drop returned");
}

//Called from ISR when a UART is ready to send/receive. Returns to exit from ISR.
void entry_isr_uart(int which)
{
	devs_isr_uart(which);
}

//Called from ISR when an Ethernet device is ready to send/receive. Returns to exit from ISR.
void entry_isr_enet(int which)
{
	devs_isr_enet(which);
}

//Called from ISR when the video system enters blanking. Returns to exit from ISR.
void entry_isr_vbl(void)
{
	fb_isr();
}

//Called from ISR when new input events are available.
void entry_isr_input(void)
{
	con_isr_input();
}

//Called from ISR when another millisecond has elapsed. Returns to exit from ISR.
void entry_isr_tick(void)
{

}

