//fb.c
//Framebuffer handling in kernel
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "fb.h"
#include "m_cache.h"
#include "m_video.h"
#include "m_spl.h"
#include "kassert.h"
#include "waiters.h"
#include "kstd.h"

//Bitmap of which framebuffers are free
static uint32_t fb_freemap;

//Threads waiting to show a framebuffer (probably only need 1 here for any sane case)
pid_t fb_waiters[8];

//Spinlock for allocating and deallocating framebuffers
m_spl_t fb_spl;

void fb_init(void)
{
	//Mark all framebuffers as free
	fb_freemap = 0;
	for(size_t bb = 0; bb < m_video_fbcount(); bb++)
	{
		fb_freemap |= (1ul << bb);
	}
	KASSERT(fb_freemap != 0);
}

void fb_isr(void)
{
	waiters_release(fb_waiters, sizeof(fb_waiters));		
}

uintptr_t fb_alloc(int size)
{
	if(size < 0)
		return 0;
	if(size > (int)m_video_fbsize())
		return 0;

	m_spl_acq(&fb_spl);
	for(size_t bb = 0; bb < m_video_fbcount(); bb++)
	{
		size_t idx = (m_video_fbcount() - 1) - bb; //Allocate from the back, so boot-up looks nicer (don't clobber bootloader's buffer)
		uint64_t bit = (1ul << idx);
		if(fb_freemap & bit)
		{
			fb_freemap &= ~bit;
			m_spl_rel(&fb_spl);
			return m_video_fbaddr(idx);
		}
	}
	m_spl_rel(&fb_spl);
	return 0;
}

void fb_free(uintptr_t addr)
{
	m_spl_acq(&fb_spl);

	uint32_t bit = 0;
	for(uint32_t idx = 0; idx < m_video_fbcount(); idx++)
	{
		if(addr == m_video_fbaddr(idx))
		{
			//Found it
			bit = (1ul << idx);
			break;
		}
	}
	KASSERT(bit != 0);
	KASSERT((fb_freemap & bit) == 0);
	fb_freemap |= bit;

	m_spl_rel(&fb_spl);
	return;
}

int fb_flip(uintptr_t addr)
{	
	//Validate the buffer address
	int addridx = -1;
	for(uint32_t idx = 0; idx < m_video_fbcount(); idx++)
	{
		if(addr == m_video_fbaddr(idx))
		{
			addridx = idx;
			break;
		}
	}
	if(addridx == -1)
		return -EINVAL;
	
	//Put us on the list of waiters first, so we get unpaused if the framebuffer is changed while we check...
	waiters_addcur(fb_waiters, sizeof(fb_waiters));
	
	//Check if the buffer is already displayed
	if(m_video_getbuf() == addr)
		return 0;
	
	//Request the buffer change and notify us when it's done
	m_video_setbuf(addr);
	return -EAGAIN;
}

