//file.c
//Table of open files in kernel
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "kstd.h"
#include "file.h"
#include "ramfs.h"
#include "m_spl.h"
#include "kassert.h"
#include "sc.h"
#include "devs.h"

//All files currently open on the system.
#define FILE_MAX 1024
static file_t file_table[FILE_MAX];

//Gets a free entry from the file table. Returns it with the lock held.
static file_t *file_lockfree(void)
{
	for(int ff = 0; ff < FILE_MAX; ff++)
	{
		//Don't bother fighting over locks - a file entry already locked is already in use.
		file_t *fptr = &(file_table[ff]);
		if(m_spl_try(&(fptr->spl)))
		{
			if(fptr->refs == 0)
			{
				//Got an unused file entry locked.
				return fptr;
			}
			
			//File entry already used - let go of lock and keep looking
			m_spl_rel(&(fptr->spl));
		}
	}	
	
	//Didn't find any free entries.
	return NULL;
}

int file_make(file_t *dir, const char *name, mode_t mode, dev_t special, int access, file_t **file_out)
{
	if(dir == NULL)
		return -EBADF;
	
	if(!(dir->access & _SC_ACCESS_W))
		return -EBADF;
	
	if((dir->mode & _SC_S_IFMT) != _SC_S_IFDIR)
		return -ENOTDIR;
	
	if(name == NULL || name[0] == '\0')
		return -EINVAL;
	
	if((!strcmp(name, ".")) || (!strcmp(name, "..")))
		return -EINVAL;
	
	for(const char *nn = name; *nn != '\0'; nn++)
	{
		//Of the low 128 characters, only allow POSIX portable filenames
		if(*nn >= 'A' && *nn <= 'Z')
			continue;
		if(*nn >= 'a' && *nn <= 'z')
			continue;
		if(*nn >= '0' && *nn <= '9')
			continue;
		if(*nn == '.' || *nn == '_' || *nn == '-')
			continue;
		
		//Allow '[' because shells need to run that as a command sometimes
		if(*nn == '[')
			continue;
		
		//Allow upper-128 characters encoding UTF-8.
		if((uint8_t)(*nn) & 0x80)
			continue;
		
		return -EINVAL;
	}
	
	//Restrict creation of file-types to those we support
	switch(mode & _SC_S_IFMT)
	{
		case _SC_S_IFREG:
		case _SC_S_IFDIR:
		case _SC_S_IFCHR:
		case _SC_S_IFLNK:
		case _SC_S_IFIFO:
		case _SC_S_IFPTY:
			break;
		default:
			return -EINVAL;
	}
	
	//Make room for the new open file entry.
	file_t *newfile = file_lockfree();
	if(newfile == NULL)
		return -ENFILE;

	//Try to make the inode in the filesystem
	ramfs_lock();
	int ino_made = 0;
	int make_err = ramfs_make(dir->ino, name, mode, special, &ino_made);
	if(make_err < 0)
	{		
		ramfs_unlock();
		m_spl_rel(&(newfile->spl));
		return make_err;
	}
	
	//Made the inode - store reference to it.
	newfile->ino = ino_made;
	ramfs_inc(ino_made, access);
	
	//Done with RAM FS at this point
	ramfs_unlock();
	
	//If we're making a device node, make sure the device is OK with that. Close the file otherwise.
	if((mode & _SC_S_IFMT) == _SC_S_IFCHR)
	{
		int dev_err = devs_open(special);
		if(dev_err < 0)
		{
			//Device said "no"
			ramfs_lock();
			ramfs_dec(ino_made, access);
			ramfs_unlock();
			
			newfile->ino = 0;
			m_spl_rel(&(newfile->spl));
			return dev_err;
		}
	}
	
	newfile->mode = mode;
	newfile->special = special;
	newfile->off = 0;
	newfile->access = access;
	newfile->refs = 1;
	
	*file_out = newfile;
	return 0;
}

int file_find(file_t *dir, const char *name, int access, file_t **file_out)
{
	//Make sure we've got room for the new file.
	file_t *newfile = file_lockfree();
	if(newfile == NULL)
	{
		//No room in file table.
		return -ENFILE;
	}
	
	//See what we're trying to find...
	ramfs_lock(); //Lock before lookup so files don't disappear underneath us
	int found_ino = 0;
	if(name != NULL && name[0] == '/' && name[1] == '\0')
	{
		//Looking for "/". Directory doesn't matter. Always find inode 0.
		found_ino = 0;
	}
	else
	{
		//Other cases require a valid starting point to refer to.
		if(dir == NULL)
		{
			ramfs_unlock();
			m_spl_rel(&(newfile->spl));
			return -EINVAL;
		}
		
		if(name == NULL || name[0] == '\0')
		{
			//Looking for empty-string. Return another reference to the given file, even if it's not a dir.
			found_ino = dir->ino;
		}
		else
		{
			//Actually doing a directory lookup.
			int find_err = ramfs_find(dir->ino, name, &found_ino);			
			if(find_err < 0)
			{
				ramfs_unlock();
				m_spl_rel(&(newfile->spl));
				return find_err;
			}
		}
	}
		
	//Alright, found the file. Add a reference and get its info.
	newfile->ino = found_ino;
	ramfs_inc(found_ino, access);
	
	_sc_stat_t st;
	int stat_err = ramfs_stat(found_ino, &st);
	KASSERT(stat_err == 0);
	
	//Done with filesystem at this point - we hold reference to the file, it won't go away.
	ramfs_unlock();
	
	//If we're opening a device, make sure the device is OK with that. Close the file otherwise.
	if((st.mode & _SC_S_IFMT) == _SC_S_IFCHR)
	{
		int dev_err = devs_open(st.rdev);
		if(dev_err < 0)
		{
			//Device said "no"
			ramfs_lock();
			ramfs_dec(found_ino, access);
			ramfs_unlock();
			
			newfile->ino = 0;
			m_spl_rel(&(newfile->spl));
			return dev_err;
		}
	}

	newfile->mode = st.mode;
	newfile->special = st.rdev;
	newfile->off = 0;
	newfile->access = access;
	newfile->refs = 1;
	
	//Return the new file with one reference, still locked.
	*file_out = newfile;
	return 0;
}

int file_unlink(file_t *dir, const char *name, file_t *rmfile)
{
	if(dir == NULL)
		return -EBADF;
	
	if(!(dir->access & _SC_ACCESS_W))
		return -EBADF;
	
	if((dir->mode & _SC_S_IFMT) != _SC_S_IFDIR)
		return -ENOTDIR;
	
	ramfs_lock();
	int result = ramfs_unlink(dir->ino, name, (rmfile != NULL) ? rmfile->ino : 0);
	ramfs_unlock();
	
	return result;
}

ssize_t file_read(file_t *file, void *buf, ssize_t nbytes)
{
	if(!(file->access & _SC_ACCESS_R))
		return -EBADF;
	
	if((file->mode & _SC_S_IFMT) == _SC_S_IFCHR)
		return devs_read(file->special, buf, nbytes);
	
	ramfs_lock();
	ssize_t retval = ramfs_read(file->ino, file->off, buf, nbytes, (file->access & _SC_ACCESS_BACK));
	ramfs_unlock();
	
	if(retval < 0)
		return retval;
	
	file->off += retval;
	return retval;
}

ssize_t file_write(file_t *file, const void *buf, ssize_t nbytes)
{
	if(!(file->access & _SC_ACCESS_W))
		return -EBADF;
	
	if((file->mode & _SC_S_IFMT) == _SC_S_IFCHR)
		return devs_write(file->special, buf, nbytes);
	
	if((file->mode & _SC_S_IFMT) == _SC_S_IFDIR)
		return -EISDIR;
	
	ramfs_lock();
	ssize_t retval = ramfs_write(file->ino, file->off, buf, nbytes, (file->access & _SC_ACCESS_BACK));
	ramfs_unlock();
	
	if(retval < 0)
		return retval;
	
	file->off += retval;
	return retval;
}

off_t file_trunc(file_t *file, off_t size)
{
	if(!(file->access & _SC_ACCESS_W))
		return -EBADF;
	
	switch(file->mode & _SC_S_IFMT)
	{
		case _SC_S_IFDIR:
			return -EISDIR;
		case _SC_S_IFIFO:
		case _SC_S_IFPTY:
			return -ESPIPE;
		case _SC_S_IFCHR:
			return -ENOTTY;
		default:
			break;
	}
	
	ramfs_lock();
	off_t retval = ramfs_trunc(file->ino, size);
	ramfs_unlock();
	
	return retval;
}

int file_stat(file_t *file, _sc_stat_t *st)
{
	ramfs_lock();
	int retval = ramfs_stat(file->ino, st);
	ramfs_unlock();
	return retval;
}

off_t file_seek(file_t *file, off_t offset, int whence)
{
	switch(file->mode & _SC_S_IFMT)
	{
		case _SC_S_IFIFO:
		case _SC_S_IFCHR:
		case _SC_S_IFPTY:
			return -ESPIPE;
		default:
			break;
	}
	
	off_t whence_val = 0;
	switch(whence)
	{
		case _SC_SEEK_SET:
		{
			whence_val = 0;
			break;
		}
		case _SC_SEEK_CUR:
		{
			whence_val = file->off;
			break;
		}
		case _SC_SEEK_END:
		{
			_sc_stat_t st;
			ramfs_lock();
			int stat_err = ramfs_stat(file->ino, &st);
			ramfs_unlock();
			
			if(stat_err < 0)
				return stat_err;
			
			whence_val = st.size;
			break;
		}
		default:
			return -EINVAL;
	}
	
	file->off = whence_val + offset;
	if(file->off < 0)
		file->off = 0;
	
	return file->off;
}

void file_lock(file_t *file)
{
	m_spl_acq(&(file->spl));
}

void file_unlock(file_t *file)
{
	if(file->refs == 0)
	{
		if((file->mode & _SC_S_IFMT) == _SC_S_IFCHR)
			devs_close(file->special);
		
		ramfs_lock();
		ramfs_dec(file->ino, file->access);
		ramfs_unlock();
		
		file->ino = 0;
		file->off = 0;
		file->mode = 0;
		file->access = 0;
		file->special = 0;
	}
	
	m_spl_rel(&(file->spl));
}



