//file.h
//Table of open files in kernel
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#ifndef FILE_H
#define FILE_H

#include "kstd.h"
#include "m_spl.h"
#include <sc_file.h>

//Open file
typedef struct file_s
{
	m_spl_t spl; //Spinlock protecting the structure
	int refs; //Number of references to this file (0 means a free entry)
	
	int access; //Type of access opened for - reading, writing, or just stat (0)
	
	ino_t ino; //Inode we refer to
	off_t off; //File pointer - where in the file we will next read/write
	mode_t mode; //Mode of file at time of open
	dev_t special; //Special-number of file at time of open (dev number if character/block dev)
	
} file_t;


//Makes a new file. Outputs a pointer to the new open file, with the lock held, and one reference.
int file_make(file_t *dir, const char *name, mode_t mode, dev_t special, int access, file_t **file_out);

//Finds an existing file. Outputs a pointer to the new open file, with the lock held, and one reference.
//Pass NULL for "dir" to search the root directory. Pass empty-string as "name" to get a reference to the same file.
//Returns 0 on success or a negative error number.
int file_find(file_t *dir, const char *name, int access, file_t **file_out);

//Unlinks a file from the given directory file.
int file_unlink(file_t *dir, const char *name, file_t *rmfile);

//Reads data from the open file.
ssize_t file_read(file_t *file, void *buf, ssize_t nbytes);

//Writes data into the open file.
ssize_t file_write(file_t *file, const void *buf, ssize_t nbytes);

//Changes the size of the given open file.
off_t file_trunc(file_t *file, off_t size);

//Returns status information about the given open file.
int file_stat(file_t *file, _sc_stat_t *st);

//Changes the file-pointer in the given open file, altering where the next read or write occurs.
off_t file_seek(file_t *file, off_t offset, int whence);

//Acquires the lock on the given file.
void file_lock(file_t *file);

//Releases the lock on a file. Frees it if there are no references when unlocked.
void file_unlock(file_t *file);

#endif //FILE_H
