//mem.c
//User memory space tracking
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "mem.h"
#include "kassert.h"
#include "fb.h"
#include "m_frame.h"
#include "m_uspc.h"
#include "kstd.h"

//Bubble-sorts a newly inserted memory region in the given memory space.
void mem_bubble(mem_t *mem, int placed)
{
	int sptr_idx = placed;
	while(1)
	{
		if(sptr_idx > 0)
		{
			if((mem->segs[sptr_idx-1].size == 0) || (mem->segs[sptr_idx].vaddr < mem->segs[sptr_idx-1].vaddr))
			{
				mem_seg_t temp = mem->segs[sptr_idx];
				mem->segs[sptr_idx] = mem->segs[sptr_idx - 1];
				mem->segs[sptr_idx - 1] = temp;
				sptr_idx--;
				continue;
			}
		}
		
		if(sptr_idx < MEM_SEG_MAX - 1)
		{
			if((mem->segs[sptr_idx+1].size != 0) && (mem->segs[sptr_idx].vaddr > mem->segs[sptr_idx+1].vaddr))
			{
				mem_seg_t temp = mem->segs[sptr_idx];
				mem->segs[sptr_idx] = mem->segs[sptr_idx + 1];
				mem->segs[sptr_idx + 1] = temp;
				sptr_idx++;
				continue;
			}
		}
		
		break;
	}
	return;
}

void mem_clear(mem_t *mem)
{
	size_t pagesize = m_frame_size();
	
	for(int ss = 0; ss < MEM_SEG_MAX; ss++)
	{
		uintptr_t start = mem->segs[ss].vaddr;
		uintptr_t end = start + mem->segs[ss].size;

		if(end > start)
		{
			KASSERT(start % pagesize == 0);
			KASSERT(end % pagesize == 0);
			KASSERT(mem->uspc != 0);
			
			for(uintptr_t pp = start; pp < end; pp += pagesize)
			{
				uintptr_t frame = m_uspc_get(mem->uspc, pp);
				KASSERT(frame != 0);
				m_uspc_set(mem->uspc, pp, 0, 0);
				
				if(mem->segs[ss].disp == 0)
				{
					//Normal memory - return to frame allocator
					m_frame_free(frame);
				}
			}
		}
		
		if(mem->segs[ss].disp != 0)
		{
			//Return framebuffer to framebuffer allocator
			fb_free(mem->segs[ss].disp);
			mem->segs[ss].disp = 0;
		}
			
		mem->segs[ss].vaddr = 0;
		mem->segs[ss].size = 0;
		mem->segs[ss].prot = 0;
	}
		
	if(mem->uspc != 0)
	{
		m_uspc_delete(mem->uspc);
		mem->uspc = 0;
	}
}

int mem_add(mem_t *mem, uintptr_t vaddr, size_t size, int prot)
{	
	//Location and size must be page-aligned
	size_t pagesize = m_frame_size();
	if(vaddr % pagesize != 0)
		return -EINVAL;
	
	if(size % pagesize != 0)
		size += pagesize - (size % pagesize);
	
	//Find place to store the segment
	mem_seg_t *sptr = NULL;
	int sptr_idx = -1;
	for(int ss = 0; ss < MEM_SEG_MAX; ss++)
	{
		if(mem->segs[ss].size == 0)
		{
			sptr = &(mem->segs[ss]);
			sptr_idx = ss;
			break;
		}
	}
	if(sptr == NULL)
	{
		//No room for more segment references
		return -EMFILE;
	}
	
	//Make the userspace paging structures, if none exist
	if(mem->uspc == 0)
	{
		mem->uspc = m_uspc_new();
		if(mem->uspc == 0)
		{
			//Failed to allocate paging structures.
			return -ENOMEM;
		}
	}
	
	//Make sure the virtual space is free
	for(uintptr_t pp = vaddr; pp < vaddr + size; pp += pagesize)
	{
		uintptr_t oldframe = m_uspc_get(mem->uspc, pp);
		if(oldframe != 0)
			return -EBUSY;
	}
	
	//Try to fill the requested range
	for(uintptr_t pp = vaddr; pp < vaddr + size; pp += pagesize)
	{
		uintptr_t newframe = m_frame_alloc();
		if(newframe != 0)
		{
			bool mapped = m_uspc_set(mem->uspc, pp, newframe, prot);
			if(mapped)
			{
				//Success - keep going
				continue;
			}
		}
		
		//Failed to allocate or map a frame here. Unwind and fail.
		if(newframe != 0)
		{
			m_frame_free(newframe);
			newframe = 0;
		}
		
		while(pp > vaddr)
		{
			pp -= pagesize;
			uintptr_t unwind = m_uspc_get(mem->uspc, pp);
			KASSERT(unwind != 0);
			m_uspc_set(mem->uspc, pp, 0, 0);
			m_frame_free(unwind);
		}
		
		return -ENOMEM;
	}
	
	//Success. Insert into the free index.
	sptr->vaddr = vaddr;
	sptr->size = size;
	sptr->prot = prot;
	sptr->disp = 0; //Not a framebuffer
	
	//Bubble-sort with the other indexes, to keep them in-order.
	mem_bubble(mem, sptr_idx);
	
	//Combine adjacent entries
	for(int ss = 0; ss < MEM_SEG_MAX - 1; ss++)
	{
		mem_seg_t *first = &(mem->segs[ss]);
		mem_seg_t *second = &(mem->segs[ss + 1]);
		if((first->size == 0) || (second->size == 0))
			continue; //Don't have two adjacent entries
		
		if(first->disp || second->disp)
			continue; //Can't combine displayable regions
		
		if(first->prot != second->prot)
			continue; //Different permissions
		
		if(first->vaddr + first->size != second->vaddr)
			continue; //Not contiguous
		
		//Alright, we found two contiguous segments. Combine them and get rid of the second entry.
		first->size += second->size;
		for(int bump = ss + 1; bump < MEM_SEG_MAX - 1; bump++)
		{
			mem->segs[bump] = mem->segs[bump+1];
		}
		memset(&(mem->segs[MEM_SEG_MAX-1]), 0, sizeof(mem->segs[MEM_SEG_MAX-1]));
	}

	return 0;
}

int mem_disp(mem_t *mem, uintptr_t vaddr, size_t size)
{
	//Location must be page-aligned
	size_t pagesize = m_frame_size();
	if(vaddr % pagesize != 0)
		return -EINVAL;
	
	if(size % pagesize != 0)
		size += pagesize - (size % pagesize);
	
	//Always read/write
	const int prot = M_USPC_PROT_R | M_USPC_PROT_W;
	
	//Find place to store the segment
	mem_seg_t *sptr = NULL;
	int sptr_idx = -1;
	for(int ss = 0; ss < MEM_SEG_MAX; ss++)
	{
		if(mem->segs[ss].size == 0)
		{
			sptr = &(mem->segs[ss]);
			sptr_idx = ss;
			break;
		}
	}
	if(sptr == NULL)
	{
		//No room for more segment references
		return -EMFILE;
	}
	
	//Make the userspace paging structures, if none exist
	if(mem->uspc == 0)
	{
		mem->uspc = m_uspc_new();
		if(mem->uspc == 0)
		{
			//Failed to allocate paging structures.
			return -ENOMEM;
		}
	}
	
	//Make sure the virtual space is free
	for(uintptr_t pp = vaddr; pp < vaddr + size; pp += pagesize)
	{
		uintptr_t oldframe = m_uspc_get(mem->uspc, pp);
		if(oldframe != 0)
			return -EBUSY;
	}
	
	//Allocate a framebuffer to use, of at least the given size
	uintptr_t fbaddr = fb_alloc(size);
	if(fbaddr == 0)
	{
		//No more framebuffers available
		return -ENOMEM;
	}
	
	//Fill the requested range with the given framebuffer
	for(uintptr_t pp = vaddr; pp < vaddr + size; pp += pagesize)
	{
		uintptr_t paddr = fbaddr + (pp - vaddr);
		KASSERT(paddr != 0);

		bool mapped = m_uspc_set(mem->uspc, pp, paddr, prot);
		if(mapped)
		{
			//Success - keep going
			continue;
		}
	
		//Failed to map a frame here (likely out of pagetable space). Unwind and fail.
		while(pp > vaddr)
		{
			pp -= pagesize;
			uintptr_t unwind = m_uspc_get(mem->uspc, pp);
			KASSERT(unwind != 0);
			m_uspc_set(mem->uspc, pp, 0, 0);
		}
		
		fb_free(fbaddr);
		return -ENOMEM;
	}
	
	//Success. Insert into the free index, and note that it's a framebuffer.
	sptr->vaddr = vaddr;
	sptr->size = size;
	sptr->prot = prot;
	sptr->disp = fbaddr;
	
	//Bubble-sort with the other indexes, to keep them in-order.
	mem_bubble(mem, sptr_idx);
	
	return 0;
}

intptr_t mem_avail(mem_t *mptr, uintptr_t around, size_t size)
{
	if(size <= 0)
		return EINVAL;
	
	size_t pagesize = m_frame_size();
	if(size % pagesize != 0)
		size += pagesize - (size % pagesize);
	
	//Get overall bounds of userspace
	uintptr_t uspc_start = 0;
	uintptr_t uspc_end = 0;
	m_uspc_range(&uspc_start, &uspc_end);
	
	KASSERT((size % pagesize) == 0);
	KASSERT((uspc_start % pagesize) == 0);
	KASSERT((uspc_end % pagesize) == 0);
	
	//Segments are stored in-order.
	//This implies that the free spaces can be found in the gaps between them as stored.
	//Search the free regions for a region of at least the given size, closest to the given address.
	
	//If there's nothing at all mapped, then we have a trivial problem
	if((mptr->segs[0].vaddr + mptr->segs[0].size) == 0)
	{
		if( (around >= uspc_start) && ((around + size) <= uspc_end) )
			return around;
		else if(size <= uspc_end - uspc_start)
			return uspc_start;
		else
			return 0;
	}
	
	//Okay, we have at least one segment mapped already.
	//There's more than one place we might put this. Search for the area closest to the requested address.
	uintptr_t best_start = 0;
	uintptr_t best_diff = ~0ul;
	
	//Consider the gaps around each segment
	for(int gg = 0; gg < MEM_SEG_MAX + 1; gg++)
	{
		//Gap gg is the space between memory segments gg-1 and gg
		//If gg is 0 then it's the space from the beginning to segment gg (the first segment)
		//If gg is MEM_SEG_MAX then it's the space from segment gg-1 (the last segment) to the end
		uintptr_t gap_start = (gg == 0) ? uspc_start : ((mptr->segs[gg-1].vaddr) + (mptr->segs[gg-1].size));
		uintptr_t gap_end = ( (gg == MEM_SEG_MAX) || (mptr->segs[gg].size == 0) ) ? uspc_end : (mptr->segs[gg].vaddr);
		
		//Check if the gap is big enough for the proposed region at all
		if(gap_end - gap_start >= size)
		{
			//See what placement would be closest to the proposed address
			uintptr_t best_start_in_gap = 0;
			if(around < gap_start)
			{
				//Wanted an address before the gap - closest we'll get is the beginning
				best_start_in_gap = gap_start;
			}
			else if(around > gap_end - size)
			{
				//Wanted a range that ends after the gap - closest we'll get is the end
				best_start_in_gap = gap_end - size;
			}
			else
			{
				//Can satisfy exactly the request in this gap
				best_start_in_gap = around;
			}
			
			uintptr_t diff = (best_start_in_gap > around) ? (best_start_in_gap - around) : (around - best_start_in_gap);
			if(diff < best_diff)
			{
				best_start = best_start_in_gap;
				best_diff = diff;
			}
		}
		
		if(mptr->segs[gg].size == 0)
			break;
	}
	
	if(best_start == 0)
		return -ENOMEM;
	
	return (intptr_t)best_start;
}

int mem_copy(mem_t *dst, const mem_t *src)
{
	size_t pagesize = m_frame_size();
		
	for(int ss = 0; ss < MEM_SEG_MAX; ss++)
	{
		const mem_seg_t *sptr = &(src->segs[ss]);
		if(sptr->size <= 0)
			continue;
		
		if(sptr->disp != 0)
		{
			//This is a displayable memory. We need to get another framebuffer to copy it.
			int alloc_err = mem_disp(dst, sptr->vaddr, sptr->size);
			if(alloc_err < 0)
			{
				mem_clear(dst);
				return alloc_err;
			}
		}
		else
		{
			//This is normal memory. We just need to allocate it.
			int alloc_err = mem_add(dst, sptr->vaddr, sptr->size, sptr->prot);
			if(alloc_err < 0)
			{
				mem_clear(dst);
				return alloc_err;
			}
		}
		
		//Copy all the frames of the segment, now that we've got a segment in the new space.
		for(uintptr_t ff = sptr->vaddr; ff < sptr->vaddr + sptr->size; ff += pagesize)
		{
			uintptr_t old_frame = m_uspc_get(src->uspc, ff);
			uintptr_t new_frame = m_uspc_get(dst->uspc, ff);
			
			KASSERT(old_frame != 0);
			KASSERT(new_frame != 0);
			
			m_frame_copy(new_frame, old_frame);
		}
	}
	
	return 0;
}

int mem_validate(mem_t *mem, uintptr_t vaddr, size_t size, int prot)
{
	if(size == 0)
		return 0;
	
	//Work through memory ranges until we find appropriate coverage for this whole range.
	uintptr_t need_start = vaddr;
	uintptr_t need_end = vaddr + size;
	for(int ss = 0; ss < MEM_SEG_MAX; ss++)
	{
		const mem_seg_t *sptr = &(mem->segs[ss]);
		if(sptr->size <= 0)
			continue; //No segment in this entry
		
		int perm_match = ((sptr->prot & prot) == prot);
		if(!perm_match)
			continue; //Segment doesn't have the permission we're looking for
		
		//Okay, this segment would be appropriate. Does it overlap the range we want?
		uintptr_t seg_start = sptr->vaddr;
		uintptr_t seg_end = sptr->vaddr + sptr->size;
		if(seg_start <= need_start && seg_end >= need_end)
			return 0; //Segment covers the whole remaining range - success
		
		if(seg_end <= need_start)
			continue; //Segment ends before the range we care about
		
		if(seg_start >= need_end)
			continue; //Segment starts after the range we care about
		
		//Segment partially covers the area we need...
		//Should we actually allow this case? I can't see where it's reasonable.
		//Who's rightly copying stuff in the middle of a segment boundary?
		//if(seg_start <= need_start && seg_end < need_start)
		//	need_start = seg_end;
		//else if(seg_end >= need_end && need_end > seg_start)
		//	need_end = seg_start;
	}
	
	//Didn't find segments to cover the whole requested range.
	return -EFAULT;
}

int mem_valstr(mem_t *mem, const char *strptr)
{
	//We read the memory to find the end of the string, so it has to be in-use currently.
	KASSERT(mem->uspc == m_uspc_current());
	
	//Start looking for the beginning of the string.
	uintptr_t needaddr = (uintptr_t)strptr;
	int len = 0;
	
	//Segments are stored in-order, so linear-search them
	for(int ss = 0; ss < MEM_SEG_MAX; ss++)
	{
		const mem_seg_t *sptr = &(mem->segs[ss]);
		if(sptr->size <= 0)
			continue; //No segment here
		
		if(!(sptr->prot & M_USPC_PROT_R))
			continue; //Not readable
		
		uintptr_t seg_start = sptr->vaddr;
		uintptr_t seg_end = seg_start + sptr->size;
		
		if(seg_start > needaddr)
		{
			//This segment starts after the address we looked for.
			//We must have already passed it.
			return -EFAULT;
		}
		
		if(seg_end <= needaddr)
		{
			//This segment ends before the address we looked for.
			//Keep looking.
			continue;
		}
		
		//Okay, the segment overlaps the address we're looking for. We can see if this is the end of the string.
		while(1)
		{
			char ch = *((char*)(needaddr));
			if(ch == '\0')
			{
				//This is the end of the string and we read it all successfully.
				return len;
			}
			
			//We read this far, but the string keeps going.
			len++;
			needaddr++;
			
			//Are we still in the same segment?
			//Technically we could allow this and keep going (i.e. 'break;' instead of 'return -EFAULT;').
			//However, I don't think it's reasonable for the user to expect that this works.
			//There's no decent reason to be reading a string across a segment boundary.
			if(needaddr >= seg_end)
				return -EFAULT;
		}
	}
	
	//Didn't find the segment containing the string
	return -EFAULT;
}
