//mem.h
//User memory space tracking
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#ifndef MEM_H
#define MEM_H

#include "kstd.h"
#include "m_uspc.h"

//One segment of memory allocated in a memory space
typedef struct mem_seg_s
{
	uintptr_t vaddr; //Virtual address where the segment begins
	size_t size; //Total size, in bytes, of the segment
	int prot; //Access permissions for the user program when accessing the segment
	uintptr_t disp; //If this is a displayable memory region, its physical address
	
} mem_seg_t;

//Memory space
typedef struct mem_s
{
	//Kernel-side tracking of which segments are allocated
	#define MEM_SEG_MAX 16
	mem_seg_t segs[MEM_SEG_MAX];
	
	//Machine-specific paging structures
	m_uspc_t uspc;
} mem_t;

//Frees all memory in a memory space. The memory space can then be zeroed.
void mem_clear(mem_t *mem);

//Allocates new memory and adds it to a memory space.
int mem_add(mem_t *mem, uintptr_t vaddr, size_t size, int prot);

//Allocates displayable memory and adds it to a memory space.
int mem_disp(mem_t *mem, uintptr_t vaddr, size_t size);

//Returns a free address where the given amount of bytes could be mapped.
intptr_t mem_avail(mem_t *mem, uintptr_t around, size_t size);

//Copies a memory space.
int mem_copy(mem_t *dst, const mem_t *src);

//Checks whether access is permitted to the entire given range.
//Returns 0 on success or a negative error number.
int mem_validate(mem_t *mem, uintptr_t vaddr, size_t size, int prot);

//Checks whether a string could be read from the given-and-ACTIVE memory space.
//Returns the length of the string on success, or a negative error number.
int mem_valstr(mem_t *mem, const char *strptr);

#endif //MEM_H
