//namei.c
//Iterative directory lookups
//Bryan E. Topp <betopp@betopp.com> 2022

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "m_spl.h"
#include "kstd.h"
#include "file.h"
#include "kassert.h"
#include "process.h"

//Buffer for namei path being constructed.
//Grows when a symlink is encountered and inserted.
#define NAMEI_PATH_MAX 32768
static char namei_path_buf[NAMEI_PATH_MAX];
static int namei_path_pos;

//How many links have been followed
static int namei_followed;

//Type of access for lookup
static int namei_access;

//Directory to resolve next step in
static file_t *namei_dir; 

//Spinlock so only one thread accesses these
static m_spl_t namei_spl; 

//Performs one step in intermediate pathname resolution.
int namei_step(void)
{	
	//Skip leading slashes. We already determined whether we want relative or absolute paths.
	while(namei_path_buf[namei_path_pos] == '/')
		namei_path_pos++;
	
	//Pull the next pathname component out
	char *remain = namei_path_buf + namei_path_pos;
	char *nextslash = strchr(remain, '/');
	size_t complen = 0;
	if(nextslash != NULL)
	{
		//There's still a slash in the path. Extract characters up to the next slash.
		complen = nextslash - remain;
	}
	else
	{
		//No further slashes. The next pathname component is the whole remaining path.
		complen = strlen(remain);
	}
	
	char component[256];
	if(complen >= sizeof(component))
	{
		//Pathname component was too long
		return -ENAMETOOLONG;
	}
	memcpy(component, remain, complen);
	component[complen] = '\0';
	
	//Look up that component - even if it's the final one.
	//This is necessary because the final component, as we know it now, may be a symlink.
	file_t *found = NULL;
	int find_result = file_find(namei_dir, component, 0x7, &found);
	if(find_result < 0 && find_result != -ENOENT)
		return find_result; //Actual error in lookup
	
	if(found != NULL)
		found->refs = 0;
	
	//If this is the final component at the moment, it may be the one we return to the caller.
	if(nextslash == NULL)
	{
		if(find_result < 0)
		{
			//Final component and didn't find it
			return 0; 
		}
		
		if(namei_access & _SC_ACCESS_NOFOLLOW)
		{
			//Final component and we don't want to follow a link
			file_unlock(found);
			return 0;
		}
		
		if((found->mode & _SC_S_IFMT) != _SC_S_IFLNK)
		{
			//Final component and it's not a link
			file_unlock(found);
			return 0;
		}
	}
	
	//Component is not the one the caller will look up. Consume it.
	namei_path_pos += complen;
	
	//At this point, we need some way to continue looking up paths
	if(find_result < 0)
		return find_result;
	
	switch(found->mode & _SC_S_IFMT)
	{
		case _SC_S_IFLNK: //symlink
		{
			//Make sure we're not in a symlink loop
			namei_followed++;
			if(namei_followed > 10)
			{
				file_unlock(found);
				return -ELOOP;
			}
			
			//Continue by inserting the target of the symlink in place of the symlink's name.
			
			//Check if the resulting path will fit.
			_sc_stat_t st = {0};
			int stat_err = file_stat(found, &st);
			KASSERT(stat_err == 0);
			size_t symlink_target_len = st.size;
			size_t after_symlink_len = strlen(namei_path_buf + namei_path_pos);
			if(symlink_target_len + after_symlink_len >= sizeof(namei_path_buf))
			{
				//No room for the target of the link + whatever else remained
				file_unlock(found);
				return -ENAMETOOLONG;
			}
			
			//Scoot the remaining path down, to allow for the link
			memmove(namei_path_buf + symlink_target_len, namei_path_buf + namei_path_pos, after_symlink_len);
			namei_path_buf[symlink_target_len + after_symlink_len] = '\0';
			
			//Read the link in after it
			int link_read = file_read(found, namei_path_buf, (int)symlink_target_len);
			KASSERT(link_read == (int)symlink_target_len);
			
			//Continue at the beginning - with the contents of the link
			namei_path_pos = 0;
			file_unlock(found);
			return 1;
		}
			
		case _SC_S_IFDIR: //directory
		{
			//Continue by looking up in this directory next time.
			file_unlock(namei_dir);
			namei_dir = found;
			return 1;
		}

		default:
		{
			//Can't continue otherwise
			file_unlock(found);
			return -ENOTDIR;
		}
	}
}

//Looks up intermediate components of the given path.
//Returns the directory and name of the final lookup.
int namei(file_t *dir, const char *path, int access, file_t **dir_out, char *name_out256)
{
	KASSERT(dir != NULL);
	KASSERT(path != NULL);
	KASSERT(dir_out != NULL);
	KASSERT(name_out256 != NULL);
	
	m_spl_acq(&namei_spl);
	
	//Fill the top-level path to look up
	int copyin = process_strget(namei_path_buf, path, sizeof(namei_path_buf));
	if(copyin < 0)
	{
		//Failed to get the user's path string
		m_spl_rel(&namei_spl);
		return copyin;
	}
	namei_path_pos = 0;
	namei_followed = 0;
	namei_access = access;
	
	//Look up from the root directory if the given path begins with '/'.
	//Otherwise, use the specified directory as the starting point.
	if(namei_path_buf[0] == '/')
	{		
		int foundroot = file_find(NULL, "/", access, &namei_dir);
		KASSERT(foundroot == 0);
		namei_dir->refs = 0;
		
		//Skip the '/'
		namei_path_pos = 1;
	}
	else
	{
		int founddir = file_find(dir, "", access, &namei_dir);
		KASSERT(founddir == 0);
		namei_dir->refs = 0;
		
		namei_path_pos = 0;
	}
	
	//Iterate until there's no further paths to look up.
	while(1)
	{
		//Perform one lookup
		int stepped = namei_step();
		
		//Always return an updated directory pointer, so the caller can unlock it
		*dir_out = namei_dir;
		
		if(stepped < 0)
		{
			//Error while following the path.
			m_spl_rel(&namei_spl);
			return stepped;
		}
		
		if(stepped == 0)
		{
			//Success. Copy out the last pathname component.
			strncpy(name_out256, namei_path_buf + namei_path_pos, 256);
			m_spl_rel(&namei_spl);
			return 0;
		}
	}
}
