//process.c
//Process tracking in kernel
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "process.h"
#include "thread.h"
#include "kassert.h"
#include "elf.h"
#include "argenv.h"
#include "thread.h"
#include "m_tls.h"
#include "kstd.h"

process_t process_table[PROCESS_MAX];

void process_init(void)
{
	//Make initial process entry
	process_t *pptr = &(process_table[1]);
	m_spl_acq(&(pptr->spl));
	
	pptr->pid = 1;
	pptr->state = PROCESS_STATE_ALIVE;
	pptr->conchild = 1;
	
	//Open the init executable, and put a reference in our process's FD 0
	file_t *root_file = NULL;
	int root_find_err = file_find(NULL, "/", 7, &root_file);
	KASSERT(root_find_err == 0);
	
	file_t *bin_file = NULL;
	int bin_find_err = file_find(root_file, "bin", 7, &bin_file);
	KASSERT(bin_find_err == 0);
	
	file_t *init_file = NULL;
	int init_find_err = file_find(bin_file, "sterm", 7, &init_file);
	KASSERT(init_find_err == 0);
	
	pptr->pwd = root_file;
	root_file->refs = 1;
	file_unlock(root_file);
	root_file = NULL;
	
	bin_file->refs = 0;
	file_unlock(bin_file);
	bin_file = NULL;
	
	pptr->fds[0].file = init_file;
	file_unlock(init_file);
	
	//Simulate a call to exec, to load up the process with the init ELF
	uintptr_t entry = 0;
	int elf_err = elf_load(init_file, &(pptr->mem), &entry);
	KASSERT(elf_err == 0);
	
	uintptr_t argenv_loaded = 0;
	//int argenv_err = argenv_load(&(pptr->mem), (char*[]){"sinit", NULL}, (char*[]){"stump=1", NULL}, &argenv_loaded);
	//KASSERT(argenv_err == 0);
	
	//Make thread to run initial process
	thread_t *init_thread = NULL;
	int thread_err = thread_new(pptr, &init_thread);
	KASSERT(thread_err == 0);
	
	m_drop_entry(&(init_thread->drop), entry); //Single entry point
	m_drop_param(&(init_thread->drop), 0, 0); //Parameter 0 - which signal is being entered (0 for startup)
	m_drop_param(&(init_thread->drop), 1, argenv_loaded); //Parameter 1 on startup - argv/envp location
	
	pptr->entry = entry;
	
	thread_unlock(init_thread);
	process_unlock(pptr);
}

process_t *process_lockfree(void)
{
	for(int pp = 1; pp < PROCESS_MAX; pp++)
	{
		process_t *pptr = &(process_table[pp]);
		if(m_spl_try(&(pptr->spl)))
		{
			if(pptr->state == PROCESS_STATE_NONE)
			{
				//Found a free slot. Make sure it's got a new, valid ID corresponding to its place in the table.
				pptr->pid += PROCESS_MAX;
				if((pptr->pid <= 0) || ((pptr->pid % PROCESS_MAX) != pp))
					pptr->pid = pp;
				
				return pptr; //Still locked
			}
			
			//In use - keep looking
			m_spl_rel(&(pptr->spl));
		}
	}
	
	//No room
	return NULL;
}

process_t *process_lockpid(pid_t pid)
{
	if(pid < 0)
		return NULL;
	
	//Processes are ID'd based on their array index, so we know where this ID must be.
	process_t *pptr = &(process_table[pid % PROCESS_MAX]);
	m_spl_acq(&(pptr->spl));
	
	if((pptr->state == PROCESS_STATE_NONE) || (pptr->pid != pid))
	{
		//Wrong/no process here
		m_spl_rel(&(pptr->spl));
		return NULL;
	}
	
	//Got it, keep locked
	return pptr;
}

process_t *process_lockcur(void)
{
	thread_t *tptr = (thread_t*)(m_tls_get());	
	process_t *pptr = tptr->process;
	
	m_spl_acq(&(pptr->spl));
	return pptr;
}

pid_t process_curpid(void)
{
	//No need to lock, as a thread never changes process and a process never changes ID
	return ((thread_t*)(m_tls_get()))->process->pid;	
}

void process_unlock(process_t *process)
{
	m_spl_rel(&(process->spl));
}

file_t *process_lockfd(int fd, bool allow_pwd)
{
	if(fd < -1 || fd >= PROCESS_FD_MAX)
		return NULL;
	
	if(fd < 0 && !allow_pwd)
		return NULL;
	
	process_t *pptr = process_lockcur();
	KASSERT(pptr != NULL);
	
	file_t *fptr = (fd < 0) ? pptr->pwd : pptr->fds[fd].file;
	if(fptr != NULL)
		file_lock(fptr);
	
	process_unlock(pptr);
	return fptr;
}

int process_addfd(file_t *newfile)
{
	KASSERT(newfile != NULL);
	
	process_t *pptr = process_lockcur();
	KASSERT(pptr != NULL);
	
	for(int ff = 0; ff < PROCESS_FD_MAX; ff++)
	{
		if(pptr->fds[ff].file == NULL)
		{
			pptr->fds[ff].file = newfile;
			pptr->fds[ff].flags = 0;
			process_unlock(pptr);
			return ff;
		}
	}
	
	process_unlock(pptr);
	return -EMFILE;
}

int process_strget(char *kbufptr, const char *uptr, size_t kbuflen)
{
	process_t *pptr = process_lockcur();
	
	size_t strl = 0;
	while(1)
	{
		//Check before we read, that we can safely access the next byte
		//(Todo - could optimize a lot)
		int next_byte_valid = mem_validate(&(pptr->mem), ((uintptr_t)uptr) + strl, 1, M_USPC_PROT_R);
		if(next_byte_valid < 0)
		{
			process_unlock(pptr);
			return next_byte_valid;
		}
		
		if(uptr[strl] == '\0')
			break;
		else
			strl++;
	}
	
	//Okay, found the length of the string and it's all valid.
	//Is it too long for the kernel buffer?
	if(strl >= kbuflen) //Remember we need a NUL at the end
	{
		process_unlock(pptr);
		return -ENAMETOOLONG;
	}
	
	//The string is all accessible in the user process, and fits into the kernel buffer. Take it.
	memcpy(kbufptr, uptr, strl);
	kbufptr[strl] = '\0';
	
	//Success
	process_unlock(pptr);
	return 0;
}

int process_memput(void *ubufptr, const void *kbufptr, size_t len)
{
	process_t *pptr = process_lockcur();
	int valid = mem_validate(&(pptr->mem), (uintptr_t)ubufptr, len, M_USPC_PROT_W);
	if(valid < 0)
	{
		//Not allowed to write here
		process_unlock(pptr);
		return valid;
	}
	
	//Okay, process has writable memory here. Do the copy.
	memcpy(ubufptr, kbufptr, len);
	
	//Success
	process_unlock(pptr);
	return 0;
}

int process_memget(void *kbufptr, const void *ubufptr, size_t len)
{
	process_t *pptr = process_lockcur();
	int valid = mem_validate(&(pptr->mem), (uintptr_t)ubufptr, len, M_USPC_PROT_R);
	if(valid < 0)
	{
		//Not allowed to read from here
		process_unlock(pptr);
		return valid;
	}
	
	//Okay, process has readable memory here. Do the copy.
	memcpy(kbufptr, ubufptr, len);
	
	//Success
	process_unlock(pptr);
	return 0;
}

bool process_lineage(pid_t younger, pid_t elder)
{
	//Follow parent links back up from younger - see if we reach elder.
	pid_t iter = younger;
	for(int steps = 0; steps < PROCESS_MAX; steps++)
	{
		if(iter == elder)
		{
			//Reached the given elder - the younger was a descendant.
			return true;
		}
		
		process_t *pptr = process_lockpid(iter);
		if(pptr == NULL)
		{
			//Reached an invalid process ID
			return false;
		}
		
		//Move up the chain
		iter = pptr->ppid;
		process_unlock(pptr);
	}
	return false;
}
