//process.h
//Process tracking in kernel
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#ifndef PROCESS_H
#define PROCESS_H

#include "m_spl.h"
#include "m_uspc.h"
#include "file.h"
#include "mem.h"
#include "kstd.h"

//States a process can be in
typedef enum process_state_e
{
	PROCESS_STATE_NONE = 0,
	
	PROCESS_STATE_ALIVE, //Process wants to continue, has threads
	PROCESS_STATE_DETHREAD, //Some thread has caused process to exit, but others remain
	PROCESS_STATE_DEAD, //All threads have exited, process can be waited upon
	
	PROCESS_STATE_MAX
	
} process_state_t;

//File descriptor in a process
typedef struct process_fd_s
{
	file_t *file; //File described by this descriptor
	int flags; //Flags about the descriptor (i.e. close-on-exec or not)	
} process_fd_t;

//Process control block
typedef struct process_s
{
	m_spl_t spl; //Spinlock protecting the process control block
	
	//Basic information
	process_state_t state; //State of process
	int32_t scver; //System-call interface version being used
	pid_t pid; //ID of the process
	pid_t ppid; //ID of the parent of the process
	int32_t nthreads; //Number of threads that belong to this process
	int wstatus; //Status information available for parent to wait() on
	
	//Files
	#define PROCESS_FD_MAX 32 //Max file descriptors per process
	process_fd_t fds[PROCESS_FD_MAX]; //File descriptors
	file_t *pwd; //Present working directory
	
	//Memory
	mem_t mem; //Memory space that threads in this process use.
	mem_t mem_attempt; //Spare memory space if we're doing an exec, in case we fail.
	uint32_t entry; //Entry point for start of program, or threads taking signals
	
	//Console
	int conchild; //Which process should have the console, according to this one
	int conflags; //Information about console usage of this process
	
} process_t;

//All processes on system
#define PROCESS_MAX 32
extern process_t process_table[PROCESS_MAX];

//Sets up process tracking and initial process entry.
void process_init(void);

//Locks an empty entry in the process table and returns a pointer to it.
process_t *process_lockfree(void);

//Locks and returns a pointer to a process by PID. Returns NULL if it doesn't exist.
process_t *process_lockpid(pid_t pid);

//Locks the current process and returns a pointer to it.
process_t *process_lockcur(void);

//Returns the current process ID
pid_t process_curpid(void);

//Releases the lock on the given process.
void process_unlock(process_t *process);

//Looks up a file descriptor in the current process and locks the file.
//Returns NULL if no file is present there.
//Optionally allows specifying the PWD by passing -1.
file_t *process_lockfd(int fd, bool allow_pwd);

//Attempts to insert a file into the current process. Returns its file descriptor number or a negative error number.
//Does not alter the file's reference count.
int process_addfd(file_t *newfile);

//Attempts to copy a string from the current process into the kernel.
int process_strget(char *kbufptr, const char *uptr, size_t kbuflen);

//Attempts to copy a buffer from the kernel into the current process.
int process_memput(void *ubufptr, const void *kbufptr, size_t len);

//Attempts to copy a buffer from the current process into the kernel.
int process_memget(void *kbufptr, const void *ubufptr, size_t len);

//Returns whether one process (younger) is descended from another (elder)
bool process_lineage(pid_t younger, pid_t elder);

#endif //PROCESS_H
