//ramfs.c
//In-memory filesystem
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "ramfs.h"
#include "kassert.h"
#include "m_spl.h"
#include "panic.h"
#include "kstd.h"
#include "waiters.h"

//Inode in our RAM filesystem
typedef struct ramfs_ino_s
{
	//Information about the file
	mode_t mode; //Type and permissions of file
	dev_t special; //Device number if this is a device special
	off_t size; //Total size in bytes
	
	//Reference counts
	int32_t nfiles; //Number of open files (file_t) that refer to this inode
	int32_t nreaders; //Number of open files, for reading
	int32_t nwriters; //Number of open files, for writing
	int32_t nlinks; //Number of directory entries in the filesystem that refer to this inode
	
	//For FIFOs/pipes and pseudoterminals, shared read/write pointers
	#define RAMFS_FIFO_SZ 4096 //Total size we allocate for these
	int32_t fstart; //Forward-going, data start
	int32_t fcount; //Forward-going, data count
	int32_t bstart; //Backward-going, data start
	int32_t bcount; //Backward-going, data count
	pid_t waiters[8]; //Threads waiting on this FIFO or PTY
	
	//Blocks holding pointers to data blocks
	#define RAMFS_DTABLE_MAX 100
	int dtables[RAMFS_DTABLE_MAX];
	
} ramfs_ino_t;

//Reference to data blocks in RAM filesystem
typedef struct ramfs_dtable_s
{
	//Blocks holding data
	#define RAMFS_DBLOCK_MAX 128
	int blocks[RAMFS_DBLOCK_MAX];
	
} ramfs_dtable_t;

//Block in our RAM filesystem - unit of allocation
#define RAMFS_BLOCK_SIZE 512
typedef union ramfs_block_u
{
	ramfs_ino_t ino;
	ramfs_dtable_t dtable;
	int nextfree;
	uint8_t data[RAMFS_BLOCK_SIZE];
	
} ramfs_block_t;

//Space pre-allocated from kernel for storing the RAM filesystem
#define RAMFS_BLOCK_MAX ((16*1024*1024) / RAMFS_BLOCK_SIZE)
static __attribute__((noinit)) ramfs_block_t ramfs_blocks[RAMFS_BLOCK_MAX];

//Free-list of FS blocks
static int ramfs_freehead;

//Spinlock protecting the filesystem
static m_spl_t ramfs_spl;

//Directory entry as stored in filesystem
#define ramfs_dirent_t _sc_dirent_t

//Allocates a block in the filesystem. Returns its block number.
static int ramfs_alloc(void)
{
	if(ramfs_freehead == 0)
		return -ENOSPC;
	
	int64_t retval = ramfs_freehead;
	ramfs_freehead = ramfs_blocks[ramfs_freehead].nextfree;
	
	KASSERT(retval > 0 && retval < RAMFS_BLOCK_MAX);
	memset(&(ramfs_blocks[retval]), 0, sizeof(ramfs_blocks[retval]));
	return retval;
}

//Frees a block in the filesystem, putting it back on the free-list.
static void ramfs_free(int blknum)
{
	KASSERT(blknum != 0);
	ramfs_blocks[blknum].nextfree = ramfs_freehead;
	ramfs_freehead = blknum;
}

//Checks if any references remain to the given inode.
//If not, frees it.
static void ramfs_checkrefs(int ino)
{
	ramfs_ino_t *iptr = &(ramfs_blocks[ino].ino);
	if(iptr->nfiles > 0)
		return;
	if(iptr->nlinks > 0)
		return;
	
	//No links or open files. Free this inode.
	KASSERT(iptr->nreaders == 0);
	KASSERT(iptr->nwriters == 0);
	
	ramfs_trunc(ino, 0);
	iptr->mode = -1; //Poison
	iptr->size = -1; //Poison
	ramfs_free(ino);
}

//Returns the block number of the data-block for the given offset in the given inode.
//Optionally tries to allocate blocks to back the given location.
static int ramfs_getblock(int ino, off_t off, int alloc)
{
	if(off < 0)
		return -EINVAL;
	
	ramfs_ino_t *iptr = &(ramfs_blocks[ino].ino);
	
	off_t block_num = off / RAMFS_BLOCK_SIZE;
	
	off_t block_idx = block_num % RAMFS_DBLOCK_MAX;
	off_t table_idx = block_num / RAMFS_DBLOCK_MAX;
	if(table_idx >= RAMFS_DTABLE_MAX)
		return -EFBIG;
	
	if(iptr->dtables[table_idx] == 0)
	{
		if(!alloc)
			return 0;
		
		int newtable = ramfs_alloc();
		if(newtable < 0)
			return newtable;
		
		iptr->dtables[table_idx] = newtable;
	}
	
	ramfs_dtable_t *dptr = &(ramfs_blocks[iptr->dtables[table_idx]].dtable);
	if(dptr->blocks[block_idx] == 0)
	{
		if(!alloc)
			return 0;
		
		int newblock = ramfs_alloc();
		if(newblock < 0)
			return newblock;
		
		dptr->blocks[block_idx] = newblock;
	}
	
	return dptr->blocks[block_idx];
}

void ramfs_init(void)
{
	//Make sure we didn't botch the block size definition...
	KASSERT(sizeof(ramfs_block_t) == RAMFS_BLOCK_SIZE);
	
	//Build the initial free-list of all blocks except 0.
	for(int bb = 1; bb < RAMFS_BLOCK_MAX; bb++)
	{
		ramfs_free(bb);
	}
	
	//Make the inode for the root directory in block 0
	memset(&(ramfs_blocks[0]), 0, sizeof(ramfs_blocks[0]));
	ramfs_blocks[0].ino.mode = _SC_S_IFDIR | 0777;	
	
	//Write "." and ".." into root directory
	ramfs_dirent_t dot = { .ino = 0, .name = "." };
	ssize_t dot_written = ramfs_write(0, 0, &dot, sizeof(dot), false);
	if(dot_written != sizeof(dot))
		panic("ramfs_init failed making root dir");
	
	ramfs_dirent_t dotdot = { .ino = 0, .name = ".." };
	ssize_t dotdot_written = ramfs_write(0, sizeof(dot), &dotdot, sizeof(dotdot), false);
	if(dotdot_written != sizeof(dotdot))
		panic("ramfs_init failed making root dir");
	
	//Give root directory a phony reference-count that can (should) never be removed
	ramfs_blocks[0].ino.nlinks++;
}

void ramfs_lock(void)
{
	m_spl_acq(&ramfs_spl);
}

void ramfs_unlock(void)
{
	m_spl_rel(&ramfs_spl);
}

int ramfs_make(int dir, const char *name, mode_t mode, dev_t special, int *ino_out)
{
	if(name == NULL || name[0] == '\0')
		return -EINVAL;
	
	//Error to return on failure
	int err_ret = 0;
	
	//Allocate a block for the new inode.
	int inoblock = ramfs_alloc();
	if(inoblock < 0)
		return inoblock;
	
	ramfs_ino_t *newino = &(ramfs_blocks[inoblock].ino);
	newino->mode = mode;
	newino->special = special;
	newino->size = 0;
	
	//Try to find the name in the directory. It shouldn't exist.
	int find_ino = 0;
	int find_err = ramfs_find(dir, name, &find_ino);
	if(find_err >= 0)
	{
		//Found an existing file with this name in this directory.
		err_ret = -EEXIST;
		goto failure;
	}
	else if(find_err != -ENOENT)
	{
		//Some other error occurred while trying to search the directory.
		err_ret = find_err;
		goto failure;
	}
	
	//If we're making a pipe or pseudoterminal, preallocate space for its buffers.
	int prealloc = 0;
	switch((mode & _SC_S_IFMT))
	{
		case _SC_S_IFIFO:
			prealloc = RAMFS_FIFO_SZ;
			break;
		case _SC_S_IFPTY:
			prealloc = RAMFS_FIFO_SZ * 2;
			break;
		default:
			break;
	}
	newino->size = prealloc;
	for(int bb = 0; bb < prealloc; bb += RAMFS_BLOCK_SIZE)
	{
		int gotblock = ramfs_getblock(inoblock, bb, 1);
		if(gotblock < 0)
		{
			//No room to preallocate space for this FIFO/PTY
			err_ret = gotblock;
			goto failure;
		}
	}
	
	//If we're making a directory, write the "." and ".." entries.
	if((mode & _SC_S_IFMT) == _SC_S_IFDIR)
	{
		ramfs_dirent_t dot = { .ino = inoblock, .name = "." };
		ssize_t dot_written = ramfs_write(inoblock, 0, &dot, sizeof(dot), false);
		if(dot_written != sizeof(dot))
		{
			err_ret = (dot_written < 0) ? dot_written : -EIO;
			goto failure;
		}
		
		ramfs_dirent_t dotdot = { .ino = dir, .name = ".." };
		ssize_t dotdot_written = ramfs_write(inoblock, sizeof(dot), &dotdot, sizeof(dotdot), false);
		if(dotdot_written != sizeof(dotdot))
		{
			err_ret = (dotdot_written < 0) ? dotdot_written : -EIO;
			goto failure;
		}
	}
	
	//Make the directory entry that points to the new file, with its initial name.
	ramfs_dirent_t firstlink = { .ino = inoblock };
	if(strlen(name) >= sizeof(firstlink.name))
	{
		err_ret = -ENAMETOOLONG;
		goto failure;
	}
	strncpy(firstlink.name, name, sizeof(firstlink.name));
	firstlink.name[sizeof(firstlink.name)-1] = '\0';
	
	//Write into the directory containing the file.
	//Do this last so we don't need to un-do it on failure.
	ramfs_ino_t *dirino = &(ramfs_blocks[dir].ino);
	ssize_t firstlink_written = ramfs_write(dir, dirino->size, &firstlink, sizeof(firstlink), false);
	if(firstlink_written != sizeof(firstlink))
	{
		err_ret = (firstlink_written < 0) ? firstlink_written : -EIO;
		goto failure;
	}
	
	//The new file starts with one link in the filesystem.
	newino->nlinks = 1;
	*ino_out = inoblock;
	return 0;
	
failure:	
	if(inoblock > 0)
	{
		ramfs_trunc(inoblock, 0);
		ramfs_free(inoblock);
	}
	
	KASSERT(err_ret < 0);
	*ino_out = -1;
	return err_ret;
}

int ramfs_find(int dir, const char *name, int *ino_out)
{	
	KASSERT(dir < RAMFS_BLOCK_MAX);
	
	if(name == NULL || name[0] == '\0')
	{
		//Looking up "" - return the given file.
		*ino_out = dir;
		return 0;
	}
	
	ramfs_ino_t *dptr = &(ramfs_blocks[dir].ino);
	if((dptr->mode & _SC_S_IFMT) != _SC_S_IFDIR)
		return -ENOTDIR;
	
	//Read all directory entries and look for this name.
	off_t nextoff = 0;
	while(nextoff < dptr->size)
	{
		ramfs_dirent_t de;
		ssize_t read = ramfs_read(dir, nextoff, &de, sizeof(de), false);
		if(read < 0)
			return read;
		if(read != sizeof(de))
			return -EIO;
		
		if(!strcmp(de.name, name))
		{
			//Found it - output the inode and give it another file reference.
			*ino_out = de.ino;
			return 0;
		}
		
		//Not the file we were looking for.
		nextoff += read;
	}
	
	//Didn't find an entry with this name
	return -ENOENT;
}

int ramfs_unlink(int dir, const char *name, int rmino)
{
	if((name == NULL) || (name[0] == '\0') || (!strcmp(name, ".")) || (!strcmp(name, "..")))
		return -EINVAL;
	
	KASSERT(dir < RAMFS_BLOCK_MAX);
	ramfs_ino_t *dptr = &(ramfs_blocks[dir].ino);
	if((dptr->mode & _SC_S_IFMT) != _SC_S_IFDIR)
		return -ENOTDIR;
	
	//Search for the directory entry with this name.
	off_t nextoff = 0;
	while(nextoff < dptr->size)
	{
		ramfs_dirent_t de;
		ssize_t read = ramfs_read(dir, nextoff, &de, sizeof(de), false);
		if(read < 0)
			return read;
		if(read != sizeof(de))
			return -EIO;
		
		if(!strcmp(de.name, name))
		{
			//Found the name we want to unlink.
			//Are we trying to unlink a specific ino too?
			if(rmino != 0 && rmino != de.ino)
			{
				//The link no longer refers to the file we're trying to remove.
				return -EDEADLK; //Like FreeBSD funlinkat
			}
			
			KASSERT(de.ino < RAMFS_BLOCK_MAX);
			ramfs_ino_t *unlinked = &(ramfs_blocks[de.ino].ino);
			
			//If we're removing a directory, is it empty?
			if((unlinked->mode & _SC_S_IFMT) == _SC_S_IFDIR)
			{
				if((size_t)(unlinked->size) > 2*sizeof(ramfs_dirent_t)) //Allow for '.' and '..' entries
				{
					return -ENOTEMPTY;
				}
			}
			
			//Copy the last directory entry over the one being removed. Then truncate by one directory entry.
			ramfs_dirent_t lastde;
			ssize_t lastread = ramfs_read(dir, dptr->size - sizeof(lastde), &lastde, sizeof(lastde), false);
			KASSERT(lastread == sizeof(lastde));
			
			ssize_t overwritten = ramfs_write(dir, nextoff, &lastde, sizeof(lastde), false);
			KASSERT(overwritten == sizeof(lastde));
			
			off_t trunced = ramfs_trunc(dir, dptr->size - sizeof(lastde));
			KASSERT(trunced >= 0);
			KASSERT((size_t)dptr->size >= 2 * sizeof(ramfs_dirent_t)); //Always have "." and ".."
			
			//File being referred to, now has one less reference.
			unlinked->nlinks--;
			ramfs_checkrefs(de.ino);
			
			//Success
			return 0;
		}

		//Not the name we wanted to unlink
		nextoff += sizeof(de);
	}
	
	//Didn't find the file to unlink
	return -ENOENT;
}

//Sets up offset/length for access to a pseudoterminal (PTY) or pipe (FIFO) inode
static int ramfs_fifo_prep(ramfs_ino_t *iptr, bool write, bool secondbuf, ssize_t maxlen, off_t *off_out, ssize_t *len_out)
{	
	//Only applicable to pipes/FIFOs
	switch(iptr->mode & _SC_S_IFMT)
	{
		case _SC_S_IFIFO:
			secondbuf = false;
			break;
		case _SC_S_IFPTY:
			break;
		default:
			return 0;
	}
	
	KASSERT(maxlen >= 0);
	
	//Check for broken pipe - writing to a pipe with no readers or vice-versa
	if((iptr->nreaders == 0) && write)
		return -EPIPE;
	if((iptr->nwriters == 0) && !write)
		return -EPIPE;
	
	//Read existing data start/end, depending on which buffer we're using
	int existing_start = secondbuf ? iptr->bstart : iptr->fstart;
	int existing_end = existing_start + (secondbuf ? iptr->bcount : iptr->fcount);
	KASSERT(existing_end >= existing_start);
	
	//Writing data has to stop at end of available space. Reading stops at end of already-written data.
	int limit = write ? RAMFS_FIFO_SZ : existing_end;
	KASSERT(limit >= existing_start && limit >= existing_end);
	
	//Start operation at end of existing data for write, beginning of existing data for read
	*off_out = write ? existing_end : existing_start;
	*len_out = limit - *off_out;
	
	//Do the access relative to the second buffer as necessary
	*off_out += secondbuf ? RAMFS_FIFO_SZ : 0;
	
	//Can't do the access if there's no room.
	//For POSIX compliance, don't permit leaving less than 512 bytes of space to write.
	//This guarantees that we can always atomically take a 512-byte write.
	if(write && (*len_out <= 512))
	{
		waiters_addcur(iptr->waiters, sizeof(iptr->waiters));
		return -EAGAIN;
	}
	
	//Check for a pipe that is empty, being read, but still has writers
	if((!write) && (*len_out <= 0) && (iptr->nwriters > 0) && (iptr->nfiles > 1))
	{
		waiters_addcur(iptr->waiters, sizeof(iptr->waiters));
		return -EAGAIN;
	}
	
	//Cap effective read length at amount of the original request -  only ever lower it
	if(*len_out > maxlen)
		*len_out = maxlen;
	
	return 0;
}

//Updates FIFO/PTY status after an access
static void ramfs_fifo_update(ramfs_ino_t *iptr, bool written, bool secondbuf, ssize_t amt)
{	
	//Only applicable to pipes/FIFOs
	switch(iptr->mode & _SC_S_IFMT)
	{
		case _SC_S_IFIFO:
			secondbuf = false;
			break;
		case _SC_S_IFPTY:
			break;
		default:
			return;
	}
	
	KASSERT(amt > 0);
	KASSERT(iptr->fstart + iptr->fcount <= RAMFS_FIFO_SZ);
	KASSERT(iptr->bstart + iptr->bcount <= RAMFS_FIFO_SZ);
	
	if(written)
	{
		//Wrote "amt" more bytes into the queue.
		if(secondbuf)
		{
			iptr->bcount += amt;
		}
		else
		{
			iptr->fcount += amt;
		}
	}
	else
	{
		//Removed "amt" bytes from the beginning of the queue. Start moves up, queue gets shorter.
		if(secondbuf)
		{
			iptr->bstart += amt;
			iptr->bcount -= amt;
		}
		else
		{
			iptr->fstart += amt;
			iptr->fcount -= amt;
		}
	}
	
	//When all data written has been read, the position of new data can be reset.
	//(There's nothing meaningful in the buffer at all, so nobody cares if we clobber the beginning.)
	if(iptr->fcount == 0)
		iptr->fstart = 0;
	if(iptr->bcount == 0)
		iptr->bstart = 0;
	
	//Imprecise but I don't want to waste space storing multiple arrays for front/back readers/writers
	waiters_release(iptr->waiters, sizeof(iptr->waiters));
}

ssize_t ramfs_read(int ino, off_t off, void *buf, ssize_t len, bool back)
{
	if(off < 0)
		return -EINVAL;
	
	KASSERT(ino < RAMFS_BLOCK_MAX);
	ramfs_ino_t *iptr = &(ramfs_blocks[ino].ino);
	
	//If reading from a FIFO/PTY, ignore offset and read next written data
	int fifoerr = ramfs_fifo_prep(iptr, false, back, len, &off, &len);
	if(fifoerr < 0)
		return fifoerr;

	//Handle a block at a time
	uint8_t *buf_bytes = (uint8_t*)buf;
	ssize_t completed = 0;
	while(len > 0 && off < iptr->size)
	{
		//Cannot read past the end of the block, nor the end of the file in bytes, nor the requested length.
		ssize_t left_in_block = RAMFS_BLOCK_SIZE - (off % RAMFS_BLOCK_SIZE);
		ssize_t left_in_file = iptr->size - off;
		ssize_t read_len = len;
		if(read_len > left_in_block)
			read_len = left_in_block;
		if(read_len > left_in_file)
			read_len = left_in_file;
		
		//Find where the data for this block is - or return 0s if it's a hole (no data yet).
		int dataloc = ramfs_getblock(ino, off, 0);
		if(dataloc < 0)
			return (completed > 0) ? completed : dataloc;
		
		if(dataloc == 0)
		{
			//No data here - read zeroes.
			memset(buf_bytes, 0, read_len);
		}
		else
		{
			//Data here - copy it out
			memcpy(buf_bytes, ramfs_blocks[dataloc].data + (off % RAMFS_BLOCK_SIZE), read_len);
		}
		
		completed += read_len;
		buf_bytes += read_len;
		off += read_len;
		len -= read_len;
		
		//Update status of shared pointers, if this is a FIFO/PTY
		ramfs_fifo_update(iptr, false, back, read_len);
	}
	
	KASSERT(len >= 0);
	return completed;
}

ssize_t ramfs_write(int ino, off_t off, const void *buf, ssize_t len, bool back)
{
	if(off < 0)
		return -EINVAL;
	
	KASSERT(ino < RAMFS_BLOCK_MAX);
	ramfs_ino_t *iptr = &(ramfs_blocks[ino].ino);
	
	//If writing to a FIFO/PTY, ignore offset and write end of queue
	int fifoerr = ramfs_fifo_prep(iptr, true, !back, len, &off, &len);
	if(fifoerr < 0)
		return fifoerr;
	
	//Handle a block at a time
	const uint8_t *buf_bytes = (uint8_t*)buf;
	ssize_t completed = 0;
	while(len > 0)
	{
		ssize_t left_in_block = RAMFS_BLOCK_SIZE - (off % RAMFS_BLOCK_SIZE);
		ssize_t write_len = len;
		if(write_len > left_in_block)
			write_len = left_in_block;
		
		//Allocate new data block if there's none here yet
		int dataloc = ramfs_getblock(ino, off, 1);
		if(dataloc < 0)
			return (completed > 0) ? completed : dataloc;
		
		KASSERT(dataloc > 0);

		//Copy the data in
		memcpy(ramfs_blocks[dataloc].data + (off % RAMFS_BLOCK_SIZE), buf_bytes, write_len);
		
		completed += write_len;
		buf_bytes += write_len;
		off += write_len;
		len -= write_len;
		
		//Expand filesize if we've written past its old end
		if(off > iptr->size)
			iptr->size = off;
		
		//Update status of shared pointers, if this is a FIFO/PTY
		ramfs_fifo_update(iptr, true, !back, write_len);
	}
	
	KASSERT(len >= 0);
	return completed;	
}

off_t ramfs_trunc(int ino, off_t size)
{	
	if(size < 0)
		return -EINVAL;
	
	if(size > (off_t)(1ul * RAMFS_DTABLE_MAX * RAMFS_DBLOCK_MAX * RAMFS_BLOCK_SIZE))
		return -EFBIG;
	
	//Change size of file
	KASSERT(ino < RAMFS_BLOCK_MAX);
	ramfs_ino_t *iptr = &(ramfs_blocks[ino].ino);
	iptr->size = size;
	
	//Ditch any tables that are totally off the end of the file
	const size_t table_size = RAMFS_BLOCK_SIZE * RAMFS_DBLOCK_MAX;
	size_t tables_needed = (size + table_size - 1) / table_size;
	for(size_t tt = tables_needed; tt < RAMFS_DTABLE_MAX; tt++)
	{
		if(iptr->dtables[tt] == 0)
			continue;
		
		ramfs_dtable_t *dptr = &(ramfs_blocks[iptr->dtables[tt]].dtable);
		for(size_t bb = 0; bb < RAMFS_DBLOCK_MAX; bb++)
		{
			if(dptr->blocks[bb] == 0)
				continue;
			
			ramfs_free(dptr->blocks[bb]);
			dptr->blocks[bb] = 0;
		}
		
		ramfs_free(iptr->dtables[tt]);
		iptr->dtables[tt] = 0;
	}
	
	//Ditch unused blocks from any partially-used table at the end of the file
	size_t blocks_needed = (size + RAMFS_BLOCK_SIZE - 1) / RAMFS_BLOCK_SIZE;
	if((tables_needed > 0) && (blocks_needed % RAMFS_DBLOCK_MAX))
	{
		size_t partial_table = tables_needed - 1;
		size_t first_block = partial_table * RAMFS_DBLOCK_MAX;
		if(iptr->dtables[partial_table] != 0)
		{
			ramfs_dtable_t *dptr = &(ramfs_blocks[iptr->dtables[partial_table]].dtable);
			for(size_t bb = blocks_needed - first_block; bb < RAMFS_DBLOCK_MAX; bb++)
			{
				if(dptr->blocks[bb] == 0)
					continue;
				
				ramfs_free(dptr->blocks[bb]);
				dptr->blocks[bb] = 0;
			}
		}
	}
	
	//If we truncated to zero, we should have no blocks referenced from this inode.
	if(size == 0)
	{
		for(int tt = 0; tt < RAMFS_DTABLE_MAX; tt++)
		{
			KASSERT(iptr->dtables[tt] == 0);
		}
	}
	
	return iptr->size;
}

int ramfs_stat(int ino, _sc_stat_t *st)
{
	KASSERT(ino < RAMFS_BLOCK_MAX);
	ramfs_ino_t *iptr = &(ramfs_blocks[ino].ino);
	
	memset(st, 0, sizeof(*st));
	st->ino = ino;
	st->mode = iptr->mode;
	st->rdev = iptr->special;
	st->size = iptr->size;
	return 0;
}

void ramfs_inc(int ino, int access)
{
	KASSERT(ino < RAMFS_BLOCK_MAX);
	ramfs_ino_t *iptr = &(ramfs_blocks[ino].ino);
	
	iptr->nfiles++;
	KASSERT(iptr->nfiles > 0);
	
	if(access & _SC_ACCESS_R)
		iptr->nreaders++;
	if(access & _SC_ACCESS_W)
		iptr->nwriters++;
}

void ramfs_dec(int ino, int access)
{
	KASSERT(ino < RAMFS_BLOCK_MAX);
	ramfs_ino_t *iptr = &(ramfs_blocks[ino].ino);
	
	iptr->nfiles--;
	KASSERT(iptr->nfiles >= 0);	
	
	if(access & _SC_ACCESS_R)
		iptr->nreaders--;
	if(access & _SC_ACCESS_W)
		iptr->nwriters--;
	
	ramfs_checkrefs(ino);
}
