//syscalls.c
//System call declarations on kernel-side
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "kstd.h"
#include "panic.h"
#include "syscalls.h"
#include "kassert.h"
#include "argenv.h"
#include "m_video.h"
#include "m_frame.h"
#include "m_intr.h"
#include "process.h"
#include "thread.h"
#include "file.h"
#include "elf.h"
#include "fb.h"
#include "con.h"
#include "namei.h"

int k_sc_version(int version)
{
	//Version numbers start at 1
	if(version < 1)
		return -EINVAL;
	
	//Don't allow versions newer than this kernel supports
	if(version > 1)
		version = 1;
	
	process_t *pptr = process_lockcur();
	pptr->scver = version;	
	process_unlock(pptr);
	
	return 1;
}

void k_sc_none(void)
{
	
}

void k_sc_exit(int exitcode, int signal)
{
	//Mark the process as wanting to exit, and store the exit status
	process_t *pptr = process_lockcur();
	if(pptr->state == PROCESS_STATE_DETHREAD)
	{
		//Some other thread already made our process quit.
		process_unlock(pptr);
		return;
	}
	
	KASSERT(pptr->state == PROCESS_STATE_ALIVE);
	pptr->state = PROCESS_STATE_DETHREAD;
	if(signal)
		pptr->wstatus = _SC_STATUS_SIGNALED_BIT | ((signal << _SC_STATUS_TERMSIG_SHIFT) & _SC_STATUS_TERMSIG_MASK);
	else
		pptr->wstatus = _SC_STATUS_EXITED_BIT | (exitcode & 0xFF);
	
	process_unlock(pptr);
	
	//When we try to return from the system call, we'll see that this thread belongs to a dethreading process.
	//The thread will be removed and cleaned-up, and if it's the last one, so will the process.
}

int k_sc_getpid(void)
{
	process_t *pptr = process_lockcur();
	pid_t retval = pptr->pid;
	process_unlock(pptr);
	return retval;
}

int k_sc_gettid(void)
{
	thread_t *tptr = thread_lockcur();
	pid_t retval = tptr->tid;
	thread_unlock(tptr);
	return retval;
}

int k_sc_getppid(void)
{
	process_t *pptr = process_lockcur();
	pid_t retval = pptr->ppid;
	process_unlock(pptr);
	return retval;
}

int k_sc_fork(void)
{
	//Error returned on failure
	int err_ret = 0;
	
	//Cleaned up on failure
	thread_t *childthread = NULL;
	process_t *child = NULL;
	
	//Get current process/thread
	process_t *pptr = process_lockcur();
	KASSERT(pptr != NULL);
	
	thread_t *tptr = thread_lockcur();
	KASSERT(tptr != NULL);
		
	if(pptr->nthreads > 1)
	{
		//More than one thread active in calling process
		err_ret = -EBUSY;
		goto cleanup;
	}
	
	//Find space for new process
	child = process_lockfree();
	if(child == NULL)
	{
		//No room for more processes
		err_ret = -EAGAIN;
		goto cleanup;
	}
	
	//Find space for new thread
	int thread_err = thread_new(child, &childthread);
	if(thread_err < 0)
	{
		//No room for child's first thread
		err_ret = thread_err;
		goto cleanup;
	}
	
	//Copy memory to new process
	KASSERT(child->mem.uspc == 0);
	KASSERT(child->mem_attempt.uspc == 0);
	int copy_err = mem_copy(&(child->mem), &(pptr->mem));
	if(copy_err < 0)
	{
		//Failed to copy memory space
		err_ret = copy_err;
		goto cleanup;
	}
	
	//Copy process settings
	child->entry = pptr->entry;
	child->scver = pptr->scver;
	
	//Copy state of calling thread and put new thread in its process, except the child thread returns 0.
	memcpy(&(childthread->drop), &(tptr->drop), sizeof(childthread->drop));
	m_drop_retval(&(childthread->drop), 0);
	
	//The child thread is traced by the same thread as the parent thread.
	//If it's being traced, it stops due to a fork.
	childthread->trace_host = tptr->trace_host;
	childthread->trace_stopped = _SC_TRACE_REASON_FORKED;
	
	//Copy file descriptors and PWD
	//Note - done last because we don't have code to un-do it.
	for(int ff = 0; ff < PROCESS_FD_MAX; ff++)
	{
		if(pptr->fds[ff].file != NULL)
		{
			file_lock(pptr->fds[ff].file);
			child->fds[ff] = pptr->fds[ff];
			pptr->fds[ff].file->refs++;
			file_unlock(pptr->fds[ff].file);
		}
	}
	
	if(pptr->pwd != NULL)
	{
		file_lock(pptr->pwd);
		pptr->pwd->refs++;
		child->pwd = pptr->pwd;
		file_unlock(pptr->pwd);
	}
	
	//Copy the state of the console
	child->conflags = pptr->conflags;
	
	//If the parent wanted to hold the console, so does the child.
	//Otherwise, the child doesn't participate in the console.
	if(pptr->conchild == pptr->pid)
		child->conchild = child->pid;
	else
		child->conchild = 0;
	
	//Success.
	child->ppid = pptr->pid;
	child->state = PROCESS_STATE_ALIVE;
	
	pid_t child_pid = child->pid;

	thread_unlock(childthread);
	thread_unlock(tptr);
	process_unlock(child);
	process_unlock(pptr);
	
	return child_pid;
	
cleanup:	
	
	if(child != NULL)
	{
		mem_clear(&(child->mem));
		mem_clear(&(child->mem_attempt));
		child->state = PROCESS_STATE_NONE;
		process_unlock(child);
	}
	
	if(childthread != NULL)
	{
		childthread->state = THREAD_STATE_NONE;
		thread_unlock(childthread);
	}
	
	process_unlock(pptr);
	thread_unlock(tptr);
	
	KASSERT(err_ret < 0);
	return err_ret;
}

int k_sc_spoon(void)
{
	//Get current thread/process
	process_t *pptr = process_lockcur();
	thread_t *tptr = thread_lockcur();
	
	//Make a new thread
	thread_t *childthread = NULL;
	int thread_err = thread_new(pptr, &childthread);
	if(thread_err < 0)
	{
		//No room for child thread
		thread_unlock(tptr);
		process_unlock(pptr);
		return thread_err;
	}

	//Copy the user context over - child returns from the syscall with a return of 0.
	memcpy(&(childthread->drop), &(tptr->drop), sizeof(childthread->drop));
	m_drop_retval(&(childthread->drop), 0);
	
	//Child thread has same thread tracing it, if any. If traced, it stops because it's just forked.
	childthread->trace_host = tptr->trace_host;
	childthread->trace_stopped = _SC_TRACE_REASON_FORKED;
	
	//Parent will return with the thread ID of the child
	pid_t retval = childthread->tid;
	
	//Success
	thread_unlock(childthread);
	thread_unlock(tptr);
	process_unlock(pptr);
	KASSERT(retval > 0);
	return retval;
}

int k_sc_exec(int fd, char *const argv[], char *const envp[])
{
	//Stuff for cleaning up safely
	process_t *pptr = NULL;
	file_t *elf_file = NULL;
	int err_ret = 0;
	
	//Get the ELF file and try to load it into this process's spare memory space.
	elf_file = process_lockfd(fd, false);
	if(elf_file == NULL)
	{
		err_ret = -EBADF;
		goto cleanup;
	}
	
	if(!(elf_file->access & _SC_ACCESS_X))
	{
		err_ret = -EBADF;
		goto cleanup;
	}
	
	pptr = process_lockcur();
	KASSERT(pptr->mem_attempt.uspc == 0);
	if(pptr->nthreads > 1)
	{
		//Can't exec while other threads are running
		err_ret = -EBUSY;
		goto cleanup;
	}
	
	uintptr_t elf_entry = 0;
	int elf_err = elf_load(elf_file, &(pptr->mem_attempt), &elf_entry);
	if(elf_err < 0)
	{
		//Failed to load the new process image.
		err_ret = elf_err;
		goto cleanup;
	}
	
	//Done with ELF now
	file_unlock(elf_file);
	elf_file = NULL;
	
	//Try to copy the arg/env buffer into the new memory space	
	uintptr_t argenv_addr = 0;
	int argenv_err = argenv_load(&(pptr->mem), &(pptr->mem_attempt), argv, envp, &argenv_addr);
	if(argenv_err < 0)
	{
		//Failed to copy argv/envp
		err_ret = argenv_err;
		goto cleanup;
	}
	
	//Set up the new memory space successfully. Switch over and ditch the old memory.
	m_uspc_activate(pptr->mem_attempt.uspc);
	mem_clear(&(pptr->mem));
	pptr->mem = pptr->mem_attempt;
	memset(&(pptr->mem_attempt), 0, sizeof(pptr->mem_attempt));
	
	//Drop any file descriptors not flagged keep-on-exec
	for(int ff = 0; ff < PROCESS_FD_MAX; ff++)
	{
		if(pptr->fds[ff].file == NULL)
			continue;
	
		if(pptr->fds[ff].flags & _SC_FLAG_KEEPEXEC)
			continue;
		
		file_lock(pptr->fds[ff].file);
		pptr->fds[ff].file->refs--;
		file_unlock(pptr->fds[ff].file);
		pptr->fds[ff].file = NULL;
		pptr->fds[ff].flags = 0;
	}
	
	//Note the new entry point
	pptr->entry = elf_entry;
	
	//Done setting up the process...
	process_unlock(pptr);
	pptr = NULL;
	
	//Reset the user context for the calling thread, to start at the new entry point
	thread_t *tptr = thread_lockcur();
	tptr->sigmask = 0x7FFFFFFFFFFFFFFFul;
	tptr->sigpend = 0;
	m_drop_entry(&(tptr->drop), elf_entry); //Single entry point
	m_drop_param(&(tptr->drop), 0, 0); //Parameter 0 - which signal is being entered (0 for startup).
	m_drop_param(&(tptr->drop), 1, argenv_addr); //Parameter 1 on startup - argv/envp location
	
	//The calling thread is now stopped if it's being traced, due to having run an exec
	tptr->trace_stopped = _SC_TRACE_REASON_EXECED;
	
	thread_unlock(tptr);
	
	//Successo
	return 0;
	
cleanup:
	if(pptr != NULL)
	{
		mem_clear(&(pptr->mem_attempt));		
		process_unlock(pptr);
		pptr = NULL;
	}
	
	if(elf_file != NULL)
	{
		file_unlock(elf_file);
		elf_file = NULL;
	}
		
	KASSERT(err_ret < 0);
	return err_ret;
}

int k_sc_find(int dirfd, const char *name, int access)
{
	if(dirfd < -1)
		return -EINVAL;
	
	//Get starting directory reference
	file_t *dptr = process_lockfd(dirfd, true);
	if(dptr == NULL)
		return -EBADF;
	
	//Resolve all but the last component of the path
	file_t *finaldptr = NULL;
	char finalname[256];
	int namei_result = namei(dptr, name, access, &finaldptr, finalname);
	if(namei_result < 0)
	{
		file_unlock(dptr);
		return namei_result;
	}
	
	//Done with the initial directory reference
	file_unlock(dptr);
	dptr = NULL;
	
	//Look up in the resulting directory
	file_t *found = NULL;
	int find_result = file_find(finaldptr, finalname, access, &found);
	
	//Done with the penultimate directory reference
	KASSERT(finaldptr != NULL);
	file_unlock(finaldptr);
	
	if(find_result < 0)
		return find_result;

	//Try to insert into FDs for this process
	int newfd = process_addfd(found);
	if(newfd < 0)
		found->refs = 0;
	
	file_unlock(found);
	return newfd;
}

int k_sc_make(int dirfd, const char *name, int mode, int rdev, int access)
{
	if(dirfd < -1)
		return -EINVAL;
	
	//Get starting directory reference
	file_t *dptr = process_lockfd(dirfd, true);
	if(dptr == NULL)
		return -EBADF;
	
	//Resolve all but the last component of the path
	file_t *finaldptr = NULL;
	char finalname[256];
	int namei_result = namei(dptr, name, access, &finaldptr, finalname);
	if(namei_result < 0)
	{
		file_unlock(dptr);
		return namei_result;
	}
	
	//Done with the initial directory reference
	file_unlock(dptr);
	dptr = NULL;
	
	//Make in the resulting directory
	file_t *made = NULL;
	int make_result = file_make(finaldptr, finalname, mode, rdev, access, &made);
	
	//Done with the penultimate directory reference
	KASSERT(finaldptr != NULL);
	file_unlock(finaldptr);
	
	if(make_result < 0)
		return make_result;

	//Try to insert into FDs for this process
	int newfd = process_addfd(made);
	if(newfd < 0)
		made->refs = 0;
	
	file_unlock(made);
	return newfd;
}

int k_sc_read(int fd, void *buf, int len)
{
	//Make sure the given buffer is a valid memory region before doing stuff at the user's request.
	//Note - this depends on us disallowing memory to be unmapped! Otherwise we'd need to hold this lock throughout.
	process_t *pptr = process_lockcur();
	int memvalid = mem_validate(&(pptr->mem), (uintptr_t)buf, len, M_USPC_PROT_W);
	process_unlock(pptr);
	if(memvalid < 0)
		return memvalid;
	
	file_t *fptr = process_lockfd(fd, false);
	if(fptr == NULL)
		return -EBADF;
	
	ssize_t result = file_read(fptr, buf, len); //Todo - validate buffer
	file_unlock(fptr);
	return result;
}

int k_sc_write(int fd, const void *buf, int len)
{
	//Make sure the given buffer is a valid memory region before doing stuff at the user's request.
	//Note - this depends on us disallowing memory to be unmapped! Otherwise we'd need to hold this lock throughout.
	process_t *pptr = process_lockcur();
	int memvalid = mem_validate(&(pptr->mem), (uintptr_t)buf, len, M_USPC_PROT_R);
	process_unlock(pptr);
	if(memvalid < 0)
		return memvalid;
	
	file_t *fptr = process_lockfd(fd, false);
	if(fptr == NULL)
		return -EBADF;
	
	ssize_t result = file_write(fptr, buf, len); //Todo - validate buffer
	file_unlock(fptr);
	return result;
}

long int k_sc_seek(int fd, long int off, int whence)
{
	file_t *fptr = process_lockfd(fd, false);
	if(fptr == NULL)
		return -EBADF;
	
	off_t result = file_seek(fptr, off, whence);
	file_unlock(fptr);
	return result;
}

long int k_sc_trunc(int fd, long int size)
{
	file_t *fptr = process_lockfd(fd, false);
	if(fptr == NULL)
		return -EBADF;
	
	off_t result = file_trunc(fptr, size);
	file_unlock(fptr);
	return result;
}

int k_sc_unlink(int dirfd, const char *name, int rmfd)
{
	//Process can specify rmfd==-1 for "remove any file with this name".
	//They can also specify a file descriptor in rmfd - then, only that exact file may be unlinked.
	if(rmfd < -1 || dirfd < -1)
		return -EINVAL;
	
	//Get starting point for search
	file_t *dptr = process_lockfd(dirfd, true);
	if(dptr == NULL)
		return -EBADF;
	
	//Look up nonfinal pathname components
	file_t *finaldptr = NULL;
	char finalname[256];
	int namei_result = namei(dptr, name, _SC_ACCESS_W | _SC_ACCESS_NOFOLLOW, &finaldptr, finalname);
	if(namei_result < 0)
	{
		file_unlock(dptr);
		return namei_result;
	}
	
	//Done with the initial directory reference
	file_unlock(dptr);
	dptr = NULL;


	//Look up specific file to remove, if any
	file_t *rmptr = NULL;
	if(rmfd != -1)
	{
		rmptr = process_lockfd(rmfd, false);
		if(rmptr == NULL)
		{
			file_unlock(finaldptr);
			return -EBADF;
		}
	}
	
	//Try to remove it
	KASSERT(finaldptr != NULL);
	int result = file_unlink(finaldptr, finalname, rmptr);
	
	if(rmptr != NULL)
		file_unlock(rmptr);
	
	file_unlock(finaldptr);
	return result;
}

int k_sc_flag(int fd, int set, int clear)
{
	if(set < 0 || clear < 0)
		return -EINVAL;
	
	if(fd < 0 || fd >= PROCESS_FD_MAX)
		return -EBADF;
	
	process_t *pptr = process_lockcur();
	if(pptr->fds[fd].file == NULL)
	{
		process_unlock(pptr);
		return -EBADF;
	}
	
	pptr->fds[fd].flags |= set;
	pptr->fds[fd].flags &= ~(clear);
	int retval = pptr->fds[fd].flags;
	
	process_unlock(pptr);
	return retval;
}

int k_sc_close(int fd)
{
	if(fd < 0 || fd >= PROCESS_FD_MAX)
		return -EBADF;
	
	process_t *pptr = process_lockcur();
	if(pptr->fds[fd].file == NULL)
	{
		process_unlock(pptr);
		return -EBADF;
	}
	
	file_t *fptr = pptr->fds[fd].file;
	file_lock(fptr);
	pptr->fds[fd].file = NULL;
	fptr->refs--;
	file_unlock(fptr);
	process_unlock(pptr);
	
	return 0;	
}

int k_sc_chdir(int fd)
{
	if(fd < 0 || fd >= PROCESS_FD_MAX)
		return -EBADF;
	
	process_t *pptr = process_lockcur();
	if(pptr->fds[fd].file == NULL)
	{
		process_unlock(pptr);
		return -EBADF;
	}
	
	if(pptr->pwd != NULL)
	{
		file_lock(pptr->pwd);
		pptr->pwd->refs--;
		file_unlock(pptr->pwd);
		pptr->pwd = NULL;
	}
	
	file_lock(pptr->fds[fd].file);
	pptr->pwd = pptr->fds[fd].file;
	pptr->pwd->refs++;
	file_unlock(pptr->fds[fd].file);
	
	process_unlock(pptr);
	return 0;
}

int k_sc_dup(int oldfd, int min, int overwrite)
{
	if(!(overwrite == 0 || overwrite == 1))
		return -EINVAL;
	
	if(oldfd < 0 || oldfd >= PROCESS_FD_MAX)
		return -EBADF;
	
	if(min < 0 || min >= PROCESS_FD_MAX)
		return -EBADF;
	
	if( (oldfd == min) && overwrite )
		return oldfd;
	
	process_t *pptr = process_lockcur();
	if(pptr->fds[oldfd].file == NULL)
	{
		process_unlock(pptr);
		return -EBADF;
	}
	
	for(int ff = min; ff < PROCESS_FD_MAX; ff++)
	{
		if(overwrite || (pptr->fds[ff].file == NULL))
		{
			if(pptr->fds[ff].file != NULL)
			{
				file_lock(pptr->fds[ff].file);
				pptr->fds[ff].file->refs--;
				file_unlock(pptr->fds[ff].file);
				pptr->fds[ff].file = NULL;
			}
			
			file_lock(pptr->fds[oldfd].file);
			pptr->fds[ff].file = pptr->fds[oldfd].file;
			pptr->fds[oldfd].file->refs++;
			file_unlock(pptr->fds[oldfd].file);
			
			process_unlock(pptr);
			return ff;
		}
	}
	
	//Didn't have room
	process_unlock(pptr);
	return -EMFILE;
}

int k_sc_stat(int fd, _sc_stat_t *buf, int len)
{
	file_t *fptr = process_lockfd(fd, true);
	if(fptr == NULL)
		return -EBADF;
	
	_sc_stat_t kbuf = {0};
	int stat_err = file_stat(fptr, &kbuf);
	if(stat_err < 0)
	{
		file_unlock(fptr);
		return stat_err;
	}
	
	ssize_t to_copy = (len < (ssize_t)sizeof(kbuf)) ? len : (ssize_t)sizeof(kbuf);
	int buf_err = process_memput(buf, &kbuf, to_copy);
	if(buf_err < 0)
	{
		file_unlock(fptr);
		return buf_err;
	}
	
	file_unlock(fptr);
	return to_copy;
}

long int k_sc_mem_anon(int size)
{
	process_t *pptr = process_lockcur();
	
	intptr_t freespace = mem_avail(&(pptr->mem), -1, size);
	if(freespace < 0)
	{
		//No room in the caller's address space
		process_unlock(pptr);
		return freespace;
	}

	intptr_t retval = mem_add(&(pptr->mem), freespace, size, M_USPC_PROT_R | M_USPC_PROT_W);
	if(retval >= 0)
	{
		size_t pagesize = m_frame_size();
		size_t clear_size = size;
		if(clear_size % pagesize)
			clear_size += pagesize - (clear_size % pagesize);
		
		memset((void*)(uintptr_t)freespace, 0, clear_size);
		retval = freespace;
	}
	
	process_unlock(pptr);
	return retval;
}

long int k_sc_mem_disp(void)
{
	size_t size = m_video_fbsize();
	
	process_t *pptr = process_lockcur();
	
	intptr_t freespace = mem_avail(&(pptr->mem), -1, size);
	if(freespace < 0)
	{
		//No room in caller's address space
		process_unlock(pptr);
		return freespace;
	}

	intptr_t retval = mem_disp(&(pptr->mem), freespace, size);
	if(retval >= 0)
	{
		//Clear memory before giving it to the user
		memset((void*)(uintptr_t)freespace, 0, size);
		retval = freespace;
	}
	
	process_unlock(pptr);
	return retval;
}

int k_sc_wait(int idtype, int id, int options, _sc_wait_t *buf, int len)
{
	if(id < 0)
		return -EINVAL;
	
	if(len != sizeof(_sc_wait_t)) //Todo - support versioning here if _sc_wait_t changes
		return -EINVAL;
	
	//Figure out our ID - whose children we're looking for
	process_t *pptr = process_lockcur();
	pid_t ourpid = pptr->pid;
	
	//Handle special-case - zero PID/PGID means "same group as the caller"
	if( (idtype == _SC_IDTYPE_PID || idtype == _SC_IDTYPE_PGID) && (id == 0) )
	{
		idtype = _SC_IDTYPE_ALL;
		id = 0;
	}
	
	process_unlock(pptr);
	pptr = NULL;
	
	//Run through the process table once and see what's up.
	//Caller should use _sc_pause and try again if we tell them so.
	//Anyone currently in the process of posting status will poke their parent after unlocking.
	bool any_match = false;
	for(int pp = 0; pp < PROCESS_MAX; pp++)
	{
		process_t *otherproc = &(process_table[pp]);
		m_spl_acq(&(otherproc->spl));
		
		if((otherproc->state == PROCESS_STATE_NONE) || (otherproc->ppid != ourpid))
		{
			//Not one of our children
			m_spl_rel(&(otherproc->spl));
			continue;
		}
		
		//Found a child of ours. Does it match the ID we want?
		bool matches_id = false;
		switch(idtype)
		{
			case _SC_IDTYPE_PID:
				matches_id = (otherproc->pid == id);
				break;
			case _SC_IDTYPE_PGID:
				matches_id = true;
				break;
			case _SC_IDTYPE_ALL:
				matches_id = true;
				break;
			default:
				m_spl_rel(&(otherproc->spl));
				return -EINVAL;
		}
		
		if(!matches_id)
		{
			//It's one of our children, but doesn't match the ID we asked for.
			m_spl_rel(&(otherproc->spl));
			continue;
		}
		
		//There's at least one process matching the query.
		//This means we won't return -ECHILD.
		any_match = true;
		
		//See if the process has the type of status info we're looking for.
		bool has_status = false;		
		if( (options & _SC_WEXITED) && (otherproc->wstatus & (_SC_STATUS_EXITED_BIT | _SC_STATUS_SIGNALED_BIT)) )
			has_status = true;
		if( (options & _SC_WSTOPPED) && (otherproc->wstatus & (_SC_STATUS_STOPPED_BIT)) )
			has_status = true;
		if( (options & _SC_WCONTINUED) && (otherproc->wstatus & (_SC_STATUS_CONTINUED_BIT)) )
			has_status = true;
		
		if(!has_status)
		{
			//Doesn't have the kind of status information we're looking for.
			m_spl_rel(&(otherproc->spl));
			continue;
		}
		
		//Alright, found what we're waiting for.
		_sc_wait_t output = {0};
		output.status = otherproc->wstatus;
		output.pid = otherproc->pid;
		
		//Clear their status unless we were asked not to
		if(!(options & _SC_WNOWAIT))
		{			
			otherproc->wstatus = 0;
			
			if(otherproc->state == PROCESS_STATE_DEAD)
			{
				//Just waited on an otherwise dead process. 
				//It should be mostly cleaned-up already - finish it off.
				KASSERT(otherproc->mem.uspc == 0);
				KASSERT(otherproc->mem_attempt.uspc == 0);
				KASSERT(otherproc->nthreads == 0);
				for(int ff = 0; ff < PROCESS_FD_MAX; ff++)
				{
					KASSERT(otherproc->fds[ff].file == NULL);
				}
				KASSERT(otherproc->pwd == NULL);
				
				otherproc->state = PROCESS_STATE_NONE;
				otherproc->ppid = 0;
			}
		}
			
		//Done. Return the data.
		m_spl_rel(&(otherproc->spl));
		
		int copy_err = process_memput(buf, &output, len);
		if(copy_err < 0)
			return copy_err;
		
		return len;
	}
	
	//Didn't find anybody with status we wanted.
	if(!any_match)
	{
		//No children match this query - no status would ever be forthcoming.
		return -ECHILD;
	}
	
	//There are matching children, but none has status right now.
	return 0; 
}

int k_sc_getticks(void)
{
	return m_intr_ticks();
}

void k_sc_pause(void)
{
	thread_t *tptr = thread_lockcur();
	tptr->pause_called = true;
	thread_unlock(tptr);
	return;
}

void k_sc_panic(const char *str)
{
	if(process_curpid() == 1)
		panic(str);
	else
		k_sc_exit(0, _SC_SIGABRT);
}

uintptr_t syscalls_handle(uintptr_t num, uintptr_t p1, uintptr_t p2, uintptr_t p3, uintptr_t p4, uintptr_t p5)
{
	//Use macro-trick to make a switch statement by call-number.
	switch(num)
	{
		#define SYSCALL0R(num, rt,  name) \
			case num: return (uintptr_t) k##name();
		#define SYSCALL1R(num, rt,  name, p1t) \
			case num: return (uintptr_t) k##name((p1t)p1);
		#define SYSCALL2R(num, rt,  name, p1t, p2t) \
			case num: return (uintptr_t) k##name((p1t)p1, (p2t)p2);
		#define SYSCALL3R(num, rt,  name, p1t, p2t, p3t) \
			case num: return (uintptr_t) k##name((p1t)p1, (p2t)p2, (p3t)p3);
		#define SYSCALL4R(num, rt,  name, p1t, p2t, p3t, p4t) \
			case num: return (uintptr_t) k##name((p1t)p1, (p2t)p2, (p3t)p3, (p4t)p4);
		#define SYSCALL5R(num, rt,  name, p1t, p2t, p3t, p4t, p5t) \
			case num: return (uintptr_t) k##name((p1t)p1, (p2t)p2, (p3t)p3, (p4t)p4, (p5t)p5);

		#define SYSCALL0V(num, rt,  name) \
			case num: k##name(); return 0;
		#define SYSCALL1V(num, rt,  name, p1t) \
			case num: k##name((p1t)p1); return 0;
		#define SYSCALL2V(num, rt,  name, p1t, p2t) \
			case num: k##name((p1t)p1, (p2t)p2); return 0;
		#define SYSCALL3V(num, rt,  name, p1t, p2t, p3t) \
			case num: k##name((p1t)p1, (p2t)p2, (p3t)p3); return 0;
		#define SYSCALL4V(num, rt,  name, p1t, p2t, p3t, p4t) \
			case num: k##name((p1t)p1, (p2t)p2, (p3t)p3, (p4t)p4); return 0;
		#define SYSCALL5V(num, rt,  name, p1t, p2t, p3t, p4t, p5t) \
			case num: k##name((p1t)p1, (p2t)p2, (p3t)p3, (p4t)p4, (p5t)p5); return 0;

		#define SYSCALL0N(num, rt,  name) \
			case num: k##name(); return 0;
		#define SYSCALL1N(num, rt,  name, p1t) \
			case num: k##name((p1t)p1); return 0;
		#define SYSCALL2N(num, rt,  name, p1t, p2t) \
			case num: k##name((p1t)p1, (p2t)p2); return 0;
		#define SYSCALL3N(num, rt,  name, p1t, p2t, p3t) \
			case num: k##name((p1t)p1, (p2t)p2, (p3t)p3); return 0;
		#define SYSCALL4N(num, rt,  name, p1t, p2t, p3t, p4t) \
			case num: k##name((p1t)p1, (p2t)p2, (p3t)p3, (p4t)p4); return 0;
		#define SYSCALL5N(num, rt,  name, p1t, p2t, p3t, p4t, p5t) \
			case num: k##name((p1t)p1, (p2t)p2, (p3t)p3, (p4t)p4, (p5t)p5); return 0;
		
		#include "systable.h"
		
		default:
			return -ENOSYS;
	}
}

