//syscalls_con.c
//System call definitions on kernel-side - console-related
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "syscalls.h"
#include "kassert.h"
#include "argenv.h"
#include "process.h"
#include "con.h"
#include "fb.h"
#include "thread.h"
#include "kstd.h"
#include "m_cache.h"
#include "m_video.h"

int k_sc_con_fbinfo(_sc_con_fbinfo_t *info_out, int info_len)
{
	//Get info in kernel-side buffer
	_sc_con_fbinfo_t kbuf = {0};
	m_video_fbinfo(&kbuf);
	
	//Copy to the process memory
	if(info_len > (int)sizeof(kbuf))
		info_len = sizeof(kbuf);
	
	int copy_err = process_memput(info_out, &kbuf, info_len);
	if(copy_err < 0)
		return copy_err;
	
	return info_len;
}

int k_sc_con_fbscale(void *dispmem, const void *srcmem, const _sc_con_fbinfo_t *infoptr, int infolen)
{
	//Find information about destination buffer from video system
	_sc_con_fbinfo_t dst_info = {0};
	m_video_fbinfo(&dst_info);
	
	//Copy information about source buffer from userspace
	_sc_con_fbinfo_t src_info = {0};
	if(infolen > (int)sizeof(src_info))
		infolen = sizeof(src_info);
	
	int info_err = process_memget(&src_info, infoptr, infolen);
	if(info_err < 0)
	{
		//Error copying source buffer info from userspace
		return info_err;
	}
	
	//Validate source buffer parameters
	if(src_info.height < 8 || src_info.width < 8 || src_info.stride < 8)
		return -EINVAL;
	if(src_info.height >= 65536 || src_info.width >= 65536 || src_info.stride >= 65536)
		return -EINVAL;
	if(src_info.pxfmt <= _SC_CON_PXFMT_NONE || src_info.pxfmt >= _SC_CON_PXFMT_MAX)
		return -EINVAL;
	
	//Validate access to source and destination buffers
	process_t *pptr = process_lockcur();
	int src_err = mem_validate(&(pptr->mem), (uintptr_t)srcmem, src_info.height * src_info.stride, M_USPC_PROT_R);
	int dst_err = mem_validate(&(pptr->mem), (uintptr_t)dispmem, dst_info.height * dst_info.stride, M_USPC_PROT_W);
	process_unlock(pptr); pptr = NULL;
	if(src_err < 0)
		return src_err;
	if(dst_err < 0)
		return dst_err;
	
	//Scale to output size - naive 2D DDA algorithm for now
	//Todo - respect aspect ratio differences
	//Todo - pixel format conversion
	//Todo - basically everything
	for(int yo = 0; yo < dst_info.height; yo++)
	{
		//Line sizes are specified by the info buffers, as stride
		int yi = yo * src_info.height / dst_info.height;
		void *srcline = ((uint8_t*)(srcmem)) + (yi * src_info.stride);
		void *dstline = ((uint8_t*)(dispmem)) + (yo * dst_info.stride);
		
		//Pixel sizes are determined by the pixel-format
		static const int pxsize[_SC_CON_PXFMT_MAX] = 
		{
			[_SC_CON_PXFMT_Y8] = 1,
			[_SC_CON_PXFMT_RGB565] = 2,
			[_SC_CON_PXFMT_RGB888] = 3,
			[_SC_CON_PXFMT_RGBX8888] = 4,
		};
	
		uint8_t *input_pixel = ((uint8_t*)srcline);
		uint8_t *output_pixel = ((uint8_t*)dstline);
		int xrem = 0;
		for(int xo = 0; xo < dst_info.width; xo++)
		{
			//Todo - format conversion
			memcpy(output_pixel, input_pixel, pxsize[dst_info.pxfmt]);
			
			output_pixel += pxsize[dst_info.pxfmt];
			xrem += src_info.width;
			while(xrem >= dst_info.width)
			{
				xrem -= dst_info.width;
				input_pixel += pxsize[src_info.pxfmt];
			}
		}
	}

	return 0;	
}

int k_sc_con_fbdisp(const void *dispmem)
{	
	//Make sure the caller has the console
	if(!con_on(process_curpid()))
	{
		//Calling process doesn't hold the console right now - can't flip
		return -EBUSY;
	}

	process_t *pptr = process_lockcur();
	KASSERT(pptr != NULL);
	
	//Find the physical address of the framebuffer we'll be displaying, given the user's virtual address for it
	uintptr_t fbaddr = 0;
	for(int mm = 0; mm < MEM_SEG_MAX; mm++)
	{
		if(pptr->mem.segs[mm].disp == 0)
			continue; //Not a displayable segment
		
		if(pptr->mem.segs[mm].vaddr != (uintptr_t)dispmem)
			continue; //Not the segment the user specified in this call
		
		fbaddr = pptr->mem.segs[mm].disp;
		break;
	}

	if(fbaddr == 0)
	{
		//The memory address specified isn't a displayable memory region in the calling process.
		process_unlock(pptr);
		return -EFAULT;
	}
	
	//Don't need any more information from the calling process
	process_unlock(pptr);
	pptr = NULL;
	
	//Flush the cache corresponding to the user's buffer
	m_cache_finish(dispmem, m_video_fbsize());
	
	//Display the requested physical buffer at the next vertical blanking interval
	return fb_flip(fbaddr);
}

int k_sc_con_input(_sc_con_input_t *buf_ptr, int each_bytes, int buf_bytes)
{
	//Make sure the caller has the console
	if(!con_on(process_curpid()))
	{
		//Process doesn't have the console. Return no input.
		return 0;
	}
	
	//Todo - can version this based on length of input event structure
	if(each_bytes != sizeof(_sc_con_input_t))
		return -EINVAL;
	
	process_t *pptr = process_lockcur();
	
	//Ensure the memory for events is writable
	int memvalid = mem_validate(&(pptr->mem), (uintptr_t)buf_ptr, each_bytes * buf_bytes, M_USPC_PROT_W);
	if(memvalid < 0)
	{
		//Process doesn't have writable memory for events output
		process_unlock(pptr);
		return memvalid;
	}
	
	ssize_t retval = con_input(buf_ptr, buf_bytes);
	process_unlock(pptr);
	return retval;
}

int k_sc_con_pass(pid_t next)
{
	//If we're delegating the console to another process, make sure it's our child
	pid_t caller_pid = process_curpid();
	if(next != 0 && next != caller_pid)
	{
		process_t *child_pptr = process_lockpid(next);
		if(child_pptr == NULL)
		{
			//The specified process-ID doesn't exist.
			return -ESRCH;
		}
		
		if(child_pptr->ppid != caller_pid)
		{
			//The specified process ID isn't a child of the calling process.
			//Todo - check if it's a descendant at all - grandchild etc
			//process_unlock(child_pptr);
			//return -ESRCH;
		}
		
		//Child process ID OK
		process_unlock(child_pptr);
		child_pptr = NULL;
	}
		
	//Change the console delegation in the calling process
	process_t *pptr = process_lockcur();
	pid_t oldval = pptr->conchild;
	pptr->conchild = next;
	process_unlock(pptr);
	pptr = NULL;
	
	//Cheap hack - unpause everybody when changing console ownership
	if(next != oldval)
	{
		for(int tt = 0; tt < THREAD_MAX; tt++)
		{
			thread_unpause(thread_table[tt].tid);
		}
	}
	
	return 0;
}

int k_sc_con_flag(int how, int bits)
{
	if(how < 0 || bits < 0)
		return -EINVAL;
	
	process_t *pptr = process_lockcur();
	
	int oldflags = pptr->conflags;	
	switch(how)
	{
		case _SC_CON_FLAG_ENABLE:
			pptr->conflags |= bits;
			break;
		case _SC_CON_FLAG_DISABLE:
			pptr->conflags &= ~(bits);
			break;
		case _SC_CON_FLAG_SETALL:
			pptr->conflags = bits;
			break;
		default:
			process_unlock(pptr);
			return -EINVAL;
	}
	
	KASSERT(pptr->conflags >= 0);
	KASSERT(oldflags >= 0);
	
	process_unlock(pptr);
	return oldflags;
}
