//syscalls_sig.c
//System call declarations on kernel-side - signal handling
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "kstd.h"
#include "syscalls.h"
#include "thread.h"
#include "kassert.h"

int k_sc_sig_mask(int how, void *mask_ptr, int mask_len)
{
	//Cap length of signal-mask to 64-bit value
	if(mask_len < 0)
		return -EINVAL;
	if(mask_len > 8)
		mask_len = 8;
	
	//Lock thread and set aside old value
	thread_t *tptr = thread_lockcur();
	int64_t retval = tptr->sigmask;
	
	//Read new value that the user wants to set
	int64_t mask_data = 0;
	ssize_t mread = process_memget(&mask_data, mask_ptr, mask_len);
	if(mread < 0)
	{
		thread_unlock(tptr);
		return mread;
	}
	
	//Alter the value in the thread
	switch(how)
	{
		case _SC_SIG_BLOCK:
		{
			tptr->sigmask |= mask_data;
			break;
		}
		case _SC_SIG_SETMASK:
		{
			tptr->sigmask = mask_data;
			break;
		}
		case _SC_SIG_UNBLOCK:
		{
			tptr->sigmask &= ~mask_data;
			break;
		}
		default:
		{
			thread_unlock(tptr);
			return -EINVAL;
		}
	}
	
	//Disallow invalid masks - nobody's allowed to mask KILL/STOP
	tptr->sigmask &= 0x7FFFFFFFFFFFFFFFul;
	tptr->sigmask &= ~(1ul << _SC_SIGKILL);
	tptr->sigmask &= ~(1ul << _SC_SIGSTOP);
	
	//Write out the old value so the user can see what it was
	ssize_t mwrite = process_memput(mask_ptr, &retval, mask_len);
	if(mwrite < 0)
	{
		thread_unlock(tptr);
		return mwrite;
	}
	
	//Success
	thread_unlock(tptr);
	return 0;
}

void k_sc_sig_return(void)
{
	thread_t *tptr = thread_lockcur();
	
	//Restore interrupted user-context and signal-mask
	tptr->drop = tptr->presig_drop;
	tptr->sigmask = tptr->presig_sigmask;
	
	thread_unlock(tptr);
}

int k_sc_sig_send(int idtype, pid_t id, int sig)
{
	if(sig < 0 || sig >= 63)
		return -EINVAL;
	
	//Get PID/TID of the caller who's sending the signal
	thread_t *senderptr = thread_lockcur();
	int senderpid = senderptr->process->pid;
	int sendertid = senderptr->tid;
	thread_unlock(senderptr);
	senderptr = NULL;
	
	bool any_signal = false;
	for(int tt = 0; tt < THREAD_MAX; tt++)
	{
		thread_t *tptr = &(thread_table[tt]);
		m_spl_acq(&(tptr->spl));
		
		if(tptr->state == THREAD_STATE_NONE)
		{
			//No thread here
			m_spl_rel(&(tptr->spl));
			continue;
		}
		
		bool id_match = false;
		switch(idtype)
		{
			case _SC_IDTYPE_PID:   id_match = (tptr->process->pid == id); break;
			case _SC_IDTYPE_PTREE: id_match = process_lineage(tptr->process->pid, id); break;
			case _SC_IDTYPE_TID:   id_match = (tptr->tid == id); break;
			case _SC_IDTYPE_ALL:   id_match = true; break;
			default: m_spl_rel(&(tptr->spl)); return -EINVAL;
		}
		
		if(!id_match)
		{
			//Not a thread we're looking for
			m_spl_rel(&(tptr->spl));
			continue;
		}
		
		//Okay, this is a thread we want to signal. We'll consider this a success.
		any_signal = true;
		
		tptr->sigpend |= (1ul << sig);
		tptr->sigcause_addr = 0;
		tptr->sigcause_pid = senderpid;
		tptr->sigcause_tid = sendertid;
		thread_unpause(tptr->tid);
		
		m_spl_rel(&(tptr->spl));
	}
	
	return any_signal ? 0 : -ESRCH;
}
