//syscalls_ptr.c
//System call declarations on kernel-side - trace and friends
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "kstd.h"
#include "syscalls.h"
#include "thread.h"
#include "kassert.h"
#include "m_uspc.h"

int k_sc_trace_attach(int tid)
{
	thread_t *caller = thread_lockcur();
	if(caller->tid == tid)
	{
		//Thread is trying to trace itself
		thread_unlock(caller);
		return -EDEADLK;
	}
		
	thread_t *target = thread_locktid(tid);
	if(target == NULL)
	{
		//Didn't find the thread to trace
		thread_unlock(caller);
		return -ESRCH;
	}
	
	if(target->trace_host != 0 && target->trace_host != caller->tid)
	{
		//Someone else is already tracing the thread we wanted to trace
		thread_unlock(caller);
		thread_unlock(target);
		return -EBUSY;
	}
	
	target->trace_host = caller->tid;
	thread_unlock(caller);
	thread_unlock(target);
	return 0;
}

int k_sc_trace_detach(int tid)
{
	thread_t *caller = thread_lockcur();
	if(caller->tid == tid)
	{
		//Thread is trying to un-trace itself
		thread_unlock(caller);
		return -EDEADLK;
	}
	
	thread_t *target = thread_locktid(tid);
	if(target == NULL)
	{
		//Didn't find the thread to release
		thread_unlock(caller);
		return -ESRCH;
	}
	
	if(target->trace_host != caller->tid)
	{
		//Not tracing them to begin with
		thread_unlock(caller);
		thread_unlock(target);
		return -EPERM;
	}
	
	target->trace_host = 0;
	thread_unlock(caller);
	thread_unlock(target);
	return 0;
}

int k_sc_trace_stop(int tid, _sc_trace_reason_t reason)
{	
	thread_t *caller = thread_lockcur();
	if(caller->tid == tid)
	{
		//Thread is trying to trace itself
		thread_unlock(caller);
		return -EDEADLK;
	}
	
	thread_t *target = thread_locktid(tid);
	if(target == NULL)
	{
		//Can't find target thread
		thread_unlock(caller);
		return -ESRCH;
	}
	
	if(target->trace_host != caller->tid)
	{
		//Not tracing them to begin with
		thread_unlock(caller);
		thread_unlock(target);
		return -EPERM;
	}
	
	int retval = target->trace_stopped;
	
	if(reason >= 0 && reason < _SC_TRACE_REASON_MAX)
		target->trace_stopped = reason;
	
	thread_unlock(caller);
	thread_unlock(target);
	KASSERT(retval >= 0);
	return retval;
}

int k_sc_trace_write(int tid, _sc_trace_space_t space, int addr, const void *src, int len)
{
	if(space <= 0 || space >= _SC_TRACE_SPACE_MAX)
		return -EINVAL;
	
	if(len < 0)
		return -EINVAL;
	

	thread_t *caller = thread_lockcur();
	if(caller->tid == tid)
	{
		//Thread is trying to trace itself
		thread_unlock(caller);
		return -EDEADLK;
	}
	
	thread_t *target = thread_locktid(tid);
	if(target == NULL)
	{
		//Can't find target thread
		thread_unlock(caller);
		return -ESRCH;
	}
	
	if(target->trace_host != caller->tid)
	{
		//Not tracing them to begin with 
		thread_unlock(caller);
		thread_unlock(target);
		return -EPERM;
	}
	
	//Make sure the data to write is accessible to read in the calling process
	int buf_err = mem_validate(&(caller->process->mem), (uintptr_t)src, len, M_USPC_PROT_R);
	if(buf_err < 0)
	{
		thread_unlock(caller);
		thread_unlock(target);
		return buf_err;
	}
	
	//Don't need calling thread info anymore...
	thread_unlock(caller);
	caller = NULL;
	
	switch(space)
	{
		case _SC_TRACE_SPACE_MEMORY:
		{
			//Validate that the target memory can be written here
			int target_valid = mem_validate(&(target->process->mem), addr, len, M_USPC_PROT_W);
			if(target_valid < 0)
			{
				//Can't write to this location in the target's process memory space
				thread_unlock(target);
				return target_valid;
			}
			
			//Copy between the two userspaces through a temp buffer in the kernel
			m_uspc_t caller_uspc = m_uspc_current();
			m_uspc_t target_uspc = target->process->mem.uspc;
			KASSERT(caller_uspc != 0);
			KASSERT(target_uspc != 0);
			
			ssize_t len_remain = len;
			const char *src_remain = (const char*)src;
			char *dst_remain = (char*)(uintptr_t)addr;
			ssize_t copied = 0;
			while(len_remain > 0)
			{
				uint8_t kbuf[256];
				ssize_t burst = len_remain;
				if(burst > (int)sizeof(kbuf))
					burst = (int)sizeof(kbuf);
				
				m_uspc_activate(caller_uspc);
				memcpy(kbuf, src_remain, burst);
				m_uspc_activate(target_uspc);
				memcpy(dst_remain, kbuf, burst);
				
				src_remain += burst;
				dst_remain += burst;
				len_remain -= burst;
				copied += burst;
			}
			
			m_uspc_activate(caller_uspc);
			thread_unlock(target);
			return copied;
		}
		break;
		case _SC_TRACE_SPACE_REGISTER:
		{
			//Get the register data from the calling process, sanitize it, and overwrite the target's registers
			m_drop_t tempdrop = {0};
			if(len > (int)sizeof(tempdrop))
				len = sizeof(tempdrop);
			
			memcpy(&tempdrop, src, len);
			m_drop_sanitize(&tempdrop);
			memcpy(&(target->drop), &tempdrop, sizeof(target->drop));
			thread_unlock(target);
			return len;
		}
		break;		
		case _SC_TRACE_SPACE_SIGMASK:
		{
			if(len > (int)sizeof(target->sigmask))
				len = sizeof(target->sigmask);
			
			memcpy(&(target->sigmask), src, len);
			thread_unlock(target);
			return len;
		}
		break;
		case _SC_TRACE_SPACE_SIGPEND:
		{
			if(len > (int)sizeof(target->sigpend))
				len = sizeof(target->sigpend);
			
			memcpy(&(target->sigpend), src, len);
			thread_unlock(target);
			return len;
		}
		break;
		default:
			KASSERT(0);
	}
	
	return -ENOSYS;
}

int k_sc_trace_read(int tid, _sc_trace_space_t space, int addr, void *dst, int len)
{
	if(space <= 0 || space >= _SC_TRACE_SPACE_MAX)
		return -EINVAL;
	
	if(len < 0)
		return -EINVAL;
	
	thread_t *caller = thread_lockcur();
	if(caller->tid == tid)
	{
		//Thread trying to trace itself
		thread_unlock(caller);
		return -EDEADLK;
	}
	
	thread_t *target = thread_locktid(tid);
	if(target == NULL)
	{
		//No such thread
		thread_unlock(caller);
		return -ESRCH;
	}
	
	if(target->trace_host != caller->tid)
	{
		//Caller isn't tracing this thread
		thread_unlock(caller);
		thread_unlock(target);
		return -EPERM;
	}
	
	//Make sure the buffer for results is accessible for writing in the calling process.
	int buf_err = mem_validate(&(caller->process->mem), (uintptr_t)dst, len, M_USPC_PROT_W);
	if(buf_err < 0)
	{
		thread_unlock(caller);
		thread_unlock(target);
		return buf_err;
	}
		
	//Don't need calling thread info anymore...
	thread_unlock(caller);
	caller = NULL;
	
	switch(space)
	{
		case _SC_TRACE_SPACE_MEMORY:
		{
			//Validate that the target memory can be read here
			int target_valid = mem_validate(&(target->process->mem), addr, len, M_USPC_PROT_R);
			if(target_valid < 0)
			{
				//Can't read from this location in the target's process memory space
				thread_unlock(target);
				return target_valid;
			}
			
			//Copy between the two userspaces through a temp buffer in the kernel
			m_uspc_t caller_uspc = m_uspc_current();
			m_uspc_t target_uspc = target->process->mem.uspc;
			KASSERT(caller_uspc != 0);
			KASSERT(target_uspc != 0);
			
			ssize_t len_remain = len;
			const char *src_remain = (const char*)(uintptr_t)addr;
			char *dst_remain = (char*)dst;
			ssize_t copied = 0;
			while(len_remain > 0)
			{
				uint8_t kbuf[256];
				ssize_t burst = len_remain;
				if(burst > (int)sizeof(kbuf))
					burst = (int)sizeof(kbuf);
				
				m_uspc_activate(target_uspc);
				memcpy(kbuf, src_remain, burst);
				m_uspc_activate(caller_uspc);
				memcpy(dst_remain, kbuf, burst);
				
				src_remain += burst;
				dst_remain += burst;
				len_remain -= burst;
				copied += burst;
			}
			
			m_uspc_activate(caller_uspc);
			thread_unlock(target);
			return copied;
		}
		break;
		case _SC_TRACE_SPACE_REGISTER:
		{
			//Make a clean copy of the registers from the target thread
			m_drop_t tempdrop = {0};
			memcpy(&tempdrop, &(target->drop), sizeof(tempdrop));
			if(len > (int)sizeof(tempdrop))
				len = sizeof(tempdrop);
			
			//Return it to userland
			memcpy(dst, &tempdrop, len);
			thread_unlock(target);
			return len;
		}
		break;		
		case _SC_TRACE_SPACE_SIGMASK:
		{
			//Read signal mask from target
			if(len > (int)sizeof(target->sigmask))
				len = sizeof(target->sigmask);
			
			memcpy(dst, &(target->sigmask), len);
			thread_unlock(target);
			return len;
		}
		break;
		case _SC_TRACE_SPACE_SIGPEND:
		{
			//Read signals pending from target
			if(len > (int)sizeof(target->sigpend))
				len = sizeof(target->sigpend);
			
			memcpy(dst, &(target->sigpend), len);
			thread_unlock(target);
			return len;
		}
		break;
		default:
			KASSERT(0);
	}
	
	return -ENOSYS;
}
