//thread.c
//User threads in kernel
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include "thread.h"
#include "m_intr.h"
#include "m_tls.h"
#include "m_cache.h"
#include "m_drop.h"
#include "kassert.h"
#include "con.h"
#include "kstd.h"
#include "panic.h"

//All threads on the system.
thread_t thread_table[THREAD_MAX];

//How many times we've halved the time-counts in the threads, for scheduling.
static int thread_halvings;

thread_t *thread_lockfree(void)
{
	for(int tt = 1; tt < THREAD_MAX; tt++)
	{
		//Don't bother fighting over a lock - locked thread is necessarily in use.
		thread_t *tptr = &(thread_table[tt]);
		if(m_spl_try(&(tptr->spl)))
		{
			if(tptr->state == THREAD_STATE_NONE)
			{
				//Found a free spot. 
				
				//Make sure it's got a valid ID
				tptr->tid += THREAD_MAX;
				if( (tptr->tid <= 0) || ((tptr->tid % THREAD_MAX) != tt) )
					tptr->tid = tt;
				
				//Return it, still locked.
				return tptr;
			}
			
			//Not free. Unlock and keep looking.
			m_spl_rel(&(tptr->spl));
		}
	}
	
	//Didn't find a spot.
	return NULL;
}

int thread_new(process_t *process, thread_t **thread_out)
{
	thread_t *tptr = thread_lockfree();
	if(tptr == NULL)
	{
		//No room for new threads.
		return -ENOMEM;
	}
	
	thread_chstate(tptr, THREAD_STATE_SUSPEND);
	tptr->process = process;
	
	//Threads start with all signals masked, so they don't catch them before they're ready.
	tptr->sigmask = 0x7FFFFFFFFFFFFFFFul;
	tptr->sigpend = 0;
	
	memset(&(tptr->drop), 0, sizeof(tptr->drop));
	
	process->nthreads++;
	
	//Success
	*thread_out = tptr;
	return 0;
}

thread_t *thread_locktid(pid_t tid)
{
	if(tid < 0)
		return NULL;
	
	//Array index corresponds to ID
	thread_t *tptr = &(thread_table[tid % THREAD_MAX]);
	m_spl_acq(&(tptr->spl));
	
	if( (tptr->state == THREAD_STATE_NONE) || (tptr->tid != tid) )
	{
		//No/wrong thread here.
		m_spl_rel(&(tptr->spl));
		return NULL;
	}
	
	//Got the thread we wanted - return, still locked
	return tptr;
}

thread_t *thread_lockcur(void)
{
	thread_t *tptr = m_tls_get();
	m_spl_acq(&(tptr->spl));
	return tptr;
}

void thread_unlock(thread_t *thread)
{
	m_spl_rel(&(thread->spl));
}

void thread_chstate(thread_t *thread, thread_state_t newstate)
{
	KASSERT(thread->state >= THREAD_STATE_NONE && thread->state < THREAD_STATE_MAX);
	KASSERT(newstate < THREAD_STATE_MAX);
	
	int64_t tsc_last = thread->tsc_last;
	int64_t tsc_new = m_intr_ticks();
	
	thread->tsc_totals[thread->state] += (tsc_new - tsc_last);
	
	if(thread->state == THREAD_STATE_RUN || thread->state == THREAD_STATE_SYSCALL)
		thread->sched_usage += (tsc_new - tsc_last);
	
	thread->tsc_last = tsc_new;
	thread->state = newstate;
}

void thread_unpause(pid_t tid)
{
	//CAN BE CALLED FROM ISR.
	
	//This kinda races but we don't care.
	//If the thread was cleaned-up or replaced then whatever, there's no harm in unpausing someone else.
	KASSERT(tid >= 0);
	thread_t *tptr = &(thread_table[tid % THREAD_MAX]);
	
	tptr->pause_wake_triggered = true;
	
	//m_cache_finish(&(tptr->unpaused), sizeof(tptr->unpaused));
	
	//Send interprocessor interrupt after incrementing the unpauses count.
	//This way, anyone who saw the old count will receive a pending interrupt after they see it.
	//They'll have interrupts disabled while looking at the old count - so they'll wake immediately when they try to sleep.
	m_intr_wake();
}

pid_t thread_curtid(void)
{
	thread_t *tptr = m_tls_get();
	if(tptr == NULL)
		return 0;

	return 	tptr->tid;
}

void thread_cleanup(thread_t *tptr)
{
	//Set aside process that contained the thread
	pid_t oldpid = tptr->process->pid;
	
	//Clear thread structure
	thread_chstate(tptr, THREAD_STATE_NONE);
	memset(tptr->tsc_totals, 0, sizeof(tptr->tsc_totals));
	tptr->process = NULL;
	tptr->drop = (m_drop_t){0};
	tptr->presig_drop = (m_drop_t){0};
	tptr->presig_sigmask = 0;
	tptr->sigmask = 0;
	tptr->sigpend = 0;
	tptr->pause_called = false;
	tptr->trace_host = 0;
	tptr->trace_stopped = 0;
	thread_unlock(tptr);
	
	//Reduce the thread-count of the process that the thread was a part of.
	process_t *pptr = process_lockpid(oldpid);
	KASSERT(pptr != NULL);
	
	KASSERT(pptr->nthreads > 0);
	pptr->nthreads--;
	
	//If the process has no threads, it's dead all the way
	if(pptr->nthreads == 0)
	{		
		//Clean up the process.
		m_uspc_activate(0);
		mem_clear(&(pptr->mem));
		mem_clear(&(pptr->mem_attempt));
		
		for(int ff = 0; ff < PROCESS_FD_MAX; ff++)
		{
			if(pptr->fds[ff].file != NULL)
			{
				file_lock(pptr->fds[ff].file);
				pptr->fds[ff].file->refs--;
				file_unlock(pptr->fds[ff].file);
			}
			memset(&(pptr->fds[ff]), 0, sizeof(pptr->fds[ff]));
		}
		
		if(pptr->pwd != NULL)
		{
			file_lock(pptr->pwd);
			pptr->pwd->refs--;
			file_unlock(pptr->pwd);
		}
		pptr->pwd = NULL;
		
		//Somebody should have called exit() on the process, rather than just killing all its threads.
		//But, if all threads died and none wanted to kill the process, just exit with code 0.
		if(pptr->wstatus == 0)
			pptr->wstatus = _SC_STATUS_EXITED_BIT;
		
		//Process stops participating in console on death
		pptr->conchild = 0;
		pptr->conflags = 0;
		
		//Process is now dead, and only exists to be waited-upon by its parent.
		pptr->state = PROCESS_STATE_DEAD;
		pid_t ppid = pptr->ppid;
		process_unlock(pptr);
		pptr = NULL;
		
		//Unpause all threads in the parent process, so one can wait on the now-dead child process
		for(int tt = 0; tt < THREAD_MAX; tt++)
		{
			m_spl_acq(&(thread_table[tt].spl));
			if(thread_table[tt].state != THREAD_STATE_NONE)
			{
				if(thread_table[tt].process->pid == ppid)
				{
					thread_table[tt].sigpend |= (1u << _SC_SIGCHLD);
					thread_unpause(thread_table[tt].tid);
				}
			}
			m_spl_rel(&(thread_table[tt].spl));
		}
		
		//Special-case - if PID1 died...
		if(ppid == 0)
		{
			panic("PID1 died.");
			//process_init(); //Respawn it
		}
	}
	else
	{
		//Process still has threads - don't clean it up yet.
		process_unlock(pptr);
	}
}

//Checks whether the given thread, already locked, can be scheduled
static bool thread_consume_pause(thread_t *tptr)
{
	if(!(tptr->pause_called))
	{
		//Thread didn't call pause(). Allow them to run with no fuss.
		return true;
	}
	
	//Hack - I thought I'd add some safety by checking if they did something waitable before calling pause.
	//i.e. don't let them simply call pause() and then get paused forever, if nobody's coming back for them.
	//However, this doesn't mesh with the idea of signals or child-deaths unpausing a thread.
	//To wait on those events would then be impossible.
	tptr->pause_wake_queued = true;
	
	if((tptr->pause_wake_queued) && !(tptr->pause_wake_triggered))
	{
		//Thread called pause, and we think they have something to wait for.
		//It hasn't happened yet.
		return false;
	}
	
	//Thread was paused and then woken up.
	tptr->pause_called = false;
	tptr->pause_wake_queued = false;
	tptr->pause_wake_triggered = false;
	return true;
}

//Handles cleanup for a thread after the kernel has performed service on behalf of it.
static void thread_postkernel(void)
{
	//Grab the thread we just finished processing - may be "none" on first entry to the kernel
	if(m_tls_get() == NULL)
		return;

	thread_t *tptr = thread_lockcur();
	
	//Check if the thread belongs to a dying process
	if(tptr->process->state != PROCESS_STATE_ALIVE)
	{
		//Process doesn't want to be alive - threads should all exit
		thread_chstate(tptr, THREAD_STATE_DEAD);
	}
	
	//Clean up dead threads with no further mess
	if(tptr->state == THREAD_STATE_DEAD)
	{
		//Thread is dead, clean it up and find another
		m_uspc_activate(0);
		thread_cleanup(tptr);
		return;
	}
	
	//If the thread has a pending signal it can handle, scoot it into the signal handler.
	int64_t triggered = tptr->sigpend & ~(tptr->sigmask);
	if(triggered)
	{
		//Figure out which signal to trigger.
		int signum = -1;
		for(int bb = 0; bb < 64; bb++)
		{
			if(triggered & (1 << bb))
			{
				signum = bb;
				break;
			}
		}
		KASSERT(signum >= 0 && signum < 63);
		
		//Set aside info about thread at the time it was signalled
		tptr->presig_drop = tptr->drop;
		tptr->presig_sigmask = tptr->sigmask;
		
		//Mask further signals
		tptr->sigmask = 0x7FFFFFFFFFFFFFFFul;
		
		//Reset to the entry point to handle the signal.
		//Feed info about the signal being taken, into the parameter registers.
		m_drop_entry(&(tptr->drop), tptr->process->entry);
		m_drop_param(&(tptr->drop), 0, signum);
		m_drop_param(&(tptr->drop), 1, tptr->sigcause_addr);
		m_drop_param(&(tptr->drop), 2, tptr->sigcause_pid);
		m_drop_param(&(tptr->drop), 3, tptr->sigcause_tid);

		//Thread unpauses when signalled, of course
		tptr->pause_wake_triggered = true;
		
		//Signal is no longer pending
		tptr->sigpend &= ~(1u << signum);
		
		//If we're being traced, we've now stopped due to a signal
		tptr->trace_stopped = _SC_TRACE_REASON_SIGNAL;
	}
	
	//When a thread is scheduled, it gets a window where the scheduler won't run again.
	//If we're still within that window, continue the thread rather than scheduling again.
	if(tptr->tsc_resched > m_intr_ticks())
	{
		if(!(tptr->trace_host && tptr->trace_stopped))
		{
			if(thread_consume_pause(tptr))
			{
				thread_chstate(tptr, THREAD_STATE_RUN);
				thread_unlock(tptr);
				m_drop(&(tptr->drop));
			}
		}
	}
	
	//Otherwise, we're due to schedule some other thread.
	//If we're going to be leaving the thread, we can't assume its memory space remains valid.
	//If the thread gets killed on another CPU and they clean up the process, it can disappear.
	m_uspc_activate(0);
	
	thread_chstate(tptr, THREAD_STATE_SUSPEND);
	thread_unlock(tptr);
	return;
}

static void thread_sched(void)
{
	//We'll spin here until we have useful work to do
	while(1)
	{
		//Disable interrupts while trying to schedule.
		//If a thread becomes runnable while we search, then, the resulting interrupt will be waiting for us.
		m_intr_ei(false);
		
		//Look for the best thread to run
		thread_t *best = NULL;
		for(int tt = 0; tt < THREAD_MAX; tt++)
		{
			//Try to lock this thread to possibly execute it.
			thread_t *candidate = &(thread_table[tt]);
			if(!m_spl_try(&(candidate->spl)))
				continue;
			
			//Update scheduling first
			
			if(candidate->sched_usage > 1024)
			{
				//We've accumulated a whole second of usage on some thread.
				//Cut everyone's billing in half, to exponentially decay old usage.
				thread_halvings++;
			}
			
			if(candidate->sched_halvings < thread_halvings)
			{
				//We've decided to cut everyone's accumulated usage in half, but this thread wasn't hit yet.
				//Do the halving for this thread.
				candidate->sched_usage /= 2;
				candidate->sched_halvings = thread_halvings;
			}
			
			//See if we can actually run this thread
			
			if(candidate->state != THREAD_STATE_SUSPEND)
			{
				m_spl_rel(&(candidate->spl));
				continue;
			}
			
			if(candidate->trace_host && candidate->trace_stopped)
			{
				m_spl_rel(&(candidate->spl));
				continue;
			}
			
			if(!thread_consume_pause(candidate))
			{
				m_spl_rel(&(candidate->spl));
				continue;
			}
			
			//See if it's better than our best choice
			
			if(best != NULL)
			{
				//We previously found a runnable thread... is this a worse choice for scheduling?
				if(candidate->sched_usage >= best->sched_usage)
				{
					m_spl_rel(&(candidate->spl));
					continue;
				}
			}
			
			//Okay, this is a new better choice.
			if(best != NULL)
				m_spl_rel(&(best->spl));
			
			best = candidate;
		}
		
		if(best == NULL)
		{
			//No threads runnable right now.
			//Wait for an interprocessor interrupt that might indicate something to do.
			m_intr_halt();
			
			//Try again to find a runnable thread.
			m_intr_ei(true);
			continue;
		}
		
		//Got a thread, it's ready, and we've locked it.
		thread_t *tptr = best;
		
		//Note which thread we'll be running on this core, as its kernel stack/context is about to be clobbered.
		m_tls_set(tptr);
		
		//Note when we should consider kicking the thread off the CPU
		tptr->tsc_resched = m_intr_ticks() + 200;
		
		//Mark the thread as running, and resume its userspace.
		//(Assume nobody's messing with this, if the thread is marked "running", even though we release the lock)
		KASSERT(tptr->state == THREAD_STATE_SUSPEND);
		thread_chstate(tptr, THREAD_STATE_RUN);
		thread_unlock(tptr);
		
		m_uspc_activate(tptr->process->mem.uspc);
		m_drop(&(tptr->drop));
		
		//m_drop doesn't return.
		//When we re-enter the kernel, though, we'll know which thread triggered it, thanks to the TLS pointer.
	}	
}

void thread_drop(void)
{
	//If we just serviced a thread, see if there's any cleanup to be done on it.
	thread_postkernel();
	
	//Pick another thread to run.
	while(1)
	{
		thread_sched();
	}
}
