//thread.h
//User threads in kernel
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#ifndef THREAD_H
#define THREAD_H

#include "m_spl.h"
#include "m_drop.h"
#include "process.h"
#include "kstd.h"
#include <sc.h>

//State a thread can be in
typedef enum thread_state_e
{
	THREAD_STATE_NONE = 0, //None/invalid
	
	THREAD_STATE_SUSPEND, //Thread is not executing - can be scheduled if pause/ptrace data permits
	THREAD_STATE_RUN, //Thread is currently executing on some core
	THREAD_STATE_SYSCALL, //Thread is having a system-call serviced
	THREAD_STATE_DEAD, //Thread has exited, but status is not delivered to its parent yet
	
	THREAD_STATE_MAX
	
} thread_state_t;

//Thread control block
typedef struct thread_s
{
	m_spl_t spl; //Spinlock protecting the thread control block

	//Fundamental parts of a thread
	pid_t tid; //ID of the thread
	thread_state_t state; //State of thread control block
	process_t *process; //Process containing the thread
	m_drop_t drop; //User context
	
	//State of signal handling
	int64_t sigmask; //Signals blocked (1 = blocked)
	int64_t sigpend; //Signals pending (1 = pending)
	
	//Information about returning from last signal
	m_drop_t presig_drop; //Context preserved when a signal is taken
	int64_t presig_sigmask; //Old signal mask preserved when a signal is taken
	
	//Information about cause of last signal
	int32_t sigcause_addr; //The last memory address responsible for causing a signal
	int sigcause_pid; //The last process responsible for causing a signal
	int sigcause_tid; //The last thread responsible for causing a signal
	
	//Pause/unpause state
	bool pause_called; //Whether the thread just made a pause system-call
	bool pause_wake_queued; //Whether the thread has been added to some list of waiters
	bool pause_wake_triggered; //Whether the thread has experienced a wakeup event
	
	//Ptrace related state
	pid_t trace_host; //Thread that is tracing this thread
	int trace_stopped; //Whether this thread is stopped due to tracing, and why

	//Timing tracking
	int64_t tsc_totals[THREAD_STATE_MAX]; //How much time has been spent in each state
	int64_t tsc_last; //Timestamp when state was last changed
	int64_t tsc_resched; //Timestamp at which we'll make another scheduling decision
	
	//Scheduling
	int32_t sched_usage; //Accumulation of how many resources this thread has used
	int32_t sched_halvings; //How many times we've cut the "sched_usage" in half, to exponentially decay it
	
} thread_t;

//All threads in the system
#define THREAD_MAX 64
extern thread_t thread_table[THREAD_MAX];

//Makes a new thread. Outputs a pointer to it, still locked.
//Returns 0 on success or a negative error number.
int thread_new(process_t *process, thread_t **thread_out);

//Locks a free entry in the thread table and returns a pointer to it.
thread_t *thread_lockfree(void);

//Locks the thread with the given TID, if it exists, and returns a pointer to it.
//Returns NULL if the thread doesn't exist.
thread_t *thread_locktid(pid_t tid);

//Locks the current thread and returns a pointer to it.
thread_t *thread_lockcur(void);

//Unlocks the given thread.
void thread_unlock(thread_t *thread);

//Changes the state of the given thread, updating its accounting
void thread_chstate(thread_t *thread, thread_state_t newstate);

//Causes a thread to resume if waiting, or skip its next waiting if already ready.
//CAN BE CALLED FROM ISR.
void thread_unpause(pid_t tid);

//Returns the thread ID of the current thread.
pid_t thread_curtid(void);

//Cleans up the given thread.
void thread_cleanup(thread_t *thread);

//Tries to exit the kernel into a user thread that has work to do.
//Finds a runnable thread and runs it, waiting until one is runnable if necessary.
//Does not return after running the thread.
void thread_drop(void) __attribute__((noreturn));

#endif //THREAD_H
