;crt0.asm
;Entry point and machine features for C runtime on AMD64
;Bryan E. Topp <betopp@betopp.com> 2022

;This file is part of Stump; Copyright Bryan E. Topp, 2023.
;Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
;Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
;You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

section .text
	
;Entry point
global _crt_entry
_crt_entry:

	;Check if this is a signal entry. RDI indicates which signal - or 0 for initial entry.
	cmp RDI, 0
	je .initial
		;Entry due to signal.
		;Set up stack for signal handling and call into libc.
		mov RSP, _crt_sigstack.top
		extern _libc_signalled
		call _libc_signalled
		
		;Shouldn't fall through here
		hlt
	.initial:
	
	;Initial entry.
	;Set up TLS and stack for initial thread.
	mov RAX, _crt_tls0
	wrgsbase RAX
	mov RSP, _crt_stack.top
	
	;RSI points to our argv/envp pointers. Preserve it.
	mov RBX, RSI
	
	;Restore arg/env pointers and load them up as parameters to entry.
	mov RDI, 0
	mov RSI, 0
	mov RDX, 0
	
	cmp RBX, 0
	je .noparms
		mov RSI, [RBX + 0]
		mov RDX, [RBX + 8]
	.noparms:
	
	;Call libc which will call main and then exit
	extern _libc_entry
	call _libc_entry
	
	;Shouldn't fall through here
	hlt
	
;System call shims
global _sc0
_sc0:

global _sc1
_sc1:

global _sc2
_sc2:

global _sc3
_sc3:

global _sc4
_sc4:

global _sc5
_sc5:
	;We can pass most parameters as usual.
	;RCX will get clobbered by the syscall instruction, so stick that in R10.
	mov R10, RCX
	syscall
	ret

;longjumps
global _setjmp ;extern int _setjmp(jmp_buf env);
_setjmp:

	;These always happen as part of a function call.
	;So, only registers preserved across function calls must be saved/loaded.
	mov [RDI + (8 * 0)], RBX
	mov [RDI + (8 * 1)], RSP
	mov [RDI + (8 * 2)], RBP
	mov [RDI + (8 * 3)], R12
	mov [RDI + (8 * 4)], R13
	mov [RDI + (8 * 5)], R14
	mov [RDI + (8 * 6)], R15
	
	;Also preserve the return value that we're about to pop off
	mov RAX, [RSP]
	mov [RDI + (8 * 7)], RAX
	
	;First return - _setjmp returns 0.
	mov RAX, 0
	ret

global _longjmp ;extern void _longjmp(jmp_buf env, int val);
_longjmp:

	;Restore saved registers
	mov RBX, [RDI + (8 * 0)]
	mov RSP, [RDI + (8 * 1)]
	mov RBP, [RDI + (8 * 2)]
	mov R12, [RDI + (8 * 3)]
	mov R13, [RDI + (8 * 4)]
	mov R14, [RDI + (8 * 5)]
	mov R15, [RDI + (8 * 6)]
	
	;Restore the return value (now on the old stack, but possibly clobbered)
	mov RAX, [RDI + (8 * 7)]
	mov [RSP], RAX

	;Return as if _setjmp is now returning "val".
	;But - ensure that's not 0, because that means _setjmp is returning the first time.
	mov RAX, RSI
	cmp RAX, 0
	jne .noinc
		inc RAX
	.noinc:
	ret

global sigsetjmp ;int sigsetjmp(sigjmp_buf env, int savemask)
sigsetjmp:
	;store savemask value to env[0]
	mov [RDI], RSI
	
	;if savemask is nonzero, get signals to save.
	mov RDX, 0 ;default to saving nothing
	cmp RSI, 0 ;check savemask...
	je .nosave
		push RDI ;Preserve env ptr across call to _sc_sigmask
		
		extern _sc_sig_mask ;System call to alter/get signal mask
		mov RDI, 0 ;SIG_BLOCK - but we're passing a mask of zeroes, so nothing is blocked
		mov RSI, 0 ;Block nothing
		call _sc_sig_mask
		
		mov RDX, RAX ;Saved signals returned in RAX, we want them in RDX
		pop RDI ;Restore env ptr
	.nosave:
	
	;Store the signal mask values in env[1]
	mov [RDI + 8], RDX
	
	;Continue to preserving registers in env[2...]
	add RDI, 8*2
	jmp _setjmp

global siglongjmp ;void siglongjmp(sigjmp_buf env, int val)
siglongjmp:
	;get old value of savemask from env[0] - figure out if we'll restore signals
	mov RDX, [RDI]
	cmp RDX, 0
	je .nosave
		;Preserve across call to _sc_sigmask
		push RDI
		push RSI
		
		extern _sc_sig_mask ;System call to alter/get signal mask
		mov RSI, [RDI + 8] ;Mask preserved in env[1]
		mov RDI, 2 ;SIG_SETMASK
		call _sc_sig_mask
		
		pop RSI
		pop RDI
	.nosave:
	
	;Do the rest of the jump from env[2...]
	add RDI, 8*2
	jmp _longjmp
	

;TLS and spinlock implementation
global _tls
_tls:
	rdgsbase RAX
	ret

;Single-threaded for now...
global _spl_lock ;void _spl_lock(m_spl_t *spl);
_spl_lock:
	mov AX, 0x0100 ;AL = 0, AH = 1.
	lock cmpxchg [RDI], AH ;Compares AL (0) with the spinlock. If it's 0, AH (1) is stored.
	jnz .wait ;If we failed, ZF is cleared - wait and try again
		mfence ;Make sure all prior memory operations complete before we say we "have the lock".
		ret
	.wait:
		;Didn't get the lock.
		;Use non-locked accesses to wait until the lock seems to be 0. Then try again.
		pause
		cmp byte [RDI], 0
		jne .wait
		jmp _spl_lock
		
global _spl_unlock ;void _spl_unlock(m_spl_t *spl);
_spl_unlock:
	;Just set the lock value to 0
	mov byte [RDI], 0
	ret

	
section .bss
	
;Space for stack
alignb 4096
_crt_stack:
	resb 4096 * 8
	_crt_stack.top:	

;Space for signal-handling stack
alignb 4096
_crt_sigstack:
	resb 4096 * 8
	_crt_sigstack.top:
	
section .data
	
;Space for initial thread TLS
alignb 4096
_crt_tls0:
	dq _crt_tls0 ;TLS starts with pointer to itself
	times (4096 - 8) db 0
	_crt_tls0.top:
