//sc.h
//System call library - main header
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#ifndef _SC_H
#define _SC_H

#include <sc_file.h> //File descriptor calls
#include <sc_con.h> //Console calls
#include <sc_mem.h> //Memory management calls
#include <sc_sig.h> //Signal handling calls
#include <sc_info.h> //System information calls
#include <sc_trace.h> //Tracing / debugging calls

//Tells the kernel what version of the system-call interface is expected by the calling application.
//Returns the version of system-call interface that the kernel will then use.
//Returns a negative error number on failure - if the kernel and application are simply incompatible.
//System-call versions are set per-process. Currently only version "1" is supported.
int _sc_version(int version);

//Waits until anything happens to the calling thread, or has happened since the last call returned.
//This is the only way to actually "block" your thread at the kernel level.
//When we say "anything happens", it means "anything relating to a system-call you previously made".
//So, idiomatically, you would write: while(1) { try_my_thing(); if(it_finished()) { break; } else { _sc_pause(); } }
void _sc_pause(void);

//Does nothing.
void _sc_none(void);

//Terminates the calling process. Optionally reports a signal that was responsible for its demise.
void _sc_exit(int exitcode, int signal) __attribute__((noreturn));

//Returns the process ID of the calling process.
//Does not fail.
int _sc_getpid(void);

//Returns the thread ID of the calling thread.
//Does not fail.
int _sc_gettid(void);

//Returns the process ID of the parent of the calling process.
//Does not fail. Note that the PPID of PID 1 is 0, and no other process has a PPID of 0.
int _sc_getppid(void);

//Creates a copy of the calling process.
int _sc_fork(void);

//Creates a copy of the calling thread, as part of the existing calling process.
int _sc_spoon(void);

//Bits of status information returned by wait call
#define _SC_STATUS_EXITCODE_MASK 0xFFu
#define _SC_STATUS_EXITCODE_SHIFT 0
#define _SC_STATUS_STOPSIG_MASK 0xFF00u
#define _SC_STATUS_STOPSIG_SHIFT 8
#define _SC_STATUS_TERMSIG_MASK 0xFF0000u
#define _SC_STATUS_TERMSIG_SHIFT 16
#define _SC_STATUS_CONTINUED_BIT 0x1000000u
#define _SC_STATUS_EXITED_BIT 0x2000000u
#define _SC_STATUS_SIGNALED_BIT 0x4000000u
#define _SC_STATUS_STOPPED_BIT 0x8000000u
#define _SC_STATUS_COREDUMP_BIT 0x10000000u

//Types of ID defined for calls which take a Process ID or Thread ID
#define _SC_IDTYPE_ALL 0x0 //All processes / all threads
#define _SC_IDTYPE_PID 0x1 //Process ID
#define _SC_IDTYPE_PGID 0x2 //Process group ID (not used)
#define _SC_IDTYPE_PTREE 0x3 //Process ID and its entire subtree of descendants
#define _SC_IDTYPE_TID 0xB //Thread ID

//Types of events that can be queried with a wait system-call (options parameter)
#define _SC_WSTOPPED 0x2
#define _SC_WEXITED 0x4
#define _SC_WCONTINUED 0x8
#define _SC_WNOWAIT 0x1000000

//Information returned by kernel from a successful call to _sc_wait.
typedef struct _sc_wait_s
{
	int status; //Wait status
	int pid; //Other process
	int tid; //Other thread
} _sc_wait_t;

//Checks whether another process has changed state.
//Returns the size of the structure filled, or a negative error number.
//Note that this doesn't actually block - use _sc_pause to actually do the waiting.
int _sc_wait(int idtype, int id, int options, _sc_wait_t *buf, int len);

//Returns the number of milliseconds since the system was booted.
int _sc_getticks(void);

//Causes the kernel to stop, and report the error if possible.
void _sc_panic(const char *str) __attribute__((noreturn));

#endif //_SC_H
