//sc_con.h
//System call library - console functions
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#ifndef _SC_CON_H
#define _SC_CON_H

//Pixel formats that the console might use
typedef enum _sc_con_pxfmt_s
{
	_SC_CON_PXFMT_NONE = 0,
	
	_SC_CON_PXFMT_Y8       = 1, //1 byte per pixel, greyscale 8b
	_SC_CON_PXFMT_RGB565   = 2, //2 bytes per pixel, RGB 5b, 6b, 5b
	_SC_CON_PXFMT_RGB888   = 3, //3 bytes per pixel, RGB 8b, 8b, 8b
	_SC_CON_PXFMT_RGBX8888 = 4, //4 bytes per pixel, RGB 8b, 8b, 8b, with pad
	
	_SC_CON_PXFMT_MAX
} _sc_con_pxfmt_t;

//Information about a framebuffer to display on the console.
//This may describe the correct format for the user's screen, when returned by _sc_con_fbinfo.
//Or, this may describe some other format, when provided to _sc_con_scalefb.
typedef struct _sc_con_fbinfo_s
{
	//Format of pixel data within a line
	_sc_con_pxfmt_t pxfmt;

	//Active pixels in each line
	int width;

	//Active lines on the display
	int height;

	//Address difference from one pixel to the pixel below it
	int stride;

	//Intended display aspect ratio - numerator (or 0 for don't-care)
	int arn;

	//Intended display aspect ratio - denominator (or 0 for don't-care)
	int ard;

} _sc_con_fbinfo_t;

//Returns information about the format of the user's display.
//This format must be used when drawing to an image in displayable memory (displayed directly with _sc_con_fbdisp).
//Alternatively, another format may be used in normal memory, and then converted to this format with _sc_con_fbscale.
int _sc_con_fbinfo(_sc_con_fbinfo_t *info_out, int info_len);

//Shows the image from the given displayable memory region on the console.
//The buffer must be a pointer returned by _sc_mem_disp, or -EINVAL will be returned.
//The process must hold the console or -EBUSY will be returned.
//If the given image is already being scanned-out, the call will return 0 immediately.
//Otherwise, it will be scanned-out at the next exit from vertical blanking, and -EAGAIN will be returned until then.
int _sc_con_fbdisp(const void *dispmem);

//Reformats an image from the given format to the correct format for display.
//The framebuffer description at *infoptr describes the buffer at *srcmem.
//Implicitly, the buffer at *dispmem is of the format to be displayed (returned by _sc_con_fbinfo).
int _sc_con_fbscale(void *dispmem, const void *srcmem, const _sc_con_fbinfo_t *infoptr, int infolen);


//Mouse buttons that the kernel reports from its console.
typedef enum _sc_con_mbutton_e
{
	_SC_CON_MBUTTON_LEFT   = 0x01, //Left-click
	_SC_CON_MBUTTON_MIDDLE = 0x02, //Middle-click or pressed the scroll wheel
	_SC_CON_MBUTTON_RIGHT  = 0x04, //Right-click
	_SC_CON_MBUTTON_NAVF   = 0x08, //Forward navigation/side-button
	_SC_CON_MBUTTON_NAVB   = 0x10, //Backward navigation/side-button
} _sc_con_mbutton_t;

//Scancodes that the kernel reports from its console.
typedef enum _sc_con_scancode_e
{
	_SC_CON_SCANCODE_UNKNOWN = 0,
	_SC_CON_SCANCODE_A = 4,
	_SC_CON_SCANCODE_B = 5,
	_SC_CON_SCANCODE_C = 6,
	_SC_CON_SCANCODE_D = 7,
	_SC_CON_SCANCODE_E = 8,
	_SC_CON_SCANCODE_F = 9,
	_SC_CON_SCANCODE_G = 10,
	_SC_CON_SCANCODE_H = 11,
	_SC_CON_SCANCODE_I = 12,
	_SC_CON_SCANCODE_J = 13,
	_SC_CON_SCANCODE_K = 14,
	_SC_CON_SCANCODE_L = 15,
	_SC_CON_SCANCODE_M = 16,
	_SC_CON_SCANCODE_N = 17,
	_SC_CON_SCANCODE_O = 18,
	_SC_CON_SCANCODE_P = 19,
	_SC_CON_SCANCODE_Q = 20,
	_SC_CON_SCANCODE_R = 21,
	_SC_CON_SCANCODE_S = 22,
	_SC_CON_SCANCODE_T = 23,
	_SC_CON_SCANCODE_U = 24,
	_SC_CON_SCANCODE_V = 25,
	_SC_CON_SCANCODE_W = 26,
	_SC_CON_SCANCODE_X = 27,
	_SC_CON_SCANCODE_Y = 28,
	_SC_CON_SCANCODE_Z = 29,
	_SC_CON_SCANCODE_1 = 30,
	_SC_CON_SCANCODE_2 = 31,
	_SC_CON_SCANCODE_3 = 32,
	_SC_CON_SCANCODE_4 = 33,
	_SC_CON_SCANCODE_5 = 34,
	_SC_CON_SCANCODE_6 = 35,
	_SC_CON_SCANCODE_7 = 36,
	_SC_CON_SCANCODE_8 = 37,
	_SC_CON_SCANCODE_9 = 38,
	_SC_CON_SCANCODE_0 = 39,
	_SC_CON_SCANCODE_RETURN = 40,
	_SC_CON_SCANCODE_ESCAPE = 41,
	_SC_CON_SCANCODE_BACKSPACE = 42,
	_SC_CON_SCANCODE_TAB = 43,
	_SC_CON_SCANCODE_SPACE = 44,
	_SC_CON_SCANCODE_MINUS = 45,
	_SC_CON_SCANCODE_EQUALS = 46,
	_SC_CON_SCANCODE_LEFTBRACKET = 47,
	_SC_CON_SCANCODE_RIGHTBRACKET = 48,
	_SC_CON_SCANCODE_BACKSLASH = 49,
	_SC_CON_SCANCODE_NONUSHASH = 50,
	_SC_CON_SCANCODE_SEMICOLON = 51,
	_SC_CON_SCANCODE_APOSTROPHE = 52,
	_SC_CON_SCANCODE_GRAVE = 53,
	_SC_CON_SCANCODE_COMMA = 54,
	_SC_CON_SCANCODE_PERIOD = 55,
	_SC_CON_SCANCODE_SLASH = 56,
	_SC_CON_SCANCODE_CAPSLOCK = 57,
	_SC_CON_SCANCODE_F1 = 58,
	_SC_CON_SCANCODE_F2 = 59,
	_SC_CON_SCANCODE_F3 = 60,
	_SC_CON_SCANCODE_F4 = 61,
	_SC_CON_SCANCODE_F5 = 62,
	_SC_CON_SCANCODE_F6 = 63,
	_SC_CON_SCANCODE_F7 = 64,
	_SC_CON_SCANCODE_F8 = 65,
	_SC_CON_SCANCODE_F9 = 66,
	_SC_CON_SCANCODE_F10 = 67,
	_SC_CON_SCANCODE_F11 = 68,
	_SC_CON_SCANCODE_F12 = 69,
	_SC_CON_SCANCODE_PRINTSCREEN = 70,
	_SC_CON_SCANCODE_SCROLLLOCK = 71,
	_SC_CON_SCANCODE_PAUSE = 72,
	_SC_CON_SCANCODE_INSERT = 73,
	_SC_CON_SCANCODE_HOME = 74,
	_SC_CON_SCANCODE_PAGEUP = 75,
	_SC_CON_SCANCODE_DELETE = 76,
	_SC_CON_SCANCODE_END = 77,
	_SC_CON_SCANCODE_PAGEDOWN = 78,
	_SC_CON_SCANCODE_RIGHT = 79,
	_SC_CON_SCANCODE_LEFT = 80,
	_SC_CON_SCANCODE_DOWN = 81,
	_SC_CON_SCANCODE_UP = 82,
	_SC_CON_SCANCODE_NUMLOCKCLEAR = 83,
	_SC_CON_SCANCODE_KP_DIVIDE = 84,
	_SC_CON_SCANCODE_KP_MULTIPLY = 85,
	_SC_CON_SCANCODE_KP_MINUS = 86,
	_SC_CON_SCANCODE_KP_PLUS = 87,
	_SC_CON_SCANCODE_KP_ENTER = 88,
	_SC_CON_SCANCODE_KP_1 = 89,
	_SC_CON_SCANCODE_KP_2 = 90,
	_SC_CON_SCANCODE_KP_3 = 91,
	_SC_CON_SCANCODE_KP_4 = 92,
	_SC_CON_SCANCODE_KP_5 = 93,
	_SC_CON_SCANCODE_KP_6 = 94,
	_SC_CON_SCANCODE_KP_7 = 95,
	_SC_CON_SCANCODE_KP_8 = 96,
	_SC_CON_SCANCODE_KP_9 = 97,
	_SC_CON_SCANCODE_KP_0 = 98,
	_SC_CON_SCANCODE_KP_PERIOD = 99,
	_SC_CON_SCANCODE_NONUSBACKSLASH = 100,
	_SC_CON_SCANCODE_APPLICATION = 101,
	_SC_CON_SCANCODE_POWER = 102,
	_SC_CON_SCANCODE_KP_EQUALS = 103,
	_SC_CON_SCANCODE_F13 = 104,
	_SC_CON_SCANCODE_F14 = 105,
	_SC_CON_SCANCODE_F15 = 106,
	_SC_CON_SCANCODE_F16 = 107,
	_SC_CON_SCANCODE_F17 = 108,
	_SC_CON_SCANCODE_F18 = 109,
	_SC_CON_SCANCODE_F19 = 110,
	_SC_CON_SCANCODE_F20 = 111,
	_SC_CON_SCANCODE_F21 = 112,
	_SC_CON_SCANCODE_F22 = 113,
	_SC_CON_SCANCODE_F23 = 114,
	_SC_CON_SCANCODE_F24 = 115,
	_SC_CON_SCANCODE_EXECUTE = 116,
	_SC_CON_SCANCODE_HELP = 117,
	_SC_CON_SCANCODE_MENU = 118,
	_SC_CON_SCANCODE_SELECT = 119,
	_SC_CON_SCANCODE_STOP = 120,
	_SC_CON_SCANCODE_AGAIN = 121,
	_SC_CON_SCANCODE_UNDO = 122,
	_SC_CON_SCANCODE_CUT = 123,
	_SC_CON_SCANCODE_COPY = 124,
	_SC_CON_SCANCODE_PASTE = 125,
	_SC_CON_SCANCODE_FIND = 126,
	_SC_CON_SCANCODE_MUTE = 127,
	_SC_CON_SCANCODE_VOLUMEUP = 128,
	_SC_CON_SCANCODE_VOLUMEDOWN = 129,
	_SC_CON_SCANCODE_KP_COMMA = 133,
	_SC_CON_SCANCODE_KP_EQUALSAS400 = 134,
	_SC_CON_SCANCODE_INTERNATIONAL1 = 135,
	_SC_CON_SCANCODE_INTERNATIONAL2 = 136,
	_SC_CON_SCANCODE_INTERNATIONAL3 = 137,
	_SC_CON_SCANCODE_INTERNATIONAL4 = 138,
	_SC_CON_SCANCODE_INTERNATIONAL5 = 139,
	_SC_CON_SCANCODE_INTERNATIONAL6 = 140,
	_SC_CON_SCANCODE_INTERNATIONAL7 = 141,
	_SC_CON_SCANCODE_INTERNATIONAL8 = 142,
	_SC_CON_SCANCODE_INTERNATIONAL9 = 143,
	_SC_CON_SCANCODE_LANG1 = 144,
	_SC_CON_SCANCODE_LANG2 = 145,
	_SC_CON_SCANCODE_LANG3 = 146,
	_SC_CON_SCANCODE_LANG4 = 147,
	_SC_CON_SCANCODE_LANG5 = 148,
	_SC_CON_SCANCODE_LANG6 = 149,
	_SC_CON_SCANCODE_LANG7 = 150,
	_SC_CON_SCANCODE_LANG8 = 151,
	_SC_CON_SCANCODE_LANG9 = 152,
	_SC_CON_SCANCODE_ALTERASE = 153,
	_SC_CON_SCANCODE_SYSREQ = 154,
	_SC_CON_SCANCODE_CANCEL = 155,
	_SC_CON_SCANCODE_CLEAR = 156,
	_SC_CON_SCANCODE_PRIOR = 157,
	_SC_CON_SCANCODE_RETURN2 = 158,
	_SC_CON_SCANCODE_SEPARATOR = 159,
	_SC_CON_SCANCODE_OUT = 160,
	_SC_CON_SCANCODE_OPER = 161,
	_SC_CON_SCANCODE_CLEARAGAIN = 162,
	_SC_CON_SCANCODE_CRSEL = 163,
	_SC_CON_SCANCODE_EXSEL = 164,
	_SC_CON_SCANCODE_KP_00 = 176,
	_SC_CON_SCANCODE_KP_000 = 177,
	_SC_CON_SCANCODE_THOUSANDSSEPARATOR = 178,
	_SC_CON_SCANCODE_DECIMALSEPARATOR = 179,
	_SC_CON_SCANCODE_CURRENCYUNIT = 180,
	_SC_CON_SCANCODE_CURRENCYSUBUNIT = 181,
	_SC_CON_SCANCODE_KP_LEFTPAREN = 182,
	_SC_CON_SCANCODE_KP_RIGHTPAREN = 183,
	_SC_CON_SCANCODE_KP_LEFTBRACE = 184,
	_SC_CON_SCANCODE_KP_RIGHTBRACE = 185,
	_SC_CON_SCANCODE_KP_TAB = 186,
	_SC_CON_SCANCODE_KP_BACKSPACE = 187,
	_SC_CON_SCANCODE_KP_A = 188,
	_SC_CON_SCANCODE_KP_B = 189,
	_SC_CON_SCANCODE_KP_C = 190,
	_SC_CON_SCANCODE_KP_D = 191,
	_SC_CON_SCANCODE_KP_E = 192,
	_SC_CON_SCANCODE_KP_F = 193,
	_SC_CON_SCANCODE_KP_XOR = 194,
	_SC_CON_SCANCODE_KP_POWER = 195,
	_SC_CON_SCANCODE_KP_PERCENT = 196,
	_SC_CON_SCANCODE_KP_LESS = 197,
	_SC_CON_SCANCODE_KP_GREATER = 198,
	_SC_CON_SCANCODE_KP_AMPERSAND = 199,
	_SC_CON_SCANCODE_KP_DBLAMPERSAND = 200,
	_SC_CON_SCANCODE_KP_VERTICALBAR = 201,
	_SC_CON_SCANCODE_KP_DBLVERTICALBAR = 202,
	_SC_CON_SCANCODE_KP_COLON = 203,
	_SC_CON_SCANCODE_KP_HASH = 204,
	_SC_CON_SCANCODE_KP_SPACE = 205,
	_SC_CON_SCANCODE_KP_AT = 206,
	_SC_CON_SCANCODE_KP_EXCLAM = 207,
	_SC_CON_SCANCODE_KP_MEMSTORE = 208,
	_SC_CON_SCANCODE_KP_MEMRECALL = 209,
	_SC_CON_SCANCODE_KP_MEMCLEAR = 210,
	_SC_CON_SCANCODE_KP_MEMADD = 211,
	_SC_CON_SCANCODE_KP_MEMSUBTRACT = 212,
	_SC_CON_SCANCODE_KP_MEMMULTIPLY = 213,
	_SC_CON_SCANCODE_KP_MEMDIVIDE = 214,
	_SC_CON_SCANCODE_KP_PLUSMINUS = 215,
	_SC_CON_SCANCODE_KP_CLEAR = 216,
	_SC_CON_SCANCODE_KP_CLEARENTRY = 217,
	_SC_CON_SCANCODE_KP_BINARY = 218,
	_SC_CON_SCANCODE_KP_OCTAL = 219,
	_SC_CON_SCANCODE_KP_DECIMAL = 220,
	_SC_CON_SCANCODE_KP_HEXADECIMAL = 221,
	_SC_CON_SCANCODE_LCTRL = 224,
	_SC_CON_SCANCODE_LSHIFT = 225,
	_SC_CON_SCANCODE_LALT = 226,
	_SC_CON_SCANCODE_LGUI = 227,
	_SC_CON_SCANCODE_RCTRL = 228,
	_SC_CON_SCANCODE_RSHIFT = 229,
	_SC_CON_SCANCODE_RALT = 230,
	_SC_CON_SCANCODE_RGUI = 231,
	_SC_CON_SCANCODE_MODE = 257,
	_SC_CON_SCANCODE_AUDIONEXT = 258,
	_SC_CON_SCANCODE_AUDIOPREV = 259,
	_SC_CON_SCANCODE_AUDIOSTOP = 260,
	_SC_CON_SCANCODE_AUDIOPLAY = 261,
	_SC_CON_SCANCODE_AUDIOMUTE = 262,
	_SC_CON_SCANCODE_MEDIASELECT = 263,
	_SC_CON_SCANCODE_WWW = 264,
	_SC_CON_SCANCODE_MAIL = 265,
	_SC_CON_SCANCODE_CALCULATOR = 266,
	_SC_CON_SCANCODE_COMPUTER = 267,
	_SC_CON_SCANCODE_AC_SEARCH = 268,
	_SC_CON_SCANCODE_AC_HOME = 269,
	_SC_CON_SCANCODE_AC_BACK = 270,
	_SC_CON_SCANCODE_AC_FORWARD = 271,
	_SC_CON_SCANCODE_AC_STOP = 272,
	_SC_CON_SCANCODE_AC_REFRESH = 273,
	_SC_CON_SCANCODE_AC_BOOKMARKS = 274,
	_SC_CON_SCANCODE_BRIGHTNESSDOWN = 275,
	_SC_CON_SCANCODE_BRIGHTNESSUP = 276,
	_SC_CON_SCANCODE_DISPLAYSWITCH = 277,
	_SC_CON_SCANCODE_KBDILLUMTOGGLE = 278,
	_SC_CON_SCANCODE_KBDILLUMDOWN = 279,
	_SC_CON_SCANCODE_KBDILLUMUP = 280,
	_SC_CON_SCANCODE_EJECT = 281,
	_SC_CON_SCANCODE_SLEEP = 282,
	_SC_CON_SCANCODE_APP1 = 283,
	_SC_CON_SCANCODE_APP2 = 284,
	
	_SC_CON_SCANCODE_MAX = 512,
} _sc_con_scancode_t;

//Type of input event that the console reports
typedef enum _sc_con_input_type_e
{
	_SC_CON_INPUT_TYPE_NONE   = 0,
	
	_SC_CON_INPUT_TYPE_KBD       = 1,
	_SC_CON_INPUT_TYPE_MOUSEBTN  = 2,
	_SC_CON_INPUT_TYPE_MOUSEREL  = 3,
	
	_SC_CON_INPUT_TYPE_PADBTN = 4,
	_SC_CON_INPUT_TYPE_PADLAS = 5,
	_SC_CON_INPUT_TYPE_PADRAS = 6,
	
} _sc_con_input_type_t;

//Data about an input event from the console - should fit in a single 32-bit word.
//High 4 bits indicate which type of input event it is.
typedef union _sc_con_input_u
{
	struct {
		int scancode           : 15; //Scancode of the key that is now pressed or released
		unsigned state         : 1;  //Whether the key is now pressed (1) or released (0)
		int unused             : 8; //Not currently used. Should be zero.
		int player             : 4;
		int type               : 4;  //_SC_CON_INPUT_TYPE_KBD
	} kbd;
	
	struct {
		int buttons             : 8;
		int unused              : 16;
		int player              : 4;
		int type                : 4; //_SC_CON_INPUT_TYPE_MOUSEBTN
	} mousebtn;
	
	struct {
		int x                   : 12;
		int y                   : 12;
		int player              :  4;
		int type                :  4; //_SC_CON_INPUT_TYPE_MOUSEREL
	} mouserel;
	
	struct {
		int a      : 1;
		int b      : 1;
		int x      : 1;
		int y      : 1;
		int st     : 1;
		int sl     : 1;
		
		int dum1   : 2;
		
		int dl     : 1;
		int dr     : 1;
		int dd     : 1;
		int du     : 1;
		int zr     : 1;
		int r      : 1;
		int l      : 1;
		int zl     : 1;
		
		int dum3   : 7;
		int plug   : 1;
		
		int player : 4;
		int type   : 4;
		
	} padbtn;
	
	struct {
		
		unsigned x      : 8;
		unsigned y      : 8;
		unsigned t      : 8;
		int player : 4;
		int type   : 4;
		
	} padlas;
	
	struct {
		
		unsigned x      : 8;
		unsigned y      : 8;
		unsigned t      : 8;
		int player : 4;
		int type   : 4;
		
	} padras;
	
	struct {
		int nottype : 28; //Data used by each event type
		int type    : 4;  //One of _sc_con_input_type_t values.
	} type;
	
} _sc_con_input_t;

//Retrieves console input. Outputs as many events as are waiting and would fit in the buffer.
//Returns the number of events returned.
//The parameters each_bytes and buf_bytes are numbers of bytes. The return value is a number of events.
//Returns a negative error number on failure.
int _sc_con_input(_sc_con_input_t *buf_ptr, int each_bytes, int buf_bytes);

//Changes the process which has possession of the console, if the calling process has it.
//Returns a nonnegative value on success or a negative error number.
int _sc_con_pass(int next);

//Ways to alter the console flags of a process.
#define _SC_CON_FLAG_ENABLE  0 //Flags enabled where input bit is 1
#define _SC_CON_FLAG_DISABLE 1 //Flags disabled where input bit is 1
#define _SC_CON_FLAG_SETALL  2 //Flags set to exactly the input, enabled where 1

//Flags that can be set on the console. The default is all zero.
typedef enum _sc_con_flag_e
{
	_SC_CON_FLAG_RAWMOUSE = 0x0001, //Grab mouse and get raw relative motion, rather than cursor
} _sc_con_flag_t;

//Sets flags for the current process's console usage.
//The "how" parameter is the same as the signal mask call.
//Returns the OLD flags or a negative error number.
int _sc_con_flag(int how, int bits);

#endif //_SC_CON_H
