//sc_file.h
//System call library - file handling
//Bryan E. Topp <betopp@betopp.com> 2022

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#ifndef _SC_FILE_H
#define _SC_FILE_H

//Access modes that the kernel recognizes about a file descriptor.
//An access mode of 0 means that a file has been opened but only its status is available.
#define _SC_ACCESS_NOFOLLOW 0x20 //Open the symlink itself, if one is encountered
#define _SC_ACCESS_BACK     0x10 //Open the back of a pseudoterminal, if one is encountered
#define _SC_ACCESS_R        0x04
#define _SC_ACCESS_W        0x02
#define _SC_ACCESS_X        0x01

//Opens a file descriptor referring to the given existing file.
//The filename will be searched for in the directory described by dirfd.
//If dirfd is -1, the current working directory will be searched.
//The file will be opened for the access types specified in "access".
int _sc_find(int dirfd, const char *name, int access);

//Opens a file descriptor referring to a new file with the given name.
//The file's first link will be made in the directory described by dirfd.
//If dirfd is -1, the link will be made in the current working directory.
//The file will be opened for the access types specified in "access".
int _sc_make(int dirfd, const char *name, int mode, int rdev, int access);

//Creates an additional link to an existing file, which was already opened.
//int _sc_link(int dirfd, const char *name, int existingfd);

//Reads from a file. Writes up to "len" bytes into the buffer at "buf".
//Returns the number of bytes read if any are successfully read.
//Returns a negative error number if an error occurs before any bytes are read.
//Returns 0 to indicate that the end-of-file has been reached.
int _sc_read(int fd, void *buf, int len);

//Writes to a file. Reads up to "len" bytes from the buffer at "buf".
//Returns the number of bytes written if any were successfully written.
//Returns a negative error number if an error occurs before any bytes are written.
int _sc_write(int fd, const void *buf, int len);

//References whence seeking from
#define _SC_SEEK_SET 0
#define _SC_SEEK_CUR 1
#define _SC_SEEK_END 2

//Changes the file pointer in an open file.
//Returns the new file pointer, which is nonnegative, or a negative error number on failure.
long int _sc_seek(int fd, long int off, int whence);

//Changes the length of a file.
//File data are sparse, and lengthening a file may not allocate space for it all.
//Returns the new length of the file, which is nonnegative, or a negative error number on failure.
long int _sc_trunc(int fd, long int size);

//Removes a link to a file, deleting the file if no links remain.
//Returns a nonnegative value on success or a negative error number on failure.
int _sc_unlink(int dirfd, const char *name, int rmfd);

//Flags exposed by the kernel about file descriptors
#define _SC_FLAG_KEEPEXEC 1 //Keep the descriptor open across exec. File descriptors are all, by default, close-on-exec.

//Changes the flags of the given file descriptor.
//Returns the new set of flags, or a negative error number on failure.
int _sc_flag(int fd, int set, int clear);

//Closes a file descriptor.
//Returns a nonnegative value on success or a negative error number.
int _sc_close(int fd);

//Changes the working directory of the calling process.
//Returns a nonnegative value on success or a negative error number.
int _sc_chdir(int fd);

//Duplicates a file descriptor. The new descriptor points to the same open file, sharing a file pointer (offset).
//Returns the number of the new file descriptor or a negative error number on failure.
int _sc_dup(int oldfd, int min, int overwrite);

//File status returned by the kernel.
typedef struct _sc_stat_s
{
	long int ino;
	long int size;
	int rdev;
	int mode;
} _sc_stat_t;

//Outputs file status information about the file referred to by an open file descriptor.
//Returns the size of structure filled on success, or a negative error number.
int _sc_stat(int fd, _sc_stat_t *buf, int len);

//Format of directory entry returned from the kernel.
typedef struct _sc_dirent_s
{
	long int ino;   //File being referred to
	int rsvd[4];    //Reserved for future use
	char name[104]; //Name by which the file is known
} _sc_dirent_t;

//Replaces the calling process image (its memory space) with that in the given file.
//The file descriptor must describe an appropriate ELF executable file in memory.
//The calling thread must be the only thread in the calling process.
//Returns a negative error number on failure. Does not return on success.
int _sc_exec(int fd, char *const argv[], char *const envp[]);

//File mode bits as used by the kernel
#define _SC_S_IFMT   ((mode_t)0770000)
#define _SC_S_IFREG  ((mode_t)0010000)
#define _SC_S_IFDIR  ((mode_t)0020000)
#define _SC_S_IFBLK  ((mode_t)0030000)
#define _SC_S_IFCHR  ((mode_t)0040000)
#define _SC_S_IFLNK  ((mode_t)0050000)
#define _SC_S_IFSOCK ((mode_t)0060000)
#define _SC_S_IFIFO  ((mode_t)0070000)
#define _SC_S_IFPTY  ((mode_t)0100000)
#define _SC_S_ISUID  ((mode_t)0004000)
#define _SC_S_ISGID  ((mode_t)0002000)
#define _SC_S_ISVTX  ((mode_t)0001000)
#define _SC_S_IRWXU  ((mode_t)0000700)
#define _SC_S_IRUSR  ((mode_t)0000400)
#define _SC_S_IWUSR  ((mode_t)0000200)
#define _SC_S_IXUSR  ((mode_t)0000100)
#define _SC_S_IRWXG  ((mode_t)0000070)
#define _SC_S_IRGRP  ((mode_t)0000040)
#define _SC_S_IWGRP  ((mode_t)0000020)
#define _SC_S_IXGRP  ((mode_t)0000010)
#define _SC_S_IRWXO  ((mode_t)0000007)
#define _SC_S_IROTH  ((mode_t)0000004)
#define _SC_S_IWOTH  ((mode_t)0000002)
#define _SC_S_IXOTH  ((mode_t)0000001)

#endif //_SC_FILE_H
