//ctype.c
//Implementation of character-type functions for libc
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include <ctype.h>
#include <wchar.h>
#include <wctype.h>
#include <stddef.h>
#include <errno.h>
#include <string.h>


//Last I checked, these return the same boolean result as glibc for 0...255.

int isalnum(int x)
{
	return isalpha(x) || isdigit(x);
}

int isalpha(int x)
{
	if(x >= 'A' && x <= 'Z')
		return 1;
	if(x >= 'a' && x <= 'z')
		return 1;
	
	return 0;
}

int isascii(int x)
{	
	if(x >= 0 && x < 128)
		return 1;
	else
		return 0;
}

int isblank(int x)
{
	if(x == ' ')
		return 1;
	if(x == '\t')
		return 1;
	
	return 0;
}

int iscntrl(int x)
{
	if(x >= 0 && x < 32)
		return 1;
	
	if(x == 127)
		return 1;
	
	return 0;
}

int isdigit(int x)
{
	if(x >= '0' && x <= '9')
		return 1;
	
	return 0;
}

int isgraph(int x)
{
	return isprint(x) && !isspace(x);
}

int islower(int x)
{
	if(x >= 'a' && x <= 'z')
		return 1;
	
	return 0;
}

int isprint(int x)
{
	return isascii(x) && !iscntrl(x);
}

int ispunct(int x)
{
	return isgraph(x) && !isalnum(x);
}

int isspace(int x)
{
	if(x == ' ') //Space
		return 1;
	if(x == '\f') //Form-feed
		return 1;
	if(x == '\n') //Newline
		return 1;
	if(x == '\r') //Carriage return
		return 1;
	if(x == '\t') //Horizontal Tab
		return 1;
	if(x == '\v') //Vertical tab
		return 1;
	
	return 0;
}

int isupper(int x)
{
	if(x >= 'A' && x <= 'Z')
		return 1;
	
	return 0;
}

int isxdigit(int x)
{
	if(x >= '0' && x <= '9')
		return 1;
	
	if(x >= 'a' && x <= 'f')
		return 1;
	
	if(x >= 'A' && x <= 'F')
		return 1;
	
	return 0;
}

int toascii(int x)
{
	return x & 0x7F;
}

int tolower(int x)
{
	if(isupper(x))
		return x + 'a' - 'A';
	
	return x;
}

int toupper(int x)
{
	if(islower(x))
		return x + 'A' - 'a';
	
	return x;
}

//Stubs for locale-specific checks.
//We don't support locales for now.

int isalnum_l(int x, locale_t l)
{
	(void)l;
	return isalnum(x);
}

int isalpha_l(int x, locale_t l)
{
	(void)l;
	return isalpha(x);
}

int isblank_l(int x, locale_t l)
{
	(void)l;
	return isblank(x);
}

int iscntrl_l(int x, locale_t l)
{
	(void)l;
	return iscntrl(x);
}

int isdigit_l(int x, locale_t l)
{
	(void)l;
	return isdigit(x);
}

int isgraph_l(int x, locale_t l)
{
	(void)l;
	return isgraph(x);
}

int islower_l(int x, locale_t l)
{
	(void)l;
	return islower(x);
}

int isprint_l(int x, locale_t l)
{
	(void)l;
	return isprint(x);
}

int ispunct_l(int x, locale_t l)
{
	(void)l;
	return ispunct(x);
}

int isspace_l(int x, locale_t l)
{
	(void)l;
	return isspace(x);
}

int isupper_l(int x, locale_t l)
{
	(void)l;
	return isupper(x);
}

int isxdigit_l(int x, locale_t l)
{
	(void)l;
	return isxdigit(x);
}

int tolower_l(int x, locale_t l)
{
	(void)l;
	return tolower(x);
}

int toupper_l(int x, locale_t l)
{
	(void)l;
	return toupper(x);
}


int mbtowc(wchar_t * restrict wcharp, const char * restrict mbchar, size_t nbytes)
{
	if(mbchar == NULL)
		return 0; //No shift states supported
	
	if(nbytes == 0)
		return 0; //No bytes allowed?
	
	if(*mbchar == '\0')
		return 0; //NUL encountered
	
	//UTF-8 decode to 32-bit integers
	if((unsigned int)mbchar[0] < 0x80)
	{
		//One-byte character, don't care about following bytes
		*wcharp = mbchar[0];
		return 1;
	}
		
	if((unsigned int)mbchar[0] < 0xC0)
	{
		//Two-byte character - following 1 byte should be 0b10xxxxxx
		if(nbytes < 2)
		{
			errno = EILSEQ;
			return -1;
		}
		
		if( (mbchar[1] & 0xC0) != 0x80 )
		{
			errno = EILSEQ;
			return -1;
		}
		
		wchar_t result = 0;
		result = (result << 5) | (mbchar[0] & 0x1F);
		result = (result << 6) | (mbchar[1] & 0x3F);
		*wcharp = result;
		return 2;
	}
	if((unsigned int)mbchar[0] < 0xE0)
	{
		//Three-byte character - following 2 bytes should be 0b10xxxxxx
		if(nbytes < 3)
		{
			errno = EILSEQ;
			return -1;
		}
		
		if( ((mbchar[1] & 0xC0) != 0x80) || ((mbchar[2] & 0xC0) != 0x80) )
		{
			errno = EILSEQ;
			return -1;
		}
		
		wchar_t result = 0;
		result = (result << 4) | (mbchar[0] & 0x0F);
		result = (result << 6) | (mbchar[1] & 0x3F);
		result = (result << 6) | (mbchar[2] & 0x3F);
		*wcharp = result;
		return 3;
	}
	if((unsigned int)mbchar[0] < 0xF0)
	{
		//Four-byte character - following 3 bytes should be 0b10xxxxxx
		if(nbytes < 4)
		{
			errno = EILSEQ;
			return -1;
		}
		
		if( ((mbchar[1] & 0xC0) != 0x80) || ((mbchar[2] & 0xC0) != 0x80) || ((mbchar[3] & 0xC0) != 0x80) )
		{
			errno = EILSEQ;
			return -1;
		}
		
		wchar_t result = 0;
		result = (result << 3) | (mbchar[0] & 0x07);
		result = (result << 6) | (mbchar[1] & 0x3F);
		result = (result << 6) | (mbchar[2] & 0x3F);
		result = (result << 6) | (mbchar[3] & 0x3F);
		*wcharp = result;
		return 4;
	}
	
	//Reserved - could represent even longer characters in the future
	errno = EILSEQ;
	return -1;
}

int wctomb(char * restrict s, wchar_t c)
{
	return (int)(wcrtomb(s, c, NULL));
}

size_t wcrtomb(char * restrict s, wchar_t c, mbstate_t * restrict ps)
{
	//We don't use shift-states in UTF-8
	(void)ps;
	
	if(s == NULL)
		return 1; //It would take one byte to represent a NUL(?)
	
	if(c < 0x80)
	{
		s[0] = c; //Directly encode 7-bit ASCII values
		return 1;
	}
	
	if(c < 0x800)
	{
		s[0] = ((c >> 6) & 0x1F) | 0xC0; //110xxxxx first byte (2-byte sequence start)
		s[1] = ((c >> 0) & 0x3F) | 0x80; //10xxxxxx following byte
		return 2;
	}
	
	if(c < 0x10000)
	{
		s[0] = ((c >> 12) & 0x0F) | 0xE0; //1110xxxx first byte (3-byte sequence start)
		s[1] = ((c >>  6) & 0x3F) | 0x80; //10xxxxxx following byte
		s[2] = ((c >>  0) & 0x3F) | 0x80; //10xxxxxx following byte
		return 3;
	}
	
	if(c < 0x110000)
	{
		s[0] = ((c >> 18) & 0x07) | 0xF0; //11110xxx first byte (4-byte sequence start)
		s[1] = ((c >> 12) & 0x3F) | 0x80; //10xxxxxx following byte
		s[2] = ((c >>  6) & 0x3F) | 0x80; //10xxxxxx following byte
		s[3] = ((c >>  0) & 0x3F) | 0x80; //10xxxxxx following byte
		return 4;
	}
	
	return (size_t)-1;
}

typedef enum _wctype_class_e
{
	WCC_NONE = 0,
	
	WCC_ALNUM,
	WCC_CNTRL,
	WCC_IDEOGRAM,
	WCC_PRINT,
	WCC_SPACE,
	WCC_XDIGIT,
	WCC_ALPHA,
	WCC_DIGIT,
	WCC_LOWER,
	WCC_PUNCT,
	WCC_SPECIAL,
	WCC_BLANK,
	WCC_GRAPH,
	WCC_PHONOGRAM,
	WCC_RUNE,
	WCC_UPPER,
	
	WCC_MAX
} _wctype_class_t;
	
static const char *_wctype_names[] = 
{
	"alnum",
	"cntrl",
	"ideogram",
	"print",
	"space",
	"xdigit",
	"alpha",
	"digit",
	"lower",
	"punct",
	"special",
	"blank",
	"graph",
	"phonogram",
	"rune",
	"upper",
	0
};

int iswctype(wint_t wc, wctype_t charclass)
{
	if(wc < 0 || wc > 0xFF)
		return 0;
	
	switch(charclass)
	{
		case WCC_ALNUM: return isalnum(wc);
		case WCC_DIGIT: return isdigit(wc);
		case WCC_ALPHA: return isalpha(wc);
		case WCC_LOWER: return islower(wc);
		case WCC_UPPER: return isupper(wc);
		case WCC_SPACE: return isspace(wc);
		default: return 0;
	}
}

wctype_t wctype(const char *property)
{
	if(property == NULL)
		return WCC_NONE;
	
	for(int ww = WCC_NONE; ww < WCC_MAX; ww++)
	{
		if(_wctype_names[ww] != NULL && !strcmp(property, _wctype_names[ww]))
			return ww;
	}
	return WCC_NONE;
}

int iswlower(wint_t wc)
{
	return iswctype(wc, WCC_LOWER);
}

int iswupper(wint_t wc)
{
	return iswctype(wc, WCC_UPPER);
}

int iswalnum(wint_t wc)
{
	return iswctype(wc, WCC_ALNUM);
}

int iswspace(wint_t wc)
{
	return iswctype(wc, WCC_SPACE);
}

int towupper(wint_t wc)
{
	return (wc < 0 || wc > 0xFF) ? wc : toupper(wc);
}

int towlower(wint_t wc)
{
	return (wc < 0 || wc > 0xFF) ? wc : tolower(wc);
}

int wcwidth(wchar_t wc)
{
	if(wc == 0)
		return 0;
	
	return 1;
}
