//dir.c
//Directory descriptors in libc
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include <dirent.h>
#include <errno.h>
#include <stddef.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#include <sc.h>

struct _DIR_s
{
	int fd; //Underlying file descriptor
	char buf[512]; //Space 
};

DIR *fdopendir(int fd)
{
	//Reopen with read permissions, in case the caller's FD was only open for stat()
	int newfd = openat(fd, ".", O_RDONLY | O_DIRECTORY | O_CLOEXEC);
	if(newfd < 0)
		return NULL; //openat sets errno
	
	//Put in a DIR structure
	DIR *retval = malloc(sizeof(DIR));
	if(retval == NULL)
	{
		close(newfd);
		errno = ENOMEM;
		return NULL;
	}
	memset(retval, 0, sizeof(*retval));
	retval->fd = newfd;
	close(fd);
	return retval;
}

int fdclosedir(DIR *dirp)
{
	free(dirp);
	return 0;
}

DIR *opendir(const char *filename)
{
	int fd = open(filename, O_RDONLY | O_DIRECTORY | O_CLOEXEC);
	if(fd < 0)
		return NULL;
	
	DIR *retval = fdopendir(fd);
	if(retval == NULL)
	{
		close(fd);
		return NULL;
	}
	
	return retval;
}

struct dirent *readdir(DIR *dirp)
{
	//Ask kernel to read the directory, getting a directory entry
	_sc_dirent_t d = {0};
	int read_result = _sc_read(dirp->fd, &d, sizeof(d));
	if(read_result < 0)
	{
		errno = -read_result;
		return NULL;
	}
	if(read_result == 0)
	{
		return NULL;
	}

	//Use the dirent buffer in the DIR structure, and clear it
	memset(dirp->buf, 0, sizeof(dirp->buf));
	struct dirent *retval = (struct dirent*)(dirp->buf);
	
	//Copy the inode number and name from the directory entry to the dirent buffer
	retval->d_ino = d.ino;
	strncpy(dirp->buf + sizeof(struct dirent), d.name, sizeof(dirp->buf) - sizeof(struct dirent) - 1);
	
	//Stat the file to find its type, and put in the dirent buffer
	struct stat st = {0};
	if(fstatat(dirp->fd, d.name, &st, AT_SYMLINK_NOFOLLOW) >= 0)
	{
		retval->d_type = st.st_mode & S_IFMT;
	}
	
	return retval;
}

int closedir(DIR *dirp)
{
	int fd = dirp->fd;
	fdclosedir(dirp);
	close(fd);
	return 0;
}
