//mmlibc/src/endian.c
//Byte-swapping functions for MMK's libc.
//Bryan E. Topp <betopp@betopp.com> 2023

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include <endian.h>

uint16_t htobe16(uint16_t h16)
{
	union {
		uint8_t bytes[2];
		uint16_t whole;
	} u;
	u.bytes[0] = (h16 >> 8) & 0xFF;
	u.bytes[1] = (h16 >> 0) & 0xFF;
	return u.whole;
}

uint16_t htole16(uint16_t h16)
{
	union {
		uint8_t bytes[2];
		uint16_t whole;
	} u;
	u.bytes[0] = (h16 >> 0) & 0xFF;
	u.bytes[1] = (h16 >> 8) & 0xFF;
	return u.whole;	
}

uint16_t be16toh(uint16_t b16)
{
	union {
		uint8_t bytes[2];
		uint16_t whole;
	} u;
	u.whole = b16;
	return	((uint16_t)u.bytes[0] <<  8l) |
		((uint16_t)u.bytes[1] <<  0l);
}

uint16_t le16toh(uint16_t l16)
{
	union {
		uint8_t bytes[2];
		uint16_t whole;
	} u;
	u.whole = l16;
	return	((uint16_t)u.bytes[1] <<  8l) |
		((uint16_t)u.bytes[0] <<  0l);	
}

uint32_t htobe32(uint32_t h32)
{
	union {
		uint8_t bytes[4];
		uint32_t whole;
	} u;
	u.bytes[0] = (h32 >> 24) & 0xFF;
	u.bytes[1] = (h32 >> 16) & 0xFF;
	u.bytes[2] = (h32 >> 8)  & 0xFF;
	u.bytes[3] = (h32 >> 0)  & 0xFF;
	return u.whole;
}

uint32_t htole32(uint32_t h32)
{
	union {
		uint8_t bytes[4];
		uint32_t whole;
	} u;
	u.bytes[0] = (h32 >> 0)  & 0xFF;
	u.bytes[1] = (h32 >> 8)  & 0xFF;
	u.bytes[2] = (h32 >> 16) & 0xFF;
	u.bytes[3] = (h32 >> 24) & 0xFF;
	return u.whole;		
}

uint32_t be32toh(uint32_t b32)
{
	union {
		uint8_t bytes[4];
		uint32_t whole;
	} u;
	u.whole = b32;
	return	((uint32_t)u.bytes[0] << 24l) |
		((uint32_t)u.bytes[1] << 16l) |
		((uint32_t)u.bytes[2] <<  8l) |
		((uint32_t)u.bytes[3] <<  0l);
}

uint32_t le32toh(uint32_t l32)
{
	union {
		uint8_t bytes[4];
		uint32_t whole;
	} u;
	u.whole = l32;
	return	((uint64_t)u.bytes[3] << 24l) |
		((uint64_t)u.bytes[2] << 16l) |
		((uint64_t)u.bytes[1] <<  8l) |
		((uint64_t)u.bytes[0] <<  0l);
}

uint64_t htobe64(uint64_t h64)
{
	union {
		uint8_t bytes[8];
		uint64_t whole;
	} u;
	u.bytes[0] = (h64 >> 24) & 0xFF;
	u.bytes[1] = (h64 >> 16) & 0xFF;
	u.bytes[2] = (h64 >> 8)  & 0xFF;
	u.bytes[3] = (h64 >> 0)  & 0xFF;
	return u.whole;	
}

uint64_t htole64(uint64_t h64)
{
	union {
		uint8_t bytes[8];
		uint64_t whole;
	} u;
	u.bytes[0] = (h64 >> 0)  & 0xFF;
	u.bytes[1] = (h64 >> 8)  & 0xFF;
	u.bytes[2] = (h64 >> 16) & 0xFF;
	u.bytes[3] = (h64 >> 24) & 0xFF;
	return u.whole;		
}

uint64_t be64toh(uint64_t b64)
{
	union {
		uint8_t bytes[8];
		uint64_t whole;
	} u;
	u.whole = b64;
	return	((uint64_t)u.bytes[0] << 56l) | 
		((uint64_t)u.bytes[1] << 48l) |
		((uint64_t)u.bytes[2] << 40l) |
		((uint64_t)u.bytes[3] << 32l) |
		((uint64_t)u.bytes[4] << 24l) |
		((uint64_t)u.bytes[5] << 16l) |
		((uint64_t)u.bytes[6] <<  8l) |
		((uint64_t)u.bytes[7] <<  0l);
}

uint64_t le64toh(uint64_t l64)
{
	union {
		uint8_t bytes[8];
		uint64_t whole;
	} u;
	u.whole = l64;
	return	((uint64_t)u.bytes[7] << 56l) | 
		((uint64_t)u.bytes[6] << 48l) |
		((uint64_t)u.bytes[5] << 40l) |
		((uint64_t)u.bytes[4] << 32l) |
		((uint64_t)u.bytes[3] << 24l) |
		((uint64_t)u.bytes[2] << 16l) |
		((uint64_t)u.bytes[1] <<  8l) |
		((uint64_t)u.bytes[0] <<  0l);
}


uint16_t bswap16(uint16_t b)
{
	return ((b>>8)&0x00FFull) | ((b<<8)&0xFF00ull);
}

uint32_t bswap32(uint32_t b)
{
	return (((uint32_t)bswap16(b)>>16)&0x0000FFFFull) | (((uint32_t)bswap16(b)<<16)&0xFFFF0000ull);
}

uint64_t bswap64(uint64_t b)
{
	return (((uint64_t)bswap32(b)>>32)&0x00000000FFFFFFFFull) | (((uint64_t)bswap32(b)<<32)&0xFFFFFFFF00000000ull);
}
