//exec.c
//Exec and friends in standard library
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include <unistd.h>
#include <sc.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>

extern char **environ;

int fexecve(int fd, char *const argv[], char *const envp[])
{
	int result = _sc_exec(fd, argv, envp);
	
	//The above syscall should only return on error.
	assert(result < 0);
	errno = -result;
	return -1;
}

int execve(const char *path, char *const argv[], char *const envp[])
{
	int fd = open(path, O_EXEC | O_CLOEXEC);
	if(fd < 0)
		return -1;
	
	return fexecve(fd, argv, envp);
}

int execv(const char *path, char *const argv[])
{
	return execve(path, argv, environ);
}

int execvp(const char *file, char *const argv[])
{
	//Search for the file in each item in the PATH environment variable
	const char *path = getenv("PATH");	
	while(path != NULL && path[0] != '\0')
	{
		//Extract next item from the PATH
		while(path[0] == ':')
			path++;
		
		char item[256] = {0};
		size_t itemlen = 0;
		while(path[0] != '\0' && path[0] != ':')
		{
			if(itemlen < sizeof(item))
			{
				item[itemlen] = *path;
				itemlen++;
			}
			path++;
		}
		
		if(itemlen == 0)
		{
			//No more items in path
			break;
		}
		
		if(itemlen >= sizeof(item))
		{
			//Overlong entry in PATH - ignore it
			continue;
		}
		
		//Open the directory from the PATH
		int pathitem_fd = _sc_find(-1, item, _SC_ACCESS_R);
		if(pathitem_fd < 0)
		{
			//Bad entry in the PATH - ignore it
			continue;
		}
		
		//See if the file is found there, and then close the directory
		int find_at_pathitem = _sc_find(pathitem_fd, file, _SC_ACCESS_X);
		_sc_close(pathitem_fd);
		
		if(find_at_pathitem < 0)
		{
			//Didn't find the file at this entry in the PATH
			continue;
		}
		
		//Try to execute it
		int result = _sc_exec(find_at_pathitem, argv, environ);
	
		//The above syscall should only return on error.
		assert(result < 0);
		
		_sc_close(find_at_pathitem);
		errno = -result;
		return -1;	
	}
	
	//Didn't find it at all - just try the name by itself
	return execv(file, argv);
}
