//fcntl.c
//File descriptor functions for standard library
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include <fcntl.h>
#include <errno.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include <sys/stat.h>

#include <sc.h>

//Bits masked-off from mode when creating new files.
//We stuff this in an environment variable so it persists across exec, as required by posix.
//It would be really silly to have the kernel concern itself with this - unless I'm missing something.
mode_t _umask_cached = -1;
void _umask_set(mode_t modeclr)
{
	//Set in environment and normal variable.
	_umask_cached = modeclr;
	char buf[16] = {0};
	snprintf(buf, 15, "0%o", modeclr);
	setenv("UMASK", buf, 1);
}
mode_t _umask_get()
{
	//Restore variable from environment if it's unset.
	if(_umask_cached == -1)
	{
		char *umask_str = getenv("UMASK");
		if(umask_str && *umask_str)
			_umask_cached = strtol(umask_str, NULL, 8);
		else
			_umask_cached = 022;
	}
	
	//Return from variable after looking up once.
	return _umask_cached;
}

mode_t umask(mode_t modeclr)
{
	mode_t old = _umask_get();
	_umask_set(modeclr);
	return old;
}

//Non-variadic version of openat.
int _openatm(int fd, const char *path, int flags, mode_t mode)
{
	if(fd == AT_FDCWD)
		fd = -1;
	
	int access_desired = 0;
	if(flags & O_RDONLY)
		access_desired |= _SC_ACCESS_R;
	if(flags & O_EXEC)
		access_desired |= _SC_ACCESS_X;
	if(flags & O_WRONLY)
		access_desired |= _SC_ACCESS_W;
	if(flags & O_RDWR)
		access_desired |= _SC_ACCESS_W | _SC_ACCESS_R; //Not stricly necessary since RDWR == RDONLY | WRONLY'
	if(flags & O_NOFOLLOW)
		access_desired |= _SC_ACCESS_NOFOLLOW;
	
	//Try to create the file first, if creating it is permitted.
	//This will fail if it already exists, but we'll then try to open it.
	//If we tried to open the existing file first, we would have a race-condition with someone else also trying that.
	
	int create_result = -ENOSYS;
	if(flags & O_CREAT)
	{
		if( (flags & O_DIRECTORY) && !S_ISDIR(mode) )
		{
			//Don't attempt to make a non-directory if we were asked to only open a directory.
			errno = ENOTDIR;
			return -1;
		}
		
		//If they didn't specify a type of file, default to a "regular file" mode.
		if((mode & S_IFMT) == 0)
			mode |= S_IFREG;
		
		create_result = _sc_make(fd, path, mode & ~_umask_get(), 0, access_desired);
	}
	
	//If we didn't create the file (or didn't try), and don't want only to create the file, try getting the existing one.
	int find_result = -ENOSYS;
	if((create_result < 0) && !((flags & O_CREAT) && (flags & O_EXCL)))
	{
		find_result = _sc_find(fd, path, access_desired);
	}

	if(create_result < 0 && find_result < 0)
	{
		if(create_result != -ENOSYS)
			errno = -create_result; //maybe? how do we pick?
		else
			errno = -find_result;
		
		return -1; //Could neither open nor create this file.
	}
	
	//Alright, we've got the file open.
	int fd_ret = -1;
	if(create_result > 0)
	{
		assert(find_result < 0);
		fd_ret = create_result;
	}
	else
	{
		assert(find_result >= 0);
		fd_ret = find_result;
	}
	
	//If we only wanted to open a directory, make sure it is.
	if(flags & O_DIRECTORY)
	{
		_sc_stat_t st = {0};
		int st_err = _sc_stat(fd_ret, &st, sizeof(st));
		if(st_err < 0)
		{
			//Failed to stat the file we just opened...?
			errno = -st_err;
			_sc_close(fd_ret);
			fd_ret = -1;
		}
		
		if(!S_ISDIR(st.mode))
		{
			//Wanted to only open a directory, but this isn't one.
			errno = ENOTDIR;
			_sc_close(fd_ret);
			fd_ret = -1;
		}
	}
	
	//If we wanted to append, go to the end of the file
	if(flags & O_APPEND)
	{
		_sc_seek(fd_ret, 0, SEEK_END);
	}
	
	//The kernel defaults to close-on-exec, with a flag for keep-through-exec.
	//This is the opposite of the conventional behavior. Set keep-through-exec unless asked for close-on-exec.
	if((fd_ret >= 0) && !(flags & O_CLOEXEC))
		_sc_flag(fd_ret, _SC_FLAG_KEEPEXEC, 0);
			
	return fd_ret;
}

//Non-variadic version of open.
int _openm(const char *path, int flags, mode_t mode)
{
	//open is equivalent to openat with fd of AT_FDCWD
	return _openatm(AT_FDCWD, path, flags, mode);
}

int open(const char *path, int flags, ...)
{
	//Kill variadic parameter and call non-variadic version.
	mode_t mode = 0;
	if(flags & O_CREAT)
	{
		va_list ap;
		va_start(ap, flags);
		mode = va_arg(ap, mode_t);
		va_end(ap);
	}
	
	return _openm(path, flags, mode);
}

int openat(int fd, const char *path, int flags, ...)
{
	//Kill variadic parameter and call non-variadic version.
	mode_t mode = 0;
	if(flags & O_CREAT)
	{
		va_list ap;
		va_start(ap, flags);
		mode = va_arg(ap, mode_t);
		va_end(ap);
	}
	
	return _openatm(fd, path, flags, mode);
}

ssize_t write(int fd, const void *buf, size_t nbytes)
{
	//Kernel doesn't handle blocking on write. Try as long as it says so.
	//Todo - nonblocking if the user asks for it.
	while(1)
	{
		ssize_t result = _sc_write(fd, buf, nbytes);
		if(result == -EAGAIN)
		{
			_sc_pause();
			continue;
		}
		if(result < 0)
		{
			errno = -result;
			return -1;
		}
		return result;
	}
}

ssize_t read(int fd, void *buf, size_t nbytes)
{
	//Similar to synchronization of write.
	while(1)
	{
		ssize_t result = _sc_read(fd, buf, nbytes);
		if(result == -EAGAIN)
		{
			_sc_pause();
			continue;
		}
		if(result < 0)
		{
			errno = -result;
			return -1;
		}
		return result;
	}
}


int dup(int oldfd)
{
	int result = _sc_dup(oldfd, 0, false);
	if(result < 0)
	{
		errno = -result;
		return -1;
	}
	_sc_flag(result, _SC_FLAG_KEEPEXEC, 0);
	return result;
}

int dup2(int oldfd, int newfd)
{
	int result = _sc_dup(oldfd, newfd, true);
	if(result < 0)
	{
		errno = -result;
		return -1;
	}
	assert(result == newfd);
	_sc_flag(result, _SC_FLAG_KEEPEXEC, 0);
	return result;
}

int fcntl(int fd, int cmd, ...)
{
	//Check command and resolve third parameter.
	int arg_int = 0;
	void *arg_ptr = NULL;
	va_list ap;
	switch(cmd)
	{
		case F_DUPFD:
		case F_DUPFD_CLOEXEC:
		case F_SETFD:		
		case F_SETFL:	
		case F_SETOWN:			
			va_start(ap, cmd);
			arg_int = va_arg(ap, int);
			va_end(ap);
			break;
		
		case F_GETLK:
		case F_SETLK:
		case F_SETLKW:
			va_start(ap, cmd);
			arg_ptr = va_arg(ap, void*);
			va_end(ap);
			break;
		
		case F_GETOWN:
		case F_GETFL:
		case F_GETFD:
			break;
		
		default:
			errno = EINVAL;
			return -1;
	}
	
	//Act on command
	switch(cmd)
	{
		case F_DUPFD:
		{
			int result = _sc_dup(fd, arg_int, false);
			if(result < 0)
			{
				errno = -result;
				return -1;
			}
			//Set keep-through-exec flag if we weren't asked to close-on-exec.
			_sc_flag(result, _SC_FLAG_KEEPEXEC, 0);
			return result;
		}
		case F_DUPFD_CLOEXEC:
		{
			int dup_result = _sc_dup(fd, arg_int, false);
			if(dup_result < 0)
			{
				errno = -dup_result;
				return -1;
			}
			//Asked to close-on-exec - don't set keep-through-exec flag.
			return dup_result;
		}
		case F_GETFD:
		{
			int k_flags = _sc_flag(fd, 0, 0);
			if(k_flags < 0)
			{
				errno = -k_flags;
				return -1;
			}
			
			if(k_flags & _SC_FLAG_KEEPEXEC)
				return 0;
			else
				return FD_CLOEXEC;
		}
		case F_SETFD:
		{
			int result = 0;
			if(arg_int & FD_CLOEXEC)
				result = _sc_flag(fd, 0, _SC_FLAG_KEEPEXEC); //Close-on-exec wanted - Clear keep-on-exec
			else
				result = _sc_flag(fd, _SC_FLAG_KEEPEXEC, 0); //Close-on-exec not wanted - set keep-on-exec.
			
			if(result < 0)
			{
				errno = -result;
				return -1;
			}
			return result;
		}
		default:
			(void)arg_ptr;
			errno = -EINVAL;
			return -1;
	}
}

off_t lseek(int fd, off_t off, int whence)
{
	off_t result = _sc_seek(fd, off, whence);
	if(result < 0)
	{
		errno = -result;
		return -1;
	}
	return result;
}

int close(int fd)
{
	int result = _sc_close(fd);
	if(result < 0)
	{
		errno = -result;
		return -1;
	}
	return 0;
}

int fstat(int fd, struct stat *out)
{
	//Performing a stat on an open file is the way our kernel works natively.
	_sc_stat_t _sc_stat_buf = {0};
	int stat_err = _sc_stat(fd, &_sc_stat_buf, sizeof(_sc_stat_buf));
	if(stat_err < 0)
	{
		errno = -stat_err;
		return -1;
	}
	
	//Convert kernel stat struct to libc struct.
	//Todo - is it worth defining these to be the same?
	//Seems like there might be reasons not to, but I can't come up with any.
	memset(out, 0, sizeof(*out));
	out->st_dev = 0;
	out->st_ino = _sc_stat_buf.ino;
	out->st_mode = _sc_stat_buf.mode;
	out->st_size = _sc_stat_buf.size;
	out->st_rdev = _sc_stat_buf.rdev;
	
	return 0;
}

int fstatat(int at_fd, const char *path, struct stat *out, int flag)
{
	//Our kernel has the concept of "open for stat", without needing permission for read nor write nor exec.
	//Try to open the given path for stat.
	int oflag = O_STAT | O_CLOEXEC;
	if(flag & AT_SYMLINK_NOFOLLOW)
		oflag |= O_NOFOLLOW;
	
	int fd = _openatm(at_fd, path, oflag, 0);
	if(fd < 0)
		return -1; //_openatm sets errno
	
	//Got the file open for stat. Stat it.
	int fstat_err = fstat(fd, out); //can set errno on failure

	//Either success or failure, just close the file and return what happened.
	_sc_close(fd);
	return fstat_err;
}

int stat(const char *path, struct stat *out)
{
	return fstatat(AT_FDCWD, path, out, 0);
}

int lstat(const char *path, struct stat *out)
{
	return fstatat(AT_FDCWD, path, out, AT_SYMLINK_NOFOLLOW);
}

ssize_t readlinkat(int at_fd, const char *path, char *buf, size_t len)
{
	int fd = _openatm(at_fd, path, O_RDONLY | O_CLOEXEC | O_NOFOLLOW, 0);
	if(fd < 0)
		return -1; //_openatm sets errno
	
	ssize_t read_result = _sc_read(fd, buf, len);
	_sc_close(fd);
	
	if(read_result < 0)
	{
		errno = -read_result;
		return -1;
	}
	return read_result;
}

ssize_t readlink(const char *path, char *buf, size_t len)
{
	return readlinkat(AT_FDCWD, path, buf, len);
}

int fchdir(int fd)
{
	//Make sure this is a directory.
	//The kernel will let us retain whatever file we want as our current directory,
	//but everything will start to fail with ENOTDIR in that case.
	_sc_stat_t st = {0};
	ssize_t stat_err = _sc_stat(fd, &st, sizeof(st));
	if(stat_err < 0)
	{
		errno = -stat_err;
		return -1;
	}
	
	if(!S_ISDIR(st.mode))
	{
		errno = ENOTDIR;
		return -1;
	}
	
	int result = _sc_chdir(fd);
	if(result < 0)
	{
		errno = -result;
		return -1;
	}
	return 0;
}

int chdir(const char *path)
{
	//Process needs exec permission on a directory to change to it. Try to open the path for exec.
	int fd = _openatm(AT_FDCWD, path, O_EXEC | O_CLOEXEC | O_RDWR, 0);
	if(fd < 0)
		return -1; //_openatm sets errno
	
	int result = fchdir(fd);
	_sc_close(fd);
	return result;
}

int ftruncate(int fd, off_t length)
{
	int err = _sc_trunc(fd, length);
	if(err < 0)
	{
		errno = -err;
		return -1;
	}
	return 0;
}

int truncate(const char *path, off_t length)
{
	int fd = _openatm(AT_FDCWD, path, O_WRONLY | O_CLOEXEC, 0);
	if(fd < 0)
		return -1; //_openatm sets errno
	
	int err = _sc_trunc(fd, length);
	if(err < 0)
	{
		errno = -err;
		return -1;
	}
	
	_sc_close(fd);
	return 0;
}

//This is a recent and nonstandard addition from the FreeBSD API. 
//It solves an important problem though ("unlink exactly this file I've been looking at") so we copy it.
int funlinkat(int dfd, const char *path, int fd, int flag)
{
	if(dfd == AT_FDCWD)
		dfd = -1;
	
	(void)flag;
	int result = _sc_unlink(dfd, path, fd);
	if(result < 0)
	{
		errno = -result;
		return -1;
	}

	return 0;
}

int unlinkat(int dfd, const char *path, int flag)
{
	return funlinkat(dfd, path, -1, flag);
}

int unlink(const char *path)
{
	return unlinkat(AT_FDCWD, path, 0);
}

int rmdir(const char *path)
{
	return unlinkat(AT_FDCWD, path, AT_REMOVEDIR);
}

int mkdirat(int fd, const char *path, mode_t mode)
{
	//Make "mode" refer to the given mode, as a directory.
	//Make sure no conflicting mode bits were set.
	mode |= S_IFDIR;
	if(!S_ISDIR(mode))
	{
		errno = EINVAL;
		return -1;
	}
	
	int made = _openatm(fd, path, O_CLOEXEC | O_CREAT | O_EXCL, mode);
	if(made == -1)
		return -1;
	
	close(made);
	return 0;
}

int mkdir(const char *path, mode_t mode)
{
	return mkdirat(AT_FDCWD, path, mode);
}

int faccessat(int fd, const char *path, int mode, int flag)
{
	int check_access = 0;
	if(mode & R_OK)
		check_access |= O_RDONLY;
	if(mode & W_OK)
		check_access |= O_WRONLY;
	if(mode & X_OK)
		check_access |= O_EXEC;
	
	int testfd = _openatm(fd, path, check_access | flag | O_CLOEXEC, 0);
	if(testfd < 0)
		return -1; //_openatm sets error
	
	close(testfd);
	return 0;
}

int access(const char *path, int mode)
{
	return faccessat(AT_FDCWD, path, mode, 0);
}

int eaccess(const char *path, int mode)
{
	return faccessat(AT_FDCWD, path, mode, AT_EACCESS);
}

int chmod(const char *path, mode_t mode)
{
	return fchmodat(AT_FDCWD, path, mode, 0);
}

int lchmod(const char *path, mode_t mode)
{
	return fchmodat(AT_FDCWD, path, mode, O_NOFOLLOW);
}

int fchmodat(int dirfd, const char *path, mode_t mode, int flag)
{
	//Look up path
	int pathfd = openat(dirfd, path, O_CLOEXEC | O_WRONLY | flag, 0);
	if(pathfd < 0)
		return -1; //fopenat sets error
	
	//Pass to fchmod to do the work
	int result = fchmod(pathfd, mode);
	close(pathfd);
	return result;
}

int fchmod(int fd, mode_t mode)
{
	//Need to implement a syscall for this
	(void)fd;
	(void)mode;
	errno = ENOSYS;
	return -1;
}

int renameat(int fromdir, const char *from, int todir, const char *to)
{
	//Need to implement a syscall for this
	(void)fromdir;
	(void)todir;
	(void)from;
	(void)to;
	errno = ENOSYS;
	return -1;
}

int rename(const char *from, const char *to)
{
	return renameat(AT_FDCWD, from, AT_FDCWD, to);
}

int mknodat(int fd, const char *path, mode_t mode, dev_t dev)
{
	int result = _sc_make(fd, path, mode & ~_umask_get(), dev, _SC_ACCESS_W | _SC_ACCESS_R);
	if(result < 0)
	{
		errno = -result;
		return -1;
	}
	_sc_close(result);
	return 0;	
}

int mknod(const char *path, mode_t mode, dev_t dev)
{
	return mknodat(-1, path, mode, dev);
}

int fsync(int fd)
{
	(void)fd;
	return 0;
}

int fdatasync(int fd)
{
	(void)fd;
	return 0;
}

int linkat(int fd1, const char *name1, int fd2, const char *name2, int flag)
{
	//Need to implement a syscall for this
	(void)fd1;
	(void)name1;
	(void)fd2;
	(void)name2;
	(void)flag;
	errno = ENOSYS;
	return -1;
}

int link(const char *name1, const char *name2)
{
	return linkat(AT_FDCWD, name1, AT_FDCWD, name2, 0);
}

int symlinkat(const char *name1, int fd, const char *name2)
{
	int oresult = openat(fd, name2, O_CLOEXEC | O_CREAT | O_EXCL | O_WRONLY, S_IFLNK | 0777);
	if(oresult < 0)
		return oresult; //openat sets errno
	
	write(oresult, name1, strlen(name1));
	close(oresult);
	return 0;
}

int symlink(const char *name1, const char *name2)
{
	return symlinkat(name1, AT_FDCWD, name2);
}

