//getline.c
//getline and friends in libc
//Bryan E. Topp <betopp@betopp.com> 2023

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

ssize_t getdelim(char ** restrict linep, size_t * restrict linecapp, int delimiter,  FILE * restrict stream)
{
	//Start with buffer as given
	char *buf_ptr = *linep;
	ssize_t buf_len = *linecapp;
	ssize_t buf_used = 0;
	
	//If no buffer was given, allocate one initially
	if(buf_ptr == NULL || buf_len <= 0)
	{
		buf_ptr = calloc(1, 16);
		if(buf_ptr == NULL)
		{
			*linep = NULL;
			*linecapp = 0;
			errno = ENOMEM;
			return -1;
		}
		buf_len = 16;
	}
	
	//Read characters into the buffer until we see the delimiter or an error
	while(1)
	{
		//Expand buffer if needed - allow for 2 characters (at most, a delimiter and a NUL)
		while(buf_used + 2 > buf_len)
		{
			size_t new_len = buf_len * 2;
			char *new_ptr = realloc(buf_ptr, new_len);
			if(new_ptr == NULL)
			{
				*linep = buf_ptr;
				*linecapp = buf_len;
				errno = ENOMEM;
				return -1;
			}
			buf_ptr = new_ptr;
			buf_len = new_len;
		}
		
		//Try to read the next character and put it in the buffer
		int ch = fgetc(stream);
		if(ch < 0)
		{
			*linep = buf_ptr;
			*linecapp = buf_len;
			return -1; //fgetc sets errno
		}
		
		buf_ptr[buf_used] = ch;
		buf_used++;
		
		//If we saw the delimiter, NUL-terminate the buffer and return it.
		if(ch == delimiter)
		{
			buf_ptr[buf_used] = '\0';
			*linep = buf_ptr;
			*linecapp = buf_len;
			return buf_used; //Excluding the terminating NUL
		}
	}
}

ssize_t getline(char ** restrict linep, size_t * restrict linecapp, FILE * restrict stream)
{
	return getdelim(linep, linecapp, '\n', stream);
}
