//poll.c
//File polling in libc
//Bryan E. Topp <betopp@betopp.com> 2023

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include <poll.h>
#include <stddef.h>
#include <errno.h>
#include <sc.h>

int poll(struct pollfd fds[], nfds_t nfds, int timeout)
{		
	//Initialize returned-events to 0
	for(nfds_t ff = 0; ff < nfds; ff++)
	{
		fds[ff].revents = 0;
	}
	
	//Decide when we'll timeout, if any
	int done_time = 0;
	if(timeout >= 0)
		done_time = _sc_getticks() + timeout;
	
	//Look for events until timeout
	int retval = 0;
	while(1)
	{
		//Scan all files for all specified events
		for(nfds_t ff = 0; ff < nfds; ff++)
		{
			if(fds[ff].fd < 0)
			{
				//Negative file descriptor specified - ignore entry
				continue;
			}
			
			if(fds[ff].events & POLLIN)
			{
				//Try a zero-length read to see if anything is available
				int nread = _sc_read(fds[ff].fd, NULL, 0);
				if(nread == -EAGAIN)
				{
					//No data available, otherwise fine
				}
				else if(nread == -EINVAL || nread == -EBADF)
				{
					//Bad file
					fds[ff].revents |= POLLNVAL;
				}
				else if(nread < 0)
				{
					//Other errors
					fds[ff].revents |= POLLERR;
				}
				else
				{
					//Data available
					fds[ff].revents |= POLLIN;
				}
			}
			
			if(fds[ff].events & POLLOUT)
			{
				//Try a zero-length write to see if space is available
				int nwritten = _sc_write(fds[ff].fd, NULL, 0);
				if(nwritten == -EAGAIN)
				{
					//No space available, otherwise fine
				}
				else if(nwritten == -EINVAL || nwritten == -EBADF)
				{
					//Bad file
					fds[ff].revents |= POLLNVAL;
				}
				else if(nwritten < 0)
				{
					//Other errors
					fds[ff].revents |= POLLERR;
				}
				else
				{
					//Space available
					fds[ff].revents |= POLLOUT;
				}
			}	
		}
		
		//Count how many files have returned events
		retval = 0;
		for(nfds_t ff = 0; ff < nfds; ff++)
		{
			if(fds[ff].revents != 0)
				retval++;
		}
		
		//If we've got events, stop polling
		if(retval > 0)
			break;
		
		//If we're timed-out, stop polling
		if(timeout >= 0 && _sc_getticks() >= done_time)
			break;
		
		//Pause if nothing happened
		_sc_pause();
	}
	
	return retval;
}