//printf.c
//Definitions for printf family for libc
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include <stdio.h>
#include <stdarg.h>
#include <stdint.h>
#include <errno.h>
#include <stddef.h>
#include <stdbool.h>
#include <sys/types.h>
#include <limits.h>

//vfprintf is stolen from musl, implemented elsewhere

int printf(const char *format, ...)
{
	//Set up va_list and call "v" version.
	va_list args;
	va_start(args, format);
	int retval = vprintf(format, args);
	va_end(args);
	return retval;
}

int fprintf(FILE *stream, const char *format, ...)
{
	//Set up va_list and call "v" version.
	va_list args;
	va_start(args, format);
	int retval = vfprintf(stream, format, args);
	va_end(args);
	return retval;
}

int dprintf(int fd, const char *format, ...)
{
	//Set up va_list and call "v" version.
	va_list args;
	va_start(args, format);
	int retval = vdprintf(fd, format, args);
	va_end(args);
	return retval;
}

int sprintf(char *str, const char *format, ...)
{
	//Set up va_list and call "v" version.
	va_list args;
	va_start(args, format);
	int retval = vsprintf(str, format, args);
	str[retval] = '\0';
	va_end(args);
	return retval;
}

int snprintf(char *str, size_t size, const char *format, ...)
{
	//Set up va_list and call "v" version.
	va_list args;
	va_start(args, format);
	int retval = vsnprintf(str, size, format, args);
	if(retval >= 0 && (size_t)retval < size)
		str[retval] = '\0';
	
	va_end(args);
	return retval;
}

int vprintf(const char *format, va_list ap)
{
	//Equivalent to calling vfprintf and specifying stdout as the stream.
	return vfprintf(stdout, format, ap);
}

int vdprintf(int fd, const char *format, va_list ap)
{
	//Phony FILE for writing directly to a file descriptor.
	FILE fd_stream = {
		.streamtype = _FILE_STREAMTYPE_RAWFD,
		.fd = fd,
	};	
	return vfprintf(&fd_stream, format, ap);
}

int vsprintf(char *str, const char *format, va_list ap)
{
	//Just do a vsnprintf with an enormous length.
	return vsnprintf(str, (INT_MAX)-1, format, ap);
}

int vsnprintf(char *str, size_t size, const char *format, va_list ap)
{
	//Phony FILE for writing to a bounded buffer.
	FILE strn_stream = {
		.streamtype = _FILE_STREAMTYPE_STRN,
		.buf_ptr = str,
		.buf_size = size,
		.buf_out = 1,
	};
	return vfprintf(&strn_stream, format, ap);
}

