//dlibc/src/common/string.c
//Entry points for string.h functions in libc
//Bryan E. Topp <betopp@betopp.com> 2023

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <ctype.h>
#include <sys/types.h>

//Define error names
#define NUM_ERR_NAMES 192
static const char *_err_names[NUM_ERR_NAMES] = 
{
	[0] = "No error",
	[E2BIG] = "Argument list too long",
	[EACCES] = "Permission denied",
	[EADDRINUSE] = "Address in use",
	[EADDRNOTAVAIL] = "Address not available",
	[EAFNOSUPPORT] = "Address family not supported",
	[EAGAIN] = "Resource unavailable, try again",
	[EALREADY] = "Connection already in progress",
	[EBADF] = "Bad file descriptor",
	[EBADMSG] = "Bad message",
	[EBUSY] = "Device or resource busy",
	[ECANCELED] = "Operation canceled",
	[ECHILD] = "No child processes",
	[ECONNABORTED] = "Connection aborted",
	[ECONNREFUSED] = "Connection refused",
	[ECONNRESET] = "Connection reset",
	[EDEADLK] = "Resource deadlock would occur",
	[EDESTADDRREQ] = "Destination address required",
	[EDOM] = "Mathematics argument out of domain of function",
	[EDQUOT] = "Reserved",
	[EEXIST] = "File exists",
	[EFAULT] = "Bad address",
	[EFBIG] = "File too large",
	[EHOSTUNREACH] = "Host is unreachable",
	[EIDRM] = "Identifier removed",
	[EILSEQ] = "Illegal byte sequence",
	[EINPROGRESS] = "Operation in progress",
	[EINTR] = "Interrupted function",
	[EINVAL] = "Invalid argument",
	[EIO] = "I/O error",
	[EISCONN] = "Socket is connected",
	[EISDIR] = "Is a directory",
	[ELOOP] = "Too many levels of symbolic links",
	[EMFILE] = "File descriptor value too large",
	[EMLINK] = "Too many links",
	[EMSGSIZE] = "Message too large",
	[EMULTIHOP] = "Reserved",
	[ENAMETOOLONG] = "Filename too long",
	[ENETDOWN] = "Network is down",
	[ENETRESET] = "Connection aborted by network",
	[ENETUNREACH] = "Network unreachable",
	[ENFILE] = "Too many files open in system",
	[ENOBUFS] = "No buffer space available",
	[ENODATA] = "No message is available on the STREAM head read queue",
	[ENODEV] = "No such device",
	[ENOENT] = "No such file or directory",
	[ENOEXEC] = "Executable file format error",
	[ENOLCK] = "No locks available",
	[ENOLINK] = "Reserved",
	[ENOMEM] = "Not enough space",
	[ENOMSG] = "No message of the desired type",
	[ENOPROTOOPT] = "Protocol not available",
	[ENOSPC] = "No space left on device",
	[ENOSR] = "No STREAM resources",
	[ENOSTR] = "Not a STREAM",
	[ENOSYS] = "Functionality not supported",
	[ENOTCONN] = "The socket is not connected",
	[ENOTDIR] = "Not a directory or a symbolic link to a directory",
	[ENOTEMPTY] = "Directory not empty",
	[ENOTRECOVERABLE] = "State not recoverable",
	[ENOTSOCK] = "Not a socket",
	[ENOTSUP] = "Not supported",
	[ENOTTY] = "Inappropriate I/O control operation",
	[ENXIO] = "No such device or address",
	//[EOPNOTSUPP] = "Operation not supported on socket (may be the same value as [ENOTSUP])",
	[EOVERFLOW] = "Value too large to be stored in data type",
	[EOWNERDEAD] = "Previous owner died",
	[EPERM] = "Operation not permitted",
	[EPIPE] = "Broken pipe",
	[EPROTO] = "Protocol error",
	[EPROTONOSUPPORT] = "Protocol not supported",
	[EPROTOTYPE] = "Protocol wrong type for socket",
	[ERANGE] = "Result too large",
	[EROFS] = "Read-only file system",
	[ESPIPE] = "Invalid seek",
	[ESRCH] = "No such process",
	[ESTALE] = "Reserved",
	[ETIME] = "Stream ioctl() timeout",
	[ETIMEDOUT] = "Connection timed out",
	[ETXTBSY] = "Text file busy",
	//[EWOULDBLOCK] = "Operation would block (may be the same value as [EAGAIN])",
	[EXDEV] = "Cross-device link",
};
//static const char *err_generic_format = "Unknown error %d";


void *memccpy(void *dst, const void *src, int c, size_t n)
{
	const unsigned char *src_bytes = src;
	unsigned char *dst_bytes = dst;	
	for(size_t i = 0; i < n; i++)
	{
		dst_bytes[i] = src_bytes[i];
		if(dst_bytes[i] == c)
			return &(dst_bytes[i+1]);
	}
	
	return NULL;
}

void *memchr(const void *s, int c, size_t n)
{
	const unsigned char *s_bytes = s;
	for(size_t i = 0; i < n; i++)
	{
		if(s_bytes[i] == c)
			return (void*)(&(s_bytes[i]));
	}
	
	return NULL;
}

int memcmp(const void *s1, const void *s2, size_t n)
{
	const unsigned char *s1_bytes = s1;
	const unsigned char *s2_bytes = s2;
	for(size_t i = 0; i < n; i++)
	{
		int diff = s1_bytes[i] - s2_bytes[i];
		if(diff != 0)
			return diff;
	}
	
	return 0;
}

//Ugly-ass memcpy to deal with ARMv4 not supporting misaligned accesses properly.
void *memcpy(void *dst, const void *src, size_t n)
{
	typedef struct _memcpy_cacheline_s { uint32_t a, b, c, d, e, f, g, h; } _memcpy_cacheline_t;

	//If we're copying less than one cacheline, just copy bytes
	char *dst_byte = (char*)dst;
	const char *src_byte = (const char*)src;
	if(n < sizeof(_memcpy_cacheline_t))
	{
		return memmove(dst, src, n);
	}
	
	//If we have different source/destination alignments, just copy bytes
	if( ((uintptr_t)dst ^ (uintptr_t)src) % sizeof(_memcpy_cacheline_t) )
	{
		return memmove(dst, src, n);
	}
	
	//Work until source and destination are cacheline-aligned
	size_t peel = ((uintptr_t)dst) % sizeof(_memcpy_cacheline_t);
	while(peel-- != 0)
	{
		*(dst_byte++) = *(src_byte++);
		n--;
	}
	
	//Do the remainder as cachelines
	_memcpy_cacheline_t *dst_line = (_memcpy_cacheline_t*)dst_byte;
	const _memcpy_cacheline_t *src_line = (_memcpy_cacheline_t*)src_byte;
	while(n >= sizeof(_memcpy_cacheline_t))
	{
		*(dst_line++) = *(src_line++);
		n -= sizeof(_memcpy_cacheline_t);
	}
	
	//Finish off any misaligned tail
	dst_byte = (char*)dst_line;
	src_byte = (const char*)src_line;
	while(n-- != 0)
	{
		*(dst_byte++) = *(src_byte++);
	}
	
	return dst;
}

void *memmove(void *dst, const void *src, size_t n)
{
	const unsigned char *src_bytes = src;
	unsigned char *dst_bytes = dst;
	
	if(dst == src || n == 0)
	{
		//No copy needed
	}
	else if(dst > src)
	{
		for(size_t i = n-1; 1; i--)
		{
			dst_bytes[i] = src_bytes[i];
			if(i == 0)
				break;
		}
	}
	else
	{
		for(size_t i = 0; i < n; i++)
		{
			dst_bytes[i] = src_bytes[i];
		}
	}
	
	return dst;
}


void *memset(void *s, int c, size_t n)
{	
	unsigned char *s_bytes = s;
	for(size_t ii = 0; ii < n; ii++)
	{
		s_bytes[ii] = c & 0xFF;
	}
	return s;
}

char *strcat(char *dst, const char *src)
{
	size_t lsrc = strlen(src);
	size_t ldst = strlen(dst);
	for(size_t i = 0; i < lsrc; i++)
	{
		dst[ldst + i] = src[i];
	}
	dst[lsrc+ldst] = '\0';
	return dst;
}

char *strchr(const char *s, int c)
{
	if(c == '\0')
		return (char*)(s + strlen(s));
	
	for(size_t i = 0; s[i] != '\0'; i++)
	{
		if(s[i] == c)
			return (char*)&(s[i]);
	}	
	return NULL;
}

int strcmp(const char *s1, const char *s2)
{
	for(size_t i = 0; (s1[i] != '\0') || (s2[i] != '\0'); i++)
	{
		int diff = s1[i] - s2[i];
		if(diff != 0)
			return diff;
	}
	
	return 0;
}

//int strcoll(const char *s1, const char *s2)
//{
	//I don't care about locales.
//	errno = ENOSYS;
//	return strcmp(s1, s2);
//}

char *strcpy(char *dst, const char *src)
{
	size_t i;
	for(i = 0; src[i] != '\0'; i++)
	{
		dst[i] = src[i];
	}
	dst[i] = '\0';
	
	return dst;
}

char *stpcpy(char *dst, const char *src)
{
	size_t i;
	for(i = 0; src[i] != '\0'; i++)
	{
		dst[i] = src[i];
	}
	dst[i] = '\0';
	return dst + i;
}

size_t strcspn(const char *s, const char *reject)
{
	size_t i;
	for(i = 0; s[i] != '\0'; i++)
	{
		for(size_t j = 0; reject[j] != '\0'; j++)
		{
			if(s[i] == reject[j])
				return i;
		}
	}
	return i;
}

char *strdup(const char *s)
{
	char *newstring = (char*)malloc(strlen(s)+1);
	strcpy(newstring, s);
	return newstring;
}

static char _strerror_non_reentrant_buffer[64];
char *strerror(int errnum)
{
	_strerror_non_reentrant_buffer[63] = '\0';
	strerror_r(errnum, _strerror_non_reentrant_buffer, 63);
	return _strerror_non_reentrant_buffer;
}

int strerror_r(int errnum, char *buf, size_t buflen)
{
	if( (errnum >= 0) && (errnum < NUM_ERR_NAMES) && (_err_names[errnum] != 0) )
	{
		strncpy(buf, _err_names[errnum], buflen);
	}
	else
	{
		strncpy(buf, "unimplemented", buflen);
		//assert(0);
		//snprintf(buf, buflen, err_generic_format, errnum);
	}
	
	return 0;
}

size_t strlen(const char *s)
{
	size_t i;
	for(i = 0; s[i] != '\0'; i++)
	{
		
	}
	return i;
}

char *strncat(char *dst, const char *src, size_t n)
{
	size_t lsrc = strnlen(src, n);
	size_t ldst = strlen(dst);
	for(size_t i = 0; i < lsrc; i++)
	{
		dst[ldst + i] = src[i];
	}
	dst[lsrc+ldst] = '\0';
	return dst;	
}

int strncmp(const char *s1, const char *s2, size_t n)
{
	for(size_t i = 0; (i < n) && (s1[i] != '\0' || s2[i] != '\0'); i++)
	{
		int diff = s1[i] - s2[i];
		if(diff != 0)
			return diff;
	}
	return 0;
}

char *strncpy(char *dst, const char *src, size_t n)
{
	size_t i;
	for(i = 0; i < n; i++)
	{
		dst[i] = src[i];
		if(src[i] == '\0')
			break;
	}
	
	for( ; i < n; i++)
	{
		dst[i] = '\0';
	}
	
	return dst;
}

size_t strnlen(const char *s, size_t n)
{
	size_t i;
	for(i = 0; i < n; i++)
	{
		if(s[i] == '\0')
			break;
	}
	return i;
}

char *strpbrk(const char *s, const char *accept)
{
	for(size_t i = 0; s[i] != '\0'; i++)
	{
		for(size_t j = 0; accept[j] != '\0'; j++)
		{
			if(s[i] == accept[j])
				return (char*)&(s[i]);
		}
	}
	return NULL;
}

char *strrchr(const char *s, int c)
{
	for(size_t i = strlen(s); 1; i--)
	{
		if(s[i] == c)
			return (char*)&(s[i]);
		
		if(i == 0)
			break;
	}
	
	return NULL;
}

size_t strspn(const char *s, const char *accept)
{
	size_t i;
	for(i = 0; s[i] != '\0'; i++)
	{
		size_t j;
		for(j = 0; accept[j] != '\0'; j++)
		{
			if(s[i] == accept[j])
				break;
		}
		
		if(accept[j] == '\0')
			return i;
	}
	return i;
}

char *strstr(const char *haystack, const char *needle)
{
	for(size_t i = 0; haystack[i] != '\0'; i++)
	{
		size_t j;
		for(j = 0; needle[j] != '\0'; j++)
		{
			if(haystack[i+j] != needle[j])
				break;
		}
		if(needle[j] == '\0')
			return (char*)&(haystack[i]);
	}
	return NULL;
}

static char *_strtok_non_reentrant_charptr = NULL;
char *strtok(char *str, const char *delim)
{
	return strtok_r(str, delim, &_strtok_non_reentrant_charptr);
}

char *strtok_r(char *str, const char *delim, char **saveptr)
{
	if(str == NULL)
		str = *saveptr;
	
	if(str == NULL || str[0] == '\0')
		return NULL;
	
	//We may have to skip over delimiters at the beginning of input -
	//so keep track of where the actual token begins.
	char *tokenstart = str;
	
	//Trivial search for delimiter characters
	for(char *sp = str; *sp != '\0'; sp++)
	{
		for(const char *dp = delim; *dp != '\0'; dp++)
		{
			if(*sp == *dp)
			{
				//Found match between string and delimiters.
				
				//Chop string.
				*sp = '\0';
				
				//If the delimiter is at the beginning of input, ignore and keep going
				if(sp == tokenstart)
				{
					tokenstart = sp + 1;
					continue;
				}
				
				//Otherwise, return the token now that we've hacked off its end.
				*saveptr = sp + 1;
				return tokenstart;
			}
		}
	}
		
	//No delimiters at all...
	*saveptr = NULL;
	return str;
}

//size_t strxfrm(char *dst, const char *src, size_t n)
//{
	//Still don't care about locales
//	errno = ENOSYS;
	
//	if(n == 0)
//		return 0;
	
//	strncpy(dst, src, n);
//	return strnlen(dst, n);
//}

void perror(const char *string)
{
	int toprint = errno;
	const char *errname = NULL;
	if(toprint < 0 || toprint >= NUM_ERR_NAMES)
		errname = "Bad error number";
	else
		errname = _err_names[toprint];
		
	if(string != NULL && *string != '\0')
	{
		fprintf(stderr, "%s: %s\n", string, errname);
	}
	else
	{
		fprintf(stderr, "%s\n", errname);
	}
}


void *memmem(const void *big, size_t big_len, const void *little, size_t little_len)
{
	const uint8_t *bigb = (const uint8_t*)big;
	const uint8_t *littleb = (const uint8_t*)little;
	for(size_t istart = 0; istart <= big_len - little_len; istart++)
	{
		bool match = true;
		for(size_t nb = 0; nb < little_len; nb++)
		{
			if(bigb[istart+nb] != littleb[nb])
			{
				match = false;
				break;
			}
		}
		if(match)
			return (void*)(bigb + istart);
	}
	return NULL;
}

char *strcasestr(const char *haystack, const char *needle)
{
	for(size_t i = 0; haystack[i] != '\0'; i++)
	{
		size_t j;
		for(j = 0; needle[j] != '\0'; j++)
		{
			if(tolower(haystack[i+j]) != tolower(needle[j]))
				break;
		}
		if(needle[j] == '\0')
			return (char*)&(haystack[i]);
	}
	return NULL;
}

char *strndup(const char *str, size_t len)
{
	//Count non-NUL characters in source string (up to "len" of them)
	size_t buflen = 0;
	while(str[buflen] != '\0' && buflen < len)
	{
		buflen++;
	}
	
	//Include room for NUL, whether source included one or not
	buflen++;
	
	char *retval = calloc(buflen, 1);
	if(retval == NULL)
		return retval;
	
	//Fill non-NUL
	memcpy(retval, str, buflen-1);
	
	//Fill NUL
	retval[buflen-1] = '\0';
	
	return retval;
}


