//sterm.c
//Stumpy terminal
//Bryan E. Topp <betopp@betopp.com> 2021

//This file is part of Stump; Copyright Bryan E. Topp, 2023.
//Stump is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//Stump is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with Stump. If not, see <https://www.gnu.org/licenses/>. 

#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sc.h>
#include <sc_con.h>
#include <signal.h>
#include <stdbool.h>
#include <ctype.h>
#include <sys/stat.h>

#include "confont.h"
#include "libtsm-3/src/libtsm.h"
#include "libtsm-3/external/xkbcommon-keysyms.h"

//Virtual terminals supported
typedef struct vt_s
{
	//Terminal state machine library - terminal state
	struct tsm_vte *vte;
	
	//Terminal state machine library - terminal screen contents
	struct tsm_screen *screen;
	
	//Last "age" returned by libTSM. If this doesn't change, nothing was drawn.
	unsigned int screen_age;
	
	//File descriptor for back of pseudoterminal
	int pty;
	
	//File descriptor for front of pseudoterminal
	int tty;
	
	//PID of child process (shell) running on this terminal
	id_t pid;
	
} vt_t;
#define VT_MAX 1
vt_t vt_table[VT_MAX];
int vt_active;

//Video backbuffer as allocated
uint16_t *fb_mem[2]; //Both buffers
uint16_t *fb_ptr; //Buffer we're drawing to
int fb_next; //Index of buffer we're drawing to
int fb_width;
int fb_height;
size_t fb_stride;

//Size of screen in character cells
int win_cols; //Width of screen in characters
int win_rows; //Height of screen in lines

//Age of last drawn character in each cell
unsigned int **win_ages;

//Keyboard modifier keys pressed
typedef enum kbd_mods_e
{
	KBD_MOD_LSHIFT = 0,
	KBD_MOD_RSHIFT = 1,
	KBD_MOD_LALT = 2,
	KBD_MOD_RALT = 3,
	KBD_MOD_LCTRL = 4,
	KBD_MOD_RCTRL = 5,
	
	KBD_MOD_SHIFT = 6,
	KBD_MOD_ALT = 7,
	KBD_MOD_CTRL = 8,
	
	KBD_MOD_MAX
} kbd_mods_t;
bool kbd_mods[KBD_MOD_MAX];

//Scancode-to-keysym mapping. Index is the standard kernel scancode enum.
//If an entry is \0, then no text character is generated for making that scancode.
//Each scancode has 4 translations - regular, shift, capslock, shift + capslock
const uint32_t keysym_usa[512][4] = 
{
	//Code                          Norm  Shft  Caps  Shft+Caps
	[_SC_CON_SCANCODE_A]           = {'a',  'A',  'A',  'a' },
	[_SC_CON_SCANCODE_B]           = {'b',  'B',  'B',  'b' },
	[_SC_CON_SCANCODE_C]           = {'c',  'C',  'C',  'c' },
	[_SC_CON_SCANCODE_D]           = {'d',  'D',  'D',  'd' },
	[_SC_CON_SCANCODE_E]           = {'e',  'E',  'E',  'e' },
	[_SC_CON_SCANCODE_F]           = {'f',  'F',  'F',  'f' },
	[_SC_CON_SCANCODE_G]           = {'g',  'G',  'G',  'g' },
	[_SC_CON_SCANCODE_H]           = {'h',  'H',  'H',  'h' },
	[_SC_CON_SCANCODE_I]           = {'i',  'I',  'I',  'i' },
	[_SC_CON_SCANCODE_J]           = {'j',  'J',  'J',  'j' },
	[_SC_CON_SCANCODE_K]           = {'k',  'K',  'K',  'k' },
	[_SC_CON_SCANCODE_L]           = {'l',  'L',  'L',  'l' },
	[_SC_CON_SCANCODE_M]           = {'m',  'M',  'M',  'm' },
	[_SC_CON_SCANCODE_N]           = {'n',  'N',  'N',  'n' },
	[_SC_CON_SCANCODE_O]           = {'o',  'O',  'O',  'o' },
	[_SC_CON_SCANCODE_P]           = {'p',  'P',  'P',  'p' },
	[_SC_CON_SCANCODE_Q]           = {'q',  'Q',  'Q',  'q' },
	[_SC_CON_SCANCODE_R]           = {'r',  'R',  'R',  'r' },
	[_SC_CON_SCANCODE_S]           = {'s',  'S',  'S',  's' },
	[_SC_CON_SCANCODE_T]           = {'t',  'T',  'T',  't' },
	[_SC_CON_SCANCODE_U]           = {'u',  'U',  'U',  'u' },
	[_SC_CON_SCANCODE_V]           = {'v',  'V',  'V',  'v' },
	[_SC_CON_SCANCODE_W]           = {'w',  'W',  'W',  'w' },
	[_SC_CON_SCANCODE_X]           = {'x',  'X',  'X',  'x' },
	[_SC_CON_SCANCODE_Y]           = {'y',  'Y',  'Y',  'y' },
	[_SC_CON_SCANCODE_Z]           = {'z',  'Z',  'Z',  'z' },
	[_SC_CON_SCANCODE_0]           = {'0',  ')',  '0',  ')' },
	[_SC_CON_SCANCODE_1]           = {'1',  '!',  '1',  '!' },
	[_SC_CON_SCANCODE_2]           = {'2',  '@',  '2',  '@' },
	[_SC_CON_SCANCODE_3]           = {'3',  '#',  '3',  '#' },
	[_SC_CON_SCANCODE_4]           = {'4',  '$',  '4',  '$' },
	[_SC_CON_SCANCODE_5]           = {'5',  '%',  '5',  '%' },
	[_SC_CON_SCANCODE_6]           = {'6',  '^',  '6',  '^' },
	[_SC_CON_SCANCODE_7]           = {'7',  '&',  '7',  '&' },
	[_SC_CON_SCANCODE_8]           = {'8',  '*',  '8',  '*' },
	[_SC_CON_SCANCODE_9]           = {'9',  '(',  '9',  '(' },
	[_SC_CON_SCANCODE_GRAVE]       = {'`',  '~',  '`',  '~' },
	[_SC_CON_SCANCODE_MINUS]       = {'-',  '_',  '-',  '_' },
	[_SC_CON_SCANCODE_EQUALS]      = {'=',  '+',  '=',  '+' },
	[_SC_CON_SCANCODE_BACKSLASH]   = {'\\', '|',  '\\', '|' },
	[_SC_CON_SCANCODE_SPACE]       = {' ',  ' ',  ' ',  ' ' },
	[_SC_CON_SCANCODE_TAB]         = {'\t', '\t', '\t', '\t'},
	[_SC_CON_SCANCODE_RETURN]      = {'\n', '\n', '\n', '\n'},
	[_SC_CON_SCANCODE_LEFTBRACKET] = {'[',  '{',  '[',  '{' },
	[_SC_CON_SCANCODE_RIGHTBRACKET]= {']',  '}',  ']',  '}' },
	
	[_SC_CON_SCANCODE_KP_DIVIDE]   = {'/',  '/',  '/',  '/' },
	[_SC_CON_SCANCODE_KP_MULTIPLY] = {'*',  '*',  '*',  '*' },
	[_SC_CON_SCANCODE_KP_MINUS]    = {'-',  '-',  '-',  '-' },
	[_SC_CON_SCANCODE_KP_PLUS]     = {'+',  '+',  '+',  '+' },
	[_SC_CON_SCANCODE_KP_ENTER]    = {'\n', '\n', '\n', '\n'},
	[_SC_CON_SCANCODE_KP_PERIOD]   = {'.',  '.',  '.',  '.' },
	[_SC_CON_SCANCODE_KP_0]        = {'0',  '0',  '0',  '0' },
	[_SC_CON_SCANCODE_KP_1]        = {'1',  '1',  '1',  '1' },
	[_SC_CON_SCANCODE_KP_2]        = {'2',  '2',  '2',  '2' },
	[_SC_CON_SCANCODE_KP_3]        = {'3',  '3',  '3',  '3' },
	[_SC_CON_SCANCODE_KP_4]        = {'4',  '4',  '4',  '4' },
	[_SC_CON_SCANCODE_KP_5]        = {'5',  '5',  '5',  '5' },
	[_SC_CON_SCANCODE_KP_6]        = {'6',  '6',  '6',  '6' },
	[_SC_CON_SCANCODE_KP_7]        = {'7',  '7',  '7',  '7' },
	[_SC_CON_SCANCODE_KP_8]        = {'8',  '8',  '8',  '8' },
	[_SC_CON_SCANCODE_KP_9]        = {'9',  '9',  '9',  '9' },
	
	[_SC_CON_SCANCODE_BACKSPACE]   = {8,    8,    8,    8   },
	[_SC_CON_SCANCODE_SLASH]       = {'/',  '?',  '/',  '?' },
	[_SC_CON_SCANCODE_PERIOD]      = {'.',  '>',  '.',  '>' },
	[_SC_CON_SCANCODE_COMMA]       = {',',  '<',  ',',  '<' },
	[_SC_CON_SCANCODE_SEMICOLON]   = {';',  ':',  ';',  ':' },
	[_SC_CON_SCANCODE_APOSTROPHE]  = {'\'', '"',  '\'', '"' },
};

//Returns the foreground RGB color for the given character attribute
static uint16_t attr_fgcolor(const struct tsm_screen_attr *attr)
{
	uint16_t fg = 0;
	fg = (fg << 5) | (attr->fr >> 3);
	fg = (fg << 6) | (attr->fg >> 2);
	fg = (fg << 5) | (attr->fb >> 3);
	
	if(attr->inverse)
		fg = ~fg;
	
	return fg;
}

//Returns the background RGB color for the given character attribute
static uint16_t attr_bgcolor(const struct tsm_screen_attr *attr)
{
	uint16_t bg = 0;
	bg = (bg << 5) | (attr->br >> 3);
	bg = (bg << 6) | (attr->bg >> 2);
	bg = (bg << 5) | (attr->bb >> 3);	
	
	if(attr->inverse)
		bg = ~bg;
	
	return bg;
}


//Draws a glyph at the given screen pixel.
static void glyph(int x, int y, char gl, uint16_t fg, uint16_t bg)
{
	for(int chy = 0; chy < confont_chy; chy++)
	{
		uint16_t *dst = (uint16_t*)(((char*)fb_ptr) + (x * sizeof(uint16_t)) + ( (y + chy) * fb_stride));
		for(int chx = 0; chx < confont_chx; chx++)
		{
			dst[chx] = (confont_bits[(uint8_t)gl][chy] & (1<<chx)) ? fg : bg;
		}
	}
}

//Handles a keyboard input.
void kbd(_sc_con_scancode_t scancode, bool press)
{
	//Look up active terminal to receive our input
	vt_t *vptr = &(vt_table[vt_active]);
	
	//Check if a modifier key is being changed
	static const _sc_con_scancode_t mod_scan[KBD_MOD_MAX] = 
	{
		[KBD_MOD_LSHIFT] = _SC_CON_SCANCODE_LSHIFT,
		[KBD_MOD_RSHIFT] = _SC_CON_SCANCODE_RSHIFT,
		[KBD_MOD_LCTRL] = _SC_CON_SCANCODE_LCTRL,
		[KBD_MOD_RCTRL] = _SC_CON_SCANCODE_RCTRL,
		[KBD_MOD_LALT] = _SC_CON_SCANCODE_LALT,
		[KBD_MOD_RALT] = _SC_CON_SCANCODE_RALT
	};
	for(int mm = 0; mm < KBD_MOD_MAX; mm++)
	{
		if(mod_scan[mm] != 0 && scancode == mod_scan[mm])
			kbd_mods[mm] = press;
	}
	
	//Recompute generic modifiers
	kbd_mods[KBD_MOD_ALT] = kbd_mods[KBD_MOD_LALT] || kbd_mods[KBD_MOD_RALT];
	kbd_mods[KBD_MOD_SHIFT] = kbd_mods[KBD_MOD_LSHIFT] || kbd_mods[KBD_MOD_RSHIFT];
	kbd_mods[KBD_MOD_CTRL] = kbd_mods[KBD_MOD_LCTRL] || kbd_mods[KBD_MOD_RCTRL];
	
	if(!press)
	{
		//This was a key release - it updates modifiers but doesn't trigger any action.
		return;
	}
	
	//Translate scancode to keysym
	int keysym_set = 0;
	if(kbd_mods[KBD_MOD_SHIFT])
		keysym_set += 1;
	uint32_t keysym = keysym_usa[scancode][keysym_set];
	
	unsigned int tsm_mods = 0;
	if(kbd_mods[KBD_MOD_ALT])
		tsm_mods |= TSM_ALT_MASK;
	if(kbd_mods[KBD_MOD_SHIFT])
		tsm_mods |= TSM_SHIFT_MASK;
	if(kbd_mods[KBD_MOD_CTRL])
		tsm_mods |= TSM_CONTROL_MASK;

	//Feed input to terminal state machine - check if it gets handled specially
	if(tsm_vte_handle_keyboard(vptr->vte, keysym, XKB_KEY_NoSymbol, tsm_mods, TSM_VTE_INVALID))
	{
		//Terminal had special handling for this keypress
		return;
	}
	
	//Pass normal keypresses
	if(keysym > 0 && keysym < 0x80)
	{
		uint8_t keyval = keysym & 0xFF;
		write(vptr->pty, &keyval, 1);
	}
}

//Write callback from libTSM
//Used when an escape sequence from the shell causes us to return terminal info to it.
void sterm_tsm_vte_write_cb(struct tsm_vte *vte, const char *u8, size_t len, void *data)
{
	(void)vte;
	
	//Data pointer points into our virtual-terminal table
	vt_t *vptr = (vt_t*)data;
	
	//Check for control characters coming through, and handle them.
	//In a system like Linux, this happens between the pty and tty, and wouldn't be part of the terminal emulator.
	//We just handle it here, because our PTY system is a straight pipe.
	if( (len > 0) && (u8[0] == 3) )
	{
		//Ctrl-C - send SIGTERM to all descendants on this virtual terminal
		_sc_sig_send(_SC_IDTYPE_PTREE, vptr->pid, _SC_SIGTERM);
		return;
	}
	
	//Write the data to the back of the pseudoterminal, sending it to the shell.
	//Note - tries once and doesn't block. If the application isn't reading its input, too bad.
	_sc_write(vptr->pty, u8, len);
}

//Draw callback from libTSM
//Used when an area of the screen is changed by console output
int sterm_tsm_screen_draw_cb(struct tsm_screen *con, uint32_t id, const uint32_t *ch, size_t len, unsigned int width,
				   unsigned int posx, unsigned int posy, const struct tsm_screen_attr *attr, tsm_age_t age, void *data)
{
	//We only call draw() on the active terminal anyway
	(void)con;
	(void)data;
	
	//Don't support long/wide characters
	(void)ch;
	(void)len;
	(void)width;
	
	//Don't handle foreign crap at the moment - draw ASCII chars, replace others with ' '.
	if(id >= 0x80)
		id = ' ';
	if(id == 0)
		id = ' ';
	if(!isprint(id))
		id = ' ';
	
	//Paint the character, if its age has changed (i.e. it's been modified)
	if(age != win_ages[posy][posx])
		glyph(posx*confont_chx, posy*confont_chy, id, attr_fgcolor(attr), attr_bgcolor(attr));
	
	win_ages[posy][posx] = age;
	return 0;
}

int main(int argc, const char **argv, const char **envp)
{
	//Ignore these for now.
	//Maybe eventually we can pass something through from the kernel command-line or whatever.
	(void)argc;
	(void)argv;
	(void)envp;
	
	//Open /dev/null as our stdin initially
	int null_fd = open("/dev/null", O_RDONLY);
	if(null_fd < 0)
	{
		perror("open /dev/null");
		abort();
	}
	dup2(null_fd, STDIN_FILENO);
	dup2(null_fd, STDOUT_FILENO);
	dup2(null_fd, STDERR_FILENO);
	
	//Set home directory and put us there
	setenv("HOME", "/home", 0);
	if(chdir("/home") < 0)
	{
		perror("chdir /home");
		abort();
	}
	
	//Set initial environment vars
	setenv("TERM", "xterm", 0);
	setenv("UNAME_s", "stump", 0);
	setenv("UNAME_n", "localhost", 0);
	setenv("UNAME_r", BUILDUSER "-" BUILDDATE, 0);
	setenv("UNAME_v", BUILDVERSION, 0);
	
	//Grab console if given to us
	_sc_con_pass(getpid());

	//Find framebuffer dimensions
	_sc_con_fbinfo_t con_fbinfo = {0};
	_sc_con_fbinfo(&con_fbinfo, sizeof(con_fbinfo));
	fb_width = con_fbinfo.width;
	fb_height = con_fbinfo.height;
	fb_stride = con_fbinfo.stride;
	win_cols = fb_width / confont_chx;
	win_rows = fb_height / confont_chy;
	
	char rcbuf[32];
	snprintf(rcbuf, sizeof(rcbuf)-1, "%d", win_rows);
	setenv("LINES", rcbuf, 0);
	snprintf(rcbuf, sizeof(rcbuf)-1, "%d", win_cols);
	setenv("COLUMNS", rcbuf, 0);
	
	//Initialize each virtual terminal
	for(int vv = 0; vv < VT_MAX; vv++)
	{
		vt_t *vptr = &(vt_table[vv]);
		
		//Make a pseudoterminal file to communicate with the shell
		char fnbuf[32] = {0};
		snprintf(fnbuf, sizeof(fnbuf)-1, "/dev/tty%d", vv);
		
		vptr->tty = open(fnbuf, O_RDWR | O_CREAT, _SC_S_IFPTY | 0666);
		if(vptr->tty < 0)
		{
			perror("open /dev/ttyX");
			abort();
		}
			
		//Open the back of that same pseudoterminal
		vptr->pty = _sc_find(vptr->tty, "", _SC_ACCESS_R | _SC_ACCESS_W | _SC_ACCESS_BACK);
		if(vptr->pty < 0)
		{
			errno = -vptr->pty;
			perror("open pty");
			abort();
		}
		
		//Initialize a libTSM screen buffer
		int screen_err = tsm_screen_new(&(vptr->screen), NULL, NULL);
		if(screen_err != 0)
		{
			errno = -screen_err;
			perror("tsm_screen_new");
			abort();
		}
		
		int screen_size_err = tsm_screen_resize(vptr->screen, win_cols, win_rows);
		if(screen_size_err != 0)
		{
			errno = -screen_size_err;
			perror("tsm_screen_resize");
			abort();
		}
		
		//Initialize a libTSM terminal state machine to update the screen
		int vte_err = tsm_vte_new(&(vptr->vte), vptr->screen, &sterm_tsm_vte_write_cb, (void*)vptr, NULL, NULL);
		if(vte_err != 0)
		{
			errno = -vte_err;
			perror("tsm_vte_new");
			abort();
		}
		
		//Start the state-machine in newline mode
		tsm_vte_input(vptr->vte, "\x1B[20h", 5);

		//Spawn the shell (spawn the getty?)
		vptr->pid = fork();
		if(vptr->pid < 0)
		{
			//Failed to fork
			perror("fork");
			abort();
		}
		
		if(vptr->pid == 0)
		{
			//Child process for this virtual terminal
			setenv("TTY", fnbuf, 0);
			
			//Install the pseudoterminal as stdin/stdout/stderr
			close(vptr->pty);
			dup2(vptr->tty, STDIN_FILENO);
			dup2(vptr->tty, STDOUT_FILENO);
			dup2(vptr->tty, STDERR_FILENO);
			
			//Show intro message (as if sent by a shell)
			printf("sterm: " BUILDVERSION " by " BUILDUSER " at " BUILDDATE "\n");
			printf("sterm: spawned pid=%d ppid=%d\n", getpid(), getppid());
			printf("sterm: exec /bin/oksh-6.9\n");
			
			//Run shell
			extern char **environ;
			execve("/bin/oksh-6.9", (char*[]){"oksh-6.9", NULL}, (char**)environ);
			perror("execve");
			abort(); //execl shouldn't return
			
			//Todo - maybe fork again in a loop so we can respawn the shell if it dies?
			//Is that what a getty does? Do we need a getty?
		}
	}
	
	//Make a buffer for the age of each character on the screen - to avoid redrawing
	win_ages = calloc(win_rows, sizeof(win_ages[0]));
	for(int rr = 0; rr < win_rows; rr++)
	{
		win_ages[rr] = calloc(win_cols, sizeof(win_ages[0][0]));
	}
	
	//Map framebuffers
	for(int bb = 0; bb < 2; bb++)
	{
		intptr_t map_result = _sc_mem_disp();
		if(map_result < 0)
		{
			errno = -map_result;
			perror("mem_disp");
			abort();
		}
		
		fb_mem[bb] = (uint16_t*)map_result;
	}
	fb_next = 0;
	fb_ptr = fb_mem[0];
	memset(fb_ptr, 0, fb_height * fb_stride);
	
	//Pump those pipes
	while(1)
	{
		//Get input until no more arrives
		while(1)
		{
			//Ask kernel for input events - bail if none are pending
			_sc_con_input_t from_con_buf[16];
			ssize_t from_con_len = _sc_con_input(from_con_buf, sizeof(from_con_buf[0]), sizeof(from_con_buf));
			if(from_con_len == -EAGAIN)
				break;

			//Work through each event, feeding to the active terminal
			for(int32_t ii = 0; ii < from_con_len; ii++)
			{
				switch(from_con_buf[ii].type.type)
				{
					case _SC_CON_INPUT_TYPE_KBD:
						kbd(from_con_buf[ii].kbd.scancode, from_con_buf[ii].kbd.state);
						break;
					default:
						break;
				}
			}
		}
		
		//Get each shell's output until no more arrives
		for(int vv = 0; vv < VT_MAX; vv++)
		{
			vt_t *vptr = &(vt_table[vv]);
			while(1)
			{
				//Ask kernel for more shell output - bail if none is pending
				char from_shell_buf[64];
				ssize_t from_shell_len = _sc_read(vptr->pty, from_shell_buf, sizeof(from_shell_buf));
				if(from_shell_len == -EAGAIN)
					break;
				
				//Feed into terminal state machine
				if(from_shell_len > 0)
					tsm_vte_input(vptr->vte, from_shell_buf, from_shell_len);
			}

			//Check for child process death
			int wstatus = 0;
			if(waitpid(vptr->pid, &wstatus, WNOHANG | WEXITED) > 0)
			{
				//Just puke and let the kernel die if the shell dies
				exit(-1);
			}
		}
		
		//Redraw screen of active terminal
		tsm_age_t new_age = tsm_screen_draw(vt_table[vt_active].screen, &sterm_tsm_screen_draw_cb, &(vt_table[vt_active].vte));
		if(new_age != vt_table[vt_active].screen_age)
		{
			vt_table[vt_active].screen_age = new_age;
			while(1)
			{
				int flip_result = _sc_con_fbdisp(fb_ptr);
				if(flip_result == -EAGAIN || flip_result == -EBUSY)
				{
					//Waiting for vsync (or not on-screen). Try again later.
					_sc_pause();
					continue;
				}
				
				if(flip_result != 0 )
				{
					//Some other error flipping.
					errno = -flip_result;
					perror("flip");
					abort();
				}
				
				//Flipped successfully.
				fb_next = (fb_next + 1) % 2;
				memcpy(fb_mem[fb_next], fb_ptr, fb_height * fb_stride);
				fb_ptr = fb_mem[fb_next];
				break;
			}
			
			//Don't go on to pause - we already sync'd to vsync this time
			continue;
		}
		
		//We read/write everything until it runs empty/full.
		//So at this point, we can try to pause - if nothing further happens since our syscalls, we'll wait.
		_sc_pause();
	}
}
